class language_circuit {
    constructor($NLP, $initLang="EN") {
        this.NLP = $NLP;

        this._pats = [];
        this._quoteCharCodes=[];

        this._prefixMap = null;
        this._maxPrefixLen = 0;
        this._minPrefixLen = 10;
        this._suffixMap = null;
        this._maxSuffixLen = 0;
        this._phSignalsMap = null;

        this.LANG = null;
        this.init_lang = $initLang;

        this._inlineLeftContext=[];
        this._inlineRightContext=[];
        this._inlineLRContext = [];
        this._lrpairs = [];

        this._innerBreakers=[];

        this._modRoles = [];
    }
    
    init = async()=>{

        if (this.init_lang =="EN") await this.initEnglish();

        this._modRoles.push(MEM.role_thing_mod.temp_nid);
        this._modRoles.push(MEM.role_actor_mod.temp_nid);
        this._modRoles.push(MEM.role_affected_thing.temp_nid);
        this._modRoles.push(MEM.role_be_thing.temp_nid);
        this._modRoles.push(MEM.role_posn_thing.temp_nid);
        this._modRoles.push(MEM.role_instrument.temp_nid);

        //TO DO.
        //quote and single quote and dash or double dash or astericks or double astericks..
        //OR ANY CREATIVE USE OF ANY CHAR
        //should be L/R context.
        // Bob "the train" is fun.
        // Bob **the train** is fun.
        // Bob <<<the train>>> is fun.

        this._setCharLRContextFlags([91,40,123], [93,41,125]);    //"[", "(", "{"
        this._setCharFlags([91, 40, 123, 93, 41, 125], ["breaker"]);  // "]", ")", "}"
        this._setCharFlags([47, 45], ["inner_breaker"]);  //  "/", "-"
        this._setCharFlags([34, 39], ["quote"]);
        this._innerBreakerAltStrs=["/"];
        this._innerBreakerPartStrs=["-"];
        this._setCharFlag(63, "question");//  "?"
        this._setCharFlag(47, "alt");   //  "/"
        this._setCharFlag(45, "part");  //  "-"
    }


    /**
     * These are for inline context that help resolve/clarify.
     * If new info, they are new relons in the parse.. but they are not
     * semantically apart of a clump.  
     * 
     * "Bill Murray (the actor) was great in What About Bob."
     * Similar to 
     * "The actor Bill Murray was great in the movie What About Bob."
     * Murray was great in What About Bob.
     * 
     * @param {*} $arrLeft 
     * @param {*} $arrRight 
     */
    _setCharLRContextFlags = function($arrLeft, $arrRight){
        for (var a=0; a <$arrLeft.length; a++) {
            var l = this.NLP._getCharByCode($arrLeft[a]);
            var r = this.NLP._getCharByCode($arrRight[a]);

            //this._inlineLRContext.push($arrLeft[a]);
            this._inlineLeftContext.push($arrLeft[a]);

            //this._inlineLRContext.push($arrRight[a]);
            this._inlineRightContext.push($arrRight[a]);

            l.left_context=true;
            r.right_context=true;
            l.r_charObj = r;
            r.l_charObj = l;
        }
    }
    
    _setCharFlagsByStr = function($arr, $flags) {
        for (var a=0; a <$arr.length; a++) {
            var charCode = $arr[a].charCodeAt(0);
            var charObj = this.NLP._getCharByCode(charCode);
                
            if (charObj==null) continue;//preload tinkerings can leave us without this.
            
            for (var f=0; f < $flags.length; f++) {
                charObj[ $flags[f] ] = true;
            }
        }
    }

    _setCharFlags = function($arr, $flags) {
        for (var a=0; a <$arr.length; a++) {
            var charObj = this.NLP._getCharByCode($arr[a]);
            for (var f=0; f < $flags.length; f++) {
                charObj[ $flags[f] ] = true;
            }
        }
    }

    _setCharFlag = function($a, $flag) {
        var charObj = this.NLP._getCharByCode($a);
        charObj[ $flag ] = true;
    }

    /**
     */
    initEnglish = async()=> {
        this.LANG = new language_circuit_english(this.NLP, this);
        await this.LANG.init();
    }

    /**
     * do this later
     */
    initBahasa = function() {
    }


    /**
     */
    initPrefixMap = function($lang=null) {
        var prefix_nOBJS = [];
        this._prefixMap = new neuron_mapper();
        MEM.LC_prefix.getAllDownlineChildren(prefix_nOBJS, 0, false, false, false);
        var pcnt=0;
        for (var p=0; p < prefix_nOBJS.length; p++) {
            var prefixNOBJ = prefix_nOBJS[p];
            if (prefixNOBJ.text.indexOf("ec:")!=-1 || prefixNOBJ.text.indexOf("ec.")!=-1) continue;
            if ($lang!=null && prefixNOBJ.bes.indexOf($lang.temp_nid)==-1) continue;
            this._prefixMap.addUniqueSingleNObj(prefixNOBJ);
            if (prefixNOBJ.text.length>this._maxPrefixLen) this._maxPrefixLen=prefixNOBJ.text.length;
            if (prefixNOBJ.text.length<this._minPrefixLen) this._minPrefixLen=prefixNOBJ.text.length;
            
            pcnt+=1;
        }
        console.log("%c(initPrefixMap) mapped prefix "+pcnt, 'background: yellow; color: black');
    }



    /**
     */
    initSuffixMap = function($lang=null) {
        var suffix_nOBJS = [];
        var suffix_objs = [];
        this._suffixMap = new neuron_mapper();
        MEM.LC_suffix.getAllDownlineChildren(suffix_nOBJS, 0, false, false, false);
        var scnt=0;
        for (var p=0; p < suffix_nOBJS.length; p++) {
            var suffixNOBJ = suffix_nOBJS[p];
            if (suffixNOBJ.text.indexOf("ec:")!=-1 || suffixNOBJ.text.indexOf("ec.")!=-1) continue;
            
            //only load suffixes of a given lang.
            //we can still parse foreign words (with respect to this lang)
            //but we will not apply morph+ending rules to new words we see in that lang.
            //NOTE: defaults will apply ENGLISH ending rules to foreign new words.. and likely
            //get it wrong... That's ok.  We would too.
            if ($lang!=null && suffixNOBJ.bes.indexOf($lang.temp_nid)==-1) continue;
            
            var sObj = this._suffixMap.addUniqueSingleNObj(suffixNOBJ);
            if (sObj.nObj.text.length>this._maxSuffixLen) this._maxSuffixLen=sObj.nObj.text.length;
            suffix_objs.push(sObj);
            scnt+=1; 
        }
        console.log("%c(initSuffixMap) mapped suffix "+scnt, 'background: yellow; color: black');
        this.LANG.alterSuffixMap(suffix_objs);
    }


    /**
     * is a neuron a resolvable? (ie, like a pn, "i" "we" "they" etc)
     * We first have to parse the word to get the pn..
     * Then we can know that concept itself (hehe) must be resolved.
     */
    _isResolvable=function($nObj) {
        var arr = this._phSignalsMap.getByText($nObj.text.toLowerCase());
        if (arr==null || arr.length==0) return false;
        for (var a=0; a < arr.length; a++) {
            if (arr[a].nObj.temp_nid==$nObj.temp_nid && arr[a].resolve==true) return true; 
        }
        return false;

    }

    /**
     * These are words in any given language that signal the beginning of a new phrase
     * 
     * Key in ANNE:
     * 
     * RESOLVABLES      (pronoun like)
     * MODES            (aka modals)
     * EXPECTATIONS     (signals for upcoming things)
     */
     initPhraseSignalsMap = async($lang=null)=> {
        var allSigs = [];
        var dupemap=[];

        var aExpections = [];  
        this._phSignalsMap = new neuron_mapper();

        //<<<<<<<<<ANYTHING THAT IS AN EXPECTATION
        MEM.ec_expection.getAllDownlineChildren(aExpections, 0, false, false, false);
        
        //MEM.ec_expect_question

        var aResolvables = [];
        MEM.ec_resolve.getAllDownlineChildren(aResolvables, 0, false, false, false);
        var rezcnt = 0;
        for (var rz=0; rz < aResolvables.length; rz++) {
            var rezNOBJ = aResolvables[rz];
            var text = rezNOBJ.text.toLowerCase();
            if (text.indexOf("ec:")!=-1 || text.indexOf("ec.")!=-1) continue;
            if (dupemap[rezNOBJ.temp_nid]==undefined) {
                //set flags in here.
                var rezObj = this._phSignalsMap.addResolveNObjArr(rezNOBJ);
                dupemap[rezNOBJ.temp_nid]=rezObj;
                allSigs.push(rezObj);
                rezcnt+=1;

                if (rezNOBJ.anyParentHasBeOf(MEM.ec_resolvableThing.temp_nid)==true) {
                    rezObj.resolve_thing=true;
                    rezObj.psbl_roles.push(MEM.role_thing.temp_nid);
                    //this._addRezThingRolesToSignal(rezObj.psbl_roles);
                    //this.ecpool_adders.push(MEM.ec_thing.temp_nid);
                }
            }
            else dupemap[rezNOBJ.temp_nid].resolve=true;
        }

        //be verb signal (no the be aux tho, that is different)
        //we want both aux and vb to be in the signals array so we get either
        //an aux role or an action role.
        var iscnt = 0;
        var aIs = [];
        MEM.ec_is.getAllDownlineChildren(aIs, 0, false, false, false);
        for (var e=0; e < aIs.length; e++) {
            var isNOBJ = aIs[e];
            var text = isNOBJ.text.toLowerCase();
            if (text.indexOf("ec:")!=-1 || text.indexOf("ec.")!=-1) continue;
            if (dupemap[isNOBJ.temp_nid]==undefined) {
                
                //set flags in here.
                var isObj = this._phSignalsMap.addIsSignalNObjArr(isNOBJ);
                dupemap[isNOBJ.temp_nid]=isObj;
                if (isObj.psbl_roles.indexOf(MEM.role_action.temp_nid==-1)) {
                    isObj.psbl_roles.push(MEM.role_action.temp_nid);
                }
                allSigs.push(isObj);
                iscnt+=1;
            }

            //make sure these are set in case the obj was in the .resolve loop
            else {
                dupemap[isNOBJ.temp_nid].is=true;  
                dupemap[isNOBJ.temp_nid].signal=true;
            }
        }

        //posn verb signal  (not the has/have/had aux, that is different)
        //we want both aux and vb to be in the signals array so we get either
        //an aux role or an action role.
        var posncnt = 0;
        var aPosn = [];
        MEM.ec_posn.getAllDownlineChildren(aPosn, 0, false, false, false);
        for (var p=0; p < aPosn.length; p++) {
            var posnNOBJ = aPosn[p];
            var text = posnNOBJ.text.toLowerCase();
            if (text.indexOf("ec:")!=-1 || text.indexOf("ec.")!=-1) continue;
            if (dupemap[posnNOBJ.temp_nid]==undefined) {
                
                //set flags in here.
                var posnObj = this._phSignalsMap.addPosnSignalNObjArr(posnNOBJ);
                dupemap[posnNOBJ.temp_nid]=posnObj;
                if (posnObj.psbl_roles.indexOf(MEM.role_action.temp_nid==-1)) {
                    posnObj.psbl_roles.push(MEM.role_action.temp_nid);
                }
                allSigs.push(posnObj);
                posncnt+=1;
            }

            //make sure these are set in case the obj was in the .resolve loop
            else {
                dupemap[posnNOBJ.temp_nid].posn=true;  
                dupemap[posnNOBJ.temp_nid].signal=true;
            }
        }

        var aModes = [];
        MEM.ec_mode.getAllDownlineChildren(aModes, 0, false, false, false);
        var mcnt = 0;
        for (var m=0; m < aModes.length; m++) {
            var mNOBJ = aModes[m];
            var text = mNOBJ.text.toLowerCase();
            if (text.indexOf("ec:")!=-1 || text.indexOf("ec.")!=-1) continue;
            if (dupemap[mNOBJ.temp_nid]==undefined) {
                
                //set flags in here.
                var mObj = this._phSignalsMap.addModeSignalNObjArr(mNOBJ);
                dupemap[mNOBJ.temp_nid]=mObj;
                mObj.psbl_roles.push(MEM.role_mode.temp_nid);
                allSigs.push(mObj);
                mcnt+=1;
            }

            //make sure these are set in case the obj was in the .resolve loop
            else {
                dupemap[mNOBJ.temp_nid].mode=true;  
                dupemap[mNOBJ.temp_nid].signal=true;
            }
        }
         
        var singulars=0;
        var plurals=0;
        var negations=0;
        var ecnt = 0;

        for (var e=0; e < aExpections.length; e++) {
            var eNOBJ = aExpections[e];
            var text = eNOBJ.text.toLowerCase();

            if (text.indexOf("ec:")!=-1 || text.indexOf("ec.")!=-1) continue;
            
            
            if (dupemap[eNOBJ.temp_nid]==undefined) {
                
                //set flags in here.
                var eobj = this._phSignalsMap.addExpectsNObjArr(eNOBJ);
                dupemap[eNOBJ.temp_nid]=eobj;
                allSigs.push(eobj);
                ecnt+=1;

                if (eNOBJ.anyParentHasBeOf(MEM.ec_expect_more_of_prev.temp_nid)==true) {
                    eobj.connector=true;
                    eobj.psbl_roles.push(MEM.role_connector.temp_nid);
                }
                                
                else if (eNOBJ.anyParentHasBeOf(MEM.ec_expect_question.temp_nid)==true) {
                    eobj.question_signal=true;
                    eobj.ec_type_tnid = MEM.ec_expect_question.temp_nid;
                    eobj.psbl_roles.push(MEM.role_question.temp_nid);    //will be question thing
                }
            }

            //make sure these are set in case the obj was in the .resolve/.mode/.signal loop
            else dupemap[mNOBJ.temp_nid].expects=true;

            //set more flags. (will use pools when we get to context/concepts)

            if (eobj.nObj.text=="to") {
                //should have expected roles here
                console.log("test to");
            }

            if (eobj.nObj.anyParentHasBeOf(MEM.ec_expect_when.temp_nid)==true) {
                eobj.x_when=true;
                x_when+=1;
                eobj.expect_pool = MEM.ec_expect_when;
                eobj.dpool_adders.push(MEM.ec_thing.temp_nid);
                if (eobj.expects_roles.indexOf(MEM.role_when.temp_nid)==-1) {
                    eobj.expects_roles.push(MEM.role_when.temp_nid);
                }
            }
            //modifies inner things
            else if (eobj.nObj.anyParentHasBeOf(MEM.ec_expect_thing_mod.temp_nid)==true) {
                eobj.x_thing_mod=true;
                eobj.mod=true;
                eobj.expect_pool = MEM.ec_expect_thing_mod;
                if (eobj.expects_roles.indexOf(MEM.role_thing_mod.temp_nid)==-1) {
                    eobj.expects_roles.push(MEM.role_thing_mod.temp_nid); 
                }
            }
            else if (eobj.nObj.anyParentHasBeOf( MEM.ec_expect_where.temp_nid)==true) {
                eobj.x_where=true;
                x_where+=1;
                eobj.expect_pool = MEM.ec_expect_where;
                eobj.dpool_adders.push(MEM.ec_thing.temp_nid);
                if (eobj.expects_roles.indexOf(MEM.role_where.temp_nid)==-1) {
                    eobj.expects_roles.push(MEM.role_where.temp_nid);
                }
            }
            //aka, aux vbs.  "has been", "does", "did", "being"
            //many aux vbs in english are also their own vbs... these are the verb signallers,
            //whatever language.
            //should also be psbl_role "action_mod"
            else if (eobj.nObj.anyParentHasBeOf(MEM.ec_expect_action.temp_nid)==true) {
                eobj.x_action=true;
                x_action+=1;
                eobj.expect_pool = MEM.ec_expect_action;
                
                //ldr(Consts.TYPE_BRAIN, n1, this.ec_expect_role, r1);
                if (eobj.expects_roles.indexOf(MEM.role_action.temp_nid)==-1) {
                    eobj.expects_roles.push(MEM.role_action.temp_nid);
                }
                eobj.dpool_adders.push(MEM.ec_action.temp_nid);
            }
            //dets.. some pns.. (her cat, their dog)..  her is a RESOLVE_THING also.
            else if (eobj.nObj.anyParentHasBeOf(MEM.ec_expect_thing.temp_nid)==true) {
                eobj.x_thing=true;
                x_thing_role+=1;
                eobj.expects_roles.push(MEM.role_thing.temp_nid);
                eobj.dpool_adders.push(MEM.ec_thing.temp_nid);
            }
            else if (eobj.nObj.anyParentHasBeOf(MEM.ec_expect_actor.temp_nid)==true) {
                eobj.x_actor=true;
                x_actor_role+=1;
                eobj.dpool_adders.push(MEM.ec_thing.temp_nid);
                if (eobj.expects_roles.indexOf(MEM.role_actor.temp_nid)==-1) {
                    eobj.expects_roles.push(MEM.role_actor.temp_nid); 
                }
            }
            else if (eobj.nObj.anyParentHasBeOf(MEM.ec_expect_theme_thing.temp_nid)==true) {
                eobj.x_theme_thing=true;
                eobj.dpool_adders.push(MEM.ec_thing.temp_nid);
                if (eobj.expects_roles.indexOf(MEM.role_theme_thing.temp_nid)==-1) {
                    eobj.expects_roles.push(MEM.role_theme_thing.temp_nid); 
                }
            }
            else if (eobj.nObj.anyParentHasBeOf(MEM.ec_expect_affected_thing.temp_nid)==true) {
                eobj.x_affected_thing=true;
                eobj.dpool_adders.push(MEM.ec_thing.temp_nid);
                if (eobj.expects_roles.indexOf(MEM.role_affected_thing.temp_nid)==-1) {
                    eobj.expects_roles.push(MEM.role_affected_thing.temp_nid);
                }
            }
            else if (eobj.nObj.anyParentHasBeOf(MEM.ec_expect_manner.temp_nid)==true) {
                eobj.x_manner=true;
                eobj.dpool_adders.push(MEM.ec_manner.temp_nid);
                if (eobj.expects_roles.indexOf(MEM.role_manner.temp_nid)==-1) {
                    eobj.expects_roles.push(MEM.role_manner.temp_nid); 
                }
            }
            else if (eobj.nObj.anyParentHasBeOf(MEM.ec_expect_actionthing.temp_nid)==true) {
                eobj.x_actionthing=true;  //not an actual role..
                eobj.dpool_adders.push(MEM.ec_thing.temp_nid);
            }
        }

        var be_manner=0;
        var be_direction=0;
        var x_when=0;
        var x_where=0;
        var x_action=0;
        var x_thing_role=0;
        var x_actor_role = 0;
        var expects_roles=0;
        var indef=0;    //KEEP WORD IN MOD OF PHRASE IT COMES IN.

        //these checks are like a limited context pool for each word..
        //pooling the relons and looking for specific EC concepts that
        //will not change lang to lang
        for (var a=0; a < allSigs.length; a++) { 
            var signal = allSigs[a];

            if (signal.nObj.anyParentHasBeOf(MEM.ec_indefiniteness.temp_nid)==true) {
                signal.indef=true;
                indef+=1;
                signal.dpool_adders.push(MEM.ec_indefiniteness.temp_nid);
            }

            //tag expected thing with singular/pluarl/negation
            //either of these, but not both.
            if (signal.nObj.anyParentHasBeOf(MEM.ec_singular.temp_nid)==true) {
                signal.be_singular=true;
                singulars+=1;
            }
            else if (signal.nObj.anyParentHasBeOf(MEM.ec_plural.temp_nid)==true) {
                signal.be_plural=true;
                plurals+=1;
            }


            if (signal.nObj.anyParentHasBeOf( MEM.ec_negation.temp_nid)==true) {
                signal.be_negation=true;
                negations+=1;
            }

            //todo: investigate why we aren't getting any of these.
            //psbl role for the wd itself.
            if (signal.nObj.anyParentHasBeOf( MEM.ec_direction.temp_nid)==true) {
                signal.be_direction=true;
                be_direction+=1;
                if (signal.psbl_roles.indexOf(MEM.role_where.temp_nid)==-1) {
                    signal.psbl_roles.push(MEM.role_where.temp_nid);
                }
                signal.dpool_adders.push(MEM.ec_thing.temp_nid);
            }

            else if (signal.nObj.anyParentHasBeOf( MEM.ec_manner.temp_nid)==true) {
                signal.be_manner=true;
                be_manner+=1;   //actor mod?
                if (signal.psbl_roles.indexOf(MEM.role_manner.temp_nid)==-1) {
                    signal.psbl_roles.push(MEM.role_manner.temp_nid);
                }
                signal.dpool_adders.push(MEM.ec_manner.temp_nid);
            }

            
            //expected roles are for signals that signal a coming role.
            //they do not serve the role -- they must be followed by another something
            //for the role they signal.   "a rabbit".  "A" signals something coming.  
            var all_relons = await MEMG.getRuntimeRelonsForContext(signal.nObj.nid, [TYPE_BRAIN, TYPE_BE]);
            var fixes = 0;

            for (var r=0; r < all_relons.length; r++) {
                var rel = all_relons[r];

                //resolve relns are typically TYPE_BRAIN and not TYPE_BE so they
                //didn't set the resolve flag in the type be check.
                if (rel.r==MEM.ec_resolve.temp_nid) {
                    if (ANNE.MEM.isRoleR(rel.t)==true) {
                        signal.psbl_roles.push(rel.t);
                    }
                }



                
                //things we know resolve to thins are psbl_roles (not expected ones to come)
                //NOTE: try to change any functions in LC/NLP to neural pools.
                //much better that way than flags/code.
                //however, KNUTH.. gotta get it solid first..
                //then we can walk it back.
                if (rel.type==TYPE_BRAIN) {
                    //ec_possessive
                    //if (rel.r==MEM.brain_assoc_strong.temp_nid && rel.t==MEM.) {
                    //}
                    //else if (rel.r==MEM.brain_assoc.temp_nid) {
                    //}
                    if (rel.r==MEM.ec_resolves_to_role.temp_nid) {
                        if (ANNE.MEM.isRoleR(rel.t)==true) {

                            //on (tuesday)
                            if (rel.t==MEM.role_when_mod.temp_nid) {
                                signal.mod = true;
                                signal.signal = true;
                                signal.resolve = false;
                                if (signal.expects_roles.indexOf(MEM.role_when.temp_nid)==-1) {
                                    signal.expects_roles.push(MEM.role_when.temp_nid);
                                }
                                //if (signal.psbl_roles.indexOf(MEM.role_when_mod.temp_nid)==-1) {
                                    //signal.psbl_roles.push(MEM.role_when_mod.temp_nid);
                                //}
                            }

                            //to (the mall)     
                            //[WHERE] -> to the mall
                            //[WHERE].mod->to
                            else if (rel.t==MEM.role_where_mod.temp_nid) {
                                signal.mod = true;
                                signal.signal = true;
                                signal.resolve = false;
                                if (signal.expects_roles.indexOf(MEM.role_where.temp_nid)==-1) {
                                    signal.expects_roles.push(MEM.role_where.temp_nid);
                                }
                                if (signal.psbl_roles.indexOf(MEM.role_where.temp_nid)==-1) {
                                    signal.psbl_roles.push(MEM.role_where.temp_nid);
                                }
                            }

                            else if (rel.t==MEM.role_aux_action.temp_nid) {
                                signal.aux = true;
                                signal.mod = true;
                                signal.signal = true;
                                //signal.resolve = true;

                                //we don't want this phrase to end for mods.
                                //we want it to parse like a prep phrase (where, when etc)
                                //where the outter phrase is the action and the inner phrase
                                //is the mod signal.
                                //**IF** the action doesn't come then the signal IS the action
                                //and not a mod!!!
                                //if (signal.expects_roles.indexOf(MEM.role_action.temp_nid)==-1) {
                                    //signal.expects_roles.push(MEM.role_action.temp_nid);
                                //} 
                                if (signal.psbl_roles.indexOf(MEM.role_aux_action.temp_nid)==-1) {
                                    signal.psbl_roles.push(MEM.role_aux_action.temp_nid);
                                }
                            }
                            else if (signal.psbl_roles.indexOf(rel.t)==-1) {
                                signal.psbl_roles.push(rel.t);
                            }
                        }
                    }
                }


                //extract some helpful relons.
                if (signal.resolve_thing!=undefined && signal.resolve_thing==true) {
                    if (signal.psbl_roles.indexOf(MEM.role_thing.temp_nid)==-1) {
                        signal.psbl_roles.push(MEM.role_thing.temp_nid);
                    }
                    if (rel.r==MEM.gender.temp_nid) {
                        var toNOBJ = MEMG.getNeuronByTempNid(rel.t);
                        signal.gender = toNOBJ.text;
                    }

                    else if (rel.r==MEM.ec_resolves_to_concept.temp_nid) {
                        var toNOBJ = MEMG.getNeuronByTempNid(rel.t);
                        //if this is speaker, we look into the context pool of the active speaker
                        //when it is time to resolve it!
                        signal.resolve_to_concept = rel.t;
                    }

                    //be-> gender->
                    //brain->this.ec_resolvableRole->
                    //brain->this.ec_resolves_to_concept->
                }
                
                if (rel.r!=MEM.ec_expect_role.temp_nid) continue;

                //get expected role.
                var roleNOBJ = MEMG.getNeuronByTempNid(rel.t);
                if (roleNOBJ==null) continue;
                if (signal.expects_roles.indexOf(roleNOBJ.temp_nid)==-1) {
                    if (signal.expects_roles.indexOf(roleNOBJ.temp_nid)==-1) {
                        signal.expects_roles.push(roleNOBJ.temp_nid);
                    }
                    expects_roles+=1;
                }
    
            }
        }
        
        console.log("%c(initPhraseSignalsMap) mapped sigs "+allSigs.length+" rez:"+rezcnt+" indef:"+indef+" expects_roles:"+expects_roles+" plurals:"+plurals+" singulars:"+singulars+" negations:"+negations+" modes:"+mcnt+" xaction:"+x_action+" xwhen:"+x_when+" xwhere:"+x_where+" xthing:"+x_thing_role+" xactor:"+x_actor_role+" be_dir:"+be_direction+" be_man "+be_manner, 'background: yellow; color: black');
    }

    
    _getPrefix = function($str) {
        return this._prefixMap.getByText($str);
    }
    _getSuffix = function($str) { 
        return this._suffixMap.getByText($str);
    }


    /**
     * Breaker maps are set for the loaded lang and each loaded lang
     * will get its own LC instance (if multiples are in play)
     * .mode, .expects 
     */
     _getPhSignals = function($str) { return this._phSignalsMap.getByText($str);}
     _getPhBreakingSignals = function($str) { 
        var sigs= this._phSignalsMap.getByText($str);
        var noncjsig = [];
        if (sigs==null) return null;
        for (var s=0; s < sigs.length; s++) {
            //if (sigs[s].cj!=undefined && sigs[s].cj==true) continue;
            //if (sigs[s].x_thing_mod!=undefined && sigs[s].x_thing_mod==true) continue;
            noncjsig.push(sigs[s]);
        }
        return noncjsig;
    }

    //may be deprecated
    _isSignalConcept = function($nObj) {
        var sigs = this._getPhSignals($nObj.text.toLowerCase());
        if (sigs==null) return false;
        for (var s=0; s < sigs.length; s++) {
            if (sigs[s].nObj.temp_nid==$nObj.temp_nid) return true;
        }
        return false;
    }


    /**
     * Break up the stream into things we will easily digest.
     * WE DO NOT CHEAT AND SCAN THIS STREAM.
     * It streams through as if it were being typed bit by bit.
     * 
     * @param {anne_nlp_parse} $nlpP 
     * @param {*} $text 
     * @returns 
     */
     firstPassPrepPunct=async($nlpP, $text)=> {
        var wds = [];
        var real_wd_cnt=0;
        var prev_last_real_wd = null;
        var wd_obj = nlp_wd.addWdBlank();

        for (var t=0; t<$text.length; t++) {
            var char = $text[t];
            var charCode = char.charCodeAt(0);

            if (char==".") {
                console.log("should be period charObj.");
            }

            //holds key flags set for chars and this lang use of chars
            var charObj = this.NLP._getCharByCode(charCode);

            //[",", ";", ":", "-"]
            if (charObj.breaker==true) {
                if (wd_obj.str.length>0) {
                    wd_obj.setWd($nlpP, this, wds, charObj);  //we want the charObj breaker in the wd.
                    //real_wd_cnt+=1;
                    prev_last_real_wd = wd_obj;
                }
                nlp_wd.addBreaker(this, wds, char, charObj);    //we also want the charObj in the char breaker itself
                wd_obj = nlp_wd.addWdBlank();
                continue;
            }

            //[".", "?", "!"]
            else if (charObj.ender==true) {
                
                if (wd_obj.str.length) {
                    
                    //sets the existing wd (now that we hit an ender)
                    wd_obj.setWd($nlpP, this, wds, charObj);   //we want the charObj ender in the wd.
                    //real_wd_cnt+=1;
                    prev_last_real_wd = wd_obj;
                }

                //adds an ender token
                var wdObj = nlp_wd.addEnder(this, wds, char);
                if (charObj.question==true) wdObj.question = true;

                wd_obj = nlp_wd.addWdBlank();
                if (char==".") {
                    this.LANG.handlePeriodFlags(wd_obj);
                }
                continue;
            }

            //[" "]
            else if (charObj.spacer==true) {
                if (wd_obj.str.length>0) {
                    wd_obj.setWd($nlpP, this, wds, charObj);  //we want the charObj breaker in the wd.
                    //real_wd_cnt+=1;
                    prev_last_real_wd = wd_obj;
                }
                nlp_wd.addBreaker(this, wds, char, charObj);    //we also want the charObj in the char breaker itself
                wd_obj = nlp_wd.addWdBlank();
                continue;

                /*
                if (wd_obj.str.length>0) {

                    //we do not want the charObj_ender set to a spacer
                    //our neural pats looking for specifc string combos do not specify 
                    //a space between neurons. It is implicit.  Other ascii chars ARE specified
                    //and are included when observed and apart of the potential pattern.
                    wd_obj.setWd($nlpP, this, wds);    
                    real_wd_cnt+=1;
                    prev_last_real_wd = wd_obj;
                }
                wd_obj = nlp_wd.addWdBlank();
                continue;
                */
            }

            //non breaker, non ender, non spacer (for this lang);
            else {
                wd_obj.str = wd_obj.str+char;
                //just move over the true char flags to the wd..

                //if (this._alphaCharCodes.indexOf(charCode)!=-1) wd_obj.has_alpha=true;
                //else if (this._numCharCodes.indexOf(charCode)!=-1) wd_obj.has_num=true;
                //else if (this._innerBreakers.indexOf(charCode)!=-1) wd_obj.has_innerbreaker=true;
                if (charObj.alpha==true) wd_obj.has_alpha=true;
                else if (charObj.num==true) wd_obj.has_num=true;
                else if (charObj.inner_breaker==true) wd_obj.has_innerbreaker=true;
                else if (char=="'") wd_obj.has_apostrophe = true;
                else wd_obj.has_nonalpha_num=true;

                if (wd_obj.has_alpha==true && wd_obj.has_num==true) wd_obj.has_mixed=true;
            }
            
           
            //if last char is not a punct mark of some kind..
            if (t==$text.length-1 && wd_obj.str.length>0) {   
                wd_obj.setWd($nlpP, this, wds);
                real_wd_cnt+=1;
                prev_last_real_wd = wd_obj;
            }
        }

        //if we had any wds with an apostrophe, let's do some lang-specific cleanup.
        //this can be vastly different depending on lang.
        //see: https://www.quora.com/What-uses-do-apostrophes-have-in-languages-other-than-English
        return {wds:wds, wd_cnt:real_wd_cnt};
    }


    //todo.. if  is_int with period then is_int, this is a number.

    /**
     */
    abbrFinder=function($wdsObj) {
        var abbr_hunt = false;
        var start_w = -1; 
        var end_w = -1;
        var new_str = "";

        var wds = $wdsObj.wds;
        for (var w=0; w < wds.length; w++) {
            var wdObj = wds[w];
            if (abbr_hunt==false && wdObj.psbl_abbreviation==true) {
                abbr_hunt = true;
                start_w = w;
                new_str = wdObj.str;
                continue;
            }
            else if (abbr_hunt==true) {
                if((wdObj.type=="punct-end" && wdObj.str==".") || wdObj.psbl_abbreviation==true) {
                    end_w = w;
                    new_str += wdObj.str;
                }
                else {
                    if ((end_w-start_w)>1 && new_str.length>1) {
                        console.log("%c abbrevation detected %c"+new_str, TALK.green, TALK.bggreen);
                        $wdsObj = this.__WdSplicer($wdsObj, new_str, start_w, end_w);
                    }
                    abbr_hunt=false;
                    new_str="";
                }
            }
        }

        //we ended this string with the abbreviation still ongoing..
        //ie,  U.S.A.
        if (new_str.length>1 && (end_w-start_w)>1) {
            console.log("%c abbrevation detected %c"+new_str, TALK.green, TALK.bggreen);
            $wdsObj = this.__WdSplicer($wdsObj, new_str, start_w, end_w);
        }

        return $wdsObj;
    }
    

    /**
     * Splice out a wd obj and mush it into another
     */
    __WdSplicer($wdsObj, $newstr, $start, $finish) {
        if ($start>$finish) return $wdsObj;
        if ($finish>$wdsObj.wds.length-1) return $wdsObj;
        var wds = $wdsObj.wds;
        var total = $finish-$start;
        var del = wds.splice($start+1, total);
        console.log("audit del",del);
        wds[$start].str = $newstr;
        for (var d=0; d < del.length; d++) {
            var wdObj = del[d];
            if (wdObj.type=="wd") {
                $wdsObj.wd_cnt-=1;
            }
        }
        return $wdsObj;
    }


    /**
     * 
     * @param {*} $wdsObj 
     * @returns 
     */
    secondPassPrepPunct = async($wdsObj)=> {
        var metaPunc = {};
        var wds = $wdsObj.wds;
        var lastIsWd = false;

        $wdsObj = this.abbrFinder($wdsObj);

        for (var w=0; w < wds.length; w++) {
            var wdObj = wds[w];
            
            if (w==wds.length-1 && wdObj.has_ender==true) {
                var lastwdObj = null;
                if (w>0 && wds[w-1].charObj_ender!=null) {//should have the ender symbol in this wd
                    console.log("it was the sentence ender");
                    
                    //last in the com signal (ie, a period for sentence)
                    //in streamer mode, this is a cheat, because you don't actually know
                    //yet when the end of the statement is.. 
                    wds[w-1].end_of_coms = true;


                    //last in the com signal (ie, a period for sentence)
                    //in streamer mode, this is a cheat, because you don't actually know
                    //yet when the end of the statement is.. 
                }
            }
            
            if (wdObj.type=="wd") {
                lastIsWd=true;
                
                //likely either/or list type usage.  "dogs/cats are good pets."
                if (wdObj.has_innerbreaker==true) {
                    //if (wdObj.has_num==false) {
                        for (var ib=0; ib < this._innerBreakerAltStrs.length; ib++) {
                            if (wdObj.str.indexOf(this._innerBreakerAltStrs[ib])!=-1) {
                                wdObj.alts = wdObj.str.split(this._innerBreakerAltStrs[ib]);
                            }      
                        }
                        for (var ib=0; ib < this._innerBreakerPartStrs.length; ib++) {
                            if (wdObj.str.indexOf(this._innerBreakerPartStrs[ib])!=-1) {
                                wdObj.parts = wdObj.str.split(this._innerBreakerPartStrs[ib]);
                            }      
                        }
                    //}
                }
                
                continue;
            }


            //if we had a splice from context we may have dupe spaces now.
            if (wdObj.str==" " && w>0 && wds[w-1].str==" ") {
                wds = wds.splice(w, 1);
                w--;
                continue;
            }
            
            //bob told me, "what are you doing".
            //austin 'texas' is great.
            //we have to be careful with this!!!!  bob's and susan's dog was mean.
            else if (this._inlineLRContext.indexOf(wdObj.charCode)!=-1) {
                if (metaPunc[wdObj.charObj.charCode]==undefined) {
                    metaPunc[wdObj.charObj.charCode]=wdObj;
                    wdObj.begin_context=wdObj.index;
                }
                else {
                    var wdo = metaPunc[wdObj.charObj.charCode];
                    wdo.end_context=wdObj.index;
                    delete metaPunc[wdObj.charObj.charCode];
                }
            }

            //we've hit an inline left context marker
            else if (this._inlineLeftContext.indexOf(wdObj.charCode)!=-1) {

                var prop = 'left_'+wdObj.charCode;
                if (metaPunc[prop]==undefined) {
                    metaPunc[prop]=[ {left:wdObj, cnt:1, l_charCode:wdObj.charCode} ];
                }
                else {
                    var arr = metaPunc[prop];

                    //push this to the front..  (ie,  nested parenthesis found (like this) )
                    arr.unshift({left:wdObj, cnt:arr.length+1});
                }

                //wdObj.begin_context=wdObj.index;
            }


            else if(this._inlineRightContext.indexOf(wdObj.charCode)!=-1) {
                
                //we have to get the left for this right code
                var prop = 'left_'+wdObj.charObj.l_charObj.charCode;
                var arr = metaPunc[prop];
                if (arr==undefined) {
                    console.log("orphaned right contextEnder!!!");
                }
                else {
                    var wdo = arr.pop();    //it has to be the last on the array.
                    wdo.r_charCode=wdObj.charCode;
                    
                    //wdo.left.end_context = wdObj.index;
                    wdo.left.end_context_depth = wdo.cnt;
                    //wdObj.start_context = wdo.left.index;
                    wdObj.start_context_depth = wdo.cnt;

                    console.log("wds before "+wds.length);

                    //the extracted slice should NOT include the front char.
                    //we need the open context to contain the  "context" property of the nested context.
                    //if a sentence begins with context then the context marker will still be present.
                    var wds_slice = wds.splice(wdo.left.index+1, (wdObj.index-wdo.left.index));
                    console.log(wds_slice);
                    console.log(wds);
                    console.log("wds after "+wds.length);
                    
                    var wdObjleft = wdo.left;

                    //pop the end char off the wds_slice.  With later advanced multi chars sigs we will treat those
                    //custom symbol combos as a single char.   ie:  Joe *** the president *** is here.
                    wds_slice.pop(); 
                    wdObjleft.context = wds_slice;
                    wdObjleft.l_charCode=wdo.l_charCode;
                    wdObjleft.r_charCode=wdo.r_charCode;
                }
            }

            //int + "." + int = number w/ decimal.
            //taking 3 wdObjs and making 1...
            else if (wdObj.is_int==true) {
                //look forward to next
                if (w < wds.length-2) {
                    var wdnext = wds[w+1];
                    var wdnextnext = wds[w+2];
                    if (wdnext.str=="." && wdnextnext.is_num==true) {
                        wdObj.is_num=true;
                        wdObj.is_int=false;
                        wdObj.charObj=null;
                        wdObj.str+=wdnext.str+wdnextnext.str;
                        wds.slice(w,2);
                        w=w-2;
                        wdObj.type="wd";
                    }
                }
            }


            lastIsWd = false;
        }


        //ptr should already be retained and this is overkill.
        $wdsObj.wds = wds;
        
        return $wdsObj;
    }


    /**
     * This works on unknowns for potential new words with known suffixes.
     * It is only a potential based on lang pats.
     * @param {*} $suffixObj 
     * @param {*} $targetNOBJTnid 
     * @returns 
     */
    _suffixIs=function($suffixObj, $targetNOBJTnid) {
        var rule = $suffixObj.rule;
        if (rule==undefined || rule==null) return false;

        var rel_action = MEMG.getRoleRelon(rule.outs,  MEM.role_action.temp_nid);
        if (rel_action==null) return false;

        var role_goal = MEMG.getRoleRelon(rule.outs,  MEM.role_goal.temp_nid);
        if (role_goal==null) return false;

        if($targetNOBJTnid==role_goal.t) return true;
        return false;
    }

    /**
     * validate a suffix obj's rule.
     * The rule should have an action role with action of CONVERT
     * The source should be in the parent line of the match.
     * The goal should be the targetNOBJTnid.
     */
     _validateEndingRule=function($suffixObj, $matchNOBJ, $targetNOBJTnid) {
        var rule = $suffixObj.rule;
        if (rule==undefined || rule==null) return false;

        var rel_action = MEMG.getRoleRelon(rule.outs,  MEM.role_action.temp_nid);
        if (rel_action==null) return false;
        //NOTE: rel action reln here should be the verb "convert".

        var rel_source = MEMG.getRoleRelon(rule.outs,  MEM.role_source.temp_nid);
        if (rel_source==null) return false;

        //match should be of the source we are converting from.
        if ($matchNOBJ.anyParentHasBeOf(rel_source.t)==false) return false;
        
        var role_goal = MEMG.getRoleRelon(rule.outs,  MEM.role_goal.temp_nid);
        if (role_goal==null) return false;

        if($targetNOBJTnid==role_goal.t) return true;
        return false;
    }


    /**
     * 
     * @param {*} $wdsObj 
     */
    signalPass = async($wdsObj)=> {
        for (var w=0; w < $wdsObj.wds.length; w++) {
            var wd = $wdsObj.wds[w];
            if (wd.type=="punct-break") continue;
            else if (wd.type=="punct-end") continue;
            var text = wd.str;
            var signal_arr = this._getPhSignals(text);
            if (signal_arr!=null) {
                wd.signal_arr = signal_arr;
            }
        }
    }




    /**
     * flat parse of parts of a clause.
     * No knowledge of context or grammar outside signal processing.
     * @param {*} $nlpP 
     */
    flatParse = async($nlpP)=> {   //x_aution is like aux.  
        var sigprops = ["resolve", "signal", "mod", "mode", "expects", "is", "posn",
                        "aux", "x_where", "x_when", "x_action", "connector",  "x_thing_mod"];

        var fph = $nlpP.newPhPartFlat("parse start");

        //will manage the last open mini nn pool for flat parse signals to help us bucket phrases
        //more accurate via expectations set by the signals.
        var _last_fppp = null;

        //this signal needs a pool for our expected next values we may encounter.
        //expect_pool
        //this pool here is ONLY created via signal words.
        function buildFlatPool($sig_fph, $wd) {
            var fppp = new flat_ph_parse_pool ($nlpP);
            fppp.initWithFlatPhExpectPool($sig_fph);
            fppp.debug = "start by sig fph "+$wd.str;
            _last_fppp = fppp;
        }

        var wds = $nlpP.wdsObj.wds;

        /**  INTERNAL *****
         * we have an open _last_fppp so fire the wd.
         * the mini pool will work things out for us..
         *
         * Firing a word is akin to a neural-net heuristic layer.  It is a different type
         * of network of neurons mostly kept to the TYPE_BRAIN dimension.
         * This leads to expectations, and later words lead to confirming those expectations.
         * Expecations of that variety touch on conceptual and semantic sides but the intent of 
         * of these circuits is to help process information at a quick-hit structural level.
         * 
         * We see "on" and we expect where/when.  If next we see "March" we've now confirmed likely when.
         * We have further expectations or potentials possible for known date patterns ---  in this case
         * quite importantly we can have a comma that is in the same phrase..  March 20, 2022.  This
         * is a known date pattern so we can process the flat phrases properly to hand this off to the 
         * next cycle of the rising spiral of awareness.
         * 
         * @param {*} $activeFlatPh 
         * @param {nlp_wd} $nextWd 
         * @returns 
         */
        var fireWd = async ($activeFlatPh, $nextWd, $bDOALT=false)=>{

            //if we have a flat ph parse pool, add this wd neuron to the pool.
            if (_last_fppp!=null) {

                //we need to return the PATTERNS
                //OR GET THEM, for this leading expectations.
                //they are key to helping us at this flat ph level to doing a proper pattern
                //type match without hardcodings.
                var status = await _last_fppp.addWdNeurons($activeFlatPh, $nextWd);
                $nextWd.fppp_status = status;

                if (status==FPPP_WD_NOT_FOUND || status==FPPP_NO_CANDIDATE) {
                    _last_fppp=null;
                    return "close-then-add-wd";
                }

                //we want to continue to add wds to this phrase.
                else if (status==AXNPP_STILL_ALIVE) {
                    return "continue";
                }

                //we want to close this phrase
                else if (status==AXNPP_MATCH) {
                    _last_fppp.printPool();
                     $activeFlatPh.setLikelyConcept(_last_fppp._leading_candidate.nObj);
                    _last_fppp=null;
                    return "include-then-close";
                }

                //we want to close this phrase
                else if (status==AXNPP_NO_MATCHES) {
                    _last_fppp=null;
                    return "close-then-add-wd";
                }
            }

            //try the naked non-expectations method
            else if ($bDOALT==true) {
                //console.log("alt word testing");//getInProgressString

                //we don't want to extend a thing phrase into the conveyor.
                //however, we want a new multi-wd we've never seen here to flow together.
                if ($activeFlatPh.wds.length>0 
                        && $nlpP.conveyor_part==null 
                        && $nextWd.psbl_manner==false 
                        && $nextWd.psbl_action==false) {
                    if ($nextWd.assessed==false) await $nextWd.assessNonSignalWd($nlpP);
                }

                //should we be testing "has_ender" here (or skip it?)
                //the pesky period is not an ender, when used in abbreviations.. so...
                //TODO: come back to this.
                
                //var highest_fire_concept = ANNE.BRAIN.pd_pool_core.testPh($nlpP, $nextWd, $activeFlatPh);
                var str = $activeFlatPh.inProgStr;
                if (str==null || str=="") str = $nextWd.str;
                else str = str + $nextWd.str;
                
                var longer_in_context = ANNE.BRAIN.getLongerByShorterString(str);
                
                if (longer_in_context.length>0) {
                    return "continue";
                }
                else {
                    if ($activeFlatPh.wd_cnt>0) {
                        return "close-then-add-wd";
                    }
                }
            }
            

            //lang specific here, but we want to find potential verbs to break up phrases further.
            //example " on March 18, 2020 Donald Trump jumped the shark."
            //Even though we are not yet in full neural conceptualizing mode, we can look up words
            //and check things out as they are coming.
            return "";//no signals in play, but continue
        }

        
        function copySigFlagsIntoFlatPh($sigs, $flatPh) {
            for (var s=0; s < $sigs.length; s++) {
                var sig = $sigs[s];
                sig.copyFlags($flatPh); //expects_roles
            }
            $flatPh.signal_arr = $sigs;
        }

        function closePhrasePartFlat($ppf) {
            $ppf.closePhrasePartFlat();
            ANNE.BRAIN.pd_pool_core.resetPoolFire();
        }

        //console.log(wds);
        var nester_rollback = [];

        //MAIN LOOP (primary look will process $nlpP.wdsObj.wds);
        for (var w=0; w < wds.length; w++) {
            var wd = wds[w];
            
            //punct chars (non alpha)
            if (wd.type=="punct-break" || wd.type=="punct-end") {
                if (wd.question==true) { fph.question=true; }

                //if we have a start_context -- this is a known nest.
                if (wd.context!=undefined && wd.context!=null && wd.context.length>0) {
                    console.log("end contact found, this is a context starter", wd);
                    console.log(wd.context);
                    nester_rollback.push({w:w, wds:wds, fph:fph}); //when we rollback we want to start AFTER this contet nest.
                    wds = wd.context.concat();  //set array to the context wds.
                    wd.context=null;    //remove context marker from this wdObj.  We don't want to accidently process it again.
                    w=-1;                //set loop to 0 of the context.
                    fph = $nlpP.newPhPartFlat("context-of-prev", true);   //ready for next thing in the list.
                    fph.isContextPart=true;
                    continue;
                }

                
                fph.addWdPunct(wd);
            
                
                //if the punct is part of the pattern we are going to want to keep it in the wds array!
                var cmd ="continue";

                //don't test spaces on this for patterns.
                //they can easily be busted by no space then.
                //_last_fppp!=null && 
                if (fph.wd_cnt>0 && wd.str!=" ") {
                    cmd = await fireWd(fph, wd, true);
                }

                //no active pat, we must break the flat phrase at the comma, :, ; here, for lists.
                //for locations or phrases with a comma native in thm, we don't want to break it.
                //Salem, Oregon USA.   or   12/12/2022   or December 12, 2020
                else {
                    if (wd.charObj!=undefined && wd.charObj!=null && wd.charObj.connector==true) {
                        closePhrasePartFlat(fph);
                        fph = $nlpP.newPhPartFlat("connector-punct", true);   //ready for next thing in the list.
                        fph.connector=true; //comma list presumed if no pats foudn with commas were used (_last_fppp)
                        fph.psbl_list=true;
                        fph.addWdFromPunctOnly(wd);
                        closePhrasePartFlat(fph);//close the char connector immediately
                        fph = $nlpP.newPhPartFlat("after punct connector");   //ready for next thing in the list.
                        continue;
                    }
                }
                
                if (wd.charObj!=undefined && wd.charObj!=null && wd.charObj.connector==true) {   //in eng "," or ";", maybe a long dash too?
                    if (cmd=="include-then-close") {
                        fph.addWdFromPunctOnly(wd);
                        closePhrasePartFlat(fph);
                        fph = $nlpP.newPhPartFlat("after punct", false);   //ready for next thing in the list.
                    }
                    else if (cmd=="close") {
                        closePhrasePartFlat(fph);
                        fph = $nlpP.newPhPartFlat("connector-punct", true);   //ready for next thing in the list.
                        fph.connector=true;
                        fph.addWdFromPunctOnly(wd);
                        closePhrasePartFlat(fph);
                        fph = $nlpP.newPhPartFlat("after punct");   //ready for next thing in the list. (don't know if it is a sig or not yet)
                    }
                    else if (cmd=="continue" || cmd==""){
                        fph.addWdFromPunctOnly(wd); //punct apart of a special pattern we have identified.
                    }
                }
                continue;
            }
            //else TALK.logFunc("parseClause", wd.str);


            //this is just like a signal.
            if (wd.possessive==true) {  //this is a punctuation possessive ('s)
                if (fph.wd_cnt==0) {
                    fph.addWd(wd);
                    fph.isSigPart=true;
                    closePhrasePartFlat(fph);
                    fph = $nlpP.newPhPartFlat("possessive-preceeded-this", true);
                    continue;
                }
                else {
                    closePhrasePartFlat(fph);   //whatever was before the possessive
                    fph = $nlpP.newPhPartFlat("possessive", true);
                    fph.isSigPart=true;
                    fph.addWd(wd);           //
                    closePhrasePartFlat(fph);
                    fph = $nlpP.newPhPartFlat("possessive-preceeded-this", true);
                    continue;
                }
            }


            /**
             * SIGNAL ARRAY PRESENT ON THIS WORD
             * MAY OR MAY NOT BE A SIGNAL BUT IF IT CAN BE WE TREAT IT LIKE IT CAN BE.
             * WE DO NOT KNOW UNTIL MORE DATA COMES IN
             */
            if (wd.signal_arr!=null) {
                var bSIGBREAKER=false;
                
                //we want the union of all the possible flags this signal plays.
                //we don't know which it is when we hear it..
                //and we don't wnat to rule out bad grammar or various levels of speak..
                //nlp should be able to parse child-speak, broken english, improper english.. etc.
                var propstrs = [];

                var lang_expectations = false;
                
                for (var s=0; s < wd.signal_arr.length; s++) {
                    var sig = wd.signal_arr[s];

                    if (sig.connector==true) bSIGBREAKER=true;
                    else if (sig.mode==true) bSIGBREAKER=true;
                    else if (sig.mod==true) bSIGBREAKER=true;
                    else if (sig.x_where==true) {
                        lang_expectations = true;
                        bSIGBREAKER=true;
                    }
                    else if (sig.x_when==true) {
                        bSIGBREAKER=true;
                        lang_expectations = true;
                    }
                    else if (sig.x_action==true) {
                        bSIGBREAKER=true;
                        lang_expectations = true;
                    }
                    else if (sig.x_thing_mod==true) {
                        bSIGBREAKER=true;
                        lang_expectations = true;
                    }
                    else if (sig.posn==true) bSIGBREAKER=true;
                    else if (sig.is==true) bSIGBREAKER=true;
                    else if (sig.resolve==true) bSIGBREAKER=true;
                    else if (sig.aux==true) bSIGBREAKER=true;
                    else if (sig.signal==true) bSIGBREAKER=true;    //is a signal-only a breaker? hmmm.
                    
                    for (var p=0; p < sigprops.length; p++) {
                        var prop = sigprops[p];
                        if (sig[prop]!=undefined && sig[prop]==true) {
                            if (propstrs.indexOf(prop)==-1) propstrs.push(prop);
                        }
                    }
                }

                //"x_where", "x_when", "x_action", "x_thing_mod"
                if (lang_expectations==true) {
                    //we should be extra keen to find these things
                    //as we have been hit a signal saying they are in play.

                    if (sig.expect_pool!=undefined && sig.expect_pool.length>0) {
                        console.log(sig.expect_pool);
                    }
                }
                
                //connectors take precedence
                if (bSIGBREAKER) { 
                    if (fph.wd_cnt>0) {
                        closePhrasePartFlat(fph);
                        var sig_fph = $nlpP.newPhPartFlat("breaker-signal", true);
                        sig_fph.addWd(wd);
                        sig_fph.isSigPart=true;
                        copySigFlagsIntoFlatPh(wd.signal_arr, sig_fph);
                                               
                        if (lang_expectations==true && ANNE.UTILS.isset(sig.expect_pool)==true) {
                            buildFlatPool(sig_fph, wd);
                        }
                        
                        closePhrasePartFlat(sig_fph);
                        fph = $nlpP.newPhPartFlat("breaker-signal-preceeded-this", false);
                        continue;
                    }
                        
                    else {
                        fph.addWd(wd);           //signal
                        fph.isSigPart=true;
                        fph.debug+=" breaker-signal";
                        copySigFlagsIntoFlatPh(wd.signal_arr, fph);
                        if (lang_expectations==true && ANNE.UTILS.isset(sig.expect_pool)==true) {
                            buildFlatPool(fph, wd);
                        }
                        
                        closePhrasePartFlat(fph);
                        fph = $nlpP.newPhPartFlat("breaker-signal-preceeded-this", false);
                        continue;
                    }
                }

                //simple signal, add to signals, not to wd.
                else {
                    fph.addWd(wd);
                    fph.isSigPart=true;
                    copySigFlagsIntoFlatPh(wd.signal_arr, fph);
                    closePhrasePartFlat(fph);
                    fph = $nlpP.newPhPartFlat("nonbreaker-signal-preceeded-this", false);
                }
            }


            /**
             * NOT A LANG STRUCTURE SIGNAL
             * NORMAL WORD
             * COULD STILL BE AN ACTION VB OR ADV HERE.  (not a modal)
             */
            else {
                var cmd = await fireWd(fph, wd, true);//on [March] 
                if (cmd=="include-then-close") {
                    fph.addWd(wd);
                    closePhrasePartFlat(fph);
                    fph = $nlpP.newPhPartFlat("found-pat-new-ph", false);
                }

                else if (cmd=="close-then-add-wd") {
                    closePhrasePartFlat(fph);
                    fph = $nlpP.newPhPartFlat("cmd-close", false);
                    
                    if (wd.assessed==false) await wd.assessNonSignalWd($nlpP);
                    if (wd.psbl_action==true || wd.psbl_manner==true) {
                        if (fph.wd_cnt>0) {
                            closePhrasePartFlat(fph);
                            if (wd.psbl_manner==true) {
                                fph = $nlpP.newPhPartFlat("psbl-manner", false); 
                            }
                            else if (wd.psbl_action==true) {
                                fph = $nlpP.newPhPartFlat("psbl-action", false);
                            }
                        }
                        else {
                            if (wd.psbl_manner==true) fph.debug=fph.debug+" psbl-manner"; 
                            else if (wd.psbl_action==true) fph.debug=fph.debug+" psbl-action";
                        }
                        fph.psbl_action=wd.psbl_action;
                        fph.psbl_manner=wd.psbl_manner;
                        fph.addWd(wd);
                        $nlpP.setPsblFlatAction(fph);
                        closePhrasePartFlat(fph);
                        fph = $nlpP.newPhPartFlat("action-closed", false);
                    }
                    else {

                        //populate the test for wd one, so we know to extend or not on wd2.
                        ANNE.BRAIN.pd_pool_core.testPh($nlpP, wd, fph);
                        fph.addWd(wd);
                    }
                }
                else if (cmd=="close" && fph.wd_cnt>0){
                    closePhrasePartFlat(fph);
                    fph = $nlpP.newPhPartFlat("cmd-close", false);
                }
                    
                else if(cmd=="continue") {
                    fph.addWd(wd);
                } 

                //no standard signals or pats in play..
                //lets look for this being an action breaker wd.
                //this lookks for actions/or manners (adverbish) direct hits
                //or via the endings of known or psbl new morphs.
                else if (cmd=="") {
                    if (wd.assessed==false) await wd.assessNonSignalWd($nlpP);
                    if (wd.psbl_action==true || wd.psbl_manner==true) {
                        if (fph.wd_cnt>0) {
                            fph.addWd(wd);
                            closePhrasePartFlat(fph);
                            if (wd.psbl_manner==true) fph = $nlpP.newPhPartFlat("psbl-manner", false); 
                            else if (wd.psbl_action==true) fph = $nlpP.newPhPartFlat("psbl-action", false);
                            continue;
                        }
                        else {
                            fph.addWd(wd);
                            fph.psbl_action = wd.psbl_action;
                            fph.psbl_manner = wd.psbl_manner;
                            closePhrasePartFlat(fph);
                            if (wd.psbl_manner==true) fph = $nlpP.newPhPartFlat("action-closed", false); 
                            else if (wd.psbl_action==true) fph = $nlpP.newPhPartFlat("action-closed", false);
                        }
                    }
                    else fph.addWd(wd);
                    
                }
            }

            
            //we were in a nester
            //step back out and resume processing the outter layer.
            if ((w==wds.length-1) && nester_rollback.length>0) {
                closePhrasePartFlat(fph);   //close the context nest.

                //restore the loop where we were before we went into the contex nest.
                var nesterR = nester_rollback.pop();
                w=nesterR.w;
                wds=nesterR.wds;
                fph=nesterR.fph;
                continue;   //resume processing.
            }
        }

        if (fph.str=="") {
                
        }
        
        //kill the dead signal placeholder (nothing followed)
        closePhrasePartFlat(fph);

        for (var p=0; p < $nlpP.phrase_parts_flat.length; p++) {
            var part = $nlpP.phrase_parts_flat[p];
            if(part.wd_cnt==0) {
                $nlpP.phrase_parts_flat.splice(p,1);
                p--;
            }
        }
        
        this.printFlatPhs($nlpP.phrase_parts_flat);
    }


    /**
     * 
     * @param {*} $flat_phs 
     */
    printFlatPhs=function($flat_phs) {
        //console.log($flat_phs);
        for (var p=0; p < $flat_phs.length; p++) {
            var fph = $flat_phs[p];
            if (fph.bCLOSED==false) fph.closePhrasePartFlat();
            var signal = "";
            if (fph.isSigPart==true) signal="<signal>";
            if (fph.flagsArr.length) {
                console.log("%c["+p+"]%c"+fph.str+" %c"+signal+" %cflags:"+fph.flagsArr.join()+" %c "+fph.debug.trim(), TALK.bgblack, TALK.bggreen, TALK.bgyellow, TALK.bgblue, TALK.bgred);
            }
            else {
                console.log("%c["+p+"]%c"+fph.str+" %c"+signal+" %c "+fph.debug.trim(), TALK.bgblack, TALK.bggreen, TALK.bgyellow, TALK.bgred);
            }
        }
        console.log("________________________________");
    }

    /**
     * piece together flat phrases to make phrases
     * @param {anne_nlp_parse} $nlpP 
     */
    phPartsParse = function($nlpP) {
        var ph_parts = [];
        var open_ph = null;
        var root_ph = null;

        function makeTop($open_ph) {
            ph_parts.push($open_ph);
        }

        //the top wrapper for the big phrase.
        open_ph = new nlp_ph($nlpP, 0);
        makeTop(open_ph);       //0 depth is container

        root_ph = open_ph;

        //TODO: nested entire sentence signal?   "that, who, which.. hmmm."
        var _lastSigPart = null;
        
        for (var p=0; p < $nlpP.phrase_parts_flat.length; p++) {
            var flat_phrase = $nlpP.phrase_parts_flat[p];

            //flat parse determined this was inline-context.
            if (flat_phrase.isContextPart==true) {
                open_ph.makeChildContext($nlpP, flat_phrase);
                continue;
            }
            
            /***************/
            //SIGNALS
            /***************/
            else if (flat_phrase.isSigPart==true) {
                //console.log("isSigPart "+flat_phrase.str);
                //we need this on its own slot but NOT level 0.
                //if we are on LVL 0 then we need to make a new PH at the top
                //we then need to set the parent for the previous no parent //|| flat_phrase.mod==true
                
                if (flat_phrase.question==true) {
                    $nlpP.question=true;
                }
                
                if (flat_phrase.connector==true 
                        || flat_phrase.is==true 
                        || flat_phrase.posn==true
                        || flat_phrase.x_action==true 
                        || flat_phrase.aux==true 
                        || flat_phrase.mode==true) {


                    //

                    //dog jumped into[sig 1] first[sig2] place
                    if (_lastSigPart!=null && (_lastSigPart.flat_ph.x_where!=undefined || _lastSigPart.flat_ph.x_when!=undefined)) {
                        open_ph.makeChildPartOfCurrentPh($nlpP, flat_phrase);  //stay in same phrase lvl
                        continue;
                    }

                    //connectors can be connecting concepts in a ph (the dog and the cat)
                    //they can connect multiple actions  (cats meow and scratch alot)
                    //they can connect multiple sentences (cats meow and dogs bark).
                    if (flat_phrase.connector==true) {

                        if (open_ph.parts.length>0) { 
                            // open_ph = open_ph.makeChildPartOfCurrentPh($nlpP); 
                            _lastSigPart = open_ph.addFlatToPh($nlpP, flat_phrase);
                            _lastSigPart.is_connector = true;
                            //open_ph = open_ph.closePh();     //should be closing the "and" phrase
                            //open_ph = open_ph.makeChildPartOfCurrentPh($nlpP);
                            //open_ph.pre_open=true;
                            //continue;
                        }
                        else {
                            //open_ph = open_ph.closePh();
                            //open_ph = open_ph.makeChildPartOfCurrentPh($nlpP); 
                            _lastSigPart = open_ph.addFlatToPh($nlpP, flat_phrase);
                            _lastSigPart.is_connector = true;
                            //open_ph = open_ph.closePh();   
                        }

                    }
                        
                    //we need this in its own ph and to break up the existing phrase.
                    //note, in english this could be an aux if another vb follow.
                    //we will have to generalize this further for desired langs after we get english working..
                    else if (flat_phrase.is==true) {
                        open_ph = open_ph.closePh();
                        open_ph = root_ph;
                        open_ph = open_ph.makeChildPartOfCurrentPh($nlpP, flat_phrase); 
                        open_ph.is_conveyor = true;
                        if(open_ph.depth==1) {
                            $nlpP.conveyor_part = open_ph;
                            $nlpP.conveyor_part_be = true;
                            open_ph.can_be_aux = true;
                        }
                        open_ph = open_ph.closePh();
                        open_ph = open_ph.makeChildPartOfCurrentPh($nlpP);    //future phrase
                        open_ph.pre_open=true;
                    }

                                        
                    else if (flat_phrase.aux==true) {
                        open_ph = open_ph.closePh();
                        open_ph = root_ph;
                        open_ph = open_ph.makeChildPartOfCurrentPh($nlpP, flat_phrase); 
                        open_ph.is_conveyor = true;
                        if(open_ph.depth==1) {
                            if (flat_phrase.posn==true || flat_phrase.has==true) $nlpP.conveyor_part_has = true;
                            $nlpP.conveyor_part = open_ph;
                            open_ph.can_be_aux = true;    
                        }
                        open_ph = open_ph.closePh();
                        open_ph = open_ph.makeChildPartOfCurrentPh($nlpP);    //future phrase
                        open_ph.pre_open=true;
                    }


                    //in english this preceeds the vb (w/ pslb advs in between)
                    //mode will be in its own phrase.   When we make the vb role (after getting the vb)
                    //we will put together mode+manner and the vb (in the next cycle of the spiral)
                    else if (flat_phrase.mode==true) {
                        open_ph = $nlpP.makeMode($nlpP, open_ph, flat_phrase);
                    }

                        
                    //if it has more than signals..
                    //we want a new ph of its parent.
                    else if (open_ph.parts.length>0) {
                        open_ph = open_ph.makeChildPartOfCurrentPh($nlpP); 
                        _lastSigPart = open_ph.addFlatToPh($nlpP, flat_phrase);  //flat part or sig part??
                        open_ph = open_ph.makeChildPartOfCurrentPh($nlpP);    //future phrase
                    }

                    //no other wds are present yet in this ph.
                    else {
                        console.log("<<add connector>> %c"+flat_phrase.str, TALK.sig_connector);
                        _lastSigPart = open_ph.addFlatToPh($nlpP, flat_phrase); //add the connector to the existing ph
                    }
                    
                    continue;
                }

                //must nest under current open_ph  (include .mod?)
                //continue loop..
                else if (flat_phrase.x_when==true || flat_phrase.x_where==true || flat_phrase.x_thing_mod==true || flat_phrase.mod==true) {
                    //NOTE: here is where we hit PREDICATE / OBJ prepositional ambiguity.
                    //we don't know via signals ONLY if we are to assign the prep phrase to the
                    //VB as a primary role, or to the NP it followed.
                    //Suzy is happy in the night.
                    //Suzy jumps dogs in the night.
                    //Suzy jumps kings of france.
                    //Suzy is a woman of her word.

                    //contrasting clues with a pattern of N is N PREP CONSTRAST N ACTION 
                    //Suzy is a homemaker [in the morning] but she turns into a party animal [at night].
                    
                    //the "wd" is still a signal within that phrase.
                    //signal phrase of new phrase [on] the mountain
                    //[my] dog. (hmmm..   [my dog] butch is the best dog in the entire universe.)

                    if (open_ph.total_parts==0) {
                        if (open_ph.pre_open==false) {

                            //future phrase  "[on the mountain]", "[over the moon]"
                            open_ph = open_ph.makeChildPartOfCurrentPh($nlpP);

                            //the signal as a mod of this new phrase
                            _lastSigPart = open_ph.addFlatToSigPh($nlpP, flat_phrase); 
                        }

                        //we already opened the ph, just add the flat to the signal.
                        else {
                            _lastSigPart = open_ph.addFlatToSigPh($nlpP, flat_phrase);  
                        }
                    }

                    //existing parts.. we could be in a prep of the phrase
                    //or a root clump prep. 
                    //store at root and resolve in context w/ neurons.
                    else {
                        //of texas
                        if (open_ph.parts.length>0 && flat_phrase.mod==true) {
                            //if (flat_phrase.x_thing_mod==true && open_ph.l1str_nosignal_wds.length>=2) {
                               // open_ph = open_ph.closePh();
                               // open_ph = root_ph;
                               // open_ph = open_ph.makeChildPartOfCurrentPh($nlpP);
                               // _lastSigPart = open_ph.addFlatToSigPh($nlpP, flat_phrase); 
                            //} 
                            //else 
                            open_ph = open_ph.makeChildModWithSignalPh($nlpP, flat_phrase); //future "of texas phrase" 
                            continue;
                        }
                            
                        var alt_parents = [open_ph];
                        open_ph = root_ph;
                        open_ph = open_ph.makeChildPartOfCurrentPh($nlpP); 
                        _lastSigPart = open_ph.addFlatToSigPh($nlpP, flat_phrase);//signal wd to this new ph
                        open_ph.alt_parents = alt_parents; 
                    }
                    continue;
                }


                /***************/
                //RESOLVES
                /***************/
                //, ie, a known replacement for a concept.  (he, she etc) 
                //some resolves cannot be solo in the subj in english.
                //they ARE signals... like  "[her] house was big."
                //in the obj they can be solo.  "I loved [her]." 
                else if (flat_phrase.resolve==true) {
                    if (open_ph.pre_open==false) open_ph = open_ph.makeChildPartOfCurrentPh($nlpP, flat_phrase);    
                    else open_ph.addFlatToPh($nlpP, flat_phrase);  //he/she..    
                    
                    if (flat_phrase.resolve_thing==true) {
                        open_ph.resolve_thing=true;
                        open_ph = open_ph.closePh();
                    }
                    continue;
                }
                
                //new basic signal at start of phrase.
                //create a future phrase (one lower)
                if (open_ph.total_parts==0) {
                    if (open_ph.pre_open==false) open_ph = open_ph.makeChildPartOfCurrentPh($nlpP);   //future phrase
                    open_ph.addFlatToSigPh($nlpP, flat_phrase);     //add the signal to the new phrase (it has only the signal)
                }

                else {
                    if (open_ph.pre_open==false) open_ph = open_ph.makeChildPartOfCurrentPh($nlpP);   //future phrase
                    open_ph.addFlatToSigPh($nlpP, flat_phrase);     //add the signal to the new phrase (it has only the signal)
                }
                continue;
            }
            else {
               // console.log(" "+flat_phrase.str);
                _lastSigPart=null;
            }


            if (flat_phrase.str=="gorge" || flat_phrase.str=="landform") {
                console.log("test here.  We want to set these resolved values since they are indef determined classes.")
            }
            
            /***************/
            //NON-SIGNALS
            /***************/
            //a [dog]
            //we do not stop a NON-SIGNAL until we get a SIGNAL.
            //that's how it works.
            //in conceptualizing we may break it up further.
            //TROUBLE CASE:   HE GAVE SUZY POTATOES.  
            //context will have to determine that suzy is seperate from pototoes and something like car tires or data graph.
            //also GAVE is special type of vb where this happens.   We can come to know this.
            if (open_ph.total_parts==0) {
     
                //Bob might have jumped.
                if (_lastSigPart==null || (_lastSigPart && _lastSigPart.is_connector==false)) {
                    if (flat_phrase.psbl_action==true && ($nlpP.conveyor_part==null || $nlpP.conveyor_part.can_be_aux==true)) {
                        open_ph = $nlpP.makeAction(open_ph, flat_phrase);
                        continue;
                    }
                    else if (flat_phrase.psbl_manner==true) {
                        open_ph = $nlpP.makeManner(open_ph, flat_phrase);
                        continue;
                    }

                    //bob or [donald] might have jumped.
                    else {
                        if (open_ph.pre_open==false) {
                            open_ph = open_ph.makeChildPartOfCurrentPh($nlpP);   //future phrase
                            open_ph.addFlatToPh($nlpP, flat_phrase);
                        }   
                        else open_ph.addFlatToPh($nlpP, flat_phrase);
                    }
                }

                //dogs and cats can bite.   (don't want cats here as the psbl_action.  we know it is and'd to dogs.)
                else {
                     if (open_ph.pre_open==false) {
                        open_ph = open_ph.makeChildPartOfCurrentPh($nlpP);   //future phrase
                        open_ph.addFlatToPh($nlpP, flat_phrase); //add thing (bob) to the part.
                    }
                    open_ph.addFlatToPh($nlpP, flat_phrase);
                }
                

                if (flat_phrase.likelyConcept!=null) {
                    open_ph.likelyConcept = flat_phrase.likelyConcept;
                }
            }

            
            else {
                //dog [jumped] into first place.
                //if we have a psbl_action and no conveyor_part yet --- OR the conveyor_part can be aux.. lets take this vb.
                if (open_ph.sig_parts.length==0 
                    && flat_phrase.psbl_action==true 
                    && ($nlpP.conveyor_part==null || $nlpP.conveyor_part.can_be_aux==true)) {
                    open_ph = $nlpP.closePhAndMakeAction(open_ph, flat_phrase);
                    continue;
                }

                //make sure to set psbl_manner from the .tnids.
                else if (flat_phrase.psbl_manner==true) {
                    open_ph = $nlpP.makeManner(open_ph, flat_phrase);
                    continue;
                }
                
                //this is where we could be possibly hitting grammar/lang specific markers.
                //in ENG, if no action/posn/be yet... a second wd here could be that.
                //be and posns are SIGNALS already -- but ec_actions must be searched for.
                //doing that here would make sense but requires we use the morphs.
                //ALSO, we could be in a mode where the VB/ACTION is known before we parse.
                //in that case we would use it as pass in to the 
                else if (open_ph.sig_parts.length > 0 && open_ph.parts.length==0) {
                    if (flat_phrase.likelyConcept!=null) {
                        open_ph.likelyConcept=flat_phrase.likelyConcept;
                        open_ph.addFlatToPh($nlpP, flat_phrase);
                        
                        //should close this .. especially after a date we parsed in the flat phase.
                        if (flat_phrase.wds.length>1 && (open_ph.x_when==true || open_ph.x_where==true)) {
                            open_ph = open_ph.closePh();
                        }
                    }
                    else open_ph.addFlatToPh($nlpP, flat_phrase);
                }

                //dog jumped into first [place].
                else {
                    //prior iteration of spiral already identified this as its own
                    //likely phrase (ie, date, person, place)
                    if (flat_phrase.likelyConcept!=null) {
                        //open_ph = open_ph.closePh();
                        if (flat_phrase.wd_cnt>1) {
                            var part_ph = open_ph.makeChildPartOfCurrentPh($nlpP);    //future phrase
                            part_ph.likelyConcept = flat_phrase.likelyConcept;
                            part_ph.addFlatToPh($nlpP, flat_phrase); 
                        }

                        //likely ph can be triggered by a known first name.. (like john)
                        //but unknown last name may come through in its own flat phrase.
                        //we want to string these together.
                        else {
                            open_ph = open_ph.makeChildPartOfCurrentPh($nlpP);    //future phrase
                            open_ph.likelyConcept = flat_phrase.likelyConcept;
                            open_ph.addFlatToPh($nlpP, flat_phrase); 
                        }
                        
                    }
                    else {
                        if (flat_phrase.psbl_action==true && ($nlpP.conveyor_part==null || $nlpP.conveyor_part.can_be_aux==true)) {
                            open_ph = $nlpP.closePhAndMakeAction(open_ph, flat_phrase);
                        }
                        else open_ph.addFlatToPh($nlpP, flat_phrase);
                    }
                }
            }
        }

        //button this one all the way up
        while (open_ph.bCLOSED==false && open_ph.depth>0) {
            open_ph = open_ph.closePh();
        }
        
        console.log(ph_parts);

        //recursive str setter.
        //we can deprecetate this ASAP. don't want strings gen'd this way
        var populateStrs=function($ph_parts) {
            var joinparts = [];
            
            for (var  p=0; p < $ph_parts.length; p++){
                var part = $ph_parts[p]; 
                part.l1str = part.l1str.trim();
                part.l1str_nosignal = part.l1str_nosignal.trim();
                
                var str2 = "";
                if (part.ordered_parts.length) {
                   for (var pp=0; pp < part.ordered_parts.length; pp++) {
                        var opart = part.ordered_parts[pp];
                        opart.l1str = opart.l1str.trim();
                        opart.l1str_nosignal = opart.l1str_nosignal.trim();
                        if (opart.l1str=="") continue;
                        if (opart.flat_ph==null) {
                            str2 = populateStrs(opart.ordered_parts);
                            if (str2=="" && opart.sig_parts.length==0 && opart.parts.length==0 && opart.mod_ph_parts.length==0) {
                                $ph_parts.splice(pp,1);
                                pp--;
                                continue;
                            }
                            joinparts.push(str2);
                            opart.full_str = str2;
                        }
                        else {
                            str2 = opart.flat_ph.str;
                            if (str2.length) {
                                joinparts.push(str2);
                                opart.full_str = str2;
                            }
                        }
                    }
                    part.full_str = joinparts.join(" ");
                }
                else {
                    if (part==undefined || part.flat_ph==null) {
                       str2 = part.l1str;
                        if (str2.length) {
                            joinparts.push(str2);
                            part.full_str = str2;
                        }
                        continue;
                    }
                    
                    str2 = part.flat_ph.str;
                    if (str2.length) {
                        joinparts.push(str2);
                        part.full_str = str2;
                    }
                }

            }
            
            return joinparts.join(" ");
        }

        $nlpP.ph_parts = ph_parts;

        //recursive str setting.
        var full = populateStrs(ph_parts);
        console.log(full);
        
        //recursive part printer
        var printStrs=function($ph_parts) {
            for (var  p=0; p < $ph_parts.length; p++){
                var part = $ph_parts[p];
                var pre = TALK['tab'+part.depth];
                var lc = "";
                var partcolor = "";
                if (part.total_parts<=1) {
                    if (part.isSigPart==true) {  //signal
                        lc = "[sig part]";
                        partcolor=TALK.part_sig;
                    }
                    else if (part.isModPart==true) {   //aux or standard mod
                        if (part.is_aux==true) {
                            lc = "[aux]"
                            partcolor=TALK.aux;
                        }
                        else {
                            lc = "[context mod part]";
                            partcolor=TALK.part_mod_ph;
                        }
                    }
                    else { //normal part
                        if (part.flat_ph!=null && part.flat_ph.connector==true) {
                            lc = "[connector]";
                            partcolor = TALK.connector;
                        }
                        else if (part.is_conveyor==true) {
                            lc = "[conveyor]";
                            partcolor = TALK.action;
                        }
                        else if (part.is_manner==true) {
                            lc = "[manner]";
                            partcolor = TALK.manner;
                        }
                        else if (part.is_mode==true) {
                            lc = "[mode]";
                            partcolor = TALK.mode;
                        }
                        else partcolor = TALK.part_full;   
                    }

                    if (part.likelyConcept!=null) {
                        lc += "<<"+part.likelyConcept.text+">>";
                    }
                    console.log("%c"+pre+"%c["+part.depth+"]"+"%c "+part.l1str+"%c"+lc, TALK.part_depth, TALK.part_cnt, partcolor, TALK.bgblue);
                }

                //multi part phrase.
                else if (part.total_parts>1) {    
                    if (part.likelyConcept!=null) {
                        lc += "<<"+part.likelyConcept.text+">>";
                    }
                    console.log("%c"+pre+"%c["+part.depth+"]"+"%c [part.FULL_STR] %c "+part.full_str+"%c"+lc, TALK.part_depth, TALK.part_cnt, TALK.bgyellow, TALK.part_full,  TALK.bgblue);
                }
                printStrs(part.sig_parts);
                printStrs(part.parts);
                printStrs(part.mod_ph_parts);
            }
        }

        printStrs(ph_parts);
            
        return;
    }

    

    /**
     * Note: this isn't a clump if this is just a phrase being parsed.
     * @param {anne_nlp_parse} $nlpP 
     */
    phToRoles = async($nlpP)=> {
        var conveyorNOBJ = null;
        var clump = $nlpP.clump;
        var conveyorPart = $nlpP.conveyor_part;
        
        //we have a problem.
        //depending on mode we are in, we may do something about it.
        if (conveyorPart==null) {
            console.log("%chelp, no explicit vb found so far!", TALK.bgbigred);
            return;
        }

        
        //NOTE: will need to know when a question is rhetorical..
        //that comes WAY later after higher spirals of awareness.
        //note for questions.. need to parse aux-thing-conveyor   (do dogs jump?)
        if ($nlpP.question==true) clump.clump_type=CLUMP_TYPE_Q;
        else clump.clump_type=CLUMP_TYPE_S;


        //psbl_new_action_morph
        var wd = conveyorPart.flat_ph.wds_no_punct[0];
        var tnids = wd.tnids;
        var convs =[];
        
        if (tnids.length==0) {
            var MATCHES = await ANNE.MEM.searchTextIndex(wd.str, true, true);
            for (var m=0; m < MATCHES.length; m++) {
                wd.tnids.push(MATCHES[m].temp_nid);
            }
            tnids = wd.tnids;
        }

        for (var n=0; n < tnids.length; n++) {
            var nObj = MEMG.getNeuronByTempNid(tnids[n]);
            if (nObj.anyParentHasBeOf(MEM.ec_is.temp_nid)==false) {
               if (nObj.anyParentHasBeOf(MEM.ec_action.temp_nid)==false) {
                    if (nObj.anyParentHasBeOf(MEM.ec_posn.temp_nid)==false) continue;
                }
            }
            console.log(nObj);
            console.log(nObj.text);
            convs.push(nObj);
        }

        //we have no valid conveyor
        if (convs.length==0) {
            console.log("%cTODO no valid conveyor", TALK.bgbigred);
            return;
        }

        //multiple conveyor options, need to do resolve with a pool to get the right one!
        //this pool should be a contextual based pool of all the phrases.
        //we should have built this along the way.  (THINK THIS THROUGH AGAIN.)
        else if (convs.length>1) {
            console.log("%cTODO conveyor splitter options", TALK.bgbigred);
        }

        //todo, if more than 1 option, resolve which one based on context and present roles.
        var conveyor = convs[0];

        var tense = conveyor.setTense($nlpP);       //sets .tenseNOBJ in the nlpP object
        
        console.log(tense);
        //TODO: did we encounter a tense-changing modal (like will?)

        //AUX + paste tense = passive voice.
        $nlpP.voice = this.LANG.setVoice($nlpP);

        //and be vb is, are, was, were am, be etc is ec_is
        if ( $nlpP.conveyor_part_be==true) {
            clump.clump_sub_type=CLUMP_SUBTYPE_DEFN;
            conveyorNOBJ = MEM.ec_is;
        }

        //and had/have/has will be having, etc, is posn.
        else if ( $nlpP.conveyor_part_has==true) {
            clump.clump_sub_type=CLUMP_SUBTYPE_POSN;
            conveyorNOBJ = MEM.ec_posn;
        }

        //we should use base form only of the conveyor.
        else {
            $nlpP.clump.clump_sub_type=CLUMP_SUBTYPE_ACTION;
            console.log(conveyor);  //test with irregular that has stored v1-v5
            if (conveyor.v1!=null) conveyor = conveyor.v1;
            conveyorNOBJ = conveyor;
        }

        //TODO: test base parts 
        //he ran and jumped in the night.  (?)
        //idomatic conveyor with multi wds?
        if (conveyorPart.parts.length>0) {
            console.log("%cTODO multiple parts in conveyor", TALK.bgbigred);
            return;
        }

        //sets default_roles and default_expected_roles
        await $nlpP.clump.setDefaultRoles(conveyorNOBJ);

        //make the conveyor role.
        var crole = clump.nlpRolesObj.makeConveyor($nlpP, conveyorPart, conveyorNOBJ, $nlpP.mode_part);
        
        //conveyor action-DEFN (defining an action) or an imperative or question
        //jumping at night is dangerous
            //to jump at night is dangerous
        //jump over that hole!
        //is jumping at night dangerous?
        if (conveyorPart.index == 0) {
        }
        
        //imperative
        //question
        //action event
        //quotation

        if (conveyorNOBJ==null) return;

        //should be primary ph (holding phrases)
        //at this stage each one is going to be a role.
        for(var p=0; p < $nlpP.ph_parts.length; p++) {
            var part = $nlpP.ph_parts[p];
            console.log("sentence: %c"+part.full_str, TALK.bggreen);
            //console.log(part);

            //loop level one parts (roles of the sentence at the top level)
            for (var pp=0; pp < part.parts.length; pp++) {

                var ppart = part.parts[pp];
                //console.log(ppart);
                if (ppart.conceptualized==true) {
                    console.log("already conceptualized %c"+ppart.l1str, TALK.bggreen);
                    continue;
                }
            
                if (ppart.id==conveyorPart.id) {
                    console.log("conveyor conceptualized %c"+ppart.l1str, TALK.bggreen);
                    continue;   //already handled
                }

                //he, she, they, we
                if (ppart.flat_ph!=null && ppart.flat_ph.resolve==true) {
                    ppart.resolvable=true;
                }
                
                //how else do we know a phrase is resolvable?
                var sigs = ppart.sig_parts;
                for (var s=0; s < sigs.length; s++) {
                    var sig = sigs[s];
                    if (sig.flat_ph!=null) {
                        var ffp = sig.flat_ph;
                        
                        //"a dog" is still resolvable if the vb is not present tense.
                        //it typically is resolved later tho in that form.
                        //"a dog jumped.   That dog was my dog."
                        if (ffp.indef==true) {
                            ppart.has_indef=true; 

                            //indef "a"/"an" not not present.  (a dog jumped)
                            //the only time this is a class is if this is an embedded referenced clause.
                            if ($nlpP.tenseNOBJ!=null && $nlpP.tenseNOBJ.temp_nid!=MEM.tense_present.temp_nid) {
                                ppart.resolvable=true;  //this is a dog in context being referenced.
                            }
                        }

                        //her car, their house, his dog (test if these come through here)
                        //should be a sig (sig part or mod part) w/ posn
                        else if (ffp.resolve==true) {
                            ppart.resolvable=true;
                        }
                    }
                }    

                //conceptuaalized the phrase into a semantic role.
                var role = this.LANG.conceptualizePh($nlpP, ppart, ppart.likelyConcept);
            }
        }

        console.log($nlpP.clump);
        $nlpP.clump.printRoles();
        return;
    }

    
    /**
     * Roles have been conceptualized.
     * Phrases have been resolved to their basic neuron.
     * Some may need additional resolution - IN CONTEXT.
     * ie, we know the neuron for "she" but we need to resolve who "she" is.
     * We know it was stated "Liberty Statue in America" but we need to resolve this to "The Statue Of Liberty" neuron.
     * We know it says "the red planet" but we want to resolve it to Mars.
     * We know it says "the capital of Texas" but we want to resolve it to Austin.
     * Note, these can only be resolved if we know these things.  This is not the way to LEARN new things.
     * @param {*} $nlpP 
     */
    resolveContext=async($nlpP)=>{
        var clump = $nlpP.clump;
        var roles = clump.nlpRolesObj._roles;

        //sets the subj/pred flag for the roles.
        $nlpP.clump.setRoleFlags($nlpP, clump);
        
        //clump has a unique id already.
        var bGOOD = await ANNE.BRAIN.ctp.processClump(clump);
        if (bGOOD==false) {
             console.log("this is just a phrase, dont store role")   
        }
        //get these in our context pool first.
        for (var r=0; r < roles.length; r++) {
            var role = roles[r];

            //this doesn't need resolution in context..
            //it sill may need resolution of the simple concept.
            //some role types haven't gotten their neuron yet up to this point yet we have identified
            //through signals that they DO NOT need resolution in context.
            if (role.needs_ric==false) {
                //resolved -- let's move this into our single- pool.
                //it can help at the thought level to resolve ambiguitities remaining.
                if (role.nObj==null) {

                    //handle pattern  (ie, date)
                    if (role.ph.bestPat!=null) {
                        console.log(role.ph.bestPat);
                        var patpool = new anne_x_neural_pat_pool();
                        await patpool.rolePatMatchWithWds(role.ph.bestPat, role);
                        continue;
                    }

                    var tnids = this.NLP.BRAIN.getTnidsByAnbl(role.ph.l1str_nosignal);
                    if (tnids.length==0) {
                       if (role.best_class_tnid!=-1) {
                            var classNOBJ= MEMG.getNeuronByTempNid(role.best_class_tnid);
                            console.log("make new generic thing neuron for %c"+nosigtext, TALK.bggreen);
                            var finalNOBJ = await ANNE.BRAIN.ctp.makeLocalNeuron(role.ph.l1str_nosignal, NTYPE_CLASS, classNOBJ);
                        } 
                        //we don't have a value yet.
                        else {
                            var finalNOBJ = await ANNE.BRAIN.ctp.makeLocalNeuron(role.ph.l1str_nosignal, ntype, MEM.ec_thing); 
                            role.updateResolved(finalNOBJ); 
                        } 
                    }
                    else {
                        if (tnids.length>1) {
                            console.log("do we really want to have to resolve this here?");
                            role.rez_pool = new nlp_role_concept_resolver(role, "rez pool for "+role.ph.l1str_nosignal);
                            role.rez_pool.fireLooseCandidates(tnids);  
                            var candidates = role.rez_pool.getHighestWeightOverThreshold(.9, "phrase rez for "+role.ph.l1str_nosignal);
                            role.setWinningCandidate(candidates);
                        }
                        else {
                            var finalNOBJ = MEMG.getNeuronByTempNid(tnids[0]);
                            role.updateResolved(finalNOBJ); 
                        }
                    }
                    
                }
                ANNE.BRAIN.ctp.addRoleToCP(clump, role);
            }
        }
        
        for (var r=0; r < roles.length; r++) {
            var role = roles[r];

            
            //we need to resolve the phrase.
            //this could be new context or rely upon existing context in the context pool (he, she, it, that, etc).
            if (role.needs_ric==true) {

                //txid:2317414450937654254[BRAIN20][Month DD, YYYY][__make_new][Gregorian Date] 
                if (role.ph.bestPat!=null) {
                    
                    console.log(role.ph.bestPat);
                    var patpool = new anne_x_neural_pat_pool();
                    await patpool.rolePatMatchWithWds(role.ph.bestPat, role);
                    continue;

                    //even knowning what this is conceptually (ie, a date, or location, or a likley name) we still
                    //need to be able to resolve it to the neuron of the date, location, or name.., 
                    //and if we can't find it, we need to create a new neuron
                    //$nlpP.resolveRoleWithPat(role, role.ph.bestPat, role.ph.l1str);

                    //var lookupstr = await axnp.lookupOrMake(role);
                    //var axnp = new anne_x_neural_pat(role.ph.bestPat);
                    //console.log(axnp);
                    //axnp.printPatPieces();
                    //var lookupstr = await axnp.lookupOrMake(role);
                    
                    //if (role.resolved_nObj!=null) {
                        //add this resolved role to the context pool.
                        //ANNE.BRAIN.ctp.addRoleToCP(clump, role);
                    //}
                }

                //need to do PNs special as they require previous context.
                //other resolution *can* use previous context if we have dualing options but doesn't require it.
                //ie, bank (turn) vs bank (save money) may be ambiguious in the single sentence..
                //but if we know we were talking about finance and not flying an airplane we can resolve it.
                //note: we can talk about both which is quite tricky.

                //how to apply logic/understanding of direction/placement/quanity.. etc?
                //ANALYZER must be used here to get legit matches.
                //FIRST-PASS MATCHES of this spiral can get PRE-ANALIZER candidates of neural circuit matches.
                //example:   liberty + statue + R_WHERE fires known statue of liberty.   
                //However, sub-sequent- "in Florida" following "liberty statue" throttles the rez of "statue of liberty"
                //QUESTION --> how to best logically apply things like "in" and "on" or "above" or "below"?
                //HOw to enforce logical language  not this but that or this and that?   This is not easy.
                //further it is especially not easy when refusing to write custom code to do that.
                //We need this also in the neurons.. the definition of the concepts.. PLUS logical circuit neurons
                //that allow us to apply/use logic.
                //test case to help think through:    
                //Are dogs and cats [both] mammals?
                //Are dogs and dolphins [both] mammals?

                //SUPER ADVANCED:
                //to actually do this right requires spacial awareness and understanding.. (and loads  of other things)
                //I am sitting at a table.  There are two cats.  The one named Pete is on a table.
                //The cat named Sal is at my feet.
                //The cat under the table is nice but the one that loves to be perched high is mean!
                role.rez_pool = new nlp_role_concept_resolver(role, "rez pool for "+role.ph.l1str);

                if (role.ph.resolve_to_concept!=null || role.ph.resolve_thing==true) {
                    
                    //speaker, such as "i".    
                    await this.NLP.BRAIN.ctp.resolveRoleResolveable(role);
                    ANNE.BRAIN.ctp.addRoleToCP(clump, role);
                    continue;
                }
                
                
                var nosigtext = role.ph.l1str_nosignal.trim();
                var anbl = this.NLP.BRAIN.getOrMakeAnbl(nosigtext);

                //kind of weak, but it works to catch the simple stuff like a name.
                if (anbl.full_tnids.length==1) {
                    var finalNOBJ = MEMG.getNeuronByTempNid(anbl.full_tnids[0]);
                    role.updateResolved(finalNOBJ);
                    ANNE.BRAIN.ctp.addRoleToCP(clump, role);
                    continue;
                }

                
                console.log("resolve this", role);
                console.log("nosigtext: "+nosigtext);
                var wds = nosigtext.split(" ");
                var reqs = [];
                var clues = [];
                for (var w=0; w < wds.length; w++) {
                    if (wds[w].length<3) continue;//dont 2 letter combos.
                    var anb = this.NLP.BRAIN.getOrMakeAnbl(wds[w]);

                    //normal mode --  like [capital] of [texas] .... or [texas] [capital]
                    //here we are trying to find the concept these combine to refer to.
                    //we are essentially finding a union of each item in wds.
                    if (anb.full_tnids.length) {
                        reqs.push(anb.full_tnids);
                        //role.rez_pool.spotlight(anb.partial_tnids);
                        //add-> chase to build wider spotlight of possible hits.
                        //ie, if statue-> put in statues
                        //if capital-> put in captials.
                        if (anb.partial_tnids.length) {
                            UTILS.addUniqueTnids(clues, anb.partial_tnids);
                        }
                    }

                    //creative extended
                    else if (anb.partial_tnids.length) {
                        role.rez_pool.fireLooseCandidates(anb.partial_tnids);        
                        //points to bigger phrases..
                        //we want a partial hit ... but we need MORE than just
                        //one partial hit to fire the neuron.
                        //so..  "liberty" and "statue" could fire the statue of liberty.
                        //or, a famous landmark statue, in NYC ---> should resolve to The Statue Of Liberty.
                    }
                }


                //for inline context.  (have to test this.. intent: disambigutate)
                if (role.ph.context_ph_parts.length) {
                    for (var ct=0; ct < role.ph.context_ph_parts.length; ct++) {
                        var cpart = role.ph.context_ph_parts[ct];
                        if (cpart.flat_ph==null) continue;//dig this later.  too much effort.
                        var cwds = cpart.flat_ph.wds;
                        for (var w=0; w < cwds.length; w++) {
                            var anb = this.NLP.BRAIN.getOrMakeAnbl(wds[w]);
                             if (anb.full_tnids.length) {
                                reqs.push(anb.full_tnids);
                                if (anb.partial_tnids.length) {
                                    UTILS.addUniqueTnids(clues, anb.partial_tnids);
                                }
                            }
                            else if (anb.partial_tnids.length) {
                                role.rez_pool.fireLooseCandidates(anb.partial_tnids);        
                            }
                        }
                    } 
                }
                

                //pop in role ec (ie, if ROLE_WHERE, pop in ec_location)
                //like a game, we are looking for a location...
                //take nosigtext..
                
                if (reqs.length) {
                    await role.rez_pool.union_tnids(reqs);    
                    var candidates = role.rez_pool.getCandidatesWithFounds("phrase rez for "+role.ph.l1str);
                    
                    if (candidates.length==0) {
                        //make a new thing.
                        //check for likely concept.
                        //clues
                        //we want to chase these up. 
                        //we want to know the best top class
                        //role.rez_pool.fireLooseCandidates(clues);

                        var pd_pool_core = new pool_core();
                        var nctypeObj = pd_pool_core.fireTnids(clues);
                        var ntype = nctypeObj.ntype;
                        var nc = nctypeObj.nc;
                        var classNOBJ = null;
                        if (nc!=null) classNOBJ = nc.nObj;
                        
                        if (classNOBJ==null) {
                            if (role.best_class_tnid!=-1) {
                                classNOBJ= MEMG.getNeuronByTempNid(role.best_class_tnid);
                                console.log("make new generic thing neuron for %c"+nosigtext, TALK.bggreen);
                                finalNOBJ = await ANNE.BRAIN.ctp.makeLocalNeuron(role.ph.l1str_nosignal, ntype, classNOBJ);
                                role.updateResolved(finalNOBJ);
                            }

                            //generic thing
                            else {
                                console.log("make new generic thing (naked guess) neuron for %c"+nosigtext, TALK.bggreen);
                                finalNOBJ = await ANNE.BRAIN.ctp.makeLocalNeuron(role.ph.l1str_nosignal, ntype, MEM.ec_thing); 
                                role.updateResolved(finalNOBJ);
                                ANNE.BRAIN.ctp.addRoleToCP(clump, role); 
                            }

                            //we want to add this in as bes.
                            //to do, try (actor) for a dualing name
                            //also try (actor, director)
                            if (role.ph.context_ph_parts.length) {
                                console.log("context present, consider a be relon if it is straightforward.");
                                
                            }
                        }

                        //ensure when a new neuron is made.. 
                        //we can find this on the next time it is used.. 
                        //deeeerp.
                        else {
                            console.log("make new generic thing (best guess from clues) neuron for %c"+nosigtext, TALK.bggreen);
                            finalNOBJ = await ANNE.BRAIN.ctp.makeLocalNeuron(role.ph.l1str_nosignal, ntype, classNOBJ); 
                            role.updateResolved(finalNOBJ);
                            ANNE.BRAIN.ctp.addRoleToCP(clump, role);    
                        }
                    }

                    else {
                        //take this thing.
                        console.log("resolved neuron for %c"+nosigtext, TALK.bggreen);
                        console.log(candidates);
                        role.setWinningCandidate(candidates);
                    }
                }

                //we used the weights of the candidates to hits.
                //this is relevant when a larger concept is matched by smaller words that we don' thave as
                //individual neurons.
                //ie,  "liberty statue"
                //this will be less relevant as more neurons are known.
                else {
                    var candidates = role.rez_pool.getHighestWeightOverThreshold(1, "phrase rez for "+role.ph.l1str);
                    
                    //we need a new neuron.
                    if (candidates.length==0) {
                        console.log("make new neuron for %c"+nosigtext, TALK.bggreen);

                        //how do we know what this is?????? 
                        //inst or class? 
                        //makeLocalNeuron
                        if (role.ph.resolve_to_concept!=null) {
                            var finalNOBJ = await ANNE.BRAIN.ctp.makeLocalNeuron(role.ph.l1str, NTYPE_INST, MEM.ec_thing); 
                            role.updateResolved(finalNOBJ);
                            ANNE.BRAIN.ctp.addRoleToCP(clump, role);      
                        }

                        else {
                            var finalNOBJ = await ANNE.BRAIN.ctp.makeLocalNeuron(role.ph.l1str, NTYPE_CLASS, MEM.ec_thing); 
                            role.updateResolved(finalNOBJ);
                            ANNE.BRAIN.ctp.addRoleToCP(clump, role);         
                        }
                    }

                    //use the match we resolved to.
                    else {
                        console.log(candidates);
                        role.setWinningCandidate(candidates);
                    }
                }

                //add this resolved role to the context pool.
                ANNE.BRAIN.ctp.addRoleToCP(clump, role);
            }
        }

        /*****************************/
        console.log("**********************************");
        console.log("**********************************");
        
        $nlpP.clump.printRoles();
    }


    /**
     * The temp/local clump has not been actually stored yet in local DB mem.
     * It hasn't yet stored the updated roles yet either.
     * @param {*} $nlpP 
     */
    optimizeClump = async($nlpP)=> {
        var clump = $nlpP.clump;
        var roles = clump.nlpRolesObj._roles;
        
        //console.log(clump.clumpNOBJ);

        //unless we are in some full convo type dialogue mode..
        //we are not storing questions.
        //We can store them locally (in RAM only).
        //do this later. not a priority for the web EXT for data entry.
        if (clump.clump_type==CLUMP_TYPE_Q) {
            //we will have the TARGET question ROLE identified from the question.   
            //where, when, how much, what, who, etc...
        }
        else {
            if (clump.clump_sub_type==CLUMP_SUBTYPE_DEFN)  {
                console.log("try to just store a single TYPE_BE");
                if (clump.canBeSingleBeRelon()==true){
                    var rels = [];
                    var dupes = 0;
                    //we can later allow for this with CONNECTOR subj (like Cats and Dogs are pets).
                    var actor_roles = clump.nlpRolesObj[MEM.role_actor.temp_nid];
                    for (var a=0; a < actor_roles.length; a++) {
                        var subjNOBJ = actor_roles[a].nObj;
                        var be_thing_roles = clump.nlpRolesObj[MEM.role_be_thing.temp_nid];
                        for (var b=0; b < be_thing_roles.length; b++) {
                            var beThingNOBJ = be_thing_roles[b].nObj;
                            if (subjNOBJ.bes.indexOf(beThingNOBJ.temp_nid)!=-1) {
                                //still fire this somewhere to indicate confirmation/affirmation?
                                dupes+=1;
                                continue;//we already have this info.
                            }
                            
                            if (be_thing_roles.length>1) {
                                clump.single_relon = await ANNE.GHOSTDATA.new_ghost_be(subjNOBJ, MEM.ec_is.nid, TO_TYPE_ANNE, beThingNOBJ.nid);
                                rels.push(clump.single_relon);
                                console.log("%c clump simplified to single BE RELON (added new be relon)", TALK.bgyellow);
                                ANNE.TALK.printRelon(clump.single_relon);
                               
                            }
                            else {
                                //if this is a temp/ghost generic ec_thing, 
                                //we want to change this be parent.
                                //note, if we allow local on annodes, we will have to differentiate between them.
                                //ghosts could be updated to LOCAL (instead of annode values)
                                if ((subjNOBJ.local==true || subjNOBJ.ghost==true) && subjNOBJ.bes[0]==MEM.ec_thing.temp_nid) {
                                    clump.single_relon = await subjNOBJ.updateGhostParent(beThingNOBJ.nid, MEM.class_nid)
                                    rels.push(clump.single_relon);
                                    console.log("%c clump simplified to single BE RELON (updated existing parent)", TALK.bgyellow);
                                    ANNE.TALK.printRelon(clump.single_relon);
                                   
                                }
                                else {
                                    clump.single_relon = await ANNE.GHOSTDATA.new_ghost_be(subjNOBJ, MEM.ec_is.nid, TO_TYPE_ANNE, beThingNOBJ.nid);
                                    rels.push(clump.single_relon);
                                    console.log("%c clump simplified to single BE RELON (added new be relon)", TALK.bgyellow);
                                    ANNE.TALK.printRelon(clump.single_relon);
                                    
                                }
                            }
                        } 
                    }

                    //successful simplify to a BE RELON.
                    if (rels.length>0) {
                        //DESTROY THE CLUMPNOBJ.  We don't need it.
                        await ANNE.GHOSTDATA._annihilateGhostNeuron(clump.clumpNOBJ);
                        clump.clumpNOBJ = null;

                        var rels = await MEMP.processNeuronsFromRelons(rels, true, false,"relons", "optimizeClump");
                        ANNE.TALK.printRelons(rels);
                        return;        
                    }
                    else if (dupes>0) {
                        await ANNE.GHOSTDATA._annihilateGhostNeuron(clump.clumpNOBJ);
                        clump.clumpNOBJ = null;
                        return;
                    }

                }
            }    

                
            //simplify to a TYPE_HAS.
            //run through some "has" sentences/clumps to see how they get conceptualized.
            //a HAS will have an obj role with a mod or not.
            //a dog has 4 legs.   (legs (mod 4))  reln-> legs   to-> 4
            //a dog has a tail. (tail)   RELN=> posn.
            else if (clump.clump_sub_type==CLUMP_SUBTYPE_POSN) {
                console.log("try to just store a single TYPE_HAS");
                if (clump.canBeSingleHasRelon()==true){
                    var rels = [];
                    var dupes = 0;
                    /*
                    //we can later allow for this with CONNECTOR subj (like Cats and Dogs have claws).
                    var actor_roles = clump.nlpRolesObj[MEM.role_actor.temp_nid];
                    for (var a=0; a < actor_roles.length; a++) {
                        var subjNOBJ = actor_roles[a].nObj;
                        var role_posn_thing = clump.nlpRolesObj[MEM.role_posn_thing.temp_nid];
                        for (var b=0; b < role_posn_thing.length; b++) {
                            var posnThingNOBJ = role_posn_thing[b].nObj;
                            
                            if (role_posn_thing.length>1) {

                                var existing_has = MEMG.getHasRelonTo(subjNOBJ.outs, MEM.role_posn_thing.temp_nid);
                                if (subjNOBJ.bes.indexOf(posnThingNOBJ.temp_nid)!=-1) {
                                    //still fire this somewhere to indicate confirmation/affirmation?
                                    dupes+=1;
                                    continue;//we already have this info.
                                }
                                
                                clump.single_relon = await ANNE.GHOSTDATA.new_ghost_be(subjNOBJ, MEM.ec_is.nid, TO_TYPE_ANNE, posnThingNOBJ.nid);
                                rels.push(clump.single_relon);
                                console.log("%c clump simplified to single BE RELON (added new be relon)", TALK.bgyellow);
                                ANNE.TALK.printRelon(clump.single_relon);
                               
                            }
                            else {
                                //if this is a temp/ghost generic ec_thing, 
                                //we want to change this be parent.
                                //note, if we allow local on annodes, we will have to differentiate between them.
                                //ghosts could be updated to LOCAL (instead of annode values)
                                if ((subjNOBJ.local==true || subjNOBJ.ghost==true) && subjNOBJ.bes[0]==MEM.ec_thing.temp_nid) {
                                    clump.single_relon = await subjNOBJ.updateGhostParent(posnThingNOBJ.nid, MEM.class_nid)
                                    rels.push(clump.single_relon);
                                    console.log("%c clump simplified to single BE RELON (updated existing parent)", TALK.bgyellow);
                                    ANNE.TALK.printRelon(clump.single_relon);
                                   
                                }
                                else {
                                    clump.single_relon = await ANNE.GHOSTDATA.new_ghost_be(subjNOBJ, MEM.ec_is.nid, TO_TYPE_ANNE, posnThingNOBJ.nid);
                                    rels.push(clump.single_relon);
                                    console.log("%c clump simplified to single BE RELON (added new be relon)", TALK.bgyellow);
                                    ANNE.TALK.printRelon(clump.single_relon);
                                    
                                }
                            }
                        } 
                    }

                    //successful simplify to a BE RELON.
                    if (rels.length>0) {
                        //DESTROY THE CLUMPNOBJ.  We don't need it.
                        await ANNE.GHOSTDATA._annihilateGhostNeuron(clump.clumpNOBJ);
                        clump.clumpNOBJ = null;

                        var rels = await MEMP.processNeuronsFromRelons(rels, true, false);
                        ANNE.TALK.printRelons(rels);
                        return;        
                    }
                    else if (dupes>0) {
                        await ANNE.GHOSTDATA._annihilateGhostNeuron(clump.clumpNOBJ);
                        clump.clumpNOBJ = null;
                        return;
                    }
                    */

                }
            }
                
            else if (clump.clump_sub_type==CLUMP_SUBTYPE_ACTION) {   
                //look for inbound mapppers on the conveyor.
                //these are linguistic conversions from clump speak to distill facts out.
                //Bob died on 01/01/2021 in Ogden, Utah.
                //This clump has a WHEN and WHERE role -- these linguistically map back to 
                //the concepts "deathday" and "death place" in simple relon fact form.  ie: "bob->deathday->01/01/2021"
                //this can be done client side OR clumps can be sent to ANNE, and if accepted, ANNE can extract these
                //factual learnings to build the graph.
                //These types of clumps provide multiple data points -- that are not dependent upon eachother for their
                //semantic meaning.   Bob died on,  and bob died at/in  are two seperate facts that can stand independently valid on their own.

                //FACT FINDER.
                //i will write the code here, then move it to the ANNODE, upon HANAI ANNE will utilize this methodology to 
                //extract facts from the clumps.
                var conveyorNOBJ = clump.getConveyorNObj();
                var arm = new anne_role_mapper(conveyorNOBJ);
                await arm.extrap(clump);
                if (clump.extraps.length) ANNE.TALK.printRelon(clump.extraps);

                //if our extrap covers all our non-action/tense/actor roles.. (ie when/where etc)
                //we do not need a clump.
            }
                
            else if (clump.clump_sub_type==CLUMP_SUBTYPE_SELF_AWAREZ) {
                console.log("Simplify action into a feelz or opinionz single relon");
            }
        }


        //common relons (via patterns we have to store in L1 and have ready for ANNE use)

        //died on =>  role_action:died, role_when
        //was born on
        //was born in [class: LOCATION]
        //died on
        //is [class: LANG] for 
        //in [class: LANG] is
        //is also known as
        //(aka blah)
        //name "blah" name 
        
        //capital of [class: LOCATION]
        //Portland, Or is a big city.
        
        var relons = [];

        for (var r=0; r < roles.length; r++) {
            var role = roles[r];
            if (role.rObj==null) {
                console.log("make role relon in clump for role:");
                role.printRole();
                console.log(role);

                //the action role is already the parent class of the clump.
                //if (role.roleNOBJ.temp_nid==MEM.role_action.temp_nid) continue;
                if (role.nObj==null) {
                    console.log("%c CAN'T MAKE ROLE, NO NEURON FOR THIS ROLE SET", TALK.bgred);
                    continue;
                }
                await ANNE.GHOSTDATA.new_ghost_role(clump.clumpNOBJ, role.roleNOBJ.nid, TO_TYPE_ANNE, role.nObj.nid);
            }
        }

        //print clump for audit.
        if (clump.clumpNOBJ!=null) {
            this.talkObj = await clump.clumpNOBJ.processActionNeuron();
            console.log(this.talkObj); 
        }
    }
}