const POOL_SORT_FIRE = 1;
const POOL_SORT_HITS = 2;

class simple_context_pool {
    constructor ($cfd=-0.80, $nctis=360) {
        this._pool = new arraymap();
        this._pool.setMapKey("nid");
        this._pool.setSecondaryIndex("text", true);
        this.COOL_FIRE_DEGRADE = $cfd;
        this.NORMAL_CYCLE_TIME_IN_SECS = $nctis;
    }
    
    getNextDummyNeuron=function($newText){
        return ANNE.MEM.makeTempNObjOnly(ANNE.USER._userNObj.nid, NTYPE_CLASS, $newText, ANNE.MEM.anbl.nid, false);
    }

    getLongerByShorterString = function($str) {
        var str = $str.toLowerCase();
        return this._pool.get2_partial(str);
    }
    
    getByText = function($str) {
        if (typeof $str != "string") {
            return null;
        }
        var str = $str.toLowerCase();
        return this._pool.get2(str);
    }

    /**
     */
    getByClass=function($classNOBJ, $max, $ntype=NTYPE_INST) {
        this.sortPool();
        var matches = [];
        
        for (var p=0; p < this._pool.arr.length; p++) {
            
            if ($ntype!=null && this._pool.arr[p].nObj.ntype!=$ntype) continue;

            if (this._pool.arr[p].nObj.is($classNOBJ)==true) {
                matches.push(this._pool.arr[p].nObj);
                if (matches.length==$max) {
                    return matches;
                }
            }
        }
        return matches;
    }


    getInstNObjs = function($classNOBJ=null, $top=20, $skipClassNObjs=[]) {
        var nObjs = [];
        this.sortPool();
        for (var p=0; p < this._pool.arr.length; p++) {
            var poolObj = this._pool.arr[p];
            if (poolObj.nObj.ntype!=NTYPE_INST) continue;
            if ($classNOBJ!=null && poolObj.nObj.is($classNOBJ)==false) continue;
            else {
                var bSKIP=false;
                for (var s=0; s < $skipClassNObjs.length; s++) {
                    if (poolObj.nObj.is($skipClassNObjs[s])==true) {
                        bSKIP=true;
                        break;
                    }
                }
                if (bSKIP==true) continue;
            }

            nObjs.push(poolObj.nObj);
            if (nObjs.length>=$top) break;
        }
        return nObjs;
    }


    /**
     * 
     * @param {*} $nid 
     * @param {*} $fireWeight 
     */
    fireConnected = function($nid, $fireWeight=2) {
        var seedNOBJ = this.addNid($nid);
        for (var p=0; p < this._pool.arr.length; p++) {
            if (this._pool.arr[p].active==false) continue;
            var fire =this._pool.arr[p].fire;
            if (this._pool.arr[p].nObj.is(seedNOBJ)==true) {
                fire+=$fireWeight;
            }
            var outs = this._pool.arr[p].nObj.outs;
            for (var r=0; r < outs.length; r++) {
                if (outs[r].reln==seedNOBJ.nid) {
                    fire+=$fireWeight/7;
                }
                if (outs[r].to_type==TO_TYPE_ANNE) {
                    if (outs[r].to_s==seedNOBJ.nid) {
                        fire+=ANNE.getQuickWeight(outs[r].r, $fireWeight)/3;
                    }
                }
                else if (outs[r].to_type==TO_TYPE_STRING){
                    //text matching fuzzyness here.
                }
            }
            this._pool.arr[p].fire = fire;
        }
    }


    /**
     * @param {*} $nobj 
     */
    sumConnected = function($candidateNOBJ) {
        var outs = $candidateNOBJ.outs;
        var firesum = 0;

        //loop our candidate outs.
        for (var r=0; r < outs.length; r++) {

            if (outs[r].to_type!=TO_TYPE_ANNE) continue;

            //the to is in the pool -- we've got a match.
            var poolObj = this._pool.get(outs[r].to_s);
            if (poolObj==null) continue;
            if (poolObj.active==false) continue;

            //
            var w=ANNE.getQuickWeight(outs[r].r, 1);
            firesum+=poolObj.fire*w;
        }
        return firesum;
    }


    addNid = function($nid, $fireWeight=1) {
        if ($nid==undefined || $nid==null) return null;
        var poolObj = this._pool.get($nid);
        if (poolObj!=null) {
            poolObj.fire+=$fireWeight;
            poolObj.hits+=1;
            poolObj.lastBump=Date.now()/1000;
            poolObj.active=true;
            return poolObj.nObj;
        }
        var nObj = MEMG.getOrMakeIfNew($nid);
        poolObj = {nObj:nObj, text:nObj.text.toLowerCase(), fire:$fireWeight, hits:1, lastBump:Date.now()/1000, active:true};
        this._pool.addUnique(poolObj, nObj.nid);
        
        return nObj;
    }


    //a string fired not tied to a real neuron.
    addString = function($str) {
        if (this.getByText($str)!=null) return;
        var dummyNOBJ = this.getNextDummyNeuron($str);
        ANNE.MEM.addToTextIndex($str, dummyNOBJ.nid);
        this.addNObj(dummyNOBJ, 1);
    }


    /**
     * 
     * @param {*} $nObj 
     * @param {*} $fireWeight 
     * @returns 
     */
    addNObj = function($nObj, $fireWeight=1) {
        if ($nObj==undefined || $nObj==null) return;
        var poolObj = this._pool.get($nObj.nid);
        if (poolObj!=null) {
            poolObj.fire+=$fireWeight;
            poolObj.hits+=1;
            poolObj.lastBump=Date.now()/1000;
            poolObj.active=true;
            return;
        }
        if (typeof $nObj.text!=="string") {
            console.log("uhhh what is going on here?");
            return;
        }
        poolObj = {nObj:$nObj, text:$nObj.text.toLowerCase(), fire:$fireWeight, hits:1, lastBump:Date.now()/1000, active:true};
        this._pool.addUnique(poolObj, $nObj.nid);
    }

    
    coolFires = function() {
        var n = Date.now();
        for (var p=0; p < this._pool.arr.length; p++) {
            if (this._pool.arr[p].active==false) continue;
            var lastGap = (n - this._pool.arr[p].lastBump) / this.NORMAL_CYCLE_TIME_IN_SECS;
            var newFire = Math.max(0, this._pool.arr[p].fire*(this.COOL_FIRE_DEGRADE*lastGap));
            if (newFire==0) {
                this._pool.arr[p].active = false;
            }
            this._pool.arr[p].fire = newFire;
        }
    }


    
    /**
     * Pass in an array of nObjs that have .fire set on them.
     * With no fire we only use the pool fire, if it exists.
     * @param {*} $aMatches 
     */
    applyPoolWeights = function($aMatches, $context=[], $bWINNERONLY=false) {
        var final = [];
        var topFire = 0;
        var topHits = 0;

        for (var m=0; m < $aMatches.length; m++) {
            var match = $aMatches[m];
            if (match==undefined) continue;
            var poolObj = this._pool.get(match.nid);
            if (match.fire==undefined) match.fire=0;

            //only using in-RAM context here..
            //not chasing anything for this quick feature.
            if ($context.length>0) {
                var additive = poolObj.nObj.contextAdditive($context);
                match.fire+=additive;    
            }
            
            if (poolObj!=null) {
                if ((poolObj.fire*match.fire) > topFire) topFire = (poolObj.fire+match.fire);
                if (poolObj.hits > topHits) topHits = poolObj.hits;
            }
        }

        for (var m=0; m < $aMatches.length; m++) {
            var match = $aMatches[m];
            var poolObj = this._pool.get(match.nid);
            if (poolObj==null) {
                var tempfire = match.fire/topFire;
                final.push( { nObj:match, fire:tempfire});
                //console.log("%ctempfire:"+tempfire+" %c"+nObj.text, ANNE.TALK.bgred, ANNE.TALK.bggreen); 
            }
            else {
                if (topFire==0) {
                    final.push( { nObj:match, fire:topFire}); 
                    //console.log("%ctempfire:"+tempfire+" %c"+match.text, ANNE.TALK.bgred, ANNE.TALK.bggreen); 
                }
                else {
                    var tempfire = (poolObj.fire*match.fire)/topFire;
                    final.push( { nObj:match, fire:tempfire}); 
                    //console.log("%ctempfire:"+tempfire+" %c"+match.text, ANNE.TALK.bgred, ANNE.TALK.bggreen); 
                }
                
            }
        }

        if (final.length==0) {
            if ($bWINNERONLY==true) return null;
            else return final;
        }
        final.sort(function($fn1, $fn2) {
            if ($fn1.fire>$fn2.fire) return -1;
            else if ($fn1.fire<$fn2.fire) return 1;
            else return 0;
        })

        //return the array they passed in but now weighted by current fire levels
        //in the pool.
        var final2 = [];
        for (var f=0; f < final.length; f++) {
            final2.push(final[f].nObj);
            //console.log("%c'"+final[f].nObj.text+"' %c"+final[f].fire, ANNE.TALK.bggreen, ANNE.TALK.bgred);
            final[f].nObj.fire=0;
        }
        if ($bWINNERONLY==true) return final2[0];
        return final2;
    }


    /**
     * 
     * @param {*} $sort 
     */
    sortPool = function($sort=POOL_SORT_FIRE) {
        if ($sort==POOL_SORT_FIRE) {
            this._pool.arr.sort( function($p1, $p2) {
                if ($p1.fire > $p2.fire) return -1;
                else if ($p1.fire < $p2.fire) return 1;
                else {
                    if ($p1.lastBump > $p2.lastBump) return -1;
                    else if ($p1.lastBump < $p2.lastBump) return 1;
                    else {
                        if ($p1.hits > $p2.hits) return -1;
                        else if ($p1.hits < $p2.hits) return 1;
                        else return 0;
                    }
                }
            });
        }
        else if ($sort==POOL_SORT_HITS) {
            this._pool.arr.sort( function($p1, $p2) {
                if ($p1.hits > $p2.hits) return -1;
                else if ($p1.hits < $p2.hits) return 1;
                else {
                    if ($p1.fire > $p2.fire) return -1;
                    else if ($p1.fire < $p2.fire) return 1;
                    else {
                        if ($p1.lastBump > $p2.lastBump) return -1;
                        else if ($p1.lastBump < $p2.lastBump) return 1;
                        else return 0;
                    }
                }
            });
        }
    }

    printPool = function($sort=POOL_SORT_FIRE, $printCntMax=100, $printFireThresh=-1, $printHitThresh=-1) {
        var nw = Date.now()/1000;
        if ($sort==POOL_SORT_FIRE) {
            this._pool.arr.sort( function($p1, $p2) {
                if ($p1.fire > $p2.fire) return -1;
                else if ($p1.fire < $p2.fire) return 1;
                else {
                    if ($p1.lastBump > $p2.lastBump) return -1;
                    else if ($p1.lastBump < $p2.lastBump) return 1;
                    else {
                        if ($p1.hits > $p2.hits) return -1;
                        else if ($p1.hits < $p2.hits) return 1;
                        else return 0;
                    }
                }
            });

            console.log("%c context pool light. %c Total neurons:"+this._pool.arr.length, ANNE.TALK.bgheading_black, ANNE.TALK.bgheading_yellow);
            var prints = 0;
            for (var p=0; p < this._pool.arr.length; p++) {
                var poolObj = this._pool.arr[p];
                if ($printFireThresh!=-1 && poolObj.fire < $printFireThresh) continue;
                if ($printHitThresh!=-1 && poolObj.hits < $printHitThresh) continue;
                console.log("#"+(p+1)+" %cname:"+poolObj.nObj.bestName()+" %cfire:"+poolObj.fire+" %chits:"+poolObj.hits+" %clast:"+Math.ceil(nw-poolObj.lastBump)+" secs ago %cactive:"+poolObj.active, ANNE.TALK.bggreen, ANNE.TALK.bgfire, ANNE.TALK.bgyellow, ANNE.TALK.bgblue, ANNE.TALK.bgred);
                prints+=1;
                if (prints>$printCntMax) return;
            }
        }
        else if ($sort==POOL_SORT_HITS) {
            this._pool.arr.sort( function($p1, $p2) {
                if ($p1.hits > $p2.hits) return -1;
                else if ($p1.hits < $p2.hits) return 1;
                else {
                    if ($p1.fire > $p2.fire) return -1;
                    else if ($p1.fire < $p2.fire) return 1;
                    else {
                        if ($p1.lastBump > $p2.lastBump) return -1;
                        else if ($p1.lastBump < $p2.lastBump) return 1;
                        else return 0;
                    }
                }
            });

            console.log("%c context pool light. %c Total neurons:"+this._pool.arr.length, ANNE.TALK.bgheading_black, ANNE.TALK.bgheading_yellow);
            var prints = 0;
            for (var p=0; p < this._pool.arr.length; p++) {
                var poolObj = this._pool.arr[p];
                if ($printFireThresh!=-1 && poolObj.fire < $printFireThresh) continue;
                if ($printFireHitThresh!=-1 && poolObj.hits < $printFireHitThresh) continue;
                console.log("#"+(p+1)+" %cname:"+poolObj.nObj.bestName()+" %chits:"+poolObj.hits+" %cfire:"+poolObj.fire+" %clast:"+Math.ceil(nw-poolObj.lastBump)+" secs ago %cactive:"+poolObj.active, ANNE.TALK.bggreen, ANNE.TALK.bgyellow, ANNE.TALK.bgfire, ANNE.TALK.bgblue, ANNE.TALK.bgred);
                prints+=1;
                if (prints>$printCntMax) return;
            }
        }
    }
}

class context_pool {
    constructor () {
        //reconfirm all of this.
        //early concepts expecations pool used for phrase making.
        //this.roles_ecpool = null;
        //this.ph_ecpool = null;
        this.concept_pool = null;
        this._pool = [];            
        this._pool_map=[];
        this._lastClump = null;
        this._primary_speaker_nid = null;
        this._active_speaker_nid = null;
        
        //ANNE, when ANNE is addressed.
        //this is how concepts tied to "this.ECS.ec_theAudience" are resolved.  (you, your, yourself)
        this._audience_nid = null;  

        this._clumps = [];

        //things we have resolved yet on a macro level
        this.pending_issues = [];   

        //this.activeSpeakerByCIndex = [];
        this._clump_index = 0;

        //sometimes we keep these, othertimes we don't.
        this.temp_new_neurons = [];

        this.start_session_context();
    }

    
    //we should do an ANNE block number here for the cycles tracker..
    //to tie "time" of clumps together (internally) when contrast to eachother.
    //we have clump index to handle that for now.
    start_session_context=function() {
        console.log("%c CONTEXT POOL STARTED", 'background: purple; color: white');  
        //this.roles_ecpool = new nlp_expectation_pool("roles pool");
        this.concept_pool = new nlp_expectation_pool("concept pool");
    }


    /**
     * A local-only (not a ghost) neuron.
     * This will be stored in local-RAM.
     * Users can choose to retain these neurons in their local store.
     * Local circuits can be used to teach ANNE.
     * Users can choose to push these learnings to global ANNE on-chain 
     * as shared resources in the ANNE worldview or they can keep them as their own.
     * 
     * As a defaut, when ANNE parses or converses (in any mode) local neurons will be created
     * for new things. (not ghosts)  ANNE does not pre-train.  There is no grammars.  ANNE is taught.
     * There are EARLY CONCEPTS which are key to ANNE having something to tie new concepts to.
     * Local neurons are the first step in creating ANNE NLP type data.
     * 
     * @param {*} $text 
     * @param {*} $parentClass 
     */
    makeLocalNeuron = async($text, $type, $parentObj)=> {
        var localNOBJ = await MEM.makeGhostNObj(USER._userNObj, $type, $text, $parentObj, false);
        localNOBJ.local = true;
        localNOBJ.bNEW = true;
        return localNOBJ;
    }

    
    /**
     * Some of these clumps will further get reduced by our on-ANNE-chain simplifier
     * once the user decides to save the data on ANNE chain.
     * A "is" clump (defn) with only actor+action+ state/be_thing is just a TYPE_BE relon.
     * NO clump is necessary.
     * Same with feelz/opinionz clumps and some has (posn) that make TYPE_HAS relons.
     * Clumps are for semantic thoughts that require more than a simple semantic triplet relon in
     * a specific data dimension.
     * @param {nlp_clump}
     */
    processClump = async($clump)=>{
        console.log("(cpt.process clump)");
        if ($clump.conveyor_role==null) return false;//this is just a phrase
        this.setActiveSpeakerFromClump($clump);    //TODO
        $clump.clumpNOBJ = await MEM.makeGhostNObj(this._active_speaker_nid, NTYPE_INST, "", $clump.conveyor_role.nObj, true, true);
        return true;
    }


    /**
     * we build it as we go so we can get roles into the context pool as we know of them (resolved) 
     */
    addRoleToCP=function($clump, $role) {
        //console.log("add role to pool ", $role);
        if ($role.nObj==null) {
            console.log("no $role.nObj we can't add this role to the context pool.");
            //gui/user remediation?  in a live ANNE EI system this would trigger a potential response...
            //a need to gain context to resolve a problem.
            return; //no neuron, we can't add this role.
        }
        
        if ($role.rObj==null && $role.roleNOBJ.nid!=MEM.role_action.nid) {
            $role.rObj = MEM.makeTempRelon($clump.clumpNOBJ, TYPE_ROLE, $role.roleNOBJ.nid, TO_TYPE_ANNE, $role.nObj.nid, true);
            $clump.clumpNOBJ.addRole($role.roleNOBJ, $role.rObj);
        }
        //this adds a role to the context pool.
        this._addRole(this._active_speaker_nid, $role.nObj, $role.roleNOBJ, this._clump_index, $clump.clump_type);
    }


    /**
     * the pool map is by resolved role neuron.
     * ie, the concept.
     * @param {*} $speaker 
     * @param {*} $nObj 
     * @param {*} $role_nObj 
     * @param {*} $clump_index 
     * @param {*} $clump_type 
     */
    _addRole = function($speaker, $nObj, $role_nObj, $clump_index, $clump_type) {
        if (this._pool_map[$nObj.temp_nid]==undefined) {
            var poolObj = new ctp_obj($speaker, $nObj, $nObj.text, $clump_index, $clump_type);
            this._pool_map[$nObj.temp_nid]=poolObj
            this._pool.unshift(poolObj);    //put at top of array (front)
            poolObj.tallySpeakerAndRole($speaker, $role_nObj);
            this.topicTally($role_nObj, $nObj);
        }
        else {
            var poolObj = this._pool_map[$nObj.temp_nid]
            poolObj.refire($clump_index, $clump_type);
            poolObj.tallySpeakerAndRole($speaker, $role_nObj);
            this.topicTally($role_nObj, $nObj);

            //_pool map is not indexes of _pool but the temp_nid of the nObj 
            //so we can directly sort our this._pool here without mucking up indexes.
            this._pool.sort(anne_nlp._sortByLastCIAndHits);
        }
    }


    /**
     * something like "i" or "we" has been encountered.
     * @returns 
     */
    getCurrentActiveSpeaker=async()=> { 
        //we should try to understand WHO THE SPEAKER IS..
        //we can just mark it as a new user... until we find out.
        if (this._active_speaker_nid==null) {
            var selfNOBJ = MEMG.getNeuronByTempNid(MEM.ANNE_ACCTID);
            var newNOBJ = await MEM.makeGhostNObj(selfNOBJ, NTYPE_UNKNOWN, "temp unk user", MEM.identity, true, true);
            newNOBJ.__functInherit(MEM.identity, "context_pool.getCurrentActiveSpeaker");
            ANNE.BRAIN.ctp.temp_new_neurons.push(newNOBJ);
            //needs verification - pending issue.. hmmm.
            return newNOBJ.nid;
        }

        return this._active_speaker_nid;
    }
    

    /**
     * 
     * @param {*} $userNid 
     */
    setPrimarySpeaker=function($userNid) {
        this._primary_speaker_nid = $userNid;
        this._active_speaker_nid = $userNid;
    }

    
    /** 
     * ANNE is listening for data to gain insight/info/learn
     * There is no intent to dialogue here (by ANNE)
     */
    initGeneralFactFinder=function() {
        console.log("%c initGeneralFactFinder mode", 'background: purple; color: white'); 
    }


    /**
     * We want to resolve a resolvable.
     * The role has been resolved (conceptualized) but 
     * we must resolve the value in context.  (ie, she, he, they, i, we etc)
     * @param {nlp_role}
     */
    resolveRoleResolveable=async($role)=> {
        console.log($role);
        console.log("%c RESOLUTION %c resolve resolvable (s.b. pn like) %c '"+$role.ph.l1str+"'", TALK.bgred, TALK.bgyellow, TALK.bggreen, );       
        
        //english "i"
        if ($role.ph.resolve_to_concept!=undefined && $role.ph.resolve_to_concept!=null) {
            if ($role.ph.resolve_to_concept==MEM.ec_theSpeaker.temp_nid) {
                var nid = await this.getCurrentActiveSpeaker();
                var nObj= MEMG.getNeuronByNid(nid);
                $role.updateResolved(nObj)
                $role.RESOLVED_TO_SPEAKER=true;
                return;
            }
        }

        var crr = new nlp_role_concept_resolver($role);
        var pnCandidates = this.getLatestThings(); //array of ctp_obj
        return await crr.resolvePNRoleAndSetWinner($role, pnCandidates)
    }

    
    /**
     * Rumi once said, "blah blah blah blah".
     * The voice should be rumi, but the worldview (owner of the clump) 
     * should be who said Rumi said that!
     * In most cases this will not be altered by the clump and will be left at
     * this._active_speaker_nid
     */
    setActiveSpeakerFromClump = function($clump) {
        console.log("to do: set speaker from clump");
    }













    /**
     * We have an extended phrase (with potential cjs, signals etc)
     * We want to try to simplify it down to a more basic concept.
     * "the dog"  (a dog in context)
     * "the man" (a man in context)
     * 
     * ie:  "the man named donald trump" 
     *       "the one they call DJT"
     *       "the 45th president of the united states"
     * ALL thes should resolve to Donald Trump... (assuming we have that neuron and know that info)
     *
     * ie:  "the 3rd planet"
     * ie:  "the planet with water and people"
     * 
     * even more difficult:  "the 44th and 45th presidents of the USA were in a photo today".
     * Eazy, to hardcode, but we want this resolve by a brain-neuron resolver, not by a rule-based
     * hardcoded list of patterns we continually have to code.  Our patterns should
     * be in neurons -- this enables the brain to grow/learn/ and be more robust as more information
     * is known.
     */
    resolveRolePhrase=async($tokensObj, $role)=> {
        console.log("%c resolveRolePhrase %c '"+$role.text+"'", TALK.bgblackyellow, TALK.bggreen);       
        
        //if the mods are adjectives.. then we do the concept resolver..
        //ie,  base: 'planet' ..   mod:  "3rd"  or "blue"  or "red spot"
        var crr = new nlp_role_concept_resolver($role);
        var bTHEREWASAWINNER = await crr.resolveAndSetWinner("(resolveRolePhrase)");
        
        if ($tokensObj.target_role_only!=null && $tokensObj.chaseClassNid!=null) {
            MEM.addModsToGhostNeuron($role, $tokensObj.chaseClassNid);
        }

    }



    //set from the outside.
    updateActiveSpeaker = function($newActiveSpeaker) {
        this._active_speaker_nid = $newActiveSpeaker;
        this.activeSpeakerTracker.push({c_index:this._clump_index, active_speaker:this._active_speaker_nid});
    }
    
    /**
     */
    nlpClumpIn=function($clump) {
        console.log('(brain.cpt.nlpClumpIn)');
        this._clumps.push($clump);
        this._clump_index = this._clumps.length;

        var bGOOD = this.processClump($clump);
        if (bGOOD==false) {
            
        }
        this.wireClumpsIntime($clump);
        this._lastClump = anneClump;

        //we want a 2 clump window for potential role expectations.
        //we can get one for the previous clump in a follow up sentence.
        this.removeRoleExpectsBefore( Math.max(this._clump_index-2, 0) );

        this.printPool();
    }

    /**
     * Inside voice
     */
    internalClumpIn=function($clump) {
    }
    wireClumpsIntime=function($clump) {
        if (this._lastClump==null) return;
    }
    addNext = function($clump) {
        console.log('wire next');
    }
    addPrev = function($clump) {
        console.log('wire prev');
    }



    //hmm.
    topicTally=function($role_nObj, $nObj){
    }


    getLatestThings=function($clumpsBack=1){
        var candidates = [];
        var minClumpIndex = Math.max(0, (this._clump_index - $clumpsBack));
        for (var p=0; p < this._pool.length; p++) {
            var pobj = this._pool[p];
            if (pobj.ci<minClumpIndex) break;
            if (pobj.roles_tally[MEM.role_actor.temp_nid]!=undefined) {
                candidates.push(pobj);
            }
            else if (pobj.roles_tally[MEM.role_theme_thing.temp_nid]!=undefined) {
                candidates.push(pobj);
            }
            else if (pobj.roles_tally[MEM.role_affected_thing.temp_nid]!=undefined) {
                candidates.push(pobj);
            }
            else if (pobj.roles_tally[MEM.role_posn_thing.temp_nid]!=undefined) {
                candidates.push(pobj);
            }
            else if (pobj.roles_tally[MEM.role_be_thing.temp_nid]!=undefined) {
                candidates.push(pobj);
            }
            else if (pobj.roles_tally[MEM.role_instrument.temp_nid]!=undefined) {
                candidates.push(pobj);
            }
        }

        return candidates;
    }



    /**
     * We have some issue that we want to flag.
     * technically, we should use an issue clump to flag issues...
     * but that is going super meta (for later)
     */
    _makeIssue = function($issueText, $signalNOBJ, $conceptToResolveNOBJ, $original_str) {
        var issue = {
                    issue_text:$issueText, 
                    signal:$signalNOBJ, 
                    unresolved_in_context:$conceptToResolveNOBJ, 
                    original_str:$original_str,
                    clump_index:this._clump_index
                };
        this.pending_issues.push(issue);
        return issue;
    }

    printPool=function(){
        console.log("%c persistent context pool", TALK.bgblackyellow);
        console.log(this._pool);
        for (var p=0; p < this._pool.length; p++) {
            this._pool[p].print();
        }
    }
}


/*************************
  NEW STUFF THAT IS DEFINITELY BEING used_tnids
*/



class pool_core {
    constructor($poolTitle=""){
        this.ncs = [];
        this.ncs_map = new Object();
        this._lastHigh = 0;
        this.poolTitle = $poolTitle;
        this.grace_misses=0;
    }

    /**
     * 
     * @param {*} $tnids 
     * @param {*} $bPRINT 
     */
    init=function($tnids, $bPRINT=true) {
        for (var t=0; t < $tnids.length; t++) {
            var nObj=MEMG.getNeuronByTempNid($tnids[t]);
            if (this.ncs_map[nObj.temp_nid]==undefined) {
                this.addCore(nObj, 1);
            }
        }
        if ($bPRINT) this.printPool(false);
    }


    /**
     * 
     * @param {*} $tnids 
     * @param {*} $weight 
     * @param {*} $bPRINT 
     */
    initHighPriority=function($tnids, $weight=2, $bPRINT=true) {
        for (var t=0; t < $tnids.length; t++) {
            var nObj=MEMG.getNeuronByTempNid($tnids[t]);
            if (this.ncs_map[nObj.temp_nid]==undefined) {
                this.addCore(nObj, $weight);
            }
            else {
                var nc = this.ncs[ this.ncs_map[nObj.temp_nid] ];
                if (nc.weight < $weight) {
                    nc.weight = $weight;
                }
            }
        }
        if ($bPRINT) this.printPool(false);
    }


    /**
     * 
     * @param {*} $tnid 
     * @param {*} $factor 
     * @returns 
     */
    bumpWeight = function($tnid, $factor) {
        var nc = this.ncs[ this.ncs_map[$tnid] ];
        if (nc==null) return;
        nc.weight = nc.weight * $factor;
    }


    printPool=function($bMUSTHAVEFIRE=true){
        console.log("******************************");
        console.log("%c pool core %c "+this.poolTitle, TALK.bgblackyellow, TALK.bgred);
        for (var p=0; p < this.ncs.length; p++) {
            if ($bMUSTHAVEFIRE==false) this.ncs[p].print();
            else if (this.ncs[p].fire>0) this.ncs[p].print();
        }
    }

    
    /**
     * This is from the flat phrase stage when discerning
     * where our general phrases begin and end.
     * We get many by signals but sometimes phrases run together without any signal words 
     * to break them.  In these cases we need to rely upon our PRIMARY DIRECTIVE mode context 
     * which tells us what type of stuff we are looking for to help us break these things up.
     * 
     * NOTE:  if we want a completely open-ended anything goes mode we will need to have a 
     * pool that is discerning what the phrases are as they come in, word by word, just as this
     * pool does --- however, that one will only have the working context of what has been discussed
     * in that convo/dialogue history w/ that entity.
     * 
     * ie, if a guy always talks baseball, then that is preloaded as context in the pool..
     * if you are talking about nature then those things are in context and can help resolve specific 
     * phrases of any given new sentence.
     * 
     * Signals and context is where it is at.
     * 
     * WE ARE NOT TRYING TO RESOLVE THE CONCEPT HERE.
     * 
     * NOTE: will set php.likelyConcept
     * @param {nlp_phrase_flat_part} $pfp 
     */
    testPh=function($nlpP, $next_wd, $activeFlatPh) {
        var tnids = [];

        //we need "columbia river" (when columbia river is unknown and colubmia university is known)
        //to still give us a likely concept of THING or GEO SITE, *NOT* entity.
        
        var fullstr = "";
        if ($activeFlatPh.wd_cnt==0) fullstr = $next_wd.str;
        else fullstr = $activeFlatPh.getInProgressString($next_wd);

        //change this to be lang specific, eng weights the right, but other langs the thing is first.
        var weight = 1+$activeFlatPh.wd_cnt*2.2;
        
        //test with a known .
        if ($activeFlatPh.wd_cnt>0 && fullstr.length > $next_wd.str.length) {
            tnids = MEMG.getTnidsByAnbl(fullstr);
            if (tnids.length) {
                $activeFlatPh.setTnids(tnids)
                console.log($activeFlatPh.tnids); //this is a likely conceptual match if this ends here.

                //how do we know this is "resolved" and doesn't need to be resolved in context ?
                //what if it refers to something else?  It is multi-wd, but this could be a multi-wd 
                //prounoun type reference.
                if (tnids.length==1) {
                    var rezNOBJ = MEMG.getNeuronByTempNid(tnids[0]);
                    if (rezNOBJ.ntype==NTYPE_INST) {
                        $activeFlatPh.setResolvedNObj(rezNOBJ);

                        var beNOBJ = MEMG.getNeuronByTempNid(rezNOBJ.bes[0]);
                        $activeFlatPh.setLikelyConcept(beNOBJ);
                    
                        return rezNOBJ;
                    }

                    //we don't know yet if this needs to be resolved, in context.
                    else {
                        $activeFlatPh.setResolveNObj(rezNOBJ);
                    }
                    
                }
            }
        }
        else $activeFlatPh.tnids = [];

        //do desperate string only attempts.
        if (tnids.length==0 && $next_wd.tnids==undefined || $next_wd.tnids==null || $next_wd.tnids.length==0) {
            //console.log("no tnids for %c"+$next_wd.str+", do crazy text string search", TALK.bggreen);
        }
        else {
            var used_tnids = [];
            for (var n=0; n < $next_wd.tnids.length; n++) {
                var neuron = MEMG.getNeuronByTempNid($next_wd.tnids[n]);
                if (used_tnids.indexOf(neuron.temp_nid)==-1) {
                    this.addFirerWd(neuron, weight); 
                    used_tnids.push(neuron.temp_nid);
                }
            }

            //if we had multi wd tnids match, get that in the pool too.
            for (var n=0; n < tnids.length; n++) {
                var neuron = MEMG.getNeuronByTempNid(tnids[n]);
                if (used_tnids.indexOf(neuron.temp_nid)==-1) {
                    this.addFirerWd(neuron, 3*weight); 
                    used_tnids.push(neuron.temp_nid);
                }
            }
        }

        //this.printPool(true);
        var lastHigh = this._lastHigh;
        var highNC = this.getHighestTop();
        if (highNC!=null) {
            this._lastHigh = highNC.fire;
            //console.log("new high "+this._lastHigh);
            if (this._lastHigh > lastHigh) {
                if ($activeFlatPh!=null) {
                    $activeFlatPh.setLikelyConcept(highNC.nObj);
                }
                return highNC.nObj;
            }
                
            //sketchy
            else if (this.grace_misses==0 && highNC.nObj!=null && highNC.nObj.temp_nid==MEM.name.temp_nid && $activeFlatPh.wds.length>0) {
                if ($activeFlatPh.likelyConveyorTest($nlpP, $next_wd)==true) {
                    return null;    
                } 
                this.grace_misses+=1;
                if ($activeFlatPh!=null) {
                    $activeFlatPh.setLikelyConcept(highNC.nObj);
                }
                return highNC.nObj;
            }
            return null;
        }
        return null;
    }



    /**
     * We need to know if these tnids are insts or classes.
     * We also need to know what their lowest most common parent class is.
     * 
     * We are essentially throwing a bunch of neurons into a firing pool..
     * and want to get a fuzzy understanding of the likliness these are a name, a person, etc.
     */
    fireTnids = function($tnids, $fireMultiplier=1) {
        for (var n=0; n < $tnids.length; n++) {
            var neuron = MEMG.getNeuronByTempNid($tnids[n]);
            this.addExploreUp(neuron, $fireMultiplier); 
        }
        this.printPool(false);
        return this.analyzeForCreation();
    }

    analyzeForCreation = function() {
        var highNC = null;
        var highW = 0;
        var d0inst = 0;
        var d0class = 0;
        for (var n=0; n < this.ncs.length; n++) {
            var ncore=this.ncs[n];
            if (ncore.depth==0) {
                if (ncore.nObj.ntype==NTYPE_INST) d0inst+=1;
                else d0class+=1;
            }
            if (ncore.depth>1) continue;//0 or 1 is ok here for what we want
            if (ncore.weight>highW) {
                highW = ncore.weight;
                highNC = ncore;
            }
        }
        var ntype = NTYPE_INST;
        if (d0class>d0inst) ntype = NTYPE_CLASS;
        return {nc:highNC, ntype:ntype};
    }



    addExploreUp = function($nObj, $startWeight=1) {
        var nc = new neuron_core($nObj, 0, $startWeight);
        this.ncs.push(nc);
        this.ncs_map[$nObj.temp_nid] = this.ncs.length-1;
        var w = $startWeight;
        for (var p=0; p < $nObj.outs.length; p++) {
            var rObj = $nObj.outs[p];
            if (rObj.type!=TYPE_BE) continue;
            if (rObj.to_type!=TO_TYPE_ANNE) continue;
            var toNOBJ = MEMG.getNeuronByTempNid(rObj.t);
            if (rObj.r==MEM.inst.temp_nid) w=$startWeight*1.3;
            else if (rObj.r==MEM.class.temp_nid) w=$startWeight*1.1;
            else w=$startWeight*.9;
            this.addUniqueSubcoreChaseUp(nc, toNOBJ, w, 1);
        }
    }


    /**
    * If we don't get any hits in our pool...
    * sometimes we can just take the highest weighted pool target.
    */
    getHighestWeight = function() {
        var highWNOBJ = null;
        var highW = 0;
        for (var n=0; n < this.ncs.length; n++) {
            var ncore=this.ncs[n];
            if (ncore.depth!=0) continue;
            if (ncore.weight>highW) {
                highW = ncore.weight;
                highWNOBJ = ncore.nObj;
            }
        }
        return highWNOBJ;
    }


    
    /**
     * 
     */
    getHighestTop=function(){
        var highNC = null;
        var highFire = 0;
        for (var n=0; n < this.ncs.length; n++) {
            var ncore=this.ncs[n];
            if (ncore.depth!=0) continue;
            if (ncore.hits==0) continue;
            if (ncore.fire>highFire) {
                highFire = ncore.fire;
                highNC = ncore;
            }
        }
        return highNC;
    }


    /**
     * 
     */
    resetPoolFire=function(){
        for (var n=0; n < this.ncs.length; n++) {
            var ncore=this.ncs[n];
            ncore.hits=0;
            ncore.fire=0;
        }
        this._lastHigh = 0;
    }

    
    /**
     */
    fireExtended = function($tnids){
    
    }

  


    /**
     * @param {*} $nObj 
     */
    addCore = function($nObj, $startWeight=1) {
        var nc = new neuron_core($nObj, 0, $startWeight);
        this.ncs.push(nc);
        this.ncs_map[$nObj.temp_nid] = this.ncs.length-1;

        for (var p=0; p < $nObj.outs.length; p++) {
            var rObj = $nObj.outs[p];

            //not needed here.
            if (rObj.type==TYPE_SYS_MSG) continue;
            if (rObj.type==TYPE_SYS) continue;
            if (rObj.type==TYPE_ROLE) continue; 
            if (rObj.type==TYPE_CLUMP) continue; 
            if (rObj.type==TYPE_BE && (rObj.r==MEM.inst.temp_nid || rObj.r==MEM.class.temp_nid)) continue;
            
            if (rObj.r==MEM.ec_is.temp_nid) continue;
            if (rObj.r==MEM.BE_CHILDREN.temp_nid) continue;

            this.addCoreFromRelon(rObj, (0.9*$startWeight), nc);  
        }
    }


    /**
     */
    addCoreFromRelon = function ($relon, $weight, $parent_nc) {
        var relnNOBJ =MEMG.getNeuronByTempNid($relon.r);
        if ($relon.to_type==TO_TYPE_STRING) {
            $parent_nc.addStr($relon.to_s);
            return;
        }
        if ($relon.to_type!=TO_TYPE_ANNE) return;
        
        if (relnNOBJ.temp_nid==MEM.ec_opposite.temp_nid || relnNOBJ.temp_nid==MEM.__NOTACTIVE.temp_nid) {
            $weight=$weight*-1;
        }
        else if (relnNOBJ.temp_nid==MEM.__throttlemin.temp_nid) {
            $weight=$weight*-.33;  
        }
        else if (relnNOBJ.temp_nid==MEM.__throttlemax.temp_nid) {
            $weight=$weight*-1;
        }
        else if (relnNOBJ.temp_nid==MEM.brain_assoc.temp_nid) {
            $weight=$weight*1.3;
        }
        else if (relnNOBJ.temp_nid==MEM.brain_assoc_strong.temp_nid) {
            $weight=$weight*2.5;
        }

        //the toNOBJ weight is altered by key BRAIN concepts.
        //assocs bump it and throttles are hurtful to matches.
        var toNOBJ = MEMG.getNeuronByTempNid($relon.t);
        //TALK.printRelon($relon);
        
        var to_ncore = this.addUniqueSubcore($parent_nc, toNOBJ, $weight, $parent_nc.depth+1);

        //we really want the ec_patterns here.
        //this will help us catch parts of patterns that tie back our targets while we are parsing.
        for (var p=0; p < toNOBJ.outs.length; p++) {
            var rel = toNOBJ.outs[p];
            if (rel.type!=TYPE_BRAIN) continue;
            if (rel.to_type!=TO_TYPE_ANNE) continue;
            if (rel.r==MEM.ec_is.temp_nid) continue;
            if (rel.r==MEM.BE_CHILDREN.temp_nid) continue;
            //TALK.printRelon(rel);
            var oneMoreNOBJ = MEMG.getNeuronByTempNid(rel.t);
            this.addUniqueSubcore($parent_nc, oneMoreNOBJ, $weight*.9, to_ncore.depth+1);
        }
    }

    /**
     * 
     * @param {*} $parent_nc 
     * @param {*} $nObj 
     * @param {*} $weight 
     * @param {*} $depth 
     * @returns 
     */
    addUniqueSubcore=function($parent_nc, $nObj, $weight, $depth=0) {
        if (this.ncs_map[$nObj.temp_nid]==undefined) {
            var nc = new neuron_core($nObj, $depth, $weight);
            this.ncs.push(nc);
            this.ncs_map[$nObj.temp_nid] = this.ncs.length-1;
            nc.parents_nc.push($parent_nc);
            return nc;
        }
        else {
            var nc = this.ncs[ this.ncs_map[$nObj.temp_nid] ];
            nc.weight+=$weight;
            for (var n=0; n < nc.parents_nc.length; n++) {
                if (nc.parents_nc[n].nObj.temp_nid==$parent_nc.nObj.temp_nid) {
                    return nc;
                }
            }
            nc.parents_nc.push($parent_nc);
            return nc;
        }
    }

    addUniqueSubcoreChaseUp=function($parent_nc, $nObj, $weight, $depth=0) {
        if (this.ncs_map[$nObj.temp_nid]==undefined) {
            var nc = new neuron_core($nObj, $depth, $weight);
            this.ncs.push(nc);
            this.ncs_map[$nObj.temp_nid] = this.ncs.length-1;
            nc.parents_nc.push($parent_nc);
            if ($nObj.be_lvl<=2) return nc;
            var beNOBJ = MEMG.getNeuronByTempNid($nObj.bes[0]);
            return this.addUniqueSubcoreChaseUp(nc, beNOBJ, $weight*.9, $depth+1);
        }
        else {
            var nc = this.ncs[ this.ncs_map[$nObj.temp_nid] ];
            nc.weight+=$weight;
            for (var n=0; n < nc.parents_nc.length; n++) {
                if (nc.parents_nc[n].nObj.temp_nid==$parent_nc.nObj.temp_nid) {
                    return nc;
                }
            }
            nc.parents_nc.push($parent_nc);
            if ($nObj.be_lvl<=2) return nc;
            var beNOBJ = MEMG.getNeuronByTempNid($nObj.bes[0]);
            return this.addUniqueSubcoreChaseUp(nc, beNOBJ, $weight*.9, $depth+1);
        }
    }


    /**
     */
    addFirerWd = function ($nObj, $firemultiplier=1) {
        var nc = this.ncs[ this.ncs_map[$nObj.temp_nid] ];
        if (nc!=undefined) {
            nc.fireMatch(nc.weight);
        }
        else {
            //what about adding THIS nObj?
        }
        
        //if its self/parent/bes/or assocs are in the pool.. fire the pool.
        for (var r=0; r < $nObj.outs.length; r++) {
            var rel  = $nObj.outs[r];  
            if (rel.to_type!=TO_TYPE_ANNE) continue;
            
            if (this.ncs_map[rel.t]==undefined) continue;

            var nc = this.ncs[ this.ncs_map[rel.t] ];
            var w=1;
            var relnNOBJ = MEMG.getNeuronByTempNid(rel.r);

            if (relnNOBJ.temp_nid==MEM.ec_opposite.temp_nid || relnNOBJ.temp_nid==MEM.__NOTACTIVE.nid) {
                w=-1;
            }
            else if (relnNOBJ.temp_nid==MEM.__throttlemin.temp_nid) {
                w=-0.33;  
            }
            else if (relnNOBJ.temp_nid==MEM.__throttlemax.temp_nid) {
                w=-1;
            }
            else if (relnNOBJ.temp_nid==MEM.brain_assoc.temp_nid) {
                w=1.3;
            }
            else if (relnNOBJ.temp_nid==MEM.brain_assoc_strong.temp_nid) {
                w=2.5;
            }
            nc.fireMatch(w*$firemultiplier);
        }    
    }


    /**
     * We adjust the role weights under certain signal conditions.
     * @param {*} $nObj 
     * @param {*} $multiplier 
     */
    weightSpecific=function($nObj, $multiplier=1) {
        var nc = this.ncs[ this.ncs_map[$nObj.temp_nid] ];
        if (nc!=undefined) {
            nc.weight = nc.weight*$multiplier;
        }
    }

    
    /**
     * for now, just assocs.
     * @param {*} $nObj 
     */
    contextualizeWithConveyor = function($nObj) {
        for (var r=0; r < $nObj.outs.length; r++) {
            var rel  = $nObj.outs[r];  
            if (rel.to_type!=TO_TYPE_BRAIN) continue;
            if (this.ncs_map[rel.t]==undefined) continue;

            var nc = this.ncs[ this.ncs_map[rel.t] ];
            var w=1;
            var relnNOBJ = MEMG.getNeuronByTempNid(rel.r);

            if (relnNOBJ.temp_nid==MEM.ec_opposite.temp_nid || relnNOBJ.temp_nid==MEM.__NOTACTIVE.nid) {
                w=-1;
            }
            else if (relnNOBJ.temp_nid==MEM.__throttlemin.temp_nid) {
                w=-0.33;  
            }
            else if (relnNOBJ.temp_nid==MEM.__throttlemax.temp_nid) {
                w=-1;
            }
            else if (relnNOBJ.temp_nid==MEM.brain_assoc.temp_nid) {
                w=1.3;
            }
            else if (relnNOBJ.temp_nid==MEM.brain_assoc_strong.temp_nid) {
                w=2.5;
            }

            //additive/subractive weight
            nc.fireMatch(w*$firemultiplier);
        }  
    }
    

    //fire a wrd, its immediate out...
    //but chase its bes uplines and fire those too.
    addFireWdUp = function ($nObj, $firemultiplier=1, $tnid_skips=[], $depth=0) {
        var nc = this.ncs[ this.ncs_map[$nObj.temp_nid] ];
        if (nc!=undefined) {
            nc.fireMatch(nc.weight);
        }

        if ($tnid_skips.indexOf($nObj.temp_nid)!==-1) return;

        if ($nObj.be_lvl<=1) return;

        $tnid_skips.push($nObj.temp_nid);
        
        //fire the be upline.
        //if we have a phrase w/ wd "car" we want to ultimately fire ec_thing.
        //This will help us resolve this role as ROLE_THING.
        //if we have "noon" we want to fire the concept of time... etc.
        //var beNOBJ = MEMG.getNeuronByTempNid($nObj.bes[0]);
        //if (beNOBJ.be_lvl>1 && $tnid_skips.indexOf(beNOBJ.temp_nid)==-1) {
           // this.addFireWdUp(beNOBJ, $firemultiplier, $tnid_skips, $depth+1);
        //}
        
        if ($depth>0) return;
        
        //if its self/parent/bes/or assocs are in the pool.. fire the pool.
        for (var r=0; r < $nObj.outs.length; r++) {
            var rel  = $nObj.outs[r];  
            if (rel.to_type!=TO_TYPE_ANNE) continue;
            if (rel.type!=TYPE_BRAIN) continue;
            
            if (this.ncs_map[rel.t]==undefined) continue;

            var nc = this.ncs[ this.ncs_map[rel.t] ];
            var w=1;
            var relnNOBJ = MEMG.getNeuronByTempNid(rel.r);

            if (relnNOBJ.temp_nid==MEM.ec_opposite.temp_nid || relnNOBJ.temp_nid==MEM.__NOTACTIVE.nid) {
                w=-1;
            }
            else if (relnNOBJ.temp_nid==MEM.__throttlemin.temp_nid) {
                w=-0.33;  
            }
            else if (relnNOBJ.temp_nid==MEM.__throttlemax.temp_nid) {
                w=-1;
            }
            else if (relnNOBJ.temp_nid==MEM.brain_assoc.temp_nid) {
                w=1.3;
            }
            else if (relnNOBJ.temp_nid==MEM.brain_assoc_strong.temp_nid) {
                w=2.5;
            }
            nc.fireMatch(w*$firemultiplier);
        }    
    }
}


class simple_disambig_pool  {
    constructor($poolTitle="", $scp=null) {
        console.log($poolTitle);
        this.scp = new simple_context_pool();
        this.scp._pool = new arraymap();
        this.scp._pool.copy($scp._pool);          
        this.cObjs = [];
    }


    /**
     * Seed the current context pool (from usage) with desired class values
     * @param {*} $candidates 
     * @param {*} $seedNObjs 
     * @param {*} $seedFire 
     */
    disambiguate = function($candidates, $seedNObjs=[], $seedFire=2.0) {
        if ($seedNObjs.length) {
            for (var s=0; s < $seedNObjs.length; s++) {
                if ($seedNObjs[s]==undefined || $seedNObjs[s]==null) continue;
                this.scp.fireConnected($seedNObjs[s].nid, $seedFire);
            }
        }

        var maxFire = -1;
        var winnerNOBJ = null;

        for (var c=0; c <$candidates.length; c++) {
            var cObj = {nObj:$candidates[c], fire:0};
            cObj.fire = this.scp.sumConnected(cObj.nObj);
            this.cObjs.push(cObj);
            if (cObj.fire>maxFire) {
                winnerNOBJ = cObj.nObj;
                maxFire = cObj.fire;
            }
        }
        console.log(this.cObjs);

        return winnerNOBJ;
    }
}


class neuron_core {
    constructor($nObj, $depth=0, $weight=1) {
        this.nObj = $nObj;
        this.depth=$depth;
        this.fire = 0;  //additive by weight here.
        this.parents_nc=[];
        this.weight=$weight;
        this.hits=0;
        this.fstrings = []; 
    }
    
    getName = function() {
        return this.nObj.text;
    }
    addStr = function($str) {
        var str = $str.toLowerCase();
        if (this.fstrings.indexOf(str)==-1) this.fstrings.push(str);
    }
    
    print=function(){
        var parents = [];
        for (var p=0; p < this.parents_nc.length; p++) {
            parents.push(this.parents_nc[p].getName());
        }
        console.log("%c "+this.nObj.text+" %cd:"+this.depth+" hits "+this.hits+"%cw:"+this.weight+" %cf:"+this.fire+" %c"+this.fstrings.join(",")+"%c "+parents.join(","), TALK.bggreen, TALK.bgyellow, TALK.bgblackyellow, TALK.bgred, TALK.bgpink, TALK.bgblue);           
    }
    flowUp=function($sum){
        if (this.depth==0) return;
        this.fire=$sum;
        this.hits+=1;
        if (this.hits>20) {
            console.log("WHAAAAAAAAAT?");
        }
        for (var p=0; p < this.parents_nc.length; p++){
            this.parents_nc[p].flowUp(this.fire);
        }
    }
    fireMatch = function($fireweight) {
        this.hits+=1;
         if (this.hits>20) {
            console.log("WHAAAAAAAAAT?");
        }
        this.fire+=$fireweight+this.weight;
        if (this.depth==0) return;
        for (var p=0; p < this.parents_nc.length; p++){
            this.parents_nc[p].fireMatch($fireweight+this.weight);
        }
    }
}


class ctp_obj {
    constructor($speaker, $nObj, $text, $ci, $hits, $clump_type) {
        this.speaker = $speaker;
        this.nObj = $nObj;
        this.text = $text;
        this.ci = $ci;
        this.hits = 1;
        this.last_ci = $ci;
        this.roles_tally = new Object();
        this.hits_tally = new Object();
        this.posn = 0;
        this.defn = 0;
        this.action = 0;
        this.question = 0;
        this.incrimentTypes($clump_type);
    }

    print = function() {
        var text = this.nObj.bestNameHTML();
        if (text=="") text = this.nObj.nid+"(no name)";
        console.log("%c ci:"+this.ci+"%c hits:"+this.hits+"%c "+this.nObj.text, TALK.bgblack, TALK.bgfire, TALK.green);
    }

    refire=function($ci, $clump_type) {
        this.hits+=1;
        this.last_ci=$ci;
    }
    incrimentTypes=function($t) {
        if ($t=="posn") this.posn+=1;
        else if ($t=="defn") this.defn+=1;
        else if ($t=="action") this.action+=1;
        else if ($t=="question") this.question+=1;
    }


    //for any given nobj (like "planet") we have a speaker and role tally.
    //who is talking about it and in what role capacity
    tallySpeakerAndRole=function($speaker, $role_nObj) {
        if ($speaker==null) return;
        var role = this.roles_tally[$role_nObj.temp_nid];
        if (role==undefined) {
            this.roles_tally[$role_nObj.temp_nid]=[];
            this.roles_tally[$role_nObj.temp_nid].push({speaker:$speaker, hits:1});
        }
        else {
            var aTally = this.roles_tally[$role_nObj.temp_nid];
            for (var t=0; t < aTally.length; t++) {
                var tally = aTally[t];
                if (tally.speaker==$speaker) {
                    tally.hits+=1;
                    this.tallyHitsBySpeaker($speaker);
                    return;
                }
            }

            //new speaker hitting this role.
            aTally.push({speaker:$speaker, hits:1});
            this.tallyHitsBySpeaker($speaker);
        }
    }

    
    tallyHitsBySpeaker=function($speaker) {
        var speaker = this.hits_tally[$speaker];
        if (this.hits_tally[$speaker]==undefined) this.hits_tally[$speaker] = {speaker:$speaker, hits:1};
        else this.hits_tally[$speaker].hits+=1;
    }
}
