class flat_ph_target {

    /**
     * 
     * @param {flat_ph_target} $parent_fpt 
     * @param {anne_neuron} $nObjTarget 
     * @param {Boolean} $bTARGET 
     * @param {String} $from_text 
     * @param {anne_neuron} $baseNOBJ 
     * @param {Number} $weight 
     */
    constructor($parent_fpt, $nObjTarget, $baseNOBJ, $weight, $depth=0) {
        this.parent_fpt             = $parent_fpt;
        this.nObj                   = $nObjTarget; 
        this.base_target_tnids      = [];
        if ($baseNOBJ!=null) {
            this.base_target_tnids  = [$baseNOBJ.temp_nid];
        }
        this.weight                 = $weight; 
        this.depth                  = $depth;
        this.hits                   = 0;
        this.fire                   = 0;
        this.txt_matches            = [];
        this.backlink_pfp           = null; //nlp_phrase_flat_part
        this.text = this.nObj.text;
        this.axnps = [];
        this._uniquePats = [];
        
    }


    /**
     * 
     * @param {*} $matchingNOBJ 
     * @returns 
     */
    fireMatch = function($matchingNOBJ) {
        this.hits+=1;
        this.fire+=this.weight;

        //console.log("(fireMatch) text:"+this.text+" hits:"+this.hits+" fire:"+this.fire)
        
        this.txt_matches.push($matchingNOBJ.text);

        if (this.depth==0) return;

        //if not a depth 0, we want to propagate up.
        if (this.parent_fpt!=null) this.parent_fpt.propagateUp($matchingNOBJ, this.weight);
    }


    /**
     * 
     */
    propagateUp = function($matchingNOBJ, $newFire) {
        this.fire+=$newFire;
        this.hits+=1;
        //console.log("(propagateUp) text:"+this.text+" hits:"+this.hits+" fire:"+this.fire)
        this.txt_matches.push($matchingNOBJ.text);
        if (this.depth==0) return;
        if (this.parent_fpt!=null) this.parent_fpt.propagateUp($matchingNOBJ, $newFire);
    }

    
    /**
     * leading expect candidate processes its pattern neurons
     * we will try to confirm any of these patterns in our incoming wds.
     * @param {*} $patNOBJ 
     */
    processPatternNeuron=function($patNOBJ, $bPRINT=false) {
        if (this._uniquePats.indexOf($patNOBJ.temp_nid)!=-1) return;
        this._uniquePats.push($patNOBJ.temp_nid);
        if ($bPRINT) console.log("%cpat name:%c"+$patNOBJ.text, TALK.bgblue, TALK.bggreen);
        var axnp = new anne_x_neural_pat($patNOBJ);
        //console.log(axnp);
        if ($bPRINT) axnp.printPatPieces();
        this.axnps.push(axnp);//we will use these later.
    }


    /**
     * A leading candidate expection has been isolated.
     * We need to check its known pats and see if we are in one of those known pats.
     * @param {nlp_phrase_flat_part} $flatPh 
     * @param {nlp_wd} $next_wd 
     */
    confirmPat=function($flatPhPart, $next_wd) {
        console.log($flatPhPart);
        console.log($next_wd);

        this.npp = new anne_x_neural_pat_pool(); //not really a pool in this capacity
        var best_status = AXNPP_NO_MATCHES;
        var best = null;
        
        for (var a=0; a < this.axnps.length; a++) {
            var axnp = this.axnps[a];
            var status = this.npp.nextWdPatMatch(axnp, $next_wd);
            
            if (status == AXNPP_MATCH) {
                return {pat:axnp, status:AXNPP_MATCH};
            }
            
            //remove this from our array.. no sense in testing anymore.
            else if (status==AXNPP_NOT_A_MATCH) {
                this.axnps.splice(a,1);
                a--;
                continue;
            }

            else if (status == AXNPP_STILL_ALIVE) {
                best_status = AXNPP_STILL_ALIVE;
                best = {pat:axnp, status:best_status};
            }
        }

        return best;
    }
}

const FPPP_WD_NOT_FOUND = "not_found";
const FPPP_NO_CANDIDATE = "no_candidate";

/**
 * Used as specialized mini-net for intra phrase parsing.
 * Focus is on identifying a phrase for an expectation that was encountered via a signal.
 */
class flat_ph_parse_pool {
    constructor($nlpP){
        this.nlpP = $nlpP;
    }

    initWithFlatPhExpectPool=function($flatPh){
        this.flatPh = $flatPh;

        //loop the expect_pool array and build out these as the top level expections.
        //we will pump words through here as we parse them to look for matches to any of these expectations.
        this._pool = [];
        this._pool_map = {};
        
        this._leading_candidate = null;
        this._greatest_last_fire = 0;

        //.exepect pool is a neuron that is on signals.  It is moved over from the signal to the flat phrase where
        //it is an array with all possible expect_pool neurons.
        if (ANNE.UTILS.isset($flatPh.expect_pool)==true) {
              for (var p=0; p < $flatPh.expect_pool.length; p++) {
                var nObj = $flatPh.expect_pool[p];
                this.addBaseTarget(nObj);
            }      
        } 
    }


    /**
     * Add a base target to the pool.  This type of pool can handle multiple targets in the same
     * pool.  A word like "on" may expect X_WHERE and X_WHEN and subsequent words will be tested against
     * possible matches.
     * @param {*} $nObj 
     */
    addBaseTarget=function($nObj, $weight=1) {
        for (var p=0; p < $nObj.outs.length; p++) {
            var rObj = $nObj.outs[p];

            //we may only need to do TYPE_BRAIN here, let's keep an eye on that.
            if (rObj.type==TYPE_SYS_MSG) continue;
            if (rObj.type==TYPE_SYS) continue;

            //we don't want the parent relon of the target.
            if (rObj.type==TYPE_BE && (rObj.r==MEM.inst.temp_nid || rObj.r==MEM.class.temp_nid)) continue;

            var parent_fpt = this.addUniqueTarget(null, $nObj, $nObj, $weight, 0);
            this.addUniqueTargetFromRelon(rObj, $weight*0.9, $nObj, parent_fpt, 0);
        }
    }


    /**
     * 
     * @param {*} $nObjTarget 
     * @param {*} $weight 
     * @param {*} $baseNOBJ 
     * @param {*} $from_text 
     * @returns 
     */
    addUniqueTarget=function($parent_fpt, $nObjTarget, $baseNOBJ, $weight, $depth=0) {
        if (this._pool_map[$nObjTarget.temp_nid]==undefined) {
            var pObj = new flat_ph_target($parent_fpt, $nObjTarget, $baseNOBJ, $weight, $depth);
            this.__addNewPoolObj(pObj);
        }
        else {
            var pObj = this.lookupPoolObj($nObjTarget.temp_nid);
            if (pObj==null) {
                alert("fix this.. this is in our map!");
                return;
            }

            //make sure this wires back up to the base_target_tnids
            if (pObj.base_target_tnids.indexOf($baseNOBJ.temp_nid)==-1){
                pObj.base_target_tnids.push($baseNOBJ.temp_nid);
            }
            pObj.weight+=$weight;
        }
        return pObj;
    }



    /**
     * Process the relons of the target to build a dart board we can try to hit with our incoming parsed words
     * of the rest of a potential phrase.
     * 
     * @param {*} $relon 
     * @param {*} $weight 
     * @param {*} $baseNOBJ 
     * @param {*} $from_text 
     * @param {*} $parentBase 
     * @returns 
     */
    addUniqueTargetFromRelon = function ($relon, $weight, $baseNOBJ, $parent_fpt=null, $depth=0) {
        var relnNOBJ = MEMG.getNeuronByTempNid($relon.r);
        //this.addUniqueTarget(relnNOBJ, weightR, $baseNOBJ);   
        
        //consider to_type txid here for more complexity.
        if ($relon.to_type!=TO_TYPE_ANNE) return;

        //fire is a summation (not clamped between 0-1)
        //as such, each match is additive.
        //we want some matches to reduce the score
        //we can achieve this by flipping the weight of the match.
        //when matched, the net total will reduce instead of go up.
        if (relnNOBJ.temp_nid==MEM.ec_opposite.temp_nid || relnNOBJ.temp_nid==MEM.__NOTACTIVE.nid) {
            $weight=$weight*-1;
        }
        else if (relnNOBJ.temp_nid==MEM.__throttlemin.temp_nid) {
            $weight=$weight*-.33;  
        }
        else if (relnNOBJ.temp_nid==MEM.__throttlemax.temp_nid) {
            $weight=$weight*-1;
        }
        else if (relnNOBJ.temp_nid==MEM.brain_assoc.temp_nid) {
            $weight=$weight*1.3;
        }
        else if (relnNOBJ.temp_nid==MEM.brain_assoc_strong.temp_nid) {
            $weight=$weight*2.5;
        }

        //the toNOBJ weight is altered by key BRAIN concepts.
        //assocs bump it and throttles are hurtful to matches.
        var toNOBJ = MEMG.getNeuronByTempNid($relon.t);
        this.addUniqueTarget($parent_fpt, toNOBJ, $baseNOBJ, $weight, $depth+1);

        //we really want the ec_patterns here.
        //this will help us catch parts of patterns that tie back our targets while we are parsing.
        for (var p=0; p < toNOBJ.outs.length; p++) {
            var rel = toNOBJ.outs[p];
            if (rel.type!=TYPE_BRAIN) continue;
            if (rel.to_type!=TO_TYPE_ANNE) continue;
            var oneMoreNOBJ = MEMG.getNeuronByTempNid(rel.t);
            this.addUniqueTarget($parent_fpt, oneMoreNOBJ, $baseNOBJ, $weight*.9, $depth+2);

            if (rel.r==MEM.ec_pattern.temp_nid) {
                var patternNOBJ = MEMG.getNeuronByTempNid(rel.t);
                $parent_fpt.processPatternNeuron(patternNOBJ);    
            }
        }
    }


    /**
    */
    lookupPoolObj=function($temp_nid) {
        var pool_index = this._pool_map[$temp_nid];
        return this._pool[pool_index];
    }
    

    /**
     * 
     * @param {*} $wd 
     */
    addWdNeurons=async($flatPh, $next_wd)=> {
        //lookup wd in mem.
        //process any matches of > than x confidence.
        var tnids =[];
        if ($next_wd.charObj!=null) {
            tnids = [$next_wd.charObj.charNOBJ.temp_nid]; 
            $next_wd.tnids = tnids;
        }
        else tnids = $next_wd.tnids; 
        
        //not a match for our purposes here.
        //we are not doing grammar, endings, prefixes, morphs etc here.
        //just looking for quick hits of knowns that may pertain to what we want.

        //NOTE: for an integer we need a reliable and fast way to create a temp integer neuron
        // with all the proper goodies.   We can emulate these quickly during a parse.
        // We need to have atleast that ability to match the class requirement of a potential pattern
        // that allows for numeric slot.
        if (tnids==undefined && tnids==null) {
            return FPPP_WD_NOT_FOUND;    
        }

        //if a hit...
        for (var n=0; n < tnids.length; n++) {
            //check the direct neuron                
            //optional (but nice)
            var nObj = MEMG.getNeuronByTempNid(tnids[n]);
            var pool_obj = this.lookupPoolObj(nObj.temp_nid);
            if (pool_obj!=undefined) {
                pool_obj.fireMatch(nObj);
            }

            //check the 1st gen BEs and BRAINS of our wd...
            for (var r=0; r < nObj.outs.length; r++) {
                var rel = nObj.outs[r];
                
                if (rel.type!=TYPE_BE && rel.type!=TYPE_BRAIN) continue;
                if (rel.to_type!=TO_TYPE_ANNE) continue;
                
                var toNOBJ = MEMG.getNeuronByTempNid(rel.t);
                //console.log("looking for "+toNOBJ.text+" in the pool");
                var pool_obj = this.__getPoolObjByTNID(rel.t);
                if (pool_obj!=undefined) {
                    pool_obj.fireMatch(toNOBJ);
                    pool_obj.backlink_pfp = $flatPh;
                }
            }
        }

        //find top
        this._leading_candidate = this.getTopFire();

        
        if (this._leading_candidate==null) {
            console.log("%c NO FIRE YET from wd "+$next_wd.str+" %c flat_ph_parse_pool FOR flat phrase %c "+this.flatPh.wds_no_punct[0].str, TALK.bgyellow, TALK.bgblack, TALK.bgyellow);
            return FPPP_NO_CANDIDATE;
        }

        if (this._leading_candidate.fire > this._greatest_last_fire) {
            this._greatest_last_fire = this._leading_candidate.fire;
        }
        
        //if > 1 length we are in confirmation mode.
        console.log($flatPh);

        //if we have AXNPP_MATCH, we need to STOP the flat phrase parsing.
        //from here we return the pat status.
        var bestObj = this._leading_candidate.confirmPat($flatPh, $next_wd);
       
        //print for debug.
        //this.printPool("wd "+$next_wd.str); 

        $flatPh.bestPat=bestObj.pat;//this informs us how to make a new object of this pat.

        return bestObj.status;
    }


    /**
     * 
     * @returns 
     */
    getTopFire = function() {
        var topFire = 0;
        var top = null;
        for (var p=0; p < this._pool.length; p++) {
            var pObj = this._pool[p];
            if (pObj.depth!=0) continue;
            if (pObj.hits==0) continue;
            if (pObj.fire > topFire) top = pObj;
        }
        return top;
    }



    /**
     * 
     * @param {*} $pObj 
     * @returns 
     */
    __addNewPoolObj = function($pObj) {
        if ($pObj.nObj==undefined || $pObj.nObj==null) {
            console.log("fix ur pool obj.. it is nulll!");
            return;
        }
        if ($pObj.fire==undefined) $pObj.fire=0;
        this._pool.push($pObj);
        this._pool_map[$pObj.nObj.temp_nid] = this._pool.length-1;
    }


    /**
     * 
     * @param {*} $temp_nid 
     * @returns 
     */
    __getPoolObjByTNID = function($temp_nid) {
        var poolIndex = this._pool_map[$temp_nid];
        if (poolIndex==undefined) return null;
        return this._pool[poolIndex];
    }



    /**
     * 
     * @param {*} $sortBy 
     */
    printPool = function($msg, $sortBy=DP_SORT_BY_FIRE) {
        function getArrStr($arr){
            var str_arr = [];
            for (var a=0; a<$arr.length; a++) { 
                var nObj = MEMG.getNeuronByTempNid($arr[a]);
                str_arr.push(nObj.text);
            }
            return str_arr.join(",")
        }
        console.log("%c "+$msg+" %c flat_ph_parse_pool FOR flat phrase %c "+this.flatPh.wds_no_punct[0].str, TALK.bgyellow, TALK.bgblack, TALK.bgyellow);
        var sortable = [];
        for (var p=0; p < this._pool.length; p++) {
            var pObj = this._pool[p];
            sortable.push(pObj);
        }

        if ($sortBy==DP_SORT_BY_FIRE) sortable.sort(anne_nlp._sortMatchObjByFire);
        else if ($sortBy==DP_SORT_BY_HITS) sortable.sort(anne_nlp._sortMatchObjByHits);


        for (var s=0; s < sortable.length; s++) {
            var sObj = sortable[s];
            var baseTarget = getArrStr(sObj.base_target_tnids);

            var from_text = "";
            if (sObj.depth==0) from_text = "";
            else {
                from_text = sObj.parent_fpt.nObj.text;
            }
            console.log("%c target %c'"+baseTarget+" %c"+sObj.nObj.text+" %cf:"+sObj.fire+"%cd:"+sObj.depth+" w:"+sObj.weight+" %cfrom:"+from_text+" %c hits:"+sObj.hits, TALK.bgblue, TALK.bgyellow, TALK.bggreen, TALK.bgred, TALK.bgblack, TALK.bgblue, TALK.bgred);
            
            //else console.log("%c target %c "+baseTarget+" %c"+sObj.nObj.text+"%cd:"+sObj.depth+" weight:"+sObj.weight, TALK.bglightblue, TALK.bgyellow, TALK.bggreen, TALK.bggray)
        }
    }
}
































/**
 * We have multiple possible valid resolutions and we need to disambiguate based on 
 * one ore more concepts.
 * 
 * Is "in February" a WHERE role or a WHEN role?   How about "in Las Vegas"?
 * Both use the prep "in" and are valid phrases that could be either.
 * 
 * We will load a target pool with concepts that are connected to the concepts.
 * We will then traverse through our candidatees and see if we can connect 
 * to anything in the target pool.  If we do we score the candidates.
 * highest candidate wins.
 * 
 * Firestarter is the base weight.  If we matched a triger word (like "to") it can be 100 (fully confident in the match).
 * A partial match will come in with lower than 100 and matching it with confidence and matching candidates to 
 * it will be more difficult even when they hit.
 */
const DP_SORT_BY_FIRE = 1;
const DP_SORT_BY_HITS = 2;

class disambig_pool {
    constructor($NLP, $poolNOBJ, $firestarter=1) {
        this.poolNOBJ = $poolNOBJ;
        this._pool = [];
        this._pool_map=[];
        this._maxCandidateGen = 0;
        this._baseCandidates = [];
        this._baseCandidatesNOBJS = [];
        this._backup_candidate_symbols = [];
        this._roleNOBJ = null;
        this._partNOBJ = null;
        this.seed_relons = null;
        this.ROLE_PART_ALIGNMENT=false;
        this.ROLE_PART_ALIGNMENT_WEIGHT = 0;
        this.NLP = $NLP;

        if ($poolNOBJ!=null) { 
            console.log("MAKING A DISAMBIG POOL FOR *******"+$poolNOBJ.text+"******");
            var pObj = {nObj:$poolNOBJ, target:true, from_text:"<<POOLTOP>>", candidate:false, hits:0, weight:$firestarter, base_target:null};
            this._pool.push(pObj);
            this._pool_map[$poolNOBJ.temp_nid] = 0;
        }
    }
    

    /** 
     * Loop the neurons in the pool and build out a target list of concepts we are trying to fire.
     * Add candidates to the pool and chase their relons attempting to connect to our target:true neurons in the pool.
     * note we may add a chase to this to get more info from the ANNODE
     * 
     * @param {Array} - an array of relon TYPES that will be chased.  Typically it is TYPE_BRAIN in this type of process. 
     *
     * @author {Mr.scatman}
     */
    buildTargetPool = async($types, $roleNOBJ=null, $langPart=null)=> {
        if (Array.isArray($types)==false) {
            console.log("TYPES MUST BE AN ARRAY");
            return;
        }

        for (var n=0; n < this._pool.length; n++) {
            var nObj = this._pool[n].nObj;
            var weight = this._pool[n].weight;
            //loop pool and fire the neurons in it for the outs of $type
            //put relns and tos in the pool.
            for (var p=0; p < nObj.outs.length; p++) {
                var rObj = nObj.outs[p];
                if (rObj.type==TYPE_SYS_MSG) continue;
                if (rObj.type==TYPE_SYS) continue;
                if ($types.indexOf(rObj.type)==-1) continue;
                this.addUniqueTargetFromRelon(rObj, weight*0.9, nObj, nObj.text);
            }
        }

        if ($roleNOBJ!=null || $langPart!=null) this.augmentTargetPoolByRoleAndPart($roleNOBJ, $langPart, $types);

        console.log(this._pool);
    }


    /**
     * We are extending the targets to chase their children lines.
     * We use this to attempt a far looser "symbol" partial match.
     */
    extendAllTargetChildren=function($maxChildren=200) {
        for (var n=0; n < this._pool.length; n++) {
            var poolObj = this._pool[n];
            if (poolObj.target==false) continue;
            var poolObj = poolObj.nObj;
            var children = [];
            poolObj.getAllDownlineChildren(children, 0, true, false, true, false);
            for (var c=0; c < Math.min(children.length, $maxChildren); c++) {
                var childNOBJ = children[c];
                this.addUniqueTarget(childNOBJ, 1, poolObj);   //relns aren't important to match here. 
            }
        }
    }


    /**
     * here we look for anything that is a target being inside
     * a bigger string that is in the candiates.
     */
    firePartials=function() {
        var candidates=[];
        var targets=[];
        for (var n=0; n < this._pool.length; n++) {
            var obj = this._pool[n];
            if (obj.target==true) targets.push(obj);
            else candidates.push(obj);
        }
        for (var t=0; t < targets.length; t++) {
            var targetPoolObj = targets[t];
            var targetText = targetPoolObj.nObj.text;
            if (targetText.length<2) continue;
            if (targetText=="day") {
                console.log("test day here");
            }
            for (var c=0; c < candidates.length; c++) {
                var candObj = candidates[c];
                if (candObj.text.indexOf(targetText)!=-1) {
                    candObj.fire+=(targetPoolObj.weight);
                    candObj.hits+=1;
                    targetPoolObj.hits+=1;
                }
            }
        }
    }



    /**
     * We load a pool up with relons in the case where 
     * we are going to use this pool to copy from and test wider options.
     * ie, the action morph gens a pool and we can test specific things about
     * the sentence.
     */
    buildTargetPoolWithRelons=function($relons) {
        this.seed_relons = $relons;
        var nObj = this._pool[0].nObj;
        var weight = this._pool[0].weight;
        for (var r=0; r < $relons.length; r++) {
            var rObj = $relons[r];
            if (rObj.type==TYPE_SYS_MSG) continue;
            if (rObj.type==TYPE_SYS) continue;
            var fromNOBJ = MEMG.getNeuronByTempNid(rObj.f);
            this.addUniqueTargetFromRelon(rObj, weight*0.9, nObj, fromNOBJ.text);
        }
        this.printPool(DP_SORT_BY_HITS);
    }

    
    /**
     * We need to augment the pool match not just by concept we want...
     * but also by any throttles/weights we have associated with this role
     */
    augmentTargetPoolByRoleAndPart = function($roleNOBJ=null, $partNOBJ=null, $types=[]) {
        this._roleNOBJ =  $roleNOBJ;
        this._partNOBJ =  $partNOBJ;

        var nObj = this._pool[0].nObj;
        var weight = this._pool[0].weight;

        if (this._roleNOBJ != null) {
            var outs = this._roleNOBJ.outs;
            for (var r=0; r < outs.length; r++) {
                var rObj = outs[r];
                if (rObj.type==TYPE_SYS_MSG) continue;
                if (rObj.type==TYPE_SYS) continue;
                if ($types.indexOf(rObj.type)==-1) continue;
                this.addUniqueTargetFromRelon(rObj, weight*1.9, nObj, this._roleNOBJ.text);
            }
        }

        if (this._partNOBJ!=null) {
            var outs = this._partNOBJ.outs;
            for (var r=0; r < outs.length; r++) {
                var rObj = outs[r];
                if (rObj.type==TYPE_SYS_MSG) continue;
                if (rObj.type==TYPE_SYS) continue;
                if ($types.indexOf(rObj.type)==-1) continue;
                
                var relnNOBJ = MEMG.getNeuronByTempNid(rObj.r);
                var toNOBJ = MEMG.getNeuronByTempNid(rObj.t);
                
                //if this part includes the role we are looking at,
                //bump it up by the relnNOBJ weight.
                var weight = 0;
                if (toNOBJ.temp_nid==this._roleNOBJ.temp_nid) {
                    if (relnNOBJ.temp_nid==MEM.brain_assoc.temp_nid) {
                        weight=1.3;
                    }
                    else if (relnNOBJ.temp_nid==MEM.brain_assoc_strong.temp_nid) {
                        weight=2.5;
                    }
                    this.ROLE_PART_ALIGNMENT=true;
                    this.ROLE_PART_ALIGNMENT_WEIGHT=weight;
                    return;
                }
            }
        }
    }


    /**
     */
    bumpAllCandidates=function() {
        for (var p=0; p < this._pool.length; p++) {
            var pObj = this._pool[p];
            if (pObj.candidate==true) {
                pObj.weight+=this.ROLE_PART_ALIGNMENT_WEIGHT;
                //pObj.hits+=1;
            }
        }
    }



    /**
     * 
     * @param {*} $relon 
     * @param {*} $weight 
     * @param {*} $baseNOBJ 
     * @param {*} $from_text 
     * @returns 
     */
    addUniqueTargetFromRelon = function ($relon, $weight, $baseNOBJ, $from_text="") {
        var weightR = $weight;
        
        //relon neurons are weighted differently if they are ANNE SYS or SYS BRAIN.
        //those are protocol level values not semantic.
        var relnNOBJ = MEMG.getNeuronByTempNid($relon.r);
        if (relnNOBJ.anyParentHasBeOf(MEM.__ANNE_SYS_CLASS.temp_nid)==true) {
            //skip the reln -- we don't need it in the pool. we do want the TO S tho!
        }
        else if (relnNOBJ.anyParentHasBeOf(MEM.__SYSBRAIN.temp_nid)==true) {
            //skip the reln -- we don't need it in the pool. we do want the TO S tho!
        }

        //the reln of the type be is class/inst or ec_is.  This shouldn't be 
        //helping us in the pool
        else {
            if ($relon.type==TYPE_BE) weightR = weightR/10;
            this.addUniqueTarget(relnNOBJ, weightR, $baseNOBJ);   //relns aren't important to match here. 
        }

        //consider to_type txid here for more complexity.
        if ($relon.to_type!=TO_TYPE_ANNE) return;

        //fire is a summation (not clamped between 0-1)
        //as such, each match is additive.
        //we want some matches to reduce the score
        //we can achieve this by flipping the weight of the match.
        //when matched, the net total will reduce instead of go up.
        if (relnNOBJ.temp_nid==MEM.ec_opposite.temp_nid || relnNOBJ.temp_nid==MEM.__NOTACTIVE.nid) {
            $weight=$weight*-1;
        }
        else if (relnNOBJ.temp_nid==MEM.__throttlemin.temp_nid) {
            $weight=$weight*-.33;  
        }
        else if (relnNOBJ.temp_nid==MEM.__throttlemax.temp_nid) {
            $weight=$weight*-1;
        }
        else if (relnNOBJ.temp_nid==MEM.brain_assoc.temp_nid) {
            $weight=$weight*1.3;
        }
        else if (relnNOBJ.temp_nid==MEM.brain_assoc_strong.temp_nid) {
            $weight=$weight*2.5;
        }

        //the toNOBJ weight is altered by key BRAIN concepts.
        //assocs bump it and throttles are hurtful to matches.
        var toNOBJ = MEMG.getNeuronByTempNid($relon.t);
        this.addUniqueTarget(toNOBJ, $weight, $baseNOBJ, $from_text);
    }


    /**
     * 
     * @param {*} $nObjTarget 
     * @param {*} $weight 
     * @param {*} $baseNOBJ 
     * @param {*} $from_text 
     * @returns 
     */
    addUniqueTarget=function($nObjTarget, $weight, $baseNOBJ, $from_text="") {
        //console.log("adding unique TARGET neuron ***"+$nObjTarget.text+"****");
        if (this._pool_map[$nObjTarget.temp_nid]==undefined) {
            var pObj = {nObj:$nObjTarget, target:true, candidate:false, from_text:$from_text, base_target:$baseNOBJ, weight:$weight, hits:0};
            this.__addNewPoolObj(pObj);
        }
        else {
            var pObj = this.__getPoolObjByTNID($nObjTarget.temp_nid);
            if (pObj==null) {
                alert("fix this.. this is in our map!");
                return;
            }
            pObj.weight+=$weight;
        }
    }


    /**
     * 
     * @param {*} $candidateMatchObjs 
     * @param {*} $bCHASEPARENTLINE 
     * @param {*} $types 
     */
    addBaseCandidates=async($candidateMatchObjs, $bCHASEPARENTLINE, $types)=> {
        console.log("%c(addBaseCandidates) "+$candidateMatchObjs.length+" candidates", 'background: white; color: green');
        console.log($candidateMatchObjs);
        for (var t=0; t < $candidateMatchObjs.length; t++) {
            var candidateMatchObj = $candidateMatchObjs[t];
            var nObj = candidateMatchObj.nObj;
            var mFire = candidateMatchObj.fire;
            
            if (this._pool_map[nObj.temp_nid]!=undefined) {
                //console.log("skipping add to pool, we already have this");
                continue;  //already have this.
            }

            //console.log("candidate: '"+nObj.text+"' nid:"+nObj.nid, nObj);

            //fire of match is the initial weighting of base candidate.
            var gen = 0;
            var candidatePObj = {
                nObj:nObj, 
                text:nObj.text.toLowerCase(), 
                weight:mFire, 
                gen:gen, 
                base_candidateTNIDS:[nObj.temp_nid], 
                target:false, 
                candidate:true, 
                hits:0, 
                fire:0,
                matchObj:candidateMatchObj};

            this._baseCandidates.push(candidatePObj);
            this._baseCandidatesNOBJS.push(nObj);

            this.__addNewPoolObj(candidatePObj);

            //console.log("get runtime relons (all inherited be line) for '"+candidateMatchObj.nObj.text+"'");
            var relons = await MEMG.getRuntimeRelonsForContext(candidateMatchObj.nObj.nid, $types);
            //console.log("found runtime '"+relons.length+"' relons",relons);
            for (var r=0; r < relons.length; r++) {
                var rel = relons[r];
                this.addUniqueCandidateGenFromRelon(rel, 1, nObj.temp_nid, gen+1)
            }

            //we have a prefix/suffix (known) on this word
            //we can use that to help us get a better match in the pool
            //from the role this._roleNOBJ
            if (candidateMatchObj.psbl_prefix!=undefined && candidateMatchObj.psbl_prefix.length>0) {
                for (var s=0; s < candidateMatchObj.psbl_prefix.length; s++) {
                    var pre = candidateMatchObj.psbl_prefix[s];
                    this.addUniqueCandidate(pre.nObj, null, 1, nObj.temp_nid, gen+1);
                    var dpool_adders = pre.dpool_adders;
                    if (dpool_adders!=undefined) this.addUniqueCandidateByArr(dpool_adders, null, 1, nObj.temp_nid, gen+1);

                }
            }
            if (candidateMatchObj.psbl_suffix!=undefined && candidateMatchObj.psbl_suffix.length>0) {
                for (var s=0; s < candidateMatchObj.psbl_suffix.length; s++) {
                    var suf = candidateMatchObj.psbl_suffix[s];
                    this.addUniqueCandidate(suf.nObj, null, 1, nObj.temp_nid, gen+1);
                    var dpool_adders = suf.dpool_adders;
                    if (dpool_adders!=undefined) this.addUniqueCandidateByArr(dpool_adders, null, 1, nObj.temp_nid, gen+1);
                }
            }

        }

        //juice up this role+matches combo (all matches in this pool will benefit)
        //this is saying "this is the right role for this language part (subj/obj)"
        if (this.ROLE_PART_ALIGNMENT==true) {
            this.bumpAllCandidates();
        }

    }


    /**
     * For candidates the relons are more relevant to match against the target since they can serve
     * as properties.   (  date->month->3 )
     */
    addUniqueCandidateGenFromRelon = function ($relon, $weight, $candidateTNID, $gen=1) {
        //console.log("(addUniqueCandidateGenFromRelon) adding relon", $relon, "weight "+$weight, "candidateTNID:"+$candidateTNID);
        var weightR = $weight;
        var relnNOBJ = MEMG.getNeuronByTempNid($relon.r);

        //we dont need targets downstream that are anne sys neurons. These are sys level relons
        //used for doing protocol-level thingies.
        if (relnNOBJ.anyParentHasBeOf(MEM.__ANNE_SYS_CLASS.temp_nid)==true) {
            //skip the reln -- we don't need it in the pool. we do want the TO S tho!
        }
        else if (relnNOBJ.anyParentHasBeOf(MEM.__SYSBRAIN.temp_nid)==true) {
            //skip the reln -- we don't need it in the pool. we do want the TO S tho!
        }
        else this.addUniqueCandidate(relnNOBJ, $relon, $weight, $candidateTNID, $gen);
         
        //flip polarity here if RELN is MEM.ec_opposite, MEM.__NOTACTIVE
        //if we have a requirement of "MALE" then a neuron assoc'd or TYPE_BE with FEMALE should be throttled. 
        if (relnNOBJ.temp_nid==MEM.ec_opposite.temp_nid || relnNOBJ.temp_nid==MEM.__NOTACTIVE.nid) {
            $weight=$weight*-1;
        }
        else if (relnNOBJ.temp_nid==MEM.__throttlemin.temp_nid) {
            $weight=$weight*-.33;  
        }
        else if (relnNOBJ.temp_nid==MEM.__throttlemax.temp_nid) {
            $weight=$weight*-1;
        }
        else if (relnNOBJ.temp_nid==MEM.brain_assoc.temp_nid) {
            $weight=$weight*1.3;
        }
        else if (relnNOBJ.temp_nid==MEM.brain_assoc_strong.temp_nid) {
            $weight=$weight*2.5;
        }

        //here we would get a text string of an alias..
        //or any other "dead-end" value that is not encoded in a relon.
        if ($relon.to_type!=TO_TYPE_ANNE) {
            this._backup_candidate_symbols.push({base_candidateTNIDS:[$candidateTNID], chunk:$relon.to_s, weight:$weight});
            return;
        }
        var toNOBJ = MEMG.getNeuronByTempNid($relon.t);
        this.addUniqueCandidate(toNOBJ, $relon, $weight, $candidateTNID, $gen);
    }

    
    /**
     * Quick add by array
     */
    addUniqueCandidateByArr = function($arr, $relon, $weight, $candidateTNID, $gen) {
        for (var c=0; c < $arr.length; c++) {
            this.addUniqueCandidate ( $arr[c], $relon ,$weight, $candidateTNID, $gen);
        }
    }


    /**
     * If this connects to any of the targets we want to incriment the fire of the base_candidateTNID.
     * This is essentially a gen-by-gen (or lvl by lvl for NNs) firing mechanism.
     * We can go as many levels deep as we like.
     */
    addUniqueCandidate = function($nObj, $relon, $weight, $candidateTNID, $gen) {
        if (this._pool_map[$nObj.temp_nid]==undefined) {
            var pObj = {nObj:$nObj, text:$nObj.text.toLowerCase(), weight:$weight, gen:$gen, candidate:true, target:false, base_candidateTNIDS:[$candidateTNID], hits:0};
            pObj.from_rel_type = $relon.type;
            this.__addNewPoolObj(pObj);
        }

        //existing neuron in the pool
        else {
            var pObj = this._pool[ this._pool_map[$nObj.temp_nid] ];
            if (pObj.target==true) {
                //console.log("%c existing neuron '"+$nObj.text+"' is a TARGET!!", 'background: ivory; color: green');
                pObj.from_rel_type = $relon.type;
                
                //mark this also as a candidate.
                pObj.candidate=true;

                //we fire the candidate that got this into the pool when we match 
                //up with an existing target in the pool (same concept)
                var candidatePObj = this.__getPoolObjByTNID($candidateTNID) 
                candidatePObj.fire+= ($weight * pObj.weight);
                candidatePObj.hits+=1;
                pObj.hits+=1;
            }

            //this is just an existing candidate neuron
            else {
                            
                //this is for candidates used for gen firings of next levels.
                //we want to know that this existing neuron is now in multiple candidate circuits.
                //if the neuron is a target, it is ALWAYS listed as target:true candidate:false first,
                //even as a candidate --- since that is how we match.  We match candidates to targets.
                //we still may need to fire another level and do the "next gen"
                if (pObj.base_candidateTNIDS.indexOf($candidateTNID)==-1) pObj.base_candidateTNIDS.push($candidateTNID);
                
                pObj.weight+=$weight;   //no guarantee of order of this before it propogates.. but want it to test for print
            }
        }

        if ($gen > this._maxCandidateGen) this._maxCandidateGen=$gen;
    }

    getFRTBe=function(){
        var nObjs = [];
        for (var p=0; p < this._pool.length; p++) {
            var pObj = this._pool[p];
            if (pObj.from_rel_type!=undefined && pObj.from_rel_type==TYPE_BE) {
                nObjs.push(pObj.nObj);
            }
        }
        console.log("(getFRTBe)",nObjs);
        return nObjs;
    }


    printPool = function($sortBy=DP_SORT_BY_FIRE) {
        //we need to print a list of candidate gen 0s.
        var candlist = ANNE.MEMG.getTextListStrFromNObjs(this._baseCandidatesNOBJS);
        var forRole = "";
        if (this._roleNOBJ!=null) forRole = "["+this._roleNOBJ.text+"]";

        var parttext = "";
        if (this._partNOBJ!=null) parttext = this._partNOBJ.text;

        var rpacolor = TALK.bggreen;
        if (this.ROLE_PART_ALIGNMENT==false) rpacolor = TALK.bgred;

        console.log("%c DP POOL FOR %c'"+this.poolNOBJ.text+"' "+forRole+" %c '"+candlist+"' %c <<"+parttext+" ROLE_PART_ALIGNMENT ("+this.ROLE_PART_ALIGNMENT+")>>", TALK.bgred, TALK.bgyellow, TALK.bgblack, rpacolor)
        var sortable = [];
        for (var p=0; p < this._pool.length; p++) {
            var pObj = this._pool[p];
            sortable.push(pObj);
        }
        if ($sortBy==DP_SORT_BY_FIRE) sortable.sort(anne_nlp._sortMatchObjByFire);
        else if ($sortBy==DP_SORT_BY_HITS) sortable.sort(anne_nlp._sortMatchObjByHits);

        for (var s=0; s < sortable.length; s++) {
            var sObj = sortable[s];
            var frt = sObj.from_rel_type;
            if (frt==undefined) frt="m/a";
            if (sObj.target==true) { 
                if (sObj.hits>0) console.log("%c target %c '"+sObj.nObj.text+"' weight:"+sObj.weight+" %cfrom:["+sObj.from_text+"] %c hits:"+sObj.hits+" frt:"+frt, TALK.bgblue, TALK.bgblack, TALK.bgblue, TALK.bggreen)
                else console.log("%c target %c '"+sObj.nObj.text+"' weight:"+sObj.weight+" frt:"+frt, TALK.bglightblue, TALK.bggray)
            }
            else if (sObj.candidate==true) {
                var froms = this._genTextListFromTNIDS(sObj.base_candidateTNIDS);
                if (sObj.fire>0) console.log("%c candidate %c '"+sObj.nObj.text+"' %c fire:"+sObj.fire+" weight:"+sObj.weight+" hits:"+sObj.hits+" %c from:"+froms+" frt:"+frt, TALK.bggreen, TALK.bgblack, TALK.bgfire, TALK.bgblack);
                else console.log("%c candidate %c '"+sObj.nObj.text+"' fire:"+sObj.fire+" weight:"+sObj.weight+" hits:"+sObj.hits+" %c from:"+froms+" frt:"+frt, TALK.bglightgreen, TALK.bggray, TALK.bgblack);
            }
        }
    }
    _genTextListFromTNIDS = function($tnids) {
        if ($tnids==undefined || $tnids==null || $tnids.length==0) return "";
        var str = [];
        for (var t=0; t <$tnids.length;t++) {
            var nObj = MEMG.getNeuronByTempNid($tnids[t]);
            if(nObj==null) continue;
            str.push(nObj.text);
        }
        if (str.length==0) return "";
        else if (str.length==1) return str[0];
        return str.join(", ");
    }


    /**
     */
    getCandidatesWithFire=function($bHITSMULTIPLIER=true){
        this.printPool();
        var candidates_with_fire=[];
        for (var p=0; p < this._pool.length; p++) {
            var pObj = this._pool[p];
            if (pObj.candidate==false) continue;
            if (pObj.fire>0) {
                if ($bHITSMULTIPLIER) pObj.fire = pObj.fire+(pObj.hits*2);
                candidates_with_fire.push(pObj);
            }
        }
        if (candidates_with_fire.length) candidates_with_fire.sort(anne_nlp._sortMatchObjByFire);
        console.log(candidates_with_fire);
        return candidates_with_fire;
    }

    __hasWeightInPool = function($nObj) {
        if (this._pool_map[$nObj.temp_nid]!=undefined) {
            var poolIndex  = (this._pool_map[$nObj.temp_nid]);
            var poolObj = this._pool[poolIndex];
            console.log(poolObj);
            return poolObj.weight;
        }
        return 0;
    }

    __addNewPoolObj = function($pObj) {
        if ($pObj.nObj==undefined || $pObj.nObj==null) {
            console.log("fix ur pool obj.. it is nulll!");
            return;
        }
        if ($pObj.fire==undefined) $pObj.fire=0;
        this._pool.push($pObj);
        this._pool_map[$pObj.nObj.temp_nid] = this._pool.length-1;
    }

    __getPoolObjByTNID = function($temp_nid) {
        var poolIndex = this._pool_map[$temp_nid];
        if (poolIndex==undefined) return null;
        return this._pool[poolIndex];
    }
}