class anne_gui_role {
    constructor($el, $rel, $guid, $bEdit=false, $bAddmod=false, $el_inject="", $newcb=null, $cancelcb=null) {
        this.el = $el, 
        this.rel = $rel;
        this.ID = $guid;
        this.bEDIT = $bEdit;
        this.bADDMOD = $bAddmod;
        this.el_inject = $el_inject;
        this.cb_new = $newcb;
        this.cb_cancel = $cancelcb;
    }

    /**
     * @param {Boolean} $bEDIT - can be changed
     */
    set edit ($bEDIT) {
        this.bEDIT = $bEDIT;
    }

    /**
     * @param {function} $cb - callback for new
     */
    set cb_new ($cb) {
        this.cb_new = $cb;
    }

    /**
     * @param {function} $cb - callback for cancel
     */
    set cb_cancel ($cb) {
        this.cb_cancel = $cb;
    }

    buildGui = function() {
        ANNE.A.ezRoleEdit(this);
    }
}



/**
 * GHOST CLUMPS are intended to be created from an existing NEURON with the ROLE_ACTION preset.
 * The action role cannot be changed in the clump as it assigns (opens up) the other roles that are available in the clump.
 * 
 * A ghost clump will be automatically created when opening up a clump with ROLE_ACTION + other neuron.
 * That ghost clump can be edited here as well.
 * 
 * Any new neurons that need to be created to support your clump will be also available from here.
 * The proper forms/handlers will inject and allow for selecting whatever is needed.
 * 
 * All items are GHOSTS and will be upgraded to real neurons when the user chooses to uprade their neuron.
 * 
 * NOTE:  in order for ANNE to HANAI a clump, all the parts of the clump must be either in ANNE's worldview or
 * be available for HANAI too.
 */
 class service_ghost_clump  {
    constructor ($pgui) {
        if ($pgui==null) return;
        this.pgui = $pgui;
        this.ID = $pgui._guid;
        if (this.ID==undefined) return;
        this.pObj = $pgui.pObj;
        if (this.pObj==null) return;

        this.misc = this.pObj.misc;
        if (this.misc==undefined || this.misc===null) return;
        console.log(this.misc);

        if (this.pObj.s_cbk) this.success_callback = this.pObj.s_cbk; 
        
        this.el_dynamic_cancel = document.getElementById("dynamic_cancel_"+this.pgui._guid);
        if (typeof this.pgui.pObj.c_cbk==="function") {
            ANNE.A.ezSquareDelete(this.el_dynamic_cancel, this.cancelThisForm, [this.pgui], "36px", "nevermind");
        }

        this.btn_parse = document.getElementById("btn_parse_"+this.ID);
        this.btn_parse.addEventListener("click", e=>this.parse(e));

        this.clump_nObj = null;
        this.clump_nid=this.pObj.p1;
        this.action_nid = null;
        this.actionNOBJ=null;  //extract this from the clump
            
        if (this.clump_nid!=null) {
            this.clump_nObj = MEMG.getNeuronByNid(this.clump_nid);

            //we are in edit mode.
            if (this.clump_nObj!=null) {
                this.actionNOBJ = MEMG.getNeuronByTempNid(this.clump_nObj.bes[0]);
            }
        }
        else {
            this.action_nid = this.pObj.p2;
            if (this.action_nid!=null) {
                this.actionNOBJ = MEMG.getNeuronByNid(this.action_nid);
            }
            else {
                alert("action nObj must be set if the clumpNOBJ is not set!");
                this.actionNOBJ=null;  //extract this from the clump
                return;
            }
        }

        this.misc = this.pObj.misc; //used for .default_roles on new clumps
        
        if (this.misc.ntype!=undefined) this.ntype = this.misc.ntype;
        else this.ntype = NTYPE_INST;
        
        this.contextNOBJ = null;
        if (this.misc.obj_role_context_nid!=undefined){
            this.contextNOBJ = MEMG.getNeuronByNid(this.misc.obj_role_context_nid);
            PAGES.updateLastHistMiscProp("obj_role_context_nid", this.misc.obj_role_context_nid);
        }        
        
        this.bNEW = this.misc.bNEW;

        //not to be confused with the default roles an action assigns.
        //these are default VALUES of those roles we want to prepopulate in our gui.
        //the caller of this class sets those as it likes to help the user.
        this.default_roles_values = this.misc.default_roles;

        this.el_make_new_ghost = document.getElementById("make_new_ghost_btn_"+this.ID);
        ANNE.A.hide(this.el_make_new_ghost);
        this.el_upgrade_ghost = document.getElementById("upgrade_ghost_neuron_btn_"+this.ID);
        this.el_upgrade_ghost.addEventListener("click", e=>this.upgradeGhostClump(e));
        ANNE.A.hide(this.el_upgrade_ghost);

        this.el_delete_ghost_btn =  document.getElementById("delete_ghost_btn_"+this.ID);
        ANNE.A.hide(this.el_delete_ghost_btn);
        this.el_delete_ghost_btn.addEventListener("click", e=>this.deleteGhostClump(e));
        
        this.el_inject_form = document.getElementById("inject_form_"+this.ID);
        this.el_roles_inject = document.getElementById("roles_inject_"+this.ID);

        this.el_parse_results = document.getElementById("parse_results_"+this.ID);
    }
    
    /**
     * 
     * @param {*} $e 
     */
    parse = async($e)=>{
        var text = ANNE.A.getValue("sentence_input_"+this.ID);
        console.log("parse "+text);

        //we know the ACTOR.
        //we know the action.

        //parse all the pieces and variants to give us all the options of the parse.
        //let the user pick which pieces are correct to build the properly conceptualized sentence.

        ANNE.A.empty(this.el_parse_results);
    }


    /**
     * We don't want the dossier page to auto launch this event view if we go back to this page later.
     * We DO want that if we click a neuron from this view and are taken away from it.
     * @param {*} $e 
     */
    cancelThisForm = async($e)=>{
        PAGES.updateLastHistMiscPropDelete("event_clump_nid");
        PAGES.updateLastHistMiscPropDelete("obj_role_context_nid");
        this.pgui.pObj.c_cbk($e);   
    }

    /**
     * 
     * @param {page_gui} $pgui 
     */
    initASYNC = async($pgui)=>{
        console.log(this.default_roles_values);

        //we cannot have no clump and no action here.
        if (this.clump_nObj==null && this.actionNOBJ==null) return;


        //we are in edit mode.
        if (this.clump_nObj!=null && this.actionNOBJ==null) {
            console.log("this is a ghost clump, we should have the parent for it.  No searching will help.")
            return;
            //we have ot try to chase the the clump and get the action (parent)
            //var types = [TYPE_SYS, TYPE_BE, TYPE_PARAM, TYPE_HAS, TYPE_CLUMP, TYPE_ROLE];
            //var done = await CHASE.queryForOuts(this.clump_nObj, true, types)
            //this.actionNOBJ = MEMG.getNeuronByTempNid(this.clump_nObj.bes[0]);
        }


        this.talkObj = await this.actionNOBJ.processActionNeuron();


        //create a temp neuron in memory.  THIS IS NOT YET A GHOST!  It is a temp=true neuron.
        if (this.clump_nObj===null) {
            var temp_name = "action:";
            if (this.actionNOBJ!=null) temp_name += " "+this.actionNOBJ.text;
            if (this.contextNOBJ!=null) temp_name += " "+this.contextNOBJ.shortName();
            this.clump_nObj = await ANNE.GHOSTDATA.makeGhostNeuron(this.actionNOBJ, temp_name, NTYPE_INST);
            console.log(this.clump_nObj);
            await this.clump_nObj.processActionNeuron();

            this.buildDefaultGui();

            //note 
            //presumed WHERE mod -- at.
            //presumed WHEN mod -- on.
            //if different, mod role must be present.  
            //examples: ([mod_where]behind [where]The Great Sphinx..)  ([mod_when]before [when]noon)
            for (var d=0; d < this.default_roles_values.length; d++) {
                var dr = this.default_roles_values[d];
                console.log(dr);
                //role_text, reln, to_s, to_type
                this._initGhostRoleValue(dr.reln, dr.to_s, dr.to_type);
            }

            //if action has no v2 v3 v4 or v5 set  we presume it is v1.
            //if it is v1, we want to lookup the v2.  If it exists, use that action.
            //if it does NOT exist, we want to add an auto ROLE_TENSE past to the ghost clump.
        }

        else {
            await this.clump_nObj.processActionNeuron();
            console.warn("************* ROLES **************")
            console.log(this.clump_nObj._roles);
            
            this.buildDefaultGui();

            //update default gui with existing values.
            var processed_txids = [];
            for (var ro=0; ro < this.clump_nObj._roles.length; ro++) {
                var role_rel = this.clump_nObj._roles[ro];
                if (processed_txids.indexOf(role_rel.txid)!=-1) continue;
                this.displayExistingRoleValue(role_rel.reln, role_rel.to_s, role_rel.to_type, role_rel.txid);
                processed_txids.push(role_rel.txid);
            }
        }

        PAGES.updateLastHistMiscProp("event_clump_nid", this.clump_nObj.nid);
        
        this.__assessBtns();
        await this.displaySentence();
    }


    
    /**
     * 
     */
    buildDefaultGui = function() {
        console.log("shift lang "+this.actionNOBJ.shift_lang);
        console.log("tense "+this.actionNOBJ.tense);
        var bTENSENEEDED=true;
        if (this.actionNOBJ.shift_lang==false && this.actionNOBJ.tense!="present tense") bTENSENEEDED=false;
        //build gui with default roles associated with this action.
        //for now we are going to do a standard type of order/pattern.. w/ less gui flexibility.
        //if these roles exist in the role PARAMS, they will show up.
    
        this.createRoleBtnFromDefaultRole(MEM.role_actor, false, false, true);
        this.createRoleBtnFromDefaultRole(MEM.role_action, false, false, false, MEM.role_action);    //make this no matter what (it isn't assigned by the action, it IS the action)
        
        if (bTENSENEEDED==true) this.createRoleBtnFromDefaultRole(MEM.role_tense, true, false, false, MEM.role_tense);    //make this no matter what (it isn't assigned by the action, it IS the action)
        this.createRoleBtnFromDefaultRole(MEM.role_affected_thing, true, false, true, null, MEM.ec_thing.nid);
        this.createRoleBtnFromDefaultRole(MEM.role_theme_thing, true, false, true, null, MEM.ec_thing.nid);
        
        //we aren't doing "in the morning" type ROLE_WHEN here, only dates.
        //this EXT interface at this point is interested in events in-time that have a date on them.
        //The protocol allows for all nuances but we are limiting it here intentionally.
        this.createRoleBtnFromDefaultRole(MEM.role_when, true, true, false, null, MEM.date_gregorian.nid);

        //we are doing locations here.. cities, places, sites, geo-sites, etc.
        //if the verb is particular on one of those types it will have a more refined
        //default class relon or TYPE_SYS _default_class_nid relon pting to the param w/ the sys_ref_txid
        this.createRoleBtnFromDefaultRole(MEM.role_where, true, true, false, null, MEM.location.nid);

    }

    displaySentence = async()=> {
        var outputOBJ = await this.clump_nObj.processAsClumpNeuron();
        ANNE.A.setHTML("dynamic_msg_"+this.ID, outputOBJ.output);
    }


    /**
     * 
     * @param {anne_neuron} $roleNOBJ 
     * @param {boolean} $bEDIT 
     * @param {boolean} $bMOD 
     * @param {anne_neuron} $roleNOBJ 
     * @param {string} $default_class_nid 
     * @returns {element}
     */
     createRoleBtnFromDefaultRole = function($roleNOBJ, $bEDIT=true, $bMOD=false, $bMULTIPLES=false, $defaultRoleNOBJ=null, $default_class_nid=null) {
        
        //reminder - outs contains the neurons direct outs.  The _param_roles contains all roles form the BEM line including inherited ones.
        var param_rel = MEMG.getParamRelon(this.actionNOBJ._param_roles, $roleNOBJ.temp_nid);
        
        var roleNOBJ = null;
        if (param_rel==null) {
            if ($defaultRoleNOBJ==null) return;
            roleNOBJ = $defaultRoleNOBJ;
        }
        else roleNOBJ = $roleNOBJ;

        var _role =roleNOBJ.text.toLowerCase().replaceAll("_", "-");
        var div = document.createElement("div");
        div.classList = "role2 "+_role+"-outter";
        this.el_roles_inject.append(div);
        var span1 = document.createElement("span");
        span1.classList = _role+"-"+"lbl";
        div.append(span1);
        var span2 = document.createElement("span");
        span2.classList = "r-inner";
        span2.id="val_"+roleNOBJ.temp_nid+"_"+this.ID;
        span2.setAttribute("role_tnid", roleNOBJ.temp_nid);
        var mo = 0;
        if ($bMULTIPLES==true) mo=1;
        span2.setAttribute("multiples_ok", mo);
        div.append(span2);

        if ($bEDIT==true) {
            var btnAdd = document.createElement("button");
            btnAdd.innerHTML = "+add";
            btnAdd.id="add_"+roleNOBJ.temp_nid+"_"+this.ID;
            div.append(btnAdd);
            btnAdd.addEventListener("click", e=>this.launchSelector(span2));
            span2.setAttribute("delete_ok", 1);
        }
        if ($bMOD==true) {
            var btnmod = document.createElement("button");
            btnmod.innerHTML = "+mod";
            btnmod.id="mod_"+roleNOBJ.temp_nid+"_"+this.ID;
            btnAdd.addEventListener("click",e=>function(e){ alert("todo");});
            div.append(btnmod);
        }

        //if ($default_class_nid==null) return span2;
        var ptxid = "";
        var to_s = "";
        if (param_rel!=null) {
            ptxid = param_rel.txid;
            //the default restricted class for the param relon.
            //this can be updated with the ptxid if we have a TYPE_SYS relon overriding this default class.
            to_s=param_rel.to_s; 
        }
        
        var default_class_nid = null;
        if (ptxid!="") {
            default_class_nid = this.actionNOBJ.getDefaultClassForParam(ptxid);
        }

        if (default_class_nid!=null) span2.setAttribute("classRestrictNid", default_class_nid); 
        else if (to_s!="") span2.setAttribute("classRestrictNid", to_s);
        else if ($default_class_nid!=null) span2.setAttribute("classRestrictNid", $default_class_nid);

        return span2;
    }


    /**
     * Edit or set the value of a role element.
     * Clicking it should launch the appropriate page and use an appropriate picker in the injected div.
     * @param {element} $span2 
     */
     launchSelector = async($span2)=>{
        var tnid = $span2.getAttribute("role_tnid");
        var roleNOBJ = MEMG.getNeuronByTempNid(tnid);

        var misc = {};
        misc.smartDefocus = this.smartDefocus;
        misc.bCALLBACKONLY=true;
        var pObj = new page_params(this.newRoleSuccess, this.cancelInject, null, roleNOBJ.nid);
        pObj.misc = misc;
        
        var contextNOBJ = null;
        var classRestrictNid = $span2.getAttribute("classRestrictNid");
        if (classRestrictNid!=undefined) {
            pObj.misc.classRestrictNid = classRestrictNid;
            contextNOBJ = MEMG.getNeuronByNid(classRestrictNid);
        }

        var page = PAGES.getDefaultSelectorPageForClass(roleNOBJ, null, contextNOBJ);
        
        pObj.misc.from_nid = this.clump_nObj.nid;   //all TYPE_ROLE relons will be put on the clump.
        pObj.misc.type = TYPE_ROLE;
        pObj.misc.reln_nid = roleNOBJ.nid;

        ANNE.A.empty(this.el_inject_form);
        pObj.injection_div = this.el_inject_form.id;
        await this.pgui.loadPart(page, pObj);
    }




    /**
     * This already exists.  Set the value for display.
     * If we put a delete button for easy erase (or if they edit)
     * we can delete this old ghost rel.
     * @param {*} $roleNid 
     * @param {*} $roleValue 
     * @param {*} $to_type 
     * @param {*} $txid 
     */
    displayExistingRoleValue = async($roleNid, $roleValue, $to_type, $txid)=> {
        var roleNOBJ = MEMG.getNeuronByNid($roleNid);
        var el_role = document.getElementById("val_"+roleNOBJ.temp_nid+"_"+this.ID);
        var pre = "";
        if ($to_type==TO_TYPE_ANNE) {
            EXT_BTNS.makeNeuronLink_colorSquares(el_role, $roleValue, "", pre, false, "", "dossier", "sq2-");
        }
        else if ($to_type==TO_TYPE_TXID) {
            EXT_BTNS.makeRelonViewLink(el_role, $roleValue);
        }
        else ANNE.A.setHTML(el_role, $roleValue);
        el_role.setAttribute("txid", $txid);
    
        var delete_ok = el_role.getAttribute("delete_ok");
        if (delete_ok==undefined) delete_ok = null;

        if (delete_ok==true) {
            var btndelete = document.createElement("button");
            btndelete.innerHTML = "[x]";
            btndelete.id="delete_"+roleNOBJ.temp_nid+"_"+this.ID;
            el_role.append(btndelete);
            btndelete.setAttribute("txid", $txid);
            btndelete.addEventListener("click", e=>this.deleletRelon(el_role, btndelete))
        }

        var multiples_ok = el_role.getAttribute("multiples_ok");
        if (multiples_ok==false) {
            ANNE.A.hide("add_"+roleNOBJ.temp_nid+"_"+this.ID);
        }
        else ANNE.A.show("add_"+roleNOBJ.temp_nid+"_"+this.ID);
    }



    /**
     * Delete a ghost relon.
     * @param {Element} $span2 
     * @returns 
     */
     deleletRelon = async($span2, $btn)=>{
        var txid = $btn.getAttribute("txid");
        if (ANNE.UTILS.isset(txid)==false) return;
        ANNE.A.empty($span2);
        var role_tnid = $span2.getAttribute("role_tnid");
        var roleNOBJ = MEMG.getNeuronByTempNid(role_tnid);
        ANNE.A.show("add_"+roleNOBJ.temp_nid+"_"+this.ID);
        var remaining_rels_of_this_role = await this.clump_nObj.deleteRolefromClump(txid);
        if (remaining_rels_of_this_role==null) return;
        for (var r=0; r < remaining_rels_of_this_role.length; r++) {
            var role_rel = remaining_rels_of_this_role[r];
            this.displayExistingRoleValue(role_rel.reln, role_rel.to_s, role_rel.to_type, role_rel.txid);
        }
        await this.displaySentence();
    }



    /**
     * A new role has been added to the ghost clump.
     * Display the new goodies, update the sentence.
     * @param {Object} $ge 
     * @param {Object} $new_nid 
     * @param {Object} $new_relon 
     * @returns 
     */
    newRoleSuccess=async($ge, $new_nid, $new_relon=null)=> {
        console.log("newRoleSuccess");
        console.log($ge, $new_nid, $new_relon);
        await this.displayExistingRoleValue($new_relon.reln, $new_nid, $new_relon.to_type, $new_relon.txid);
        await this.smartDefocus(false);
        await PAGES.closeInjectedDetails($ge);
        await this.displaySentence();
        return;
    }


    /**
     * @param {Boolean} $bDEFOCUS 
     */
    smartDefocus = async($bDEFOCUS=true)=> {
        if ($bDEFOCUS==true) {
            var el = ANNE.A.addClass("fuzzyWuzzy_"+this.ID, "defocus");
            el.setAttribute("disabled",true);
        }
        else {
            var el = ANNE.A.removeClass("fuzzyWuzzy_"+this.ID, "defocus");
            el.removeAttribute("disabled",false);
        }
    }


    /**
     * 
     * @param {Object} $ge 
     * @returns 
     */
    cancelInject = async($ge)=> {
        console.log($ge);
        await this.smartDefocus(false);
        await PAGES.closeInjectedDetails($ge);
    }

    /**
     * 
     * @param {Element} $el 
     * @returns 
     */
    makeRoleDiv = function($el) {
        var el = ANNE.A.__getEl($el);
        if (el==null) return;
        var div = document.createElement("div");
        div.classList = "role role-actor";
        el.appendChild(div);
        var span1 = document.createElement("span");
        span1.classList.add("role-affected-thing")
    }



    /**
     * Will only delete ghost clumps that were being worked on.
     * Will delete all connecting links, but NOT any ghost neurons that were referenced.
     */
    deleteGhostClump = async($e)=> {
        var tnid = this.clump_nObj.temp_nid;
        await ANNE.GHOSTDATA._annihilateGhostNeuron(this.clump_nObj);
        ANNE.removeSmartTnid(tnid);
        this.cancelInject($e);
    }


    /**
     */
    upgradeGhostClump = async($e)=> {
        ANNE.A.hide(this.el_upgrade_ghost);
        this.ghost_to_anneOBJ = new ghost_to_anne(this.clump_nObj);
        var bAUTOHANAI = true;
        if (ANNE.SETTINGS!=null) bAUTOHANAI = ANNE.SETTINGS.bAUTOHANAI;
        await this.ghost_to_anneOBJ.update(this.clump_nObj, 0, false, bAUTOHANAI);
        console.log(this.ghost_to_anneOBJ);
        await ANNE.SENDQ.___sendQueue(this.refreshBack);
    }


    /**
     */
    refreshBack = async($obj)=>{
        await this.smartDefocus(false);
        await PAGES.closeInjectedDetails({pgui:this.pgui, pObj:this.pObj});
        PAGES.refreshCurrent();
    }

        
    /**
     * Turn on the make ghost button if we have a temp clump.
     * If we have a ghost clump already, turn on the delete and/or upgrade.
     */
     __assessBtns = function() {
        if (this.clump_nObj.ghost==true) ANNE.A.show(this.el_upgrade_ghost);
        ANNE.A.show(this.el_delete_ghost_btn); 
    }


    /**
     * 
     * @param {string} $roleNid - nid long string 
     * @param {*} $roleValue - misc value (string or number) 
     * @param {int} $to_type - TO_TYPE int
     */
     _initGhostRoleValue = async($roleNid, $roleValue, $to_type)=> {
        var roleNOBJ = MEMG.getNeuronByNid($roleNid);
        var el_role = document.getElementById("val_"+roleNOBJ.temp_nid+"_"+this.ID);
        
        if (el_role==null) return; //??

        var pre = "";

        if ($to_type==TO_TYPE_ANNE) {
            if ($roleNid==MEM.role_action.nid) {
                var toNOBJ = MEMG.getNeuronByNid($roleValue);
            }
            EXT_BTNS.makeNeuronLink_colorSquares(el_role, $roleValue, "", pre, false, "", "dossier", "sq2-");
        }
        else if ($to_type==TO_TYPE_TXID) {
            EXT_BTNS.makeRelonViewLink(el_role, $roleValue);
        }
        else ANNE.A.setHTML(el_role, $roleValue);
        

        //add the delete per set relon in the role.
        //we can have roles that have multiple values.
        //in this GUI we are not handling logical operands (this OR that.. this, but not that)
        //we are simply allowing for multiple relons of the same role with indicates a presumed AND.
        //For more robust handling the NP would be its own neuron with the logical operands (if they existed)
        //That would be its own concept..  (This, but not that)
        var delete_ok = el_role.getAttribute("delete_ok");
        if (delete_ok==undefined) delete_ok = null;

        //we don't store action roles.. they are displayed from the PARENT VALUE of the inst.
        //they ARE instances of the action.
        if ($roleNid!=MEM.role_action.nid) {
            var ghost_rel = await GHOSTDATA.new_ghost_role(this.clump_nObj, $roleNid, $to_type, $roleValue);
            el_role.setAttribute("txid", ghost_rel.txid);
        }

        if (delete_ok==true) {
            var btndelete = document.createElement("button");
            btndelete.innerHTML = "[x]";
            btndelete.id="delete_"+roleNOBJ.temp_nid+"_"+this.ID;
            el_role.append(btndelete);
            btndelete.addEventListener("click", e=>this.deleletRelon(el_role, btndelete));
            btndelete.setAttribute("txid", ghost_rel.txid);
        }

        var multiples_ok = el_role.getAttribute("multiples_ok");
        if (multiples_ok==false) {
            ANNE.A.hide("add_"+roleNOBJ.temp_nid+"_"+this.ID);
        }
        else ANNE.A.show("add_"+roleNOBJ.temp_nid+"_"+this.ID);
    }


}