var DB = {};
DB.db = null;
DB.dbName = "ANNODE";
DB.bSUPPORTED = false;
DB.ERROR_INSERTS = 0;
DB.DO_THE_ONETIME_PRELOAD=false;

/**
 * relonsMinBoot is a table that is entirely loaded into RAM upon bootup.
 * ALL RELONS IN THIS TABLE WILL GO INTO RAM.
 * 
 * The relons table is accumulated over all app usages and will get VERY large.
 * We strategically load relons out of this table into RAM on-demand in key places throughout the apps.
 * Typically this is along-side the API queries to go get info from the ANNODE.
 * If relons return we save them into the relons table along with updating our history optimizer which is 
 * intended as a query filter for future ANNODE queries.  It documents the last time that query was used
 * and saves the last block (height) for which results were found.  The next time it queries it will use
 * that historic information as a filter to avoid redundant data pulls from the ANNODE.
 */

/**
 * needs to now add in the relonsMinBoot  (lolz) 
 * @param {*} $tbls 
 * @returns 
 */
DB.getCnt=async($tbls)=> {
    var total = 0;
    for (var t=0; t < $tbls.length; t++) {
        var tbl = $tbls[t];
        var objs = await DB.loadAll(tbl)
        total+=objs.length;
    }
    return total;
}


/**
 * @author {Mr.Scatman}
 */
 DB.___init=async($onLoadCallback=null, $callbackCallback=null)=> {
    if (this._annodeDetails!=null && this._annodeDetails.network=="ANNE-TESTNET") {
        DB.dbName = "ANNODE_testnet";
    }

    //big trouble for this app...
    if (DB.bSUPPORTED==false) {
        if ($onLoadCallback!=null) $onLoadCallback(null, null, $callbackCallback);
        return;
    }

    //open existing (or new, if first time) DB.
    var done = await DB.openConnection(DB.dbName, 2, $onLoadCallback, $callbackCallback);
    if (done==true) {
        var aFullPullHistory = await DB.loadAll("fullPullHistory");
        var aSiteSettings = await DB.loadAll("settings");

        //initial load of RAM is limited to hypergraph_preload.
        var aRelonsToProcess = await DB.loadAll("relonsMinBoot");
        var aR = await DB.loadAll("relons");

        //just get it done, all now, in-mem.
        if (ANNE.SETTINGS!=null && ANNE.SETTINGS.bMAXIMIZE_JS_MEM==true) {
            aRelonsToProcess = aRelonsToProcess.concat(aR);
        }
        
        for (var a=0; a < aR.length; a++) {
            if (aR[a].to_type!=TO_TYPE_STRING) continue;
            if (aR[a].type==TYPE_SYS_MSG) continue;

            if (aR[a].dead!=undefined) {
                console.warn("HOW DID THIS GET IN HERE????");
                delete aR[a].dead;
            }
            if (aR[a].temp!=undefined) {
                console.warn("HOW DID THIS GET IN HERE????");
                delete aR[a].temp; 
            }

            ANNE.MEM.addToTextIndex(aR[a].to_s, aR[a].from_nid);
        }

        if ($onLoadCallback!=null) {
            $onLoadCallback(aRelonsToProcess, aFullPullHistory, aSiteSettings, $callbackCallback);
        }
        
        else {
            var response = {};
            response.relons = aRelonsToProcess;
            response.history = aFullPullHistory;
            response.settings = aSiteSettings;
            response.status = "ok";
            var retObj = {status:'ok', func:"init_db", fromclaim:"annelocalstorage", results:response};
            return retObj;
        }
    }
    else if ($onLoadCallback!=null) {
        $onLoadCallback([], [], [], $callbackCallback);
    }

    else {
        var response = {};
        response.relons = [];
        response.history = [];
        response.settings = [];
        response.status = "ok";
        var retObj = {status:'ok', func:"init_db", fromclaim:"annelocalstorage", results:response};
        return retObj;
    }
}


/**
 * ???????????????????????????????????????
 * 
 * We can run this query to load hypergraph MEM anytime we go external
 * to load new ANNE relons.  If we aren't getting new ANNE relons we don't have
 * to look to local indexdb for more relons.
 * 
 * Note: We will want to do this, however, even when there are no external ANNODE relons in our query.
 * 
 * Filters are 
 * .type
 * .from_nid 
 * .reln  
 * .to_type
 * .to_s
 * .time
 * .be
 * .sponsored_by
 * .relon_sponsor
 * @param {*} $filters 
 */
DB.___geRelonsQuery=async()=>{
    console.time("getRelonsQuery");
    var aRelonsToProcess = await DB.loadAll("relons");
    console.timeEnd("getRelonsQuery");
    return aRelonsToProcess;
}


/**
 * @author {Mr.Scatman}
 */
DB.___deleteEntireDB=async()=> {
    if (DB.bSUPPORTED==false) return;
    if (DB.db==null) return;
    return new Promise((resolve, reject) => {
        DB.db.close();
        var request = indexedDB.deleteDatabase(DB.dbName, 2);
        request.onblocked = function () {
            console.log("Delete blocked.");
            return resolve(false);
        }
        request.onerror = function () {
            console.log("Error deleting the DB");
            return resolve(false);
        };
        request.onsuccess = function () {
            console.log("Deleted OK.");
            return resolve(true);
        }   
    });             
}


DB.getDBSize=async()=> {
    var size = await DB.getDBTblSize("fullPullHistory");
    size+= await DB.getDBTblSize("settings");
    size+= await DB.getDBTblSize("relons");
    return size;
}

DB.getDBTblSize=async($tbl, $method="readwrite")=> {
    var transaction  = DB.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    return new Promise((resolve, reject) => {
        //FULL PULL HISTORY RETREIVAL
        if ('getAll' in objectStore) {
            objectStore.getAll().onsuccess = function(event) {
                var arr = event.target.result;
                var json = JSON.stringify(arr);
                size = json.length;
                return resolve(size);
            };
        } 
    });
}



/**
 * @author {Mr.Scatman}
 */
DB.openConnection=async($dbName, $version)=> {
    if (DB.db!=null) return true;
    return new Promise((resolve, reject) => {
        var request = indexedDB.open($dbName, $version);
        request.onerror = function(event) {
            console.log("openConnection request.Error");
            return resolve(false);
        };

        request.onsuccess = function(event) {
            DB.db = event.target.result;
            return resolve(true);
        };
        
        request.onupgradeneeded = function(event) {
            console.log("openConnection request.onUpradeneeded");
            DB.db = event.target.result;

            //our min boot RAM is loaded in this table one-time.
            //we will add to it only if we find very particular relons we want in PRE-BOOT.
            var relonsMinBootObjStore = DB.db.createObjectStore("relonsMinBoot", { keyPath: "txid" });
            relonsMinBootObjStore.createIndex("txid", "txid", { unique: true });

            var relonObjStore = DB.db.createObjectStore("relons", { keyPath: "txid" });
            relonObjStore.createIndex("txid", "txid", { unique: true });

            var fullPullHistoryStore = DB.db.createObjectStore("fullPullHistory", { keyPath: "pullid" });
            fullPullHistoryStore.createIndex("pullid", "pullid", { unique: true });

            var settingsObjStore = DB.db.createObjectStore("settings", { keyPath:  ["uniqueID", "misc_value", "user_nid"] });
            settingsObjStore.createIndex("valclassuser", ["uniqueID", "misc_value", "user_nid"], { unique: true });
            return resolve(false);  //no load needed
        };
    });
}



/**
 * 
 * @param {*} $tbl 
 * @param {*} $method 
 * @returns 
 */
DB.loadAll=async($tbl, $method="readwrite")=> {
    var transaction  = DB.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    return new Promise((resolve, reject) => {
        if ('getAll' in objectStore) {
            objectStore.getAll().onsuccess = function(event) {
                return resolve(event.target.result);
            };
        } 
        else {
            var arr = [];

            // Fallback to the traditional cursor approach if getAll isn't supported.
            objectStore.openCursor().onsuccess = function(event) {
                var cursor = event.target.result;
                if (cursor) {     
                    arr.push(cursor.value);       
                    cursor.continue();
                } 
                else return resolve(arr);
            };
        }
    });
}


/**
 * On min-boot load we are passing this a tbl of relonsMinBoot
 * @param {*} $objs 
 * @param {*} $tbl 
 * @param {*} $method 
 * @returns 
 */
DB.___insert = async($objs, $tbl="relons", $method="readwrite")=> {
    if (DB.bSUPPORTED==false) return;
    if ($objs.length==0) return;

    if (DB.db==null) {
        console.log("db was not initiated!");
        return;
    }

    if (DB.DO_THE_ONETIME_PRELOAD==true) {
        //init aware here.
        if (ANNE.AWARE) await ANNE.AWARE.init(); //this will require some nuanced handling.
    }

    //we may need to insert some of these into the relonsMinBoot.  Yikes.
    if ($tbl=="relons") {
        var minBootRelons = [];
        var minBootFromNids = [];
        var finalRelons = [];
        var mbs = 0;

        for (var r=0; r < $objs.length; r++) {
            var mbRObj = $objs[r];
            mbRObj.minboot = false;

            //no intrablock pending.  DO NOT SAVE THIS!
            if (mbRObj.reln==MEM.INTRA_BLOCK_PENDING.nid) continue;

            //any relon with a from nid we have in our arrays.
            if (minBootFromNids.indexOf(mbRObj.from_nid)!=-1) {
                mbRObj.minboot=true;
                mbs+=1;
                continue;
            }

            //put these in min boot?  yes.
            if (mbRObj.type==TYPE_INTERNAL_PAY || mbRObj.type==TYPE_BROADCAST_PAY) {
                mbRObj.minboot=true;
                mbs+=1;
                continue;
            }

            //countries
            else if (mbRObj.type==TYPE_BE && mbRObj.to_s==MEM.country.nid) {
                var fOBJ = MEMG.getNeuronByNid(mbRObj.from_nid);
                if (ANNE.isANNE(fOBJ.owned_by)==true && minBootFromNids.indexOf(fOBJ.nid)==-1) {
                    minBootFromNids.push(fOBJ.nid);
                }        
            }

            //states
            else if (mbRObj.type==TYPE_BE && mbRObj.to_s==MEM.state.nid) {
                var fOBJ = MEMG.getNeuronByNid(mbRObj.from_nid);
                if (ANNE.isANNE(fOBJ.owned_by)==true && minBootFromNids.indexOf(fOBJ.nid)==-1) {
                    minBootFromNids.push(fOBJ.nid);
                }        
            }

            //integers
            else if (mbRObj.type==TYPE_BE && mbRObj.to_s==MEM.integer.nid) {
                var fOBJ = MEMG.getNeuronByNid(mbRObj.from_nid);
                if (ANNE.isANNE(fOBJ.owned_by)==true && minBootFromNids.indexOf(fOBJ.nid)==-1) {
                    minBootFromNids.push(fOBJ.nid);
                }        
            }

            else if (mbRObj.type==TYPE_BE && ANNE.MEM.isDateAny(mbRObj.to_s)==true) {
                var fOBJ = MEMG.getNeuronByNid(mbRObj.from_nid);
                if (ANNE.isANNE(fOBJ.owned_by)==true && minBootFromNids.indexOf(fOBJ.nid)==-1) {
                    minBootFromNids.push(fOBJ.nid);
                }        
            }


            else if (mbRObj.type==TYPE_BE && mbRObj.to_s==MEM.ASCII.nid) {
                var fOBJ = MEMG.getNeuronByNid(mbRObj.from_nid);
                if (ANNE.isANNE(fOBJ.owned_by)==true && minBootFromNids.indexOf(fOBJ.nid)==-1) {
                    minBootFromNids.push(fOBJ.nid);
                }        
            }
            


            //BRAIN dimension puts feelz in buckets.
            else if (mbRObj.type==TYPE_BRAIN && mbRObj.to_type==TO_TYPE_ANNE && mbRObj.to_s==MEM.feeling.nid) {
               var fOBJ = MEMG.getNeuronByNid(mbRObj.from_nid);
                if (ANNE.isANNE(fOBJ.owned_by)==true && minBootFromNids.indexOf(fOBJ.nid)==-1) {
                    minBootFromNids.push(fOBJ.nid);
                }
            }

            //opinionz buckets outs
            else if (ANNE.AWARE!=null && ANNE.AWARE.fzOpNids!=null && ANNE.AWARE!=null && ANNE.AWARE.fzOpNids.length>0 && ANNE.AWARE.fzOpNids.indexOf(mbRObj.from_nid)!=-1) {
                var fOBJ = MEMG.getNeuronByNid(mbRObj.from_nid);
                if (ANNE.isANNE(fOBJ.owned_by)==true && minBootFromNids.indexOf(fOBJ.nid)==-1) {
                    minBootFromNids.push(fOBJ.nid);
                }
            }

            //any anne class
            else if (ANNE.isANNE(mbRObj.owned_by)==true) {
                var fOBJ = MEMG.getNeuronByNid(mbRObj.from_nid);
                if (fOBJ.ntype==NTYPE_CLASS && minBootFromNids.indexOf(fOBJ.nid)==-1) {
                    minBootFromNids.push(fOBJ.nid);
                }
                else if (fOBJ.is(MEM.fileX)==true && minBootFromNids.indexOf(fOBJ.nid)==-1) {
                    //console.log("minboot file neuron "+fOBJ.nid);
                    minBootFromNids.push(fOBJ.nid);
                }
            }

            else if (ANNE.isIdentity(mbRObj.from_nid)==true) {
                var fOBJ = MEMG.getNeuronByNid(mbRObj.from_nid);
                if (fOBJ.ntype==NTYPE_CLASS && minBootFromNids.indexOf(fOBJ.nid)==-1) {
                    minBootFromNids.push(fOBJ.nid);
                }
            }
        }

        //one more pass (lol) to get the final minBootRelons.
        if (minBootFromNids.length || mbs>0) {
            for (var r=0; r < $objs.length; r++) {
                var mbRObj = $objs[r];
                if (mbRObj.minboot==true) {
                    minBootRelons.push(mbRObj);
                }

                //this relon goes in our min-boot tbl to preload on startup
                else if (minBootFromNids.indexOf(mbRObj.from_nid)!=-1) {
                    minBootRelons.push(mbRObj);
                }

                //this goes in the relons tbl.
                else {
                    if (mbRObj.type==TYPE_BROADCAST_PAY || mbRObj.type==TYPE_INTERNAL_PAY) {
                        console.warn("WTHECK");
                        minBootRelons.push(mbRObj);
                        delete mbRObj.minboot;
                        continue;
                    }
                    finalRelons.push(mbRObj);
                }

                delete mbRObj.minboot;
            }

            //update objs for relons tbl
            $objs = finalRelons;

            //if we had minbootRelons, do those now in the min boot tbl.
            if (minBootRelons.length) {

                //these don't get sifted for update status.
                
                await DB.____finalInsert(minBootRelons, "relonsMinBoot", $method);
            }
        }
    }

    //do the normal relons insert.
    if ($objs.length) {
        return await DB.____finalInsert($objs, $tbl, $method);
    }
    else return [];
}



/**
 * @param {*} $objs 
 * @param {*} $tbl 
 * @param {*} $method 
 * @returns 
 */
DB.____finalInsert=async($objs, $tbl, $method)=> {
    DB.ERROR_INSERTS = 0;
    DB.reroutes_needed = [];

    //some willy-wonkas are required here.
    //we try to optimize RAM by pulling only relons we think we need.
    //Sometimes the ANNODE returns relons we already have in indexedb but we 
    //HAVE NOT YET LOADED in RAM, even with our RAM pull prior to the indexedb pull.
    //This can happen when we pull the getAllMy$tuff and look in indexedb for our sponsors.
    //We may not be documented yet as the sponsor of a neuron so we won't get that relon...
    //yet the ANNODE API call gets that relon which is already in indexedb.  Since we don't have it 
    //yet in RAM we will think it is new and try to insert it into indexed
    //In ADD NEW mode, the MEM processor will attempt to save in IDB all relons we didn't have in ram.
    //This will cause a key error..  We capture that key error and then reroute it through the ___update
    DB.reroutes_needed = [];

    //normal processing here...
    if (DB.DO_THE_ONETIME_PRELOAD==false) {
        for (var o=0; o < $objs.length; ++o) {
            var updateReroute = await DB.____insertRecord($tbl, $method, $objs[o]);
            if (updateReroute!=null) {
                if ($objs[o].type==TYPE_INTERNAL_PAY || $objs[o].type==TYPE_BROADCAST_PAY) {
                    if (ANNE.SETTINGS.debug()==true) console.log("no need to update internal/broadcast pay relon "+$objs[o].txid);
                }
                else await DB.___update($objs[o], $tbl);
            }
        }
    }


    //one-time preload batch insert.. fast.
    else {
        
        DB.insertPRELOAD_ONETIME_Batch($tbl, $method, $objs, 0);
        DB.DO_THE_ONETIME_PRELOAD = false;
    }

    if (DB.ERROR_INSERTS>0) {
        console.log("%c error inserts requiring updates: %c"+DB.ERROR_INSERTS, ANNE.TALK.bgblack, ANNE.TALK.bgred);
    }

    DB.ERROR_INSERTS=0;
    return;
}



/**
 * We need to update the proper place when we update a relon.
 * If we updated a min-boot record, update the relon in that table.
 * Otherwise, just update our normal relons table.
 */
DB.__handleRelonUpdate=async($obj, $method="readwrite")=>{
    if (DB.bSUPPORTED==false) return null;

    var tbl = "relons";

    if ($obj.type==TYPE_INTERNAL_PAY || $obj.type==TYPE_BROADCAST_PAY) {
        tbl = "relonsMinBoot";
    }
            
    //BRAIN dimension puts feelz in buckets.
    else if ($obj.type==TYPE_BRAIN && $obj.to_type==TO_TYPE_ANNE && $obj.to_s==MEM.feeling.nid) {
        var fOBJ = MEMG.getNeuronByNid($obj.from_nid);
        if (fOBJ==null) return null;
        if (ANNE.isANNE(fOBJ.owned_by)==true) {
            tbl = "relonsMinBoot";
        }
    }

    //opinionz buckets outs
    else if (ANNE.AWARE!=null && ANNE.AWARE.fzOpNids.length>0 && ANNE.AWARE.fzOpNids.indexOf($obj.from_nid)!=-1) {
        var fOBJ = MEMG.getNeuronByNid($obj.from_nid);
        if (fOBJ==null) return null;
        if (ANNE.isANNE(fOBJ.owned_by)==true) {
            tbl = "relonsMinBoot";
        }
    }

    //countries (already requires anne or user worldview)
    else if (ANNE.isCountry($obj.from_nid)==true) {
        tbl = "relonsMinBoot";      
    }

    //states (already requires anne or user worldview)
    else if (ANNE.isState($obj.from_nid)==true) {
        tbl = "relonsMinBoot";       
    }

    else if (ANNE.isInteger($obj.from_nid)==true) {
        tbl = "relonsMinBoot";       
    }

    else if (ANNE.isDateAny($obj.from_nid)==true) {
        tbl = "relonsMinBoot";
    }

    //any anne class OR instance of ASCII char
    else if (ANNE.isANNE($obj.owned_by)==true) {
        var fOBJ = MEMG.getNeuronByNid($obj.from_nid);
        if (fOBJ==null) return null;
        if (fOBJ.ntype==NTYPE_CLASS || fOBJ.bes[0]==ANNE.MEM.ASCII.temp_nid) {
            tbl = "relonsMinBoot";
        }
        else if (fOBJ.is(MEM.fileX)==true) {
            tbl = "relonsMinBoot";
        }
    }

    else if (ANNE.isIdentity($obj.from_nid)==true) {
        var fOBJ = MEMG.getNeuronByTempNid($obj.f);
        if (fOBJ==null) return null;
        if (fOBJ.ntype==NTYPE_CLASS) {
            tbl = "relonsMinBoot";
        }
    }

    var transaction  = DB.db.transaction([tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore(tbl);
    if (objectStore==null) return;

    var sanitized_rObj = {
        txid:$obj.txid, 
        type:parseInt($obj.type),
        from_nid:$obj.from_nid,
        reln:$obj.reln,
        to_type:parseInt($obj.to_type),
        to_s:$obj.to_s,
        time:$obj.time,
        firstseen:$obj.firstseen,
        freeze:$obj.freeze
    }

    if ($obj.old_txid!=undefined) sanitized_rObj.old_txid = $obj.old_txid;
    if (sanitized_rObj.txid[0]=="z") {
        console.warn("BAD DATA -- DON'T STORE GHOST/TEMP TXIDS in relons tbl!!");
        console.warn(sanitized_rObj);
        alert("hunt this down!");
        return null;
    }

    if ($obj.owned_by!=undefined && $obj.owned_by!=null) {
        sanitized_rObj.owned_by=$obj.owned_by;
    }
    else $obj.owned_by="";
    if ($obj.sponsored_by!=undefined && $obj.sponsored_by!=null) {
        sanitized_rObj.sponsored_by=$obj.sponsored_by;
    }
    else $obj.sponsored_by="";

    if ($obj.annies!=undefined) sanitized_rObj.annies=$obj.annies;
    if ($obj.parent!=undefined) sanitized_rObj.parent=$obj.parent;
    if ($obj.origin!=undefined) sanitized_rObj.origin=$obj.origin;
    if ($obj.rorigin!=undefined) sanitized_rObj.rorigin=$obj.rorigin;
    if ($obj.relon_sponsor!=undefined) sanitized_rObj.relon_sponsor=$obj.relon_sponsor;
    if ($obj.from_text!=undefined) sanitized_rObj.from_text=$obj.from_text;
    if ($obj.ntype!=undefined) sanitized_rObj.ntype=$obj.ntype;
    if ($obj.sys_ref_txid!=undefined) sanitized_rObj.sys_ref_txid=$obj.sys_ref_txid;
    if ($obj.annies!=undefined) sanitized_rObj.annies=$obj.annies; 
    if ($obj.decrypted!=undefined) sanitized_rObj.decrypted=$obj.decrypted;
    if ($obj.flathexint!=undefined) {
        sanitized_rObj.flathexint = $obj.flathexint;
    }

    return new Promise((resolve, reject) => {
        request = objectStore.put(sanitized_rObj);
        request.onsuccess = function(event) {
            return resolve(true);
        };
        request.onerror = function(event) {
            console.log(event);
            return resolve(false);
        }
    });
}



/**
 * 
 * @param {*} $tbl 
 * @param {*} $method 
 * @param {*} $rObjs 
 * @param {*} $startIndex 
 * @returns 
 */
DB.insertPRELOAD_ONETIME_Batch=function($tbl, $method, $rObjs, $startIndex=0) {
    console.log("%c insertPRELOAD_ONTIME_Batch %c"+$tbl, ANNE.TALK.bgblack, ANNE.TALK.bgyellow)
    const batchSize = 40000;  
    const endIndex = Math.min($startIndex + batchSize, $rObjs.length);

    var transaction  = DB.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;

    var txids = [];
    
    for (var i = $startIndex; i < endIndex; i++) {
        var $rObj = $rObjs[i];

        //wiillllllllllllly wonka, yet again.
        if (txids.indexOf($rObj.txid)!=-1) continue;
        txids.push($rObj.txid);

        var sanitized_rObj = {
            txid:$rObj.txid, 
            type:parseInt($rObj.type),
            from_nid:$rObj.from_nid,
            reln:$rObj.reln,
            to_type:parseInt($rObj.to_type),
            to_s:$rObj.to_s,
            time:$rObj.time,
            firstseen:$rObj.firstseen,
            freeze:$rObj.freeze
        }

        //DONT SAVE THIS, EVER
        if ($rObj.reln==MEM.INTRA_BLOCK_PENDING.nid) return;

        //for bad ghost cleanup fixing
        if ($rObj.old_txid!=undefined) sanitized_rObj.old_txid = $rObj.old_txid;
        if (sanitized_rObj.txid==undefined || sanitized_rObj.txid==null) {
            console.warn("bad data, no txid");
            alert("hunt this down!");
            resolve(null);
        }
        if (sanitized_rObj.txid[0]=="z") {
            console.warn("BAD DATA -- DON'T STORE GHOST/TEMP TXIDS in relons tbl!!");
            console.warn(sanitized_rObj);
            alert("hunt this down!");
            resolve(null);
        }

        if ($rObj.anbl_id!=undefined) sanitized_rObj.anbl_id=$rObj.anbl_id;

        if ($rObj.owned_by!=undefined && $rObj.owned_by!=null) {
            sanitized_rObj.owned_by=$rObj.owned_by;
        }
        else $rObj.owned_by="";
        if ($rObj.sponsored_by!=undefined && $rObj.sponsored_by!=null) {
            sanitized_rObj.sponsored_by=$rObj.sponsored_by;
        }
        else $rObj.sponsored_by="";
        if ($rObj.parent!=undefined) sanitized_rObj.parent=$rObj.parent;
        if ($rObj.origin!=undefined) sanitized_rObj.origin=$rObj.origin;
        if ($rObj.rorigin!=undefined) sanitized_rObj.rorigin=$rObj.rorigin;
        if ($rObj.relon_sponsor!=undefined) sanitized_rObj.relon_sponsor=$rObj.relon_sponsor;
        if ($rObj.from_text!=undefined) sanitized_rObj.from_text=$rObj.from_text;
        if ($rObj.ntype!=undefined) sanitized_rObj.ntype=$rObj.ntype;
        if ($rObj.sys_ref_txid!=undefined) sanitized_rObj.sys_ref_txid=$rObj.sys_ref_txid;
        if ($rObj.annies!=undefined) sanitized_rObj.annies=$rObj.annies;
        if ($rObj.decrypted!=undefined) sanitized_rObj.decrypted=$rObj.decrypted;
        if ($rObj.flathexint!=undefined) {
            sanitized_rObj.flathexint = $rObj.flathexint;
        }
        objectStore.add(sanitized_rObj);
    }

    transaction.oncomplete = function() {
        //console.log("WHEN DOES THIS GET PRINTED??");
        if (endIndex < $rObjs.length) {
            DB.insertBatch($tbl, $method, $rObjs, $endIndex);
        } 
        else {
            //console.log('All records inserted successfully!');
            return true;
        }
    };

    transaction.onerror = function(event) {
        console.error('Error inserting records:', event.target.error);
        return null;
    };
}

/**
 * 
 * @param {*} $tbl 
 * @param {*} $method 
 * @param {*} $rObj 
 * @returns 
 */
DB.____insertRecord_RR=async($tbl, $method, $rObj)=> {

    return new Promise((resolve, reject) => {   
        var transaction  = DB.db.transaction([$tbl], $method);
        if (transaction==null) return;
        var objectStore = transaction.objectStore($tbl);
        if (objectStore==null) return;
        var sanitized_rObj = {
            txid:$rObj.txid, 
            type:parseInt($rObj.type),
            from_nid:$rObj.from_nid,
            reln:$rObj.reln,
            to_type:parseInt($rObj.to_type),
            to_s:$rObj.to_s,
            time:$rObj.time,
            firstseen:$rObj.firstseen,
            freeze:$rObj.freeze
        }

        //DONT SAVE THIS, EVER
        if ($rObj.reln==MEM.INTRA_BLOCK_PENDING.nid) return;

        //for bad ghost cleanup fixing
        if ($rObj.old_txid!=undefined) sanitized_rObj.old_txid = $rObj.old_txid;
        if (sanitized_rObj.txid==undefined || sanitized_rObj.txid==null) {
            console.log("test");
            resolve(null);
        }
        if (sanitized_rObj.txid[0]=="z") {
            console.warn("BAD DATA -- DON'T STORE GHOST/TEMP TXIDS in relons tbl!!");
            console.warn(sanitized_rObj);
            alert("hunt this down!");
            resolve(null);
        }

        if ($rObj.anbl_id!=undefined) sanitized_rObj.anbl_id=$rObj.anbl_id;

        if ($rObj.owned_by!=undefined && $rObj.owned_by!=null) {
            sanitized_rObj.owned_by=$rObj.owned_by;
        }
        else $rObj.owned_by="";
        if ($rObj.sponsored_by!=undefined && $rObj.sponsored_by!=null) {
            sanitized_rObj.sponsored_by=$rObj.sponsored_by;
        }
        else $rObj.sponsored_by="";
        if ($rObj.parent!=undefined) sanitized_rObj.parent=$rObj.parent;
        if ($rObj.origin!=undefined) sanitized_rObj.origin=$rObj.origin;
        if ($rObj.rorigin!=undefined) sanitized_rObj.rorigin=$rObj.rorigin;
        if ($rObj.relon_sponsor!=undefined) sanitized_rObj.relon_sponsor=$rObj.relon_sponsor;
        if ($rObj.from_text!=undefined) sanitized_rObj.from_text=$rObj.from_text;
        if ($rObj.ntype!=undefined) sanitized_rObj.ntype=$rObj.ntype;
        if ($rObj.sys_ref_txid!=undefined) sanitized_rObj.sys_ref_txid=$rObj.sys_ref_txid;
        if ($rObj.annies!=undefined) sanitized_rObj.annies=$rObj.annies;
        if ($rObj.decrypted!=undefined) sanitized_rObj.decrypted=$rObj.decrypted;
        if ($rObj.flathexint!=undefined) {
            sanitized_rObj.flathexint = $rObj.flathexint;
        }
        var request = objectStore.add(sanitized_rObj);
        request.onsuccess = function(event) {
            resolve(null);
        };
        request.onerror = function(event) {
            DB.ERROR_INSERTS+=1;
            console.warn("insert error", $rObj);
            DB.reroutes_needed.push(sanitized_rObj)
            reject(null);
        }
    });
}


/**
 * 
 * @param {*} $tbl 
 * @param {*} $method 
 * @param {*} $rObj 
 * @returns 
 */
DB.____insertRecord=async($tbl, $method, $rObj)=> {
    
    //never insert this relon.
    if ($rObj.reln==MEM.INTRA_BLOCK_PENDING.nid) return;
    
    return new Promise((resolve, reject) => {   
        var transaction  = DB.db.transaction([$tbl], $method);
        if (transaction==null) return;
        var objectStore = transaction.objectStore($tbl);
        if (objectStore==null) return;
        var sanitized_rObj = {
            txid:$rObj.txid, 
            type:parseInt($rObj.type),
            from_nid:$rObj.from_nid,
            reln:$rObj.reln,
            to_type:parseInt($rObj.to_type),
            to_s:$rObj.to_s,
            time:$rObj.time,
            firstseen:$rObj.firstseen,
            freeze:$rObj.freeze
        }

        //DONT SAVE THIS, EVER
        if ($rObj.reln==MEM.INTRA_BLOCK_PENDING.nid) return resolve(null);

        //for bad ghost cleanup fixing
        if ($rObj.old_txid!=undefined) sanitized_rObj.old_txid = $rObj.old_txid;
        if (sanitized_rObj.txid==undefined || sanitized_rObj.txid==null) {
            console.log("test");
            return resolve(null);
        }
        if (sanitized_rObj.txid[0]=="z") {
            console.warn("BAD DATA -- DON'T STORE GHOST/TEMP TXIDS in relons tbl!!");
            console.warn(sanitized_rObj);
            alert("hunt this down!");
            return resolve(null);
        }

        if ($rObj.anbl_id!=undefined) sanitized_rObj.anbl_id=$rObj.anbl_id;

        if ($rObj.owned_by!=undefined && $rObj.owned_by!=null) {
            sanitized_rObj.owned_by=$rObj.owned_by;
        }
        else $rObj.owned_by="";
        if ($rObj.sponsored_by!=undefined && $rObj.sponsored_by!=null) {
            sanitized_rObj.sponsored_by=$rObj.sponsored_by;
        }
        else $rObj.sponsored_by="";
        if ($rObj.parent!=undefined) sanitized_rObj.parent=$rObj.parent;
        if ($rObj.origin!=undefined) sanitized_rObj.origin=$rObj.origin;
        if ($rObj.rorigin!=undefined) sanitized_rObj.rorigin=$rObj.rorigin;
        if ($rObj.relon_sponsor!=undefined) sanitized_rObj.relon_sponsor=$rObj.relon_sponsor;
        if ($rObj.from_text!=undefined) sanitized_rObj.from_text=$rObj.from_text;
        if ($rObj.ntype!=undefined) sanitized_rObj.ntype=$rObj.ntype;
        if ($rObj.sys_ref_txid!=undefined) sanitized_rObj.sys_ref_txid=$rObj.sys_ref_txid;
        if ($rObj.annies!=undefined) sanitized_rObj.annies=$rObj.annies;
        if ($rObj.decrypted!=undefined) sanitized_rObj.decrypted=$rObj.decrypted;
        if ($rObj.flathexint!=undefined) {
            sanitized_rObj.flathexint = $rObj.flathexint;
        }
        var request = objectStore.add(sanitized_rObj);
        request.onsuccess = function(event) {
            return resolve(null);
        };
        request.onerror = function(event) {
            DB.ERROR_INSERTS+=1;
            console.warn("insert error", sanitized_rObj);
            ANNE.TALK.printRelon($rObj);
            return resolve(sanitized_rObj.txid);
        }
    });
}


DB.___insert_single = async($obj, $tbl="relons", $method="readwrite", $callaback=null)=> {
    if (DB.bSUPPORTED==false) return;
    var transaction  = DB.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    var sanitized_rObj = {}

    if ($tbl=="relons") { 
        sanitized_rObj = {
            txid:$obj.txid, 
            type:parseInt($obj.type),
            from_nid:$obj.from_nid,
            reln:$obj.reln,
            to_type:parseInt($obj.to_type),
            to_s:$obj.to_s,
            time:$obj.time,
            firstseen:$obj.firstseen,
            freeze:$obj.freeze
        }
    
        if ($obj.old_txid!=undefined) sanitized_rObj.old_txid = $obj.old_txid;

        //DO NOT SAVE THIS, EVER
        if (sanitized_rObj.reln==MEM.INTRA_BLOCK_PENDING.nid) {
            return resolve(true);
        }
        if (sanitized_rObj.txid[0]=="z") {
            console.warn("BAD DATA -- DON'T STORE GHOST/TEMP TXIDS in relons tbl!!");
            console.warn(sanitized_rObj);
            alert("hunt this down!");
            return resolve(true);
        }

        if ($obj.anbl_id!=undefined) sanitized_rObj.anbl_id=$obj.anbl_id;

        if ($obj.owned_by!=undefined && $obj.owned_by!=null) {
            sanitized_rObj.owned_by=$obj.owned_by;
        }
        else $obj.owned_by="";
        if ($obj.sponsored_by!=undefined && $obj.sponsored_by!=null) {
            sanitized_rObj.sponsored_by=$obj.sponsored_by;
        }
        else $obj.sponsored_by="";

        if ($obj.annies!=undefined) sanitized_rObj.annies=$obj.annies;
        if ($obj.parent!=undefined) sanitized_rObj.parent=$obj.parent;
        if ($obj.origin!=undefined) sanitized_rObj.origin=$obj.origin;
        if ($obj.rorigin!=undefined) sanitized_rObj.rorigin=$obj.rorigin;
        if ($obj.relon_sponsor!=undefined) sanitized_rObj.relon_sponsor=$obj.relon_sponsor;
        if ($obj.from_text!=undefined) sanitized_rObj.from_text=$obj.from_text;
        if ($obj.ntype!=undefined) sanitized_rObj.ntype=$obj.ntype;
        if ($obj.sys_ref_txid!=undefined) sanitized_rObj.sys_ref_txid=$obj.sys_ref_txid;
        
        if ($obj.annies!=undefined) sanitized_rObj.annies=$obj.annies; 
        if ($obj.decrypted!=undefined) sanitized_rObj.decrypted=$obj.decrypted;
       
        if ($obj.flathexint!=undefined) {
            sanitized_rObj.flathexint = $obj.flathexint;
        }
    }
    else sanitized_rObj = $obj;

    return new Promise((resolve, reject) => {
        var request = objectStore.add(sanitized_rObj);
        request.onsuccess = function(event) {
            return resolve(true);
        };
        request.onerror == function(event) {
            console.log(event);
            console.log(sanitized_rObj);
            return resolve(false);
        }       
    });
}


/**
 * 
 * @param {*} $objs 
 * @param {*} $tbl 
 * @param {*} $method 
 */
DB.___updateAll = async($objs, $tbl="relons", $method="readwrite")=> {
    for (var r=0; r < $objs.length; r++) {
        await DB.___update($objs[r], $tbl, $method);
    }
}



/**
 */
DB.___update = async($obj, $tbl="relons", $method="readwrite")=> {
    if ($tbl=="relons") {
        return await DB.__handleRelonUpdate($obj, $method);
    }
    if (DB.bSUPPORTED==false) return;
    var transaction  = DB.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;

    return new Promise((resolve, reject) => {

        if ($obj==null || $obj==undefined || $obj.txid==undefined) return resolve(false);
        var request = null;
        var sanitized_rObj = {}

        var keys = Object.keys($obj);
        for (var k=0; k < keys.length; k++) {
            if (typeof ($obj[keys[k]])=="function") {
                alert("woh horsey!");
                delete $obj[keys[k]];
            }
            else if (typeof ($obj[keys[k]])=="object") {
                 delete $obj[keys[k]];
            }
            else if ($obj.to_sNOBJ!=null) {
                delete $obj.to_sNOBJ;
            }
            else if ($obj.from_nidTNOBJ!=null) {
                delete $obj.from_nidTNOBJ;
            }
        }
        
        request = objectStore.put($obj);
        request.onsuccess = function(event) {
            return resolve(true);
        };
        request.onerror = function(event) {
            console.log(event);
            return resolve(false);
        }
    });
}

DB.___deleteSetting = async($unique, $val, $user_nid)=>{
    if (DB.bSUPPORTED==false) return false;
    var transaction  = DB.db.transaction(["settings"],"readwrite");
    if (transaction==null) return;
    var objectStore = transaction.objectStore("settings");
    if (objectStore==null) return;
    return new Promise((resolve, reject) => {
        console.log("delete setting "+$unique+" val "+$val+" user nid:"+$user_nid);
       var request = objectStore.delete([$unique, $val, $user_nid]);
        request.onsuccess = function(event) { return; }
        request.onerror = function(event) { return; }
        return resolve(true);
    });
}
DB.___delete = async ($txids, $tbl="relons", $method="readwrite")=> {
    if (DB.bSUPPORTED==false) return false;
    if ($txids.length==0) return false;

    var transaction  = DB.db.transaction([$tbl], $method);
    if (transaction==null) return;

    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;

    return new Promise((resolve, reject) => {

        //we only delete one at time
        $txids.forEach(function($txid) {
            if ($txid==undefined || typeof $txid!=='string') return;   //just like continue in the forEach
            var request = objectStore.delete($txid);
            request.onsuccess = function(event) { return; }
            request.onerror = function(event) { return; }
        });

        return resolve(true);
    });
  }