/* 
MEM_UDF space is for site-specific neuron pre-loads and local DB pre-processing. 
THIS FILE MUST BE LOADED AFTER the main memory handler!!!!!

REMEMBER:  functions named _udf_XXXX are referenced external to this class 
and are pointed to by in-ANNE interface functions.  They can be changed to do 
as you please or left as default.
*/
class mem_udf {
    constructor($nObj) {
        this.appName="ANNE_EXT";
        this.THIS_APP = null;
        this._collab_relons = [];           //collab relons
        this._ghost_rels = [];
        this.aSettingsToProcess = [];

        this._collab_hlc = 0;

        this._bDOSMARTLINKUPDATES = true;
        this._bWORLDVIEWBIPASS = false;
        this._lookupPreloadedOuts = true;   //should we spend a lookup chasing the outs of a preloaded neuron?
        this._lookupPreloadedIns = true;    //should we spend a lookup chasing the ins of a preloaded neuron?
    
        //holds types of events we want to focus on for this app
        //other apps can create their own event pat neurons for specific classes.
        this._event_pats = []; 

        this._bONETIMEINIT=false;
        
        this._bSHOWINGTHREEJS=false;

        this.bGETDEDSATBOOTUP = true;       //referenced in mem

        this.el_modal = null;
        this._modal_callback_confirm = null;
        this._model_callback_cancel = null;

		this._basic_demographic_nids = [
			"7771124117310918099", "777925711019642749", "7770927790880535113", 
            "7777627441463363769", "7770855606276797271", "7770982642614293125",
            "7772878965553112082", "7775045073476665086"
    	];

        this.debug_ws_externals = [];

        this.bANUMON=true;

        this.chaseWIKIPEDIA = [];   //any array of {tnid, ntype} objs.

        this._recentsLukatSearches = null;
        this.last_wiki_search_pages = null;
    }


    isWikipedaChaseable=function($nObj){

        if ($nObj.bes[0]!=ANNE.MEM.person.temp_nid && $nObj.bes[0]!=ANNE.MEM.name.temp_nid && ANNE.USER._userNObj.nid!=ANNE.MEM.ANNE_ACCTID) {
            console.log("wikipedia queries of non-persons only available to ANNE ACCT, for now.");
            return false;
        }

        if ($nObj.is(MEM.fileX)==true) return false;
        
        if (this.chaseWIKIPEDIA.length==0) return true;
        
        if (ANNE.MEMG.isBesOfs2($nObj, this.chaseWIKIPEDIA)==false) return false;
        return true;
    }
    

    /**
     * @param {wiki_results} $wk 
     */
    setBigWikiPages=function($wk) {
        this.last_wiki_search_pages = 
            {   str:$wk.searchString, 
                pages:$wk.pObj.misc.data, 
                targetClassNOBJ:$wk.targetClassNOBJ,
                dataMap:$wk.dataMap,
                extras:$wk.extras
            };
    }
    getBigWikiPages=function($str){
        if (this.last_wiki_search_pages==null) return null;
        if (this.last_wiki_search_pages.str==$str) {
            return this.last_wiki_search_pages;
        }
        return null;
    }


    /**
     * @author {Mr.Scatman} 
     */
    init = async()=> {

        this._recentsLukatSearches = new arraymap();
        this._recentsLukatSearches.setMapKey("text");

        //allows explore of potential big data pulls and going back..
        //to get the wiki results.
        this.last_wiki_search_result = null;    ///wiki_results


    //we will expand this later.
        this.chaseWIKIPEDIA=[
            {tnid:MEM.person.temp_nid, ntype:NTYPE_INST}, 
            {tnid:MEM.company.temp_nid, ntype:NTYPE_INST}, 
            {tnid:MEM.group.temp_nid, ntype:NTYPE_INST},  
            {tnid:MEM.group_event.temp_nid, ntype:NTYPE_INST}, 
            {tnid:MEM.location.temp_nid, ntype:NTYPE_INST},  
            {tnid:MEM.name.temp_nid, ntype:NTYPE_CLASS}, 
            {tnid:MEM.religion.temp_nid, ntype:NTYPE_INST}, 
            {tnid:MEM.political_party.temp_nid, ntype:NTYPE_INST}, 
            {tnid:MEM.ec_thing.temp_nid, ntype:NTYPE_INST},
            {tnid:MEM.ec_living.temp_nid, ntype:NTYPE_INST}, 
            {tnid:MEM.ec_nonliving.temp_nid, ntype:NTYPE_INST},
            {tnid:MEM.service_term.temp_nid, ntype:NTYPE_INST},
            ];

            //TODO:
            //{tnid:MEM.ec_nonliving.temp_nid, ntype:NTYPE_INST},  INST or CLASS HERE?  CHECK.
            

        //a list of valid identity classes (this app accepts) when a new user is created.
        this.valid_persona_tnids = [MEM.application_identity.temp_nid, 
                            MEM.corporation_identity.temp_nid, 
                            MEM.individual_identity.temp_nid,
                            MEM.miner_identity.temp_nid,
                            MEM.annode_identity.temp_nid,
                            MEM.oracle_identity.temp_nid];
    


        var ANNENOBJ = MEMG.getNeuronByNid(MEM.ANNE_ACCTID);
        
        MEM.addToAppWideActiveWorldViews(ANNENOBJ.temp_nid);    

        this.el_modal = ANNE.A.__getEl("anne_modals");
        this.el_modal_cancel = ANNE.A.__getEl("btn_anne_modal_close");
        this.el_modal_cancel.addEventListener("click", e=>this.popupElCancel(e));
        this.el_modal_confirm = ANNE.A.__getEl("btn_anne_modal_confirm");
        this.el_modal_confirm.addEventListener("click", e=>this.popupElConfirm(e));
        this.el_site = ANNE.A.__getEl("anne_site");
        
        //for collab/ghosts/ and  prefs not stored on the ANNODE.
        if (typeof DB_local != "undefined") {
            ANNE.DB_local = DB_local;
            if (!('indexedDB' in window)) {
                console.log('This browser does not support IndexedDB');
                ANNE.DB_local.bSUPPORTED=false;
                ANNE.DB_local=null;
            }
            else {	
                ANNE.DB_local.bSUPPORTED=true;
                return await ANNE.DB_local.___init(MEM_UDF._finisheAppDBPreInit); 
            }
        }
        else ANNE.DB_local=null;

      
        return;
    }


    bootup_complete = async()=> {
        ANNE.BRAIN.initLight();
    }

    fireNid = function($nid, $fireWeight=1) {
        ANNE.BRAIN.ctp_light.addNid($nid, $fireWeight);
    }
    fireNeuron = function($nObj, $fireWeight=1) {
        if (ANNE.BRAIN.ctp_light==null) return;
        ANNE.BRAIN.ctp_light.addNObj($nObj, $fireWeight);
    }
    coolFires = function() {
        ANNE.BRAIN.ctp_light.coolFires();
    }
    printPool=function($sort=POOL_SORT_FIRE) {
        ANNE.BRAIN.ctp_light.printPool($sort);
    }


    resetUser = async()=> {
        this._collab_relons = [];           //collab relons
        this._ghost_rels = [];
        this.aSettingsToProcess = [];

        //rewire this to use our madeNew
        USER.__loginCheck(this.swappedUsers, this.noANNEID);
        
    }
    

    /**
     * 
     */
    doConfirmModalWithEl=async($cloneEl, $cb_confirm=null, $cb_cancel=null, $title="continue?")=>{
        ANNE.A.setHTML("anne_modal_title", $title); 
        ANNE.A.show(this.el_modal_confirm);
        await this.popupEl($cloneEl);
        this._modal_callback_confirm = $cb_confirm;
        this._model_callback_cancel = $cb_cancel;
        ANNE.playSound("confirm");
    }

    /**
     * populate the modal div with the $removeEl div.
     * @param {*} $cloneThisEl 
     */
    popupEl=async($cloneThisEl, $cb_confirm=null, $cb_cancel=null, $title="")=>{
        ANNE.A.empty(this.el_modal.id+"_body");
        ANNE.A.setHTML("anne_modal_title", $title); 
        ANNE.A.show(this.el_modal_confirm);
        this._modal_callback_confirm = $cb_confirm;
        this._model_callback_cancel = $cb_cancel;

        var newel = ANNE.A.cloneAndAppend($cloneThisEl, this.el_modal.id+"_body");
        ANNE.A.show(newel);
        ANNE.A.show(this.el_modal);
        ANNE.A.show(this.el_modal.id+"_blocker");
        ANNE.A.addClass(this.el_site, "blur-it");
        this.el_site.setAttribute("disabled", true);
    }

    /**
     */
    popupElCancel=async($e, $bPLAYSOUND=false)=>{    
        console.log("%cmodal cancelled", TALK.bgred);
        ANNE.A.empty(this.el_modal.id+"_body", true);
        ANNE.A.hide(this.el_modal.id+"_blocker");
        ANNE.A.hide(this.el_modal);
        ANNE.A.removeClass(this.el_site, "blur-it");
        this.el_site.setAttribute("disabled", false);
        
        if ($bPLAYSOUND==true) ANNE.playSound("nope");

        if (this._model_callback_cancel!=null) {
            this._model_callback_cancel();
        }
        this._model_callback_cancel = null;
        this._modal_callback_confirm = null;
    }


    /**
     * popup confirm
     */
    popupElConfirm=async($e)=>{
        console.log("%cmodal confirmed", TALK.bggreen);
        ANNE.A.empty(this.el_modal.id+"_body", true);
        ANNE.A.hide(this.el_modal.id+"_blocker");
        ANNE.A.hide(this.el_modal);
        ANNE.A.removeClass(this.el_site, "blur-it");
        this.el_site.setAttribute("disabled", false);
        if (this._modal_callback_confirm!=null) {
            this._modal_callback_confirm();
        }
        this._model_callback_cancel = null;
        this._modal_callback_confirm = null;
    }


    /**
     * 
     */
    loadAnnodeSummaryDetails = async()=> {
        var aObj = await ANNE.API.annodeDetails();
        var el = ANNE.A.__getEl("main_recap");
        if (el==null) return;
        el.style['z-index']=9999;
        
        //link to to pullup the annode props and change this?
        ANNE.A.setHTML("main_network", aObj.network);
        ANNE.A.setHTML("main_uri", ANNE._annodesLS.last_uri);
        ANNE.A.setHTML("main_height", ANNE.UTILS.commas(aObj.height));
        ANNE.A.setHTML("main_neurons", ANNE.UTILS.commas(aObj.cnt_neurons));
        ANNE.A.setHTML("main_relons", ANNE.UTILS.commas(aObj.cnt_relons));
        ANNE.A.setHTML("main_awareness_relons", ANNE.UTILS.commas(aObj.cnt_awareness_relons));
        ANNE.A.setHTML("main_payment_relons", ANNE.UTILS.commas(aObj.cnt_payment_relons));
        ANNE.A.setHTML("main_local_neurons", ANNE.UTILS.commas(ANNE.MEM._aNeurons.length));
        ANNE.A.setHTML("main_local_relons", ANNE.UTILS.commas(ANNE.MEM._aRelons.length));

        var totalRelonsIDB = await ANNE.DB.getCnt(["relons", "relonsMinBoot"]);
        ANNE.A.setHTML("main_indexdb_relons", ANNE.UTILS.commas(totalRelonsIDB));

        ANNE.HISTORY.updateGLKB_SOLID(aObj.height);

        if (ANNE.API._AAUTH && ANNE.API._AAUTH.network==true) {
            ANNE.A.setHTML("main_newtork", "<span class='text-active'>active</span>");
        }
        else ANNE.A.setHTML("main_newtork", "<span class='text-deactive'>deactive</span>");

        //ANNDLE_active
        if (aObj.anndle_active==true) {
            ANNE.A.setHTML("main_anndle", "<span class='text-active'>active</span>");
        }
        else ANNE.A.setHTML("main_anndle", "<span class='text-deactive'>deactive</span>");

        //PDN_active
        if (aObj.pdn_active==true) {
            ANNE.A.setHTML("main_pdn", "<span class='text-active'>active</span>");
        }
        else ANNE.A.setHTML("main_pdn", "<span class='text-deactive'>deactive</span>");

                
        //ASDF_active
        if (aObj.asdf_active==true) {
            ANNE.A.setHTML("main_asdf", "<span class='text-active'>active</span>");
        }
        else ANNE.A.setHTML("main_asdf", "<span class='text-deactive'>deactive</span>");

        //BOUNTY_active
        if (aObj.bounty_active==true) {
            ANNE.A.setHTML("main_bounty", "<span class='text-active'>active</span>");
        }
        else {
            ANNE.A.setHTML("main_bounty", "<span class='text-deactive'>deactive</span>");
        }

        ///PDNS_active
        if(aObj.pdn_active==true) {
            ANNE.A.setHTML("main_pdns", "active");
        }
        else {
            ANNE.A.setHTML("main_pdns", "deactive");
        }
    }



    /**
     * Here u get whatever u store in your app-specific DB and u process it. 
     * This could be app settings for gui or any persistent data you chose  
     * to store for the user locally for whatever reason.                   
     *                                                                       
     * The aPullHistory is available for you to use to optimize any non-ANNE 
     * local app queries of non-ANNE calls to your data providers.
     * 
     * @author {Mr.Scatman} 
     */ 
    _finisheAppDBPreInit=async($aFullPullHistory, $callbackCallback)=> {
        console.log("%c (_finisheAppDBPreInit)", TALK.bgredbig);
        this.THIS_APP = MEM.ANNE_EXT_APP;
    
        this.el_logo_main = document.getElementById("main_logo");
        this.el_recap = document.getElementById("main_recap");

        if (this.el_logo_main) {
            this.el_logo_main.addEventListener("click", async($e)=>{
                if (USER._userNObj==null && USER.seedObj==null) return;    
                var rez = ANNE.A.toggle(this.el_recap);
                if (rez=="show") await this.loadAnnodeSummaryDetails();
            })
            this.el_recap.addEventListener("click", async($e)=>{
                ANNE.A.toggle(this.el_recap);
            });

        }
        return;
    }


    /**
     * Let's pull collabs everytime from the annode.
     * This is cleaner with multiple users.
     */
    preloadCollabs = async()=> {
        this._collab_hlc = 0;
        var mem_collabs = await ANNE.DB_local.__getCollabRelons();
        var bLOADONLY=true;
        if(mem_collabs.length > 0) this._processCollabs(mem_collabs, bLOADONLY); 
    }


    /**
     * Used to assess if we can do a hanai collab all here.
     * @param {*} $nid 
     * @returns 
     */
    pendingHanaisFor = function($nid) {
        var arr=[];
        for (var c=0; c < MEM_UDF._collab_relons.length; c++) {
            var crelon = MEM_UDF._collab_relons[c];
            if(crelon.status_text!="p") continue;
            if (crelon.collab_type!=CTYPE_HANAI_ONLY) continue;
            if (crelon.relon_sponsor!=$nid) continue;
            arr.push(crelon);
        }
        return arr;
    }

    pendingHanaiCollabsFor = function($nid) {
        var arr=[];
        for (var c=0; c < MEM_UDF._collab_relons.length; c++) {
            var crelon = MEM_UDF._collab_relons[c];
            if (crelon.status_text!="p") continue;
            if (crelon.collab_type!=CTYPE_COLLAB_HANAI) continue;
            if (crelon.relon_sponsor!=$nid) continue;
            arr.push(crelon);
        }
        return arr;
    }


    /**
    */
    getPendingCollabs = async()=> {
        var pending = [];
        var deletes = [];
        
        if (ANNE.SETTINGS.debug()==true) ANNE.TALK.printRelons(MEM_UDF._collab_relons);
        
        for (var c=0; c < MEM_UDF._collab_relons.length; c++) {
            var crelon = MEM_UDF._collab_relons[c];
            if (crelon.sys_ref_txid!=undefined && crelon.sys_ref_txid.length>0) {
                deletes.push(crelon.txid);
                MEM_UDF._collab_relons.splice(c, 1);
                c=c-1;
                continue;
            }

            if(crelon.status_text=="p") {
                var txid = MEM_UDF._collab_relons[c].txid;
                var bSKIP=false;

                for (var r=0; r < USER._userNObj.outs.length; r++) {
                    var rel = USER._userNObj.outs[r];
                    if (rel.to_s==txid) {
                        bSKIP=true;
                        crelon.status_txid=rel.txid;
                        deletes.push(crelon);
                        MEM_UDF._collab_relons.splice(r, 1);
                        c=c-1;
                        break;
                    }
                }
                
                if(bSKIP==true) continue;
                
                pending.push(MEM_UDF._collab_relons[c]);
            }
        }

        //cleanup first
        if (deletes.length) {
            if (ANNE.SETTINGS.debug()==true) {
                console.warn("deleting "+deletes.length+" collab relons from indexdb mem");
            }
            await DB_local.___delete(deletes, "collab_relons");
        }

        return pending;
    }
    
    
    /**
     * we don't want to put these in MEM _aRelons.. 
     * They are not real yet.. 
     * We already have upgraded ghost to real for the childRelonTX (broadcaster)
     * and without work it will look real.
     */
    _processCollabs=async($aCollabs, $bLOADONLY=false)=> {
        var adds = [];
        var deletes = [];
        for (var a=0; a < $aCollabs.length; a++) {
            var col = $aCollabs[a];
            col.ghost=false;
            col.temp=false;
            col.andy=true;
            col.rorigin=RORIGIN_COLLAB_BROADCAST;
            
            if (col.collaborator!=undefined) col.relon_sponsor = col.collaborator;
           
            if (col.height_last_change!=undefined && 
                col.height_last_change!=null
                && parseInt(col.height_last_change)>this._collab_hlc) {
                this._collab_hlc=parseInt(col.height_last_change);
            }

            
            var broadcastRELON = MEMG.lookupByTXID(col.txid);
            
            //we really don't want status_text on broadcast REL mem.
            if (broadcastRELON!=null) {
                delete broadcastRELON.status_text;
                delete broadcastRELON.status_txid;
            }
            
            //broadcast RELON is the collab/andy req sent from the user.
            //need to figure a cleaner way to manage this.
            //if (broadcastRELON!=null && broadcastRELON.ghost==false) {
               // MEM_UDF._collab_relons = UTILS.removeRelonByTxid(MEM_UDF._collab_relons, col.txid);
               // deletes.push(col.txid);
            // note:  status_txid of collab s.b. equal to sys_ref_txid of collab.
               // continue;
            //}

            var bDELETEIT = false;
            
            if (col.status_text!="a") {
                col.isAndyRel = true;

                //a helpful backlink to the andy rel for any given broadcast relon that has andy
                //admin impacts.
                if (broadcastRELON!=null) {
                    MEM_UDF.setCollabRelIndexes(col); //arrrg.
                    broadcastRELON.andy_rel = col;   
                }
            }
            else {
                deletes.push(col.txid);
                bDELETEIT=true;
            }

            
            var existingCollabRObj = null;
            for (var aa=0; aa < MEM_UDF._collab_relons.length; aa++) {
                if (MEM_UDF._collab_relons[aa].txid==col.txid) {
                    if (bDELETEIT==true) {
                        MEM_UDF._collab_relons.splice(aa,1);
                        break;
                    }
                    existingCollabRObj = MEM_UDF._collab_relons[aa];
                    break;
                }
            }

            
            
            //if we don't have an existing collabRObj
            //set the status_txid and status_text fields
            if (col.status_txid==null || col.status_txid==undefined) col.status_txid="";
            if (col.status_text==null || col.status_text==undefined) col.status_text="";
            
            if (bDELETEIT==false && existingCollabRObj==null) {
                adds.push(col);
                MEM_UDF._collab_relons.push(col);
            }

            //if we have new info (height change on edit or status text change)..
            //update the local INDEXDB collab record.  We don't want the old info
            //next time we pull local collabs from the local db.
            else if (bDELETEIT==false && ((col.height_last_change!= existingCollabRObj.height_last_change) 
                    || (existingCollabRObj.status_text!=col.status_text && col.status_txid==""))) {
                await DB_local.___update(col, "collab_relons");         //edit one by one
            }
            
            var fromNOBJ = MEMG.getOrMakeIfNew(col.from_nid);

            //store andy requests on their fromNOBJ so they are visible in
            //the neuron viewer as pending items that require action
            //(when do we clean up these run-time tho?  lol.)
            if (bDELETEIT==false && fromNOBJ.andy_rels==undefined) {
                fromNOBJ.andy_rels=[col];
            }
            else if (bDELETEIT==false) {
                ANNE.UTILS.addUniqueRel(fromNOBJ.andy_rels, col);
            }
            else if (bDELETEIT==true) {
                if (fromNOBJ.andy_rels!=undefined && fromNOBJ.andy_rels.length>0) {
                   ANNE.UTILS.findValAndRemoveInObjnArr(fromNOBJ.andy_rels, "txid", col.txid);
               }
            }
        }

        if (adds.length && $bLOADONLY==false) {
            await DB_local.___insert(adds, "collab_relons");//insert in bulk.
        }
        
        if (deletes.length) {
            await DB_local.___delete(deletes, "collab_relons");
        }
        
        //set the relon indexes (.f, .r, [.t]) since they weren't run thru the proper mem processor.
        //this should have handled our andy_rel ptr but apparently it doesn't
        //thus we do this above as well.  (silly rubes)
        for (var a=0; a < MEM_UDF._collab_relons.length; a++) {
            MEM_UDF.setCollabRelIndexes(MEM_UDF._collab_relons[a]);
        }

        return;
    }


    /**
     * We broadcast a new collab and/or hanai relon.
     * Loosely add it in to our RAM array here.
     * 
     * For auto-hanai/collab this will be updated with status_txid whenever
     * the relon is mined and repulled.
     * 
     */
    addCollabFromBroadcast=function($newBroadcastRelon, $collab_type) {
        if (ANNE.MEM.isANNEADMIN(ANNE.USER._userNObj.nid)==true) return;
        for (var c=0; c < MEM_UDF._collab_relons.length; c++) {
            if (MEM_UDF._collab_relons[c].txid==$newBroadcastRelon.txid) {
                console.warn("IS THIS EVER RELEVANT?? , SEEMS LIKE IT SHOULDN'T BE");
                return;
            }
        }
        $newBroadcastRelon.isAndyRel = true;
        $newBroadcastRelon.relon_sponsor = USER._userNObj.nid;
        $newBroadcastRelon.status_text = "p";
        $newBroadcastRelon.status_txid = null;
        if ($collab_type==CTYPE_COLLAB_ONLY) {
            if ($newBroadcastRelon.r==MEM.CMD_MERGE_NEURON.temp_nid) {
                $collab_type = CTYPE_MERGE_REQUEST;
            }
        }
        $newBroadcastRelon.collab_type = $collab_type;
        MEM_UDF._collab_relons.push($newBroadcastRelon);

        if ($newBroadcastRelon.old_txid==undefined) return;

        //remove old txid from collab_relons array.  
        //(we just added the new updated one)
        //we will later remove it from ghost db mem by txid.
        var old_txid = $newBroadcastRelon.old_txid;
        for (var c=0; c < MEM_UDF._collab_relons.length; c++) {
            var rel =  MEM_UDF._collab_relons[c];
            if (rel.txid==old_txid) {
                MEM_UDF._collab_relons.splice(c, 1);
                console.warn("DELETED OLD COLLAB RELON AND REPLACED WITH NEW.")
                return;
            }
        }
    }


    /**
     */
    getCollabType=function($rObj) {
        if ($rObj.collab_type==undefined) return "";
        if ($rObj.collab_type==CTYPE_COLLAB_ONLY) return "collab";
        if ($rObj.collab_type==CTYPE_COLLAB_HANAI) return "hanai -> collab";
        if ($rObj.collab_type==CTYPE_PENDING_ANJ) return "pending anjucator";
        if ($rObj.collab_type==CTYPE_HANAI_ONLY) return "hanai";
        if ($rObj.collab_type==CTYPE_MERGE_REQUEST) return "merge";
        console.log("unhandled collab type "+$rObj.collab_type);
        return "";
    }

    /**
     */
    getCollabTypeMsg=function($rObj) {
        if ($rObj.collab_type==undefined) return "(unhandld msg)";
        if ($rObj.collab_type==CTYPE_COLLAB_ONLY) {
            var fromNOBJ = MEMG.getNeuronByNid($rObj.from_nid);
            return "This is a pending <b>collab</b> request to add data to the <b>"+fromNOBJ.text+"</b> neuron";
        }
        if ($rObj.collab_type==CTYPE_COLLAB_HANAI) {
            var fromNOBJ = MEMG.getNeuronByNid($rObj.from_nid);
            var toNOBJ = MEMG.getNeuronByNid($rObj.to_s);
            return "This is a pending <b>hanai->collab</b> request to hanai the <b>"+toNOBJ.text+"</b> neuron and add data to the <b>"+fromNOBJ.text+"</b> neuron.";
        }
        if ($rObj.collab_type==CTYPE_PENDING_ANJ) return "pending anjucator";
        if ($rObj.collab_type==CTYPE_HANAI_ONLY) {
            var fromNOBJ = MEMG.getNeuronByNid($rObj.from_nid);
            return "This is a pending <b>hanai</b> request of the <b>"+fromNOBJ.text+"</b> neuron.";
        }
        if ($rObj.collab_type==CTYPE_MERGE_REQUEST) {
            var fromNOBJ = MEMG.getNeuronByNid($rObj.from_nid);
            var toNOBJ = MEMG.getNeuronByNid($rObj.to_s);
            return "This is a pending <b>collab merge</b> request to merge the <b>"+fromNOBJ.text+"</b> and <b>"+toNOBJ.text+"</b> neurons.";
        }
        console.log("unhandled collab type "+$rObj.collab_type);
        return "";
    }


    /**
     * We do not clutter MEM._aRelons space with collabs that are 
     * pending.. (unless we sent them and it is our worldview that wants them).
     * As an ANDY (admin) we hold these in a different area.  We do pass them in
     * to relon tables -- which expect the .f, .r, and .t temp_nid indexes to be set.
     * As such, we must set them to have them.
     */
    setCollabRelIndexes = function($collabRel) {
        var fNOBJ = MEMG.getOrMakeIfNew($collabRel.from_nid);
        if (fNOBJ!=null) $collabRel.f=fNOBJ.temp_nid;
        var rNOBJ = MEMG.getOrMakeIfNew($collabRel.reln);
        if (rNOBJ!=null) $collabRel.r=rNOBJ.temp_nid;
        if ($collabRel.to_type==TO_TYPE_ANNE) {
            var tNOBJ  = MEMG.getOrMakeIfNew($collabRel.to_s);
            if (tNOBJ!=null) $collabRel.t=tNOBJ.temp_nid;
        }
    }


    /**
     * Get relons which this ANDY user has rights to approve or decline.
     * 
     * These must follow the relon table syntax of an object with rObj: and active:true.
     * This allows for the DOM filters to work (by toggline the active true/false)
     * if the DOM filter has filtered it out.  
     * @param {string} - Nid long string for the neuron we are viewing.
     * @author {Mr.Scatman} 
     * 
     */
    getAndyTblRelonsForNid = function($nid) {
        var andy_tbl_rel = [];  
        var txids = [];
        for (var c=0; c < MEM_UDF._collab_relons.length; c++) {
            var cRelon = MEM_UDF._collab_relons[c];
            cRelon.rorigin=RORIGIN_COLLAB_BROADCAST;
            if (cRelon.from_nid==$nid) {

                //accepted have real relons.
                if (cRelon.status_text=="a") continue;

                if (txids.indexOf(cRelon.txid)!=-1) continue;
                andy_tbl_rel.push(cRelon);//{rObj:cRelon, active:true});     
                txids.push(cRelon.txid);
            }
        }
        return andy_tbl_rel;
    }

    getAndyInsForNid = function($nid) {
        var andy_ins = [];  
        var txids = [];
        for (var c=0; c < MEM_UDF._collab_relons.length; c++) {
            var cRelon = MEM_UDF._collab_relons[c];
            cRelon.rorigin=RORIGIN_COLLAB_BROADCAST;
            if (cRelon.to_s==$nid) {
                //accepted have real relons.
                if (cRelon.status_text=="a") continue;
                if (txids.indexOf(cRelon.txid)!=-1) continue;
                andy_ins.push(cRelon);  
                txids.push(cRelon.txid);
            }
        }
        return andy_ins;
    }

    
    /**
     * Look in ram.. 
     * then look on the ANNE hypergraph..
     * last resort, look in the raw txs we have in ram
     */
    getAnyRelChase=async($txid)=> {
        var rel = MEMG.lookupByTXID($txid);
        if (rel==null) {
            rel = await CHASE.lookupTxid($txid);
            console.log(rel);
        }
        if (rel==null) {
            rel = this.getAndyRel($txid);
        }
        return rel;
    }


    getSysRelonWithNoSysRefTxid=function($outs, $from_nid, $relnTnid) {
        for (var r=0; r < $outs.length; r++) {
            if ($outs[r].r==$relnTnid && $outs[r].type==TYPE_SYS) {
                if ($outs[r].sys_ref_txid==null || $outs[r].sys_ref_txid=="") return $outs[r];
            }
        }
        
        for (var c=0; c < MEM_UDF._collab_relons.length; c++) {
            var cRelon = MEM_UDF._collab_relons[c];
            if (cRelon.from_nid!=$from_nid) continue;
            if (cRelon.type!=TYPE_SYS) continue;
            if (cRelon.r!=$relnTnid) continue;
            if (cRelon.sys_ref_txid==null || cRelon.sys_ref_txid=="") return cRelon;
        }

        return null;
    }

    /**
     * @param {string} $txid - txid long string.
     */
    getAnyRel=function($txid){
        var rel = this.getAndyRel($txid);
        if (rel==null) {
            rel =  MEMG.lookupByTXID($txid);
        }
        return rel;
    }

    getAndyTxidsFrom=function($f) {
        var txids = [];
        for (var c=0; c < MEM_UDF._collab_relons.length; c++) {
            var cRelon = MEM_UDF._collab_relons[c];
            if (cRelon.f==$f && ANNE.UTILS.isset(cRelon.status_txid)==false) {
                txids.push(cRelon.txid);
            }
        }
        return txids;
    }
    
    getAndyCollabRel=function($nObj) {
        for (var c=0; c < MEM_UDF._collab_relons.length; c++) {
            var cRelon = MEM_UDF._collab_relons[c];
            if (cRelon.collab_type==CTYPE_COLLAB_HANAI) {
                if (cRelon.to_type==TO_TYPE_ANNE && cRelon.t==$nObj.temp_nid) return cRelon;
            }
            else if (cRelon.collab_type==CTYPE_HANAI_ONLY) {
                if (cRelon.f==$nObj.temp_nid) return cRelon;
            }
            else continue;
        }
        return null;
    }

    getAnyHasRelsFromWith=function($nObj, $relNOBJ) {
        var andy_rels = this.getAndyRelsFromWithTo($nObj.nid, $relNOBJ.nid);
        var rels = MEMG.getHasRelons($nObj.outs, $relNOBJ.temp_nid);
        return ANNE.UTILS.concatUniqueRels(andy_rels, rels);
    }

    getAndyRelsFromWithTo=function($from_nid, $reln) {
        var rels = [];
        for (var c=0; c < MEM_UDF._collab_relons.length; c++) {
            var cRelon = MEM_UDF._collab_relons[c];
            if (cRelon.type!=TYPE_HAS) continue;
            if (cRelon.from_nid!=$from_nid) continue;
            if (cRelon.reln!=$reln) continue;
            rels.push(cRelon);
        }
        return rels;
    }

    /**
     */
    getAndyRel=function($txid){
        if (ANNE.SETTINGS.debug()==true) {
            console.log("%c lookup andy rel:%c"+$txid, TALK.bggreen, TALK.bgred);
        }
        for (var c=0; c < MEM_UDF._collab_relons.length; c++) {
            var cRelon = MEM_UDF._collab_relons[c];
            if (cRelon.txid==$txid) return cRelon;
        }
        return null;
    }

    /**
     */ 
    getMyFeelzFor=function($target_id, $to_type, $bGHOSTSONLY=false) {
        var matches = [];
        for (var u=0; u < USER._userNObj.outs.length; u++) {
            var rel = USER._userNObj.outs[u];
            if (rel.type!=TYPE_AWARENESS) continue;
            if ($bGHOSTSONLY==true && rel.ghost==false) continue;
            if (rel.to_s==$target_id && rel.to_type==$to_type) matches.push(rel);
        }
        return matches;
    }


    
    /**
     * Accept a collab relon.
     * 
     * These workflow approve/deny relons are in the TYPE_SYS_MSG dimension.
     * We don't want to add to the load required to get basic info when searching for
     * neurons and TYPE_SYS / TYPE_BE.  These are best served in the TYPE_SYS_MSG dimension.
     */ 
     _makeAndyAccept = async($guid, $txid, $action)=> {
        if ($action!=MEM.andy_collab_accepted.temp_nid) return;
        return await MEMZ.ez_new_sys_msg(USER._userNObj, MEM.andy_collab_accepted, TO_TYPE_TXID, $txid, {collab_txid:$txid, guid:$guid, status_text:"accept"});
    }


    /**
     * Reject a collab relon.
     * 
     * These workflow approve/deny relons are in the TYPE_SYS_MSG dimension.
     * We don't want to add to the load required to get basic info when searching for
     * neurons and TYPE_SYS / TYPE_BE.  These are best served in the TYPE_SYS_MSG dimension.
     */ 
     _makeAndyReject = async($guid, $txid)=> {
        if ($action!=MEM.andy_collab_rejected.temp_nid) return;
        return await MEMZ.ez_new_sys_msg(USER._userNObj, MEM.andy_collab_rejected, TO_TYPE_TXID, $txid, {collab_txid:$txid, guid:$guid, status_text:"reject"});
    }


    /**
     * Splice out the collab txid and remove it from collab memory.
     */
    removeCollab = async($collab_txid)=> {
        var delete_relon = null;
        for (var c=0; c < MEM_UDF._collab_relons.length; c++) {
            var cRelon = MEM_UDF._collab_relons[c];
            if (cRelon.txid==$collab_txid) {
                delete_relon = cRelon;
                MEM_UDF._collab_relons.splice(c, 1)
                break;
            }
        }
        if (delete_relon==null) return null;
        await DB_local.___delete([delete_relon]);
        return delete_relon;
    }


   /**
     * Optional implementation. 
     * @author {Mr.Scatman} 
     */
    udf_handleWindowResize = async()=> {
        alert("udf_handleWindowResize");
        //i will piggy-back the cheap events here.
        await NF.cheapEvents({event:"RESIZE_WINDOW", data:null});
    }



    hasANNEID=async()=> {
        if (ANNE.SETTINGS.bCONSOLE_DEBUG==true) console.log("(hasANNEID) alreadyLoggedIn");
        ANNE.A.hide("loading");

        //if we are using pages built in to the ANNE code, call the locked here.
        //this will be for the extension type app
        //
        if (ANNE.PAGES!=null) {

            if (USER.seedObj==null) {
                ANNE.PAGES.openPage("locked");
            }

            //we get here from new account creation process.
            //hasANNEID is the callback from the user mod.
            //we have a seedObj so we open anneid.
            //before we do that, init the btns and the ready the threejs stuff.
            //one-time.
            else {
                await this.unlocked_pre(false, "login");
                ANNE.PAGES.openPage("anneid", null, null, null, null, (e)=>{
                    ANNE.A.gs_alphaIn("leftMainBtnMenuXX");
                });
            }
        }
    }


    /**
     * 
     */
    swappedUsers=async()=>{
        if (ANNE.PAGES!=null) {
            if (USER.seedObj==null) {
                ANNE.PAGES.openPage("locked");
            }

            //we get here from new account creation process.
            //hasANNEID is the callback from the user mod.
            //we have a seedObj so we open anneid.
            //before we do that, init the btns and the ready the threejs stuff.
            //one-time.
            else {
                var bSWAPUSERBIPASS = true;
                await this.unlocked_pre(bSWAPUSERBIPASS, "swapped users");
                ANNE.PAGES.openPage("anneid");
                ANNE.A.show("leftMainBtnMenuXX");
            }
        }
    }


    makeRelonViewLink_SMART=function($el, $relon, $bEMPTY=true, $h=8, $imgOnly=false, $relonClass="", $txtoverride="") {
        return ANNE.EXT_BTNS.makeRelonViewLink_SMART($el, $relon, $bEMPTY, $h, $imgOnly, $relonClass, $txtoverride);
    }
    makeRelonViewLink=function($el, $txid, $bEMPTY=true, $h=8, $imgOnly=false, $relonClass="", $txtoverride="") {
        return ANNE.EXT_BTNS.makeRelonViewLink($el, $txid, $bEMPTY, $h, $imgOnly, $relonClass, $txtoverride);
    }

    makeNeuronLink=function($el, $nid, $class="", $pre="", $bADDCOPY=false, $txtoverride="", $page="dossier", $classpre="") {
        return ANNE.EXT_BTNS.makeNeuronLink($el, $nid, $class, $pre, $bADDCOPY, $txtoverride, $page, $classpre);
    }
    
    sendFeelzQueue = async($refreshCallbackTemp=null, $mustRunPreTestFunction=null, $cancelSendCB=null, $failedCB=null)=>{
        ANNE.A.hide('fzWidgetOverlayOutter_ex');

        return await this.sendQueue($refreshCallbackTemp, $mustRunPreTestFunction, $cancelSendCB, $failedCB);
    }

    sendQueue = async($refreshCallbackTemp=null, $mustRunPreTestFunction=null, $cancelSendCB=null, $failedCB=null)=>{
        return await SENDQ.___sendQueue($refreshCallbackTemp, $mustRunPreTestFunction, $cancelSendCB, $failedCB);
    }



    /**
     */
    doConfirmModalWithHTML=async($html, $addendum, $cb_confirm=null, $cb_cancel=null, $title="continue?", $ico="", $bMSGONLYOVERRIDE=false)=> {
        ANNE.A.setHTML("anne_modal_title", $title); 
        if ($ico!="") {
            var ico = GUI_STUFF.getIco($ico, 48, "");
            if (ico!=null && ico!="") {
                ANNE.A.setHTML("anne_modal_ico", ico);
                ANNE.A.show("anne_modal_ico")
            }
            else ANNE.A.hide("anne_modal_ico");
        }
        else ANNE.A.hide("anne_modal_ico");

        ANNE.A.setHTML(this.el_modal.id+"_body", $html, true);
        ANNE.A.setHTML(this.el_modal.id+"_addendum", $addendum, true);
        ANNE.A.show(this.el_modal);
        ANNE.A.show(this.el_modal.id+"_blocker");
        ANNE.A.addClass('anne_site', "blur-it");

       this.el_site.setAttribute("disabled", true);

        if ($bMSGONLYOVERRIDE==true) {
           this._modal_callback_confirm = null;
            ANNE.A.hideParent("btn_anne_modal_confirm");
            //border red
        }
        else {
            this._modal_callback_confirm = $cb_confirm;
            this._model_callback_cancel = $cb_cancel;
            ANNE.A.showParent("btn_anne_modal_confirm");
            //border yellow
        }

        ANNE.playSound("confirm");
    }

    
    /**
     * 
     */
    noANNEID=async()=>{
        //console.log("noANNEID");
        var ah = localStorage.getItem('ANNE_history');

        if (ah!=null) {
            var ANNE_HISTORY = JSON.parse(ah);
            delete ANNE_HISTORY.default_unlock_pw;
            USER.unlock_pw = "";
            localStorage.setItem('ANNE_history', JSON.stringify(ANNE_HISTORY));
            //if we have a default_unlock_pw, use it to default the unlock pw for the new acct.
            //if (ANNE_HISTORY.default_unlock_pw!=undefined) {
                //USER.unlock_pw = ANNE_HISTORY.default_unlock_pw;
            //}
            //else USER.unlock_pw = "";
        }
        
        A.hide("loading");
        await ANNE.PAGES.openPage("login");
    }


    /**
     * 
     */
    unlocked = function() {  
        if (ANNE.SETTINGS.bCONSOLE_DEBUG==true) console.log("** unlocked! **");
    }



    /**
     * Footer clicks are only available from the threejs app.
     */
    unlocked_pre = async($bSWAPUSERBIPASS=false, $fromwhat="")=> {
        console.log("%c UNLOCK.unlocked_pre %c from: "+$fromwhat, TALK.bggreen, TALK.bgyellow);
        if (this._bONETIMEINIT==true && $bSWAPUSERBIPASS==false) {
            await this.launchUDF_firstPage();
            return;    
        } 

        var el_loader = ANNE.A.show("login_loader");
        ANNE.A.hide("main_logo");
        
        this._bONETIMEINIT=true;
        ANNE.A.empty(PAGES.el_home_view);   //empty the home view.


        //we want to get beyond this block when the user swaps users..
        //we don't want to redo these events but we do want to do what is after it.
        if ($bSWAPUSERBIPASS==false) {
            var btnback = ANNE.A.__getEl('btn_back');
            btnback.addEventListener("click", (e)=>{
                e.stopPropagation();
                e.preventDefault();
                PAGES.historyGoBack()
            });
            //anneid btns (header, left overlay footer)
            var btnanneid = ANNE.A.__getEl('btn_header_anneid');
            btnanneid.addEventListener("click", (e)=>{
                e.stopPropagation();
                e.preventDefault();
                ANNE.MEM_UDF.hide3js();
                EXT_BTNS.clicked("anneid");
            });
            var btnanneid_footer = ANNE.A.__getEl('btn_footer_anneid');
            btnanneid_footer.addEventListener("click", (e)=>{
                e.stopPropagation();
                e.preventDefault();
                EXT_BTNS.clicked("anneid");
            });


            var btnannkhid = ANNE.A.__getEl('btn_header_annkh');
            btnannkhid.addEventListener("click", (e)=>{
                e.stopPropagation();
                e.preventDefault();
                ANNE.MEM_UDF.hide3js();
                EXT_BTNS.clicked("annkh")
            });
            var btnannkhid_footer = ANNE.A.__getEl('btn_footer_annkh');
            btnannkhid_footer.addEventListener("click", (e)=>{
                e.stopPropagation();
                e.preventDefault();
                EXT_BTNS.clicked("annkh");
            });


            var btnlukat = ANNE.A.__getEl('btn_header_lukat');
            btnlukat.addEventListener("click", (e)=>{
                e.stopPropagation();
                e.preventDefault();
                ANNE.MEM_UDF.hide3js();
                EXT_BTNS.clicked("lukat_search")
            });
            var btnlukat_footer = ANNE.A.__getEl('btn_footer_lukat');
            btnlukat_footer.addEventListener("click", (e)=>{
                e.stopPropagation();
                e.preventDefault();
                EXT_BTNS.clicked("lukat_search");
            });


            var btnandy = ANNE.A.__getEl('btn_header_andy');
            btnandy.addEventListener("click", (e)=>{
                e.stopPropagation();
                e.preventDefault();
                EXT_BTNS.clicked("andy")
            });
            var btnandy_footer = ANNE.A.__getEl('btn_footer_andy');
            btnandy_footer.addEventListener("click", (e)=>{
                e.stopPropagation();
                e.preventDefault();
                EXT_BTNS.clicked("andy");
            });


            /**
             * show the annet 3js hypergraph
             */
            var btnannet = ANNE.A.__getEl('btn_header_annet');
            btnannet.addEventListener("click", (e)=>{
                e.stopPropagation();
                e.preventDefault();
                ANNE.MEM_UDF.show3js();
            });

                
            //annode btns (header, left overlay footer)
            var btnannode = ANNE.A.__getEl('btn_header_annode');
            btnannode.addEventListener("click", (e)=>{
                e.stopPropagation();
                e.preventDefault();
                EXT_BTNS.clicked("annode");
            });
            var btnannode_footer = ANNE.A.__getEl('btn_footer_annode');
            btnannode_footer.addEventListener("click",  (e)=>{
                e.stopPropagation();
                e.preventDefault();
                EXT_BTNS.clicked("annode");
            });
            
            //lnot btns (header, left overlay footer)
            var btnlnot = ANNE.A.__getEl('btn_header_lnots');
            btnlnot.addEventListener("click", (e)=>{
                e.stopPropagation();
                e.preventDefault();
                EXT_BTNS.clicked("lnots");
            });
            var btnlnot_footer = ANNE.A.__getEl('btn_footer_lnots');
            btnlnot_footer.addEventListener("click", (e)=>{
                e.stopPropagation();
                e.preventDefault();
                EXT_BTNS.clicked("lnots");
            });

            //lock btns (header, left overlay footer)
            var btn_lock = ANNE.A.__getEl('btn_lock');
            btn_lock.addEventListener("click", (e)=>{
                e.stopPropagation();
                e.preventDefault();
                USER.lockExt(e);
            });

            var btn_lock_footer = ANNE.A.__getEl('btn_footer_lock');
            btn_lock_footer.addEventListener("click", (e)=>{
                e.stopPropagation();
                e.preventDefault();
                USER.lockExt(e);
            });        

            //ghost btns (header, left overlay footer)
            var btnGhosts = ANNE.A.__getEl('btn_header_ghosts');
            btnGhosts.addEventListener("click", (e)=>{
                e.stopPropagation();
                e.preventDefault();
                EXT_BTNS.clicked("ghosts");
            });
            var btnGhosts_footer = ANNE.A.__getEl('btn_footer_ghosts');
            btnGhosts_footer.addEventListener("click", (e)=>{
                e.stopPropagation();
                e.preventDefault();
                EXT_BTNS.clicked("ghosts");
            });
        }

        const customEvent = new CustomEvent(ANNE.EVENT_LOAD_HYPERGRAPH, { bubbles: true, cancelable: true, });
        el_loader.dispatchEvent(customEvent);
    }
    

    doTheRestPre = async($e)=>{
        await ANNE.UTILS.sleep(1000);
        ANNE.MEM_UDF.doTheRest($e);    //don't await it.. we want our css animation to go
    }

    doTheRest = async($e)=> {
        var el_loader = document.getElementById("login_loader");
        el_loader.removeEventListener(ANNE.EVENT_LOAD_HYPERGRAPH, ANNE.MEM_UDF.doTheRestPre);
        
        //we can proces ghosts/TXs tbls per anneid. 
        //rememeber, logged in is not a traditional login.
        //it just means an anneid is in browser mem.
        //to actually use it the user will need the pw that unlocks the seed/key
        await ANNE.GHOSTDATA.processGhosts();
    
        //do we want to do this mumbo jumbo with non-hypergraph txs?  It was an older way.
        //may be depercated -- INVESTIGATE.
        await ANNE.HISTORY_LOCAL.getAndProcessOURTXS_bootup();

        //TODO:  update this to getAllNeuronsForIdentity  ---   (get all payment outs and ins)
        //await ANNE.API.getAllNeuronsFor(USER._userNObj.nid, true);

        if (ANNE.isANNE(USER._userNObj.nid)==false) {
            if (USER._userLS.activated==false || USER._userNObj.ghost==true) {
                console.log("not activated, no query yet.");
            }
        }


        if (ANNE.USER._userNObj!=null) {
            var el_welcome = ANNE.A.setHTML("ro_welcome", ANNE.USER._userNObj.bestNameHTML());
            var pObj = new page_params(null, null, null, ANNE.USER._userNObj.nid);
            el_welcome.addEventListener("click", e=>ANNE.PAGES.openPage("dossier", pObj));

            var el_logout = document.getElementById("logout_all");
            if (el_logout!=null) {
                el_logout.addEventListener("click", e=>{
                    ANNE.USER.logoutThisAccount();
                })
            }
        }
        
        //load all non-awareness user outs data into RAM
        //await ANNE.IDB.loadIDBNidOuts_single(USER._userNObj.nid, TYPE_AWARENESS);

        //*** ANDY PULL ***/
        //WE're GOING TO HAVE TO STORE CLASSES WHEN POSSIBLE.
        //This list will otherwise get insane!
        var andy_nids = USER._userNObj.andy_nids;   //nids this user has ANDY rights to.
        if (andy_nids==undefined || andy_nids==null || andy_nids.length==0) {
            andy_nids = USER._userNObj.resetAndyNids();   //set them just in case.
        }

        //some cheat relon pts we stored that must be deleted in RAM.
        //they will be reset for the potential new user.
        for (var r=0; r < MEM._aRelons.length; r++) {
            delete MEM._aRelons[r].hanai_rights;
            delete MEM._aRelons[r].andy_rel;
            delete MEM._aRelons[r].isAndyRel;
        }
        for (var n=0; n < MEM._aNeurons.length; n++) {
            MEM._aNeurons[n]._govenor = null;
        }

        //repull collabs from db local mem, process.
        //will only pull our stuff.
        await this.preloadCollabs();

        if (ANNE.SETTINGS!=null && this.aSettingsToProcess!=null) {
            await ANNE.SETTINGS.setSettings(this.aSettingsToProcess);
        }


        ANNE.A.hide("anne_site");
        ANNE.A.show("threejs_search_bar");

        //we init this and use the gui menu.
        await ANNE.ACE.initMANUAL();

        await this.launchUDF_firstPage();
    }


    /**
     * 
     */
    launchUDF_firstPage = async()=> {
        if (ANNE.SETTINGS.debug()==true) console.log("%c launchUDF_firstPage ", TALK.bgyellow);

        ANNE.API.initMSGS();
        ANNE.A.show("centerFooterDiv");
        ANNE.A.hide("login_loader");
        ANNE.A.show("main_logo");
        
        ANNE.MSGS.initMsgs();
        if (ANNE._bFIRST_TIME_LOGIN==false) {
            ANNE.MSGS.addWelcomeBackMsg();
        }
        else if (ANNE._bFIRST_TIME_LOGIN==true) {
            console.log("new user? "+USER._bNEWUSER);
            ANNE.MSGS.addWelcomeMsg();
        }

        //what is this?  a play sound in the loadvars?
        if (ANNE.loadVars.play!=undefined) {
            console.log("new user? "+USER._bNEWUSER);
            ANNE.A.play(ANNE.loadVars.play);
        }

        if (ANNE.loadVars!=null) {
            var page = ANNE.loadVars.page;
            if (ANNE.loadVars.page=="annet") {
                if (ANNE.loadVars.nid!=undefined) {
                    var nOBJ = MEMG.getOrMakeIfNew(ANNE.loadVars.nid);
                    ANNE.MEM_UDF.show3js(nOBJ);
                }
                else  ANNE.MEM_UDF.show3js();
            }
            else if(page=="dossier") {
                if (ANNE.loadVars.nid!=undefined) {
                    var nOBJ = MEMG.getOrMakeIfNew(ANNE.loadVars.nid);
                    await ANNE.PAGES.launchNeuron(nOBJ.nid);
                }
                else ANNE.EXT_BTNS.clicked("anneid");
            }
                
            else if (ANNE.loadVars.page=="andy") {
                ANNE.EXT_BTNS.clicked("andy");
            }
                
            else if (ANNE.loadVars.page=="lnot") {
                ANNE.EXT_BTNS.clicked("lnot");
            }
                
            else {
                if(ANNE.loadVars.txid!=undefined) {
                    ANNE.PAGES.launchRelon(ANNE.loadVars.txid, false)
                }
                else ANNE.EXT_BTNS.clicked("anneid");
            }
        }
        
        else {
            var homepage = ANNE.SETTINGS.getPref("homepage");
            if (homepage!=null) {
                if (homepage.misc_value=="annet") {
                    ANNE.MEM_UDF.show3js();
                }
                else if (homepage.misc_value=="anneid")  ANNE.EXT_BTNS.clicked("anneid");
                else if (homepage.misc_value=="andy")  ANNE.EXT_BTNS.clicked("andy");
                else if (homepage.misc_value=="lnot")  ANNE.EXT_BTNS.clicked("lnot");
            }
            else ANNE.EXT_BTNS.clicked("anneid");    //let's default on the ANNEID page.
        }

        ANNE.loadVars = null;
    }

    
    /**
     * 
     */
    clean3dAfterGhostDelete = function() {

    }


    /**
     * Show the 3js
     */
    show3js = function($nObj=null, $textPreload="") {
        ANNE.ACE.showLeft3jsOverlay();
        ANNE.A.show("csscontainer");
        ANNE.A.show("container");
        if (ANNE.ANX!=null && ANNE.ANX.CIRCUIT._aCircuits.length==0) {
            var tjssb = ANNE.A.show("threejs_search_bar");
            tjssb.style.pointerEvents="all";
        }
        ANNE.A.hide("anne_site");
        ANNE.A.show("instruct_overlay");

        var cssel = ANNE.A.show("csscontainer");
        cssel.style.pointerEvents="none";
        var tjsel = ANNE.A.show("container");
        tjsel.style.pointerEvents="none";

        ANNE.ACE.refocus($nObj, $textPreload);
    }


    /**
     * hide the 3js
     */
    hide3js = function() {
        ANNE.MEM_UDF._bSHOWINGTHREEJS = false;
        ANNE.A.gs_alphaOut(ANNE.ACE.el_3js_left_overlay);
        ANNE.A.hide("csscontainer");
        ANNE.A.hide("container");
        var tjssb = ANNE.A.hide("threejs_search_bar");
        tjssb.style.pointerEvents="none";
        ANNE.A.hide("instruct_overlay");
        var cssel = ANNE.A.show("csscontainer");
        cssel.style.pointerEvents="none";
        var tjsel = ANNE.A.show("container");
        tjsel.style.pointerEvents="none";
        ANNE.ACE.disable();
    }



    /**
     * 
     * @author {Mr.Scatman}
     */
    __setTheme = async ($theme=0, $bINIT=false)=> {
        console.log("do nothing.")
    }

    
    /**
     * 
     */
    __login_check = function() {
        USER.__loginCheck(this.hasANNEID, this.noANNEID);
    }
   
    /**
     * 
     * @author {Mr.Scatman}
     */
    backup_bg = function ($bg_url="", $bg_color) {
        if ($bg_url=="") {
            var bg_url = document.body.style['background-color'];
            if (bg_url.indexOf('url')!==-1) bg_url = bg_url.replace('url(','').replace(')','').replace(/\"/gi, "");
            MEM_UDF._backup_bg_url = bg_url;
        }
        else MEM_UDF._backup_bg_url = $bg_url;

        if ($bg_color=="") {
            var bg_color = document.body.style['background-color'];
            MEM_UDF._backup_bg_color = bg_color;
        }
        else MEM_UDF._backup_bg_color = $bg_color;
    }


    /**
     * 
     * @author {Mr.Scatman}
     */
    revert_bg = function () {
        if (MEM_UDF._backup_bg_url!="") {
            ANNE.A.setBodyBackground(MEM_UDF._backup_bg_url);
        }
        if (MEM_UDF._backup_bg_color!="") {
        document.body['background-image']= MEM_UDF._backup_bg_color;
        }
    }


    /** 
     * @author {Mr.Scatman}
     */
    _udf_recourse_needsFundsIdentity=function($totalBalance, $annies_needed, $aid_that_needs_funds, $code="SENDER") {
        var funds= [];
        alert("UPDATE FOR EXT, no inline JS");
        if ($annies_needed < 5000) $annies_needed = 5000;
        funds.push("<BR>Please deposit more annies to your ANNE identity.");
        //funds.push('<a href="javascript:ANNE.___launchQRFund(\''+$annies_needed+'\', \''+$aid_that_needs_funds+'\', \''+$code+'\')">fund '+$annies_needed+'</a>');
        //funds.push('<a href="javascript:ANNE.___launchQRFund(\''+(5*$annies_needed)+'\', \''+$aid_that_needs_funds+'\', \''+$code+'\')">fund 5x '+($annies_needed*5)+'</a>');
        //funds.push('<a href="javascript:ANNE.___launchQRFund(\''+(10*$annies_needed)+'\', \''+$aid_that_needs_funds+'\', \''+$code+'\')">fund 10x '+($annies_needed*10)+'</a>');
        var rLink = "TODO";//GUI_STUFF.html_spans_from_array(funds);
        return rLink;
    }


    //UNTESTED IN NEW ANNE OSM.
    _udf_recourse_needsFundsChild=function($totalBalance, $annies_needed, $aid_that_needs_funds, $code="SENDER") {
        if ($annies_needed < 5000) $annies_needed = 5000;
        var funds = [];
        alert("UPDATE FOR EXT, no inline JS");
        funds.push("<BR>Please deposit more annies to this ANNE neuron you own so it can make the requested ANNE Transactions.");
        //funds.push('<a href="javascript:ANNE.___launchInlineFund(\''+$annies_needed+'\', \''+$aid_that_needs_funds+'\', \''+$code+'\')">fund '+$annies_needed+'</a>');
        //funds.push('<a href="javascript:ANNE.___launchInlineFund(\''+(5*$annies_needed)+'\', \''+$aid_that_needs_funds+'\', \''+$code+'\')">fund 5x '+($annies_needed*5)+'</a>');
        //funds.push('<a href="javascript:ANNE.___launchInlineFund(\''+(10*$annies_needed)+'\', \''+$aid_that_needs_funds+'\', \''+$code+'\')">fund 10x '+($annies_needed*10)+'</a>');
        var rLink =  "TODO";//GUI_STUFF.html_spans_from_array(funds);
        return rLink;
    }
}

var MEM_UDF = new mem_udf();