class ANNE_NET {
    constructor($base_uri="http://localhost:9116") {
        this.base_uri = $base_uri;
		this._nodeConnectedCallback = null;
		this._consecutive_network_errors = 0;
		this.bSSEINIT=false;
		this.last_msg_id = 0;

        if (this.base_uri[this.base_uri.length-1]=="/") {
			this.base_uri = this.base_uri.substring(0, this.base_uri.length-1);
		}
		this.TODOS = [];
		this.active_coms = false;
    }


	/**
	 * 
	 * @param {*} $funcname 
	 * @param {*} $params 
	 * @returns 
	 */
	setTODO=async($funcname, $params)=> {
		if (this.active_coms==false) {
			return await this.processToDo({functname:$funcname, params:$params});
		}
		this.TODOS.push({functname:$funcname, params:$params});
		return;
	}


	/**
	 * 
	 * @returns 
	 */
	doTODO=async()=> {
		if (this.TODOS.length==0) return;
		var todo = this.TODOS.unshift();
		await this.processToDo(todo);
	}

	processToDo=async($todo)=>{
		if ($todo.functname=="lookupNids") {
			return await ANNE.API.lookupNids($todo.nids, $odo.types);
		}
	}


	/**
	 * This is a RAM-only relon that will only be in RAM if the neuron is in RAM
	 * and was pending a change in the mempool.
	 * 
	 * Since this app is "open" it will receive said pending change and thus we 
	 * can count on being able to remove the pending RAM-only relon here.
	 * @param {*} $nid 
	 */
	cleanupIntraPending = function($nid) {
		var nOBJ = MEMG.getOrMakeIfNew($nid);
		if (nOBJ._aplex!=null) nOBJ._aplex.has_pending=false;
		
		var intrablockpending_relon = MEMG.getRelonFromWith(nOBJ.outs, TYPE_SYS_MSG, ANNE.MEM.INTRA_BLOCK_PENDING.nid);
		if (intrablockpending_relon!=null) MEMP.localOnlyArborPrune_RAM(intrablockpending_relon);
		return nOBJ;
	}

	
	/**
	 * Fix any on-screen gui data we can from our data stream.
	 * 
	 * Pass the msgObj to the page.
	 * @param {*} $nid 
	 */
	liveUpdate = async($nObj, $msgObj)=> {
		//console.log("%c NET.liveUpdate %c"+$nObj.bestNameHTML(), ANNE.TALK.funct, ANNE.TALK.bggreen);
		
		ANNE.MEM_UDF.fireNeuron($nObj, .5);
		
		if ($msgObj.eventType=="newblock") {
			
			if ($msgObj.payments!=undefined) {
				var payment_relons = [];
				for (var p=0; p < $msgObj.payments.length; p++) {
					
					//keep around payments from or to this user.
					if ($msgObj.payments[p].from_nid!=ANNE.USER._userNObj.nid
					   && $msgObj.payments[p].to_nid!=ANNE.USER._userNObj.nid) continue;
					
					payment_relons.push($msgObj.payments[p]);
				}

				//we can add the payment(s) into the table now... if the page is open.
				//if it isn't open, then we will get them when we go to this page again anyhow.
				if (PAGES._lastPgui.spacevar=="anneid") {
					await MEMP.processNeuronsFromRelons(payment_relons, true, true, "relons", "NET.liveUpdate");
					await ANNE.ANNEID.liveUpdate(payment_relons);
				}
			}
		}

		else if (PAGES._lastPgui!=null) {
			if (PAGES._lastPgui.spacevar=="dossier") {
				if ($msgObj.eventType=="annex" || $msgObj.eventType=="boosty" || $msgObj.eventType=="hanai") {
					if (PAGES._lastPgui.INST==undefined || PAGES._lastPgui.INST==null) {
						console.log("must have a pgui INST here.");
						return;
					}

					//................
					if ($msgObj.nid!=undefined && $msgObj.nid==PAGES._lastPgui.INST.pgui.pObj.p1) {
						console.log("%c dossier page open!", ANNE.TALK.bggreen);
						PAGES._lastPgui.INST.liveUpdate($msgObj.eventType, $msgObj);
					}
				} 
			}
		}
	}


	/**
	 * Here we get a stream of new data from the annode --
	 * a new ANNE block was mined.
	 * 
	 * Our block cycle IS our brain cycle ----
	 * Each block we will degrade the fire value to demphasize
	 * various neurons based on time they have last been fired.
	 * This is a cheap way to emulate neuron firing and cooling off 
	 * and it maximizes our resources.
	 * 
	 * We do not need intra-second or even second-based firing cool-offs.
	 * 
	 * @param {*} $resp 
	 * @returns 
	 */
	processBlockCycleMsgs = async($resp)=> {
		if ($resp.status!="good") return;
		if ($resp.results==0) {
			return;
		}

		ANNE.MEM_UDF.coolFires();

		var lookup_nids = [];
		var consec = 0;
		
		if ($resp.results.length>1) {
			$resp.results.sort(function($r1, $r2) {
				if ($r1.height>$r2.height) return -1;
				else if ($r1.height<$r2.height) return 1;
				return 0;
			})
		}

		var _H = ANNE.HISTORY._GREATEST_LAST_KNOWN_BLOCK;

		for (var m = $resp.result.length-1; m>=0; m--) {
			var msgObj = $resp.result[m];
			msgObj.skip = false;

			if (ANNE.SETTINGS && ANNE.SETTINGS._maxMsgBlockGap) {
				if ((_H-msgObj.height) > ANNE.SETTINGS._maxMsgBlockGap) {
					msgObj.skip = true;
				}
			}

			if (msgObj.eventType=="newblock") {
				consec+=1;
				if(consec>3) msgObj.skip=true;
				for (var m2=1; m2<=7; m2++) {
					var sm = msgObj['sm'+m2+"nid"];
					if (sm==undefined) {
						msgObj.sm=false;
						break;
					}
					if (sm==USER._userNObj.nid) {
						msgObj.sm=true;
						break;
					}
				}
			}

			if (msgObj.generatorId!=undefined) {
				if (lookup_nids.indexOf(msgObj.generatorId)==-1) {
					var NOBJ = MEMG.getOrMakeIfNew(msgObj.generatorId);
					if (NOBJ.text=="" && NOBJ.bNONAME==undefined) lookup_nids.push(msgObj.generatorId);
				}
			}
			if (msgObj.nid!=undefined) {
				if (lookup_nids.indexOf(msgObj.nid)==-1) {
					var NOBJ = MEMG.getOrMakeIfNew(msgObj.nid);
					if (NOBJ.text=="" && NOBJ.bNONAME==undefined) lookup_nids.push(msgObj.nid);
				}
			}
			if (msgObj.sponsor_nid!=undefined) {
				if (lookup_nids.indexOf(msgObj.sponsor_nid)==-1) {
					var NOBJ = MEMG.getOrMakeIfNew(msgObj.sponsor_nid);
					if (NOBJ.text=="" && NOBJ.bNONAME==undefined) lookup_nids.push(msgObj.sponsor_nid);
				}
			}
			if (msgObj.annexedfromnid!=undefined) {
				if (lookup_nids.indexOf(msgObj.annexedfromnid)==-1) {
					var NOBJ = MEMG.getOrMakeIfNew(msgObj.annexedfromnid);
					if (NOBJ.text=="" && NOBJ.bNONAME==undefined) lookup_nids.push(msgObj.annexedfromnid);
				}
			}
		}

		
		//do it now, if we can, otherwise, try later.
		if (lookup_nids.length>0) {
			var bDOQUERY=true;
			var bDOLOADER=false;
			await ANNE.API.lookupNids(lookup_nids, [0,1], bDOQUERY, bDOLOADER, "NET.processBlockCycleMsgs");
			for (var x=0; x < lookup_nids.length; x++) {
				if (lookup_nids[x]==undefined || lookup_nids[x]==null) {
					continue;
				}
				
				if (lookup_nids[x][0]=="z") continue;
				
				var nObj = MEMG.getNeuronByNid(lookup_nids[x]);
				if (nObj==null) {
					//uhhhhhhhhhhhh
					console.warn("WHAT IS "+lookup_nids[x]);
					continue;
				}
				
				if (nObj.text=="") {
					nObj.bNONAME=true; //prevent future hits to api if it is not named.
				}
			}
		}
		

		for (var m = 0; m < $resp.result.length; m++) {
			var msgObj = $resp.result[m];
			if (msgObj.skip==true) continue;
			
			if (msgObj.id>ANNE.API.last_msg_id) {
				ANNE.API.last_msg_id = msgObj.id;
			}
			
			if (msgObj.eventType=="hanai") {
				//console.log("hanai msg", msgObj);
				var innerNOBJ = ANNE.API.cleanupIntraPending(msgObj.nid);
				await ANNE.API.liveUpdate(innerNOBJ, msgObj);
				ANNE.MSGS.doHanai(msgObj, 6000);
				ANNE.MEM_UDF.fireNid(msgObj.sponsornid, .3);
			}

			else if (msgObj.eventType=="annex") {
				//console.log("annex msg", msgObj);
				var innerNOBJ = ANNE.API.cleanupIntraPending(msgObj.nid);
				await ANNE.API.liveUpdate(innerNOBJ, msgObj);
				ANNE.MSGS.doAnnex(msgObj, 6000);
				ANNE.MEM_UDF.fireNid(msgObj.sponsornid, .3);
				ANNE.MEM_UDF.fireNid(msgObj.annexedfromnid, .3);
			}

			else if (msgObj.eventType=="boosty") {
				//console.log("boosty msg", msgObj);
				var innerNOBJ = ANNE.API.cleanupIntraPending(msgObj.nid);
				await ANNE.API.liveUpdate(innerNOBJ, msgObj);
				ANNE.MSGS.doBoosty(msgObj, 5000);
				ANNE.MEM_UDF.fireNid(msgObj.sponsornid, .3);
			}

			else if (msgObj.eventType=="newblock") {
				//console.log("newblock msg!", msgObj);
				if (ANNE.HISTORY._GREATEST_LAST_KNOWN_BLOCK<msgObj.height) {
					ANNE.HISTORY.updateGLKB_SOLID(msgObj.height); 
				}
				ANNE.MSGS.doNewBlock(msgObj, 7000);
				ANNE.MEM_UDF.fireNid(msgObj.generatorId, .3);
			}
			else {
				//console.log("unhandled ", msgObj);
			}
		}


		return;
	}
	
	
	/**
	 * 20 seconds is fast enough
	 */
	initMSGS=function() {	
		if (this.bSSEINIT==true) return;
		this.bSSEINIT=true;

		//start after 10 seconds.
		//loop every 90 seconds. (this is data coming in at mining)
		this.timerID = setTimeout(() => {
			setInterval(this.getMsgs, 90000);
		}, 5000);
	}

	stopInterval=function() {
		clearInterval(this.timerID);
	}

	restartInterval=function(){
		this.timerID = setTimeout(() => {
			setInterval(this.getMsgs, 90000);
		}, 5000);
	}

	//don't await it.
	getMsgs=async()=> {
		if (ANNE.__DO_NOT_INTERRUPT==true) {
			//don't want to interrupt anything when this flag is on.
			//we wrap processes with it to protect them from
			//interruptions.
			return;
		}

		if (ANNE.ACE!=null && ANNE.SETTINGS.bDATAMESSAGES==false) {
			console.log("messages turned off. -- not gettingLatestMsgs.. (time should be nix'd)");
			return;
		}
		ANNE.API.getLatestMsgs(ANNE.API.last_msg_id); //don't await it.
	}


	/**
	 * 
	 * @param {*} $f 
	 */
	setNodeConnectedCallback=function($f) { 
		this._nodeConnectedCallback = $f; 
	}

	/** 
	 * 
	*/
	setNodeConnected=function($bNODECONNECTED=true) {
		this._bNONODE = $bNODECONNECTED;
		if (this._nodeConnectedCallback!=null) {
			this._nodeConnectedCallback(this._bNONODE);
		}
	}



	//***************************************************** */
	//////  TRANACTION CREATION / SIGNING / BROADCAST ////////
	//***************************************************** */

	/**
	 * @param {String} $functName 
	 * @param {Object} $params 
	 * @param {String} $uriOverride - the node uri to use, if not the current node ur connected to. 
	 * @returns 
	 */
	makeOnNodeSignOnJSThenBroadcast=async($functName, $params=null, $uriOverride="")=> {
		var bCHASE_FROM=false;
		var bDO_EXTRAS=false;

		if ($params!=null) {
			if ($params.secretPhrase!=undefined) delete $params.secretPhrase;
			
			//chase the froms in the broadcastTransaction so we can get
			//crucial return data that we need in the processing of the next TXs in 
			//the queue.
			if ($params.chase_from!=undefined) {
				bCHASE_FROM = $params.chase_from;
				delete $params.chase_from;
			}
			if ($params.do_extras!=undefined) {
				bDO_EXTRAS = $params.do_extras;
				delete $params.do_extras;
			}

			//we need to make the TX first (using the annode)
			//then sign it here in the EXT,
			//then broadcast it.
			if ($params.makeSignSend!=undefined && $params.makeSignSend==true) {
				$params.publicKey = USER.pubKey;
				delete $params.makeSignSend;
			}

			//this wasn't set up right and it is just a normal query/send.
			else {
				return await this.query($functName, $params);
			}

			try {
			
				var response = await this.post($functName, $params, $uriOverride);
				console.log('%capi call for '+$functName, TALK.bggreenbig);
				
				this._consecutive_network_errors = 0;

				if (response && response.height!=undefined) {
					HISTORY._GREATEST_LAST_KNOWN_BLOCK = response.height;
				}

				if (response.status!=undefined && response.status!="good") {
					console.log(response);
					var retObj =  {status:"fail", reason:response.status};
					if (response.recourse!=undefined) retObj.recourse = response.recourse;
					if (response.result!=undefined) retObj.result = response.result;
					if (response.results!=undefined) retObj.results = response.results;					
					return retObj;
				}

				if (response.unsignedTransactionBytes==undefined || response.unsignedTransactionBytes==null) {
					return {status:"fail", reason:"problem with unsignedTransactionBytes"};
				}
				var finalResp = await this.signAndBroadcastTransaction(response, bCHASE_FROM, bDO_EXTRAS, $uriOverride);
				
				//don't do this from RELONS!!!
				//we want to remove relons that are beyond this point..
				//for rollback situations.
				if (finalResp.block_height!=undefined) {
					ANNE.HISTORY.updateGLKB_SOLID(finalResp.block_height);
				}
				ANNE.API.setNodeConnected(true);

 				return finalResp;
			}

			catch(e) {
				console.warn(e);
				if (e.data=="Network Error") ANNE.API.setNodeConnected(false);
				var reason = "problem with transaction";
				if (e.data!=undefined && e.data.errorCode!=undefined) {
					if(e.reason!=undefined && e.message==undefined) e.message = e.reason;
					e.data.status="bad";
					return e.data;
				}
				if (e.message!=undefined) reason = e.message;
				var retObj ={status:"bad", reason:reason};
				if (e.recourse!=undefined) retObj.recourse = e.recourse;
				return retObj;
			}
		}

		//this was just a simple query (without params)
		return await this.query($functName);
	}



	/**
	 * 
	 * @param {*} $resp 
	 * @param {*} $bCHASEFROM 
	 * @param {*} $bDOEXTRAS 
	 * @param {*} $uriOverride 
	 * @returns 
	 */
	signAndBroadcastTransaction=async($resp, $bCHASEFROM=false, $bDOEXTRAS=false, $uriOverride="")=> {
		this.active_coms = true;
		if (USER.privKey=="" || USER.privKey==undefined) {
			var aO = USER.prepANNEAuthObjFromSeed(USER.seedObj.seed);
			USER.privKey = aO.sign_key;
		}

		var sig=null;
		
		//gen the signature
		try{
			sig = ANNE.SIGNA.crypto.generateSignature($resp.unsignedTransactionBytes, USER.privKey);
		}
		catch(e) {
			this.active_coms = false;
			return {status:"fail", reason:"(generateSignature) problem"};
		}

		//verify the signature
		try{
			var rez = ANNE.SIGNA.crypto.verifySignature(sig, $resp.unsignedTransactionBytes, USER.pubKey);
			if (rez==false) {
				this.active_coms = false;
				return {status:"fail", reason:"(verifySignature) did not pass"};
			}
		}
		catch(e){
			console.log(e);
			this.active_coms = false;
			return {status:"fail", reason:"(verifySignature) problem"};
		}

		//generate signed bytes
		var signedTransactionBytes=null;
		try {
			signedTransactionBytes = ANNE.SIGNA.crypto.generateSignedTransactionBytes($resp.unsignedTransactionBytes, sig);
		}
		catch(e){
			console.log(e);
			this.active_coms = false;
			return {status:"fail", reason:"(generateSignedTransactionBytes) problem"};
		}
		
		//broadcast the transcation.
		var broadcastParams = {
			transactionBytes:signedTransactionBytes,
			transactionJSON:$resp.transactionJSON
		}
		if ($bCHASEFROM==true) {
			broadcastParams.chase_from = true;
		}
		else if ($bDOEXTRAS==true) {
			broadcastParams.do_extras = true;
		}

		try {
			var response2 = await this.broadcastSigned(broadcastParams, $uriOverride);
			if (response2.status!=undefined && response.status!="good") {
				console.log(response2);
			}
			this.active_coms = false;
			return response2;
		}	

		catch(e) {
			console.log('ERROR: api call for broadcastTransaction', e);
			if (e.data=="Network Error") {
				this._consecutive_network_errors+=1;
			}

			var rez = {status:"fail", msg:""};
			
			if (e.data!=undefined && e.data.message!=undefined) {
				rez.msg = e.data.message;
			}
			else {
				rez.msg = e.message;
			}
			if (e.data!=undefined) {
				if (e.data.error!=undefined) rez.error = e.data.error;
				if (e.data.errorCode!=undefined) rez.errorCode = e.data.errorCode;
				if (e.data.errorDescription!=undefined) rez.errorDescription = e.data.errorDescription;
				if (e.data.errorRecourse!=undefined) rez.errorRecourse = e.data.errorRecourse;
				if (e.data.result) rez.result=e.data.result;
			}
			
			this.active_coms = false;
			return rez;
		}
	}



	/**
	 * Last step of process to gen a TX on the annode, sign it on the edge device
	 * and broadcast transaction bytes back to the annode
	 * @param {Object} $params
	 * @param {String} $uriOverride  
	 * @returns 
	 */
	broadcastSigned = async($params, $uriOverride="")=>{
		ANNE.playSound("broadcast");
		return await this.post("broadcastTransaction", $params, $uriOverride);
	}



	/**
	 * Post data (used for TX broadcasts)
	 * @param {*} $functName 
	 * @param {*} $params
	 * @param {String} $uriOverride  
	 * @returns 
	 */
	post = async($functName, $params, $uriOverride="")=>{
		this.active_coms = true;
		var uri = this.base_uri;
		if (uri==undefined) {
			uri = ANNE.setAnnodeByURI();
		}
		if ($uriOverride!="") uri = $uriOverride;
		
		var cs = ANNE.SIGNA.sig.ChainService;
		const service = new cs({ nodeHost: uri, apiRootUrl:"/anne"}, "anne");
		var response = null;
		if ($params!=null) response = await service.send($functName, $params);
		else response = await service.send($functName);
		this.active_coms = false;
		return response;
	}


	apiHTTPNOJSON=async($functName, $params)=> {
		this.active_coms = true;
		var cs = ANNE.SIGNA.sig.ChainService;
		var uri = this.base_uri;
		if (uri==undefined) {
			uri = ANNE.setAnnodeByURI();
		}
		const service = new cs({ nodeHost: uri, apiRootUrl:"/anne"}, "anne");
		if ($params!=null) {
			var rez = await service.send($functName, $params);
			this.active_coms = false;
			return rez;
		}
		else {
			var rez = await service.send($functName);
			this.active_coms = false;
			return rez;
		}
	}


    /**
     * Perhaps we lack here too from oversimplification of the signa method...
	 * may have to revert back to using the chainService here too.
	 * 
     * @param {String} $functName 
     * @param {Object} $params 
	 * @param {String} $uriOverride
     * @returns 
     */
    query = async($functName, $params=null, $uriOverride="")=>{
		if ($functName!="getLatestMsgs") {
			console.log("%cAPI.query."+$functName, ANNE.TALK.bgpurple, $params);
		}
		if ($params!=null && $params.height!=undefined) console.log("for height:"+$params.height);
		
		this.active_coms = true;
		var cs = ANNE.SIGNA.sig.ChainService;
		if (cs==undefined) console.throw("Must have ANNE.SIGNA.sig.Chainservice");
		var uri = this.base_uri;
		if (uri==undefined) {
			uri = ANNE.setAnnodeByURI();
		}
		if ($uriOverride!="" && $uriOverride!=null) uri = $uriOverride;
		try {
			const service = new cs({ nodeHost: uri, apiRootUrl:"/anne"}, "anne");
			//console.log('api call for '+$functName);

			if ($params!=null) var response = await service.send($functName, $params);
			else var response = await service.send($functName);

			

			//get height if we passed it.. helpful for other processes.
			if (response!=null && response.block_height!=undefined) {
				ANNE.HISTORY.updateGLKB_SOLID(response.block_height);
			}

			if ($functName=="getNeuronBalance") await this.__balanceHack(response, $params);
			this.active_coms = false;
			return response;
		}
			
		catch(e) {
			this.active_coms = false;
			return this.makeCatchReturn(e);
		}
    }


	/**
	 * 
	 * @param {*} $e 
	 * @returns 
	 */
	makeCatchReturn = function($e) {
		if ($e.data=="Network Error") {
			ANNE.API.setNodeConnected(false);
		}
		else if ($e.data.errorCode=="777.2.0") {
			var retObj = {status:"good-but", message:"taken"};
			if ($e.data.result!=undefined) retObj.result = $e.data.result;
			if ($e.data.results!=undefined) retObj.results = $e.data.results;
			return retObj;
		}
		return {status:"fail", reason:"problem with transaction"};
	}


	/**
	 * TODO: review this.
	 * @param {*} $response 
	 * @param {*} $params 
	 * @returns 
	 */
	__balanceHack=async($response, $params)=> {
		if ($response==undefined) { return;}
		var nObj = MEMG.getNeuronByNid($params.nid);
		
		if ($response.activated=="y") {
			nObj.bACTIVATED = true;
			if (nObj.nid==USER._userNObj.nid && USER._userLS.activated==false) {
				await USER.updateLSActivated(); 
			}
		}
		else if ($response.activated=="n") {
			nObj.bACTIVATED = false;
			if (nObj.nid==USER._userNObj.nid) {
				USER._userLS.activated=false;
				await USER.updateLSActivated(); 
			}
		}
		return $response;
	}
}