import * as THREE from './js/other/three.module.js';
import { GUI } from './js/other/dat.gui.module.js';
import { CSS3DRenderer }  from './js/other/CSS3DRenderer.js';
import {CSS3DObject } from './js/other/CSS3DRenderer.js';
import { Sky } from './js/other/Sky.js';

const audioContext = new (window.AudioContext || window.webkitAudioContext)();

var RM = {}
ANNE.RM = RM;
RM._lbls = [];
RM._rmMeshes = [];

RM._ring_camera_place_index = -1;

RM.white_material = null;
RM.white_material_low_alpha = null;
RM.green_material = null;
RM.yellow_material = null;
RM.yellow_material_low_alpha = null;
RM._lastGreenMesh = null;

RM.bSTEPONETIMEPASS=false;      //when looping, we cross this once in the intro and no to stop.
RM.bINTRODONE=false;            //intro done and mouse events activated.

RM.ff_source = null;
RM.logoheartofannegainnode = null;
RM.rotateBall = false;
RM.tick = 0;

RM._b1s = [];
RM._b2s = [];
RM._m1s = [];
RM._m2s = [];

RM.bHOAPULSE = false;
RM.bFLYTHRU = false;
RM.bCHILLMIX = false;
RM.bANNIE = false;

RM.vol = 32*255/2;

RM.track_index = 0;
RM.tracks = ["mission.mp3", "In_The_Beginning_Annie.mp3", "Sunlounger-SunnyTales(Chill).mp3"];

document.body.onload = function() {
    window.THREE = THREE;
    
    const textureLoader = new THREE.TextureLoader();
    RM.logo_texture = textureLoader.load('./imgs/anne_logo_no_ball.png');
    
    RM.listener = new THREE.AudioListener();
    const main_track_audioLoader = new THREE.AudioLoader();
    RM.maintrack = new THREE.Audio(RM.listener);
    main_track_audioLoader.load('./mp3/mission.mp3', function(buffer) {
        RM.maintrack.setBuffer(buffer);
        RM.maintrack.onEnded = function() {
            RM.loadAndPlayNextTrack()
        };
        var skyboxGeo = new THREE.BoxGeometry(10000, 10000, 10000);
        const ft = new THREE.TextureLoader().load("./imgs/skybox/corona_ft.png");
        const ft_mesh = new THREE.MeshBasicMaterial({ map: ft, side: THREE.DoubleSide });
        const bk = new THREE.TextureLoader().load("./imgs/skybox/corona_bk.png");
        const bk_mesh = new THREE.MeshBasicMaterial({ map: bk, side: THREE.DoubleSide });
        const up = new THREE.TextureLoader().load("./imgs/skybox/corona_up.png");
        const up_mesh = new THREE.MeshBasicMaterial({ map: up, side: THREE.DoubleSide });
        const downz = new THREE.TextureLoader().load("./imgs/skybox/corona_dn.png");
        const dn_mesh = new THREE.MeshBasicMaterial({ map: downz, side: THREE.DoubleSide });
        const rt = new THREE.TextureLoader().load("./imgs/skybox/corona_rt.png");
        const rt_mesh = new THREE.MeshBasicMaterial({ map: rt, side: THREE.DoubleSide });
        const lf = new THREE.TextureLoader().load("./imgs/skybox/corona_lf.png");
        const lf_mesh = new THREE.MeshBasicMaterial({ map: lf, side: THREE.DoubleSide });
        RM.skyboxMesh = new THREE.Mesh(skyboxGeo, [
            ft_mesh, bk_mesh, up_mesh, dn_mesh, rt_mesh, lf_mesh  
        ]);

        const launchButton = document.getElementById('launchButton');
        RM.show(launchButton);

        launchButton.addEventListener('click', () => {
            RM.hide(launchButton);
            const preloader = document.getElementById('preloader');
            preloader.remove();
            RM.init();
            RM.playSound("spaceintro");
            RM.maintrack.play();
            RM.analyser = new THREE.AudioAnalyser(RM.maintrack, 32);
            RM.animate();
            RM.startAnimation();
        });
    });
}


/**
 * 
 */
RM.loadAndPlayNextTrack = function() {
    RM.track_index+=1;
    if(RM.track_index>=RM.tracks.length) RM.track_index=0;
    var track = RM.tracks[RM.track_index];

    RM.maintrack.stop();
    
    const main_track_audioLoader = new THREE.AudioLoader();
    main_track_audioLoader.load("./mp3/"+track, function(buffer) {
        RM.maintrack.setBuffer(buffer);
        RM.maintrack.onEnded = function() {
            RM.loadAndPlayNextTrack()
        };
        RM.maintrack.play();
        if (RM.track_index==1) {
            RM.bANNIE=true;
            RM.swapSky();
        }

        //move sun up so we can see it.
        else if (RM.track_index==2) {
            RM.bCHILLMIX=true;
            gsap.to(RM.effectController, {
                time:2,
                elevation:.2 
            });

            /**
            for (var r=0; r < RM._rmMeshes.length; r++) {
                if(RM._rmMeshes[r].ringCSSObj==null) continue;
                RM.show(RM._rmMeshes[r].ringCSSObj.element);
            }
             */

        }
    });
}


RM.swapSky=function() {
    RM.scene.remove(RM.skyboxMesh);
    RM.renderer.toneMapping = THREE.ACESFilmicToneMapping;
	RM.renderer.toneMappingExposure = 0.8;  //causes darker tone .. darkness..
    RM.sky = new Sky();
    RM.sky.scale.setScalar( 450000 );
    RM.sky.material.uniforms.up.value.set( 0, 0, 1);
    RM.scene.add( RM.sky );
    RM.sun = new THREE.Vector3();
    RM.effectController = {
        turbidity: 4,
        rayleigh: 0.451,
        mieCoefficient: 0.076,
        mieDirectionalG: 0.524,
        elevation: 44.2,
        azimuth: -88.8,
        exposure: 0
    };

    RM.white_material_low_alpha.opacity=.1; 
    RM.yellow_material_low_alpha.opacity=.1; 

    for (var r=0; r < RM._rmMeshes.length; r++) {
        if(RM._rmMeshes[r].ringCSSObj==null) continue;
        RM.hide(RM._rmMeshes[r].ringCSSObj.element);
    }
    
    RM.effectController.exposure=0;
    RM.effectController.azimuth=-180;
    RM.updateSun()

    //tween let their be light.
   gsap.to(RM.effectController, {
        delay:21,
        turbidity:4, 
        rayleigh:.451, 
        mieCoefficient:.076, 
        mieDirectionalG:.362, 
        elevation:0, 
        azimuth:28, 
        ease:"back",
        duration: 1,
        onUpdate: function() {
            RM.bANNIE=true;
            RM.updateSun();
        },
        onComplete:function() {
            gsap.to(RM.effectController, {
                time:2,
                turbidity:0, 
                rayleigh:2.398, 
                mieCoefficient:0.026, 
                mieDirectionalG:0.372, 
                elevation:0, 
                azimuth:90.3, 
                exposure:0.6319,
                onUpdate: function() {
                    RM.updateSun();
                }
            });
        }
    });
}


RM.updateSun=function() {
    //console.log("gui changed ec.exposure:"+RM.effectController.exposure);
    const uniforms = RM.sky.material.uniforms;
    uniforms[ 'turbidity' ].value = RM.effectController.turbidity;
    uniforms[ 'rayleigh' ].value = RM.effectController.rayleigh;
    uniforms[ 'mieCoefficient' ].value = RM.effectController.mieCoefficient;
    uniforms[ 'mieDirectionalG' ].value = RM.effectController.mieDirectionalG;
    const phi = THREE.MathUtils.degToRad( 90 - RM.effectController.elevation );
    const theta = THREE.MathUtils.degToRad( RM.effectController.azimuth );
    RM.sun.setFromSphericalCoords( 1, phi, theta );
    uniforms[ 'sunPosition' ].value.copy( RM.sun );
    RM.renderer.toneMappingExposure = RM.effectController.exposure;
    RM.renderer.render( RM.scene, RM.camera );
}


/**
 * 
 */
RM.init = function() {
    var container = document.getElementById( 'container' );
    RM.renderer = new THREE.WebGLRenderer({antialias: true});
    RM.renderer.setPixelRatio( window.devicePixelRatio );
    RM.renderer.setSize( window.innerWidth, window.innerHeight );
    container.appendChild( RM.renderer.domElement );

    RM.scene = new THREE.Scene();
    RM.scene.add(RM.skyboxMesh);

    var csscontainer = document.getElementById( 'csscontainer' );
    RM.rendererCSS = new CSS3DRenderer();
    RM.renderer.setPixelRatio( window.devicePixelRatio );
    RM.rendererCSS.setSize( window.innerWidth, window.innerHeight );
    RM.rendererCSS.domElement.style.position = 'absolute';
    RM.rendererCSS.domElement.style.top = 0;
    csscontainer.appendChild( RM.rendererCSS.domElement );
    
    RM.roadmap = [];

    RM.roadmap.push({phase:0, cat:"L1", title:"ANNE pre-Blockchain", part:"annepre.html", estimate:"2006-2018", status:"historic"});
    RM.roadmap.push({phase:0, cat:"L1", title:"ANNE-DEV", part:"dev.html", estimate:"q4 2018-2023", status:"ongoing"});
    RM.roadmap.push({phase:0, cat:"L1", title:"Testnet", part:"testnet.html", estimate:"q4 2022+", status:"active"});
    RM.roadmap.push({phase:0, cat:"docs", title:"Whitepaper", part:"whitepaper.html", estimate:"q3 2023", status:"draft"});
    RM.roadmap.push({phase:1, cat:"L1", title:"Mainnet Data-Mode", part:"datamode.html", estimate:"q4 2023", status:"soon"});
    RM.roadmap.push({phase:2, cat:"L1", title:"Mainnet Live-Mode", part:"livemode.html", estimate:"q1 2024", status:"soon"});

    RM.roadmap.push({phase:2, cat:"L1", title:"Bounties", part:"bounties.html", estimate:"q1 2024", status:"waiting activation"});
    RM.roadmap.push({phase:2, cat:"community", title:"ADF Projects", part:"adf-projects.html", estimate:"starting q1 2024", status:"planning"});
    RM.roadmap.push({phase:2, cat:"community", title:"DEX/CEX Listings", part:"exchanges.html",  estimate:"late q1 2024", status:"planning"});
    
    RM.roadmap.push({phase:3, cat:"L1", title:"Private Data Neurons", part:"pdns.html", estimate:"q2 2024", status:"waiting activation"});
    RM.roadmap.push({phase:3, cat:"L1", title:"ANNE-Settled Data Futures", part:"asdfs.html",  estimate:"q2 2024", status:"waiting activation"});
    RM.roadmap.push({phase:3, cat:"L1", title:"ANNDLES", part:"anndles.html", estimate:"q2 2024", status:"waiting activation"});
    RM.roadmap.push({phase:3, cat:"L1", title:"ANNE CDN/P2P Files", part:"p2p_files.html", estimate:"q2 2024", status:"planning"});
    
    RM.roadmap.push({phase:4, cat:"L2 software", title:"AGI/AI/NLP/OSINT Tools", part:"agi_ai_nlp.html",  estimate:"start q3 2024", status:"ongoing"});
    
    RM.roadmap.push({phase:4, cat:"hardware", title:"Custom Hardware", part:"hardware.html",  estimate:"q3 2024", status:"investigating"});
    RM.roadmap.push({phase:4, cat:"L1/L2 software", title:"ANNODE Native Linux OS", part:"linux.html", estimate:"q4 2024", status:"investigating"});
    
    RM.roadmap.push({phase:5, cat:"community", title:"ANNE-U", part:"anneu.html", estimate:"q4 2024 / q1 2025", status:"brainstorming"});    

    RM.white_material = new THREE.MeshBasicMaterial({ color: 0xffffff, wireframe:true });
    RM.white_material_low_alpha = new THREE.MeshBasicMaterial({ color: 0xffffff, wireframe:true, transparent:true, opacity:.2 });

    RM.yellow_material = new THREE.MeshBasicMaterial({ color: 0xfefe17, wireframe:true });
    RM.yellow_material_low_alpha =  new THREE.MeshBasicMaterial({ color: 0xfefe17, wireframe:true, transparent:true, opacity:.2 });

    RM.green_material = new THREE.MeshBasicMaterial({ color: 0x00ff00, wireframe:true });

    RM.outter_r = 300;
    const inner_r = 60;
    const giantRingGeo = new THREE.TorusGeometry(RM.outter_r, inner_r, 64, 128);
    RM.giantRing = new THREE.Mesh(giantRingGeo, RM.white_material);
    RM.scene.add(RM.giantRing);

    // Set up the camera
    RM.camera = new THREE.PerspectiveCamera(55, window.innerWidth / window.innerHeight, 0.1, 30000);
    RM.camera.position.z = 50;
    RM.camera.add( RM.listener );

    
    const planeGeometry = new THREE.PlaneGeometry(761, 549);

    const planeMaterial = new THREE.MeshBasicMaterial({ map: RM.logo_texture, side: THREE.DoubleSide, transparent:true });
    RM.logoWithoutBall = new THREE.Mesh(planeGeometry, planeMaterial);
    RM.scene.add(RM.logoWithoutBall);
    var scale = new THREE.Vector3(0.25, 0.25, 0.25);
    RM.logoWithoutBall.scale.copy(scale);

    const logogeo = new THREE.SphereGeometry(20, 32, 16);
    RM.logoheartofanne = new THREE.Mesh(logogeo, RM.white_material);
    RM.logoheartofanne.position.y=10;
    RM.scene.add(RM.logoheartofanne);

    const audioLoader = new THREE.AudioLoader();
        const soundUrl = './mp3/forceField.mp3';
        audioLoader.load(soundUrl, function(buffer) {

            RM.ff_source = audioContext.createBufferSource();
            RM.ff_source.buffer = buffer;
            RM.ff_source.connect(audioContext.destination);
            
            // Set the volume using the `gain` property of the `AudioNode`
            RM.logoheartofannegainnode = audioContext.createGain();
            RM.logoheartofannegainnode.gain.value = 0.05; 
            RM.ff_source.connect( RM.logoheartofannegainnode);
            RM.logoheartofannegainnode.connect(audioContext.destination);

            RM.ff_source.positionX = RM.logoheartofanne.position.x;
            RM.ff_source.positionY = RM.logoheartofanne.position.y;
            RM.ff_source.positionZ = RM.logoheartofanne.position.z;
            RM.ff_source.loop = false;
        });

    //circle around the torus and place equadistant 
    RM.radStep = (Math.PI * 2) / RM.roadmap.length; 

    const pointLight = new THREE.PointLight(0xffffff, 10, 100);
    pointLight.position.set(0, 1000, 0);
    RM.camera.add(pointLight);

    RM.camera.zoom=.5;
    RM.camera.position.z = 1670;

    window.addEventListener( 'resize', RM.onWindowResize, false );

    const radius = 60;
    const tubeRadius = 12;
    const radialSegments = 16;
    const tubularSegments = 100;
    const geometry = new THREE.TorusGeometry(radius, tubeRadius, radialSegments, tubularSegments);
    
    for (var r=0; r < RM.roadmap.length; r++) {
        var rObj = RM.roadmap[r];
        var ring = new THREE.Mesh(geometry, RM.yellow_material);
        ring.mat_i = 1;
        ring.r_index = r;
        rObj.r_index = r;

        rObj.el_btn = null;
        rObj.el_template = null;

        //we will dupe this to crate a waypoint when animating between rings.
        ring.position.x = Math.cos(RM.radStep*(r)) * RM.outter_r;
        ring.position.y = Math.sin(RM.radStep*(r)) * RM.outter_r;
        ring.rotation.x = Math.PI / 2;
        ring.rotation.y = RM.radStep*(r);

        RM.roadmap[r].ring = ring;
        var ringLbl = document.createElement( 'div' );
        ringLbl.className = "ring-title";
        ringLbl.textContent = rObj.title;
        ringLbl.classList.add("show");
        const ringCSSObj = new CSS3DObject( ringLbl );
        ringCSSObj.position.set( 0, 0, 10 );
        ring.add(ringCSSObj);
        ring.rObj = rObj;
        ring.ringCSSObj = ringCSSObj;
        RM._lbls.push(ringCSSObj);
        RM.scene.add(ring);
        RM._rmMeshes.push(ring);
    }

    RM.savedRotation = new THREE.Vector3();
    RM.mouseLock = false;

    document.addEventListener("mousedown", $e=>{
        if (RM.bINTRODONE==false) return;
        if ($e.target.getAttribute("overlay")!=undefined) {
            return;
        }
        if ($e.target.parentElement!=undefined) {
            if ($e.target.parentElement.getAttribute("overlay")!=undefined) {
                console.log("skipping event from overlay")
                return;
            }
            if ($e.target.parentElement.parentElement!=undefined 
                && $e.target.parentElement.parentElement.getAttribute("overlay")!=undefined) {
                console.log("skipping event from overlay")
                return;
            }
        }
        RM.savedRotation.copy(RM.camera.rotation);
        //console.log(RM.savedRotation);
        RM.mouseLock=true;
    });


    /**
     * Must be in mouselock for this to work.
     * Otherwise it will throw us WAY off!  :)
     */
    document.addEventListener("mouseup", e=>{
        console.log("Mouse up damp back to start");
        if (RM.bINTRODONE==false || RM.mouseLock==false) return;

        RM.mouseLock=false;

        gsap.to(RM.camera.rotation, {
            x:RM.savedRotation.x,
            y:RM.savedRotation.y,
            z:RM.savedRotation.z,
            ease:"power4.inOut",
            duration: .5,
        });
    });


    document.addEventListener("mousemove", e=>{
        if (RM.bINTRODONE==false) return;
        if (RM.mouseLock==true) {
            const movementX = e.movementX || e.mozMovementX || e.webkitMovementX || 0;
            const movementY = e.movementY || e.mozMovementY || e.webkitMovementY || 0;
            RM.camera.rotation.y -= movementX * 0.004;
            RM.camera.rotation.x -= movementY * 0.002;
            //RM.camera.rotation.x = Math.max(-Math.PI / 2, Math.min(Math.PI / 2, RM.camera.rotation.x));
          }
    });

}



/**
 * 
 * @param {*} $mesh 
 * @param {*} $mag 
 */
RM.poorMansLogoShake=function() {
    var x = RM.logoWithoutBall.position.x;
    var y = RM.logoWithoutBall.position.y;
    var z = RM.logoWithoutBall.position.z;
    var ani = gsap.timeline();
    ani.to(RM.logoWithoutBall.position, {
        x:x+20,
        y:y+20,
        z:z+20,
        ease:"rough",
        duration: .05
    })
    .to(RM.logoWithoutBall.position, {
        x:x-20,
        y:y-20,
        z:z-20,
        ease:"rough",
        duration: .05
    })
    .to(RM.logoWithoutBall.position, {
        x:x,
        y:y,
        z:z,
        ease:"rough",
        duration: .05,
        onComplete: async()=> {

            RM.logoheartofanne.material = RM.yellow_material;
            RM.logoheartofanne.material.needsUpdate = true;
            RM.rotateBall=true;
            RM.playSound("boomhit");
        
        }
    });
}


RM.turnMeshWhiteLowAlpha=function($mesh) {
    $mesh.material = RM.white_material_low_alpha;
    $mesh.material.needsUpdate = true;
}
RM.turnMeshYellowLowAlpha=function($mesh) {
    $mesh.material = RM.yellow_material_low_alpha;
    $mesh.material.needsUpdate = true;
    if ($mesh.ringCSSObj!=undefined) $mesh.ringCSSObj.element.classList.remove("active"); 
}
RM.turnMeshGreen=function($mesh) {
    $mesh.material = RM.green_material;
    $mesh.material.needsUpdate = true;
    RM.playSound("sonar");
    RM._lastGreenMesh = $mesh;
    if ($mesh.ringCSSObj!=undefined) $mesh.ringCSSObj.element.classList.add("active"); 
}



/**
 * 
 * @param {*} $index 
 */
RM.highlightNextIndex=function($index) {
    
    //RM.sound.play();
    var last_mesh = RM._rmMeshes[$index];
    RM.turnMeshYellowLowAlpha(last_mesh);

    var next_index = $index+1;
    if (next_index >= RM._rmMeshes.length) next_index = 0;
    var mesh = RM._rmMeshes[next_index];
    RM.turnMeshGreen(mesh);

    RM.makeSideBtn(mesh);
}



/**
 * 
 * @param {*} $mesh 
 */
RM.makeSideBtn = function($mesh) {
    //make the menu button if we haven't yet.
    var rObj = $mesh.rObj;
    if (rObj.el_btn==null) {
        var el_roadmap_btns = document.getElementById("roadmap_btns");
        rObj.el_btn = document.createElement("div");
        rObj.el_btn.classList = "roadmap-btn";
        var span_el = document.createElement("span");
        span_el.innerHTML = "phase "+rObj.phase+" - "+rObj.title;
        rObj.el_btn.appendChild(span_el);
        el_roadmap_btns.append(rObj.el_btn);
        rObj.el_btn.addEventListener("click", e=>RM.sideBtnClick(rObj));
    }
}


/**
 * 
 * @param {*} $rObj 
 */
RM.sideBtnClick = async($rObj)=> {
    if (RM.bINTRODONE==false) {
        //console.log("make a beep indicating nope")
        return;
    }
    await RM.loadDetails($rObj);
    RM.travelToRingObj($rObj);
}


/**
 * 
 */
RM.startAnimation = function() {
    RM._ring_camera_place_index = -1;
    gsap.to(RM.camera.position, {
        z:134,
        ease:"power1.in",
        duration: 4,
        onUpdate: function() {
        },
        onComplete: async()=> {
            RM._ring_camera_place_index = 0;
            RM.poorMansLogoShake()
            RM.enterTheTorusFromLogo();
        }});
}


/**
 * From logo view, back out then enter the torus
 */
RM.enterTheTorusFromLogo = function() {
    RM._ring_camera_place_index = RM._rmMeshes.length-1;
    var ringMesh = RM._rmMeshes[RM._rmMeshes.length-1];
    RM.enterTheTorusFinish(ringMesh);   //start on the last..
    return;
}


/**
 *  We could also do a more beefy mesh once we're in it. (LOD) 
 */
RM.prepForInnerTravel = function() {
    RM.turnMeshWhiteLowAlpha(RM.giantRing);
    for (var r=0; r < RM._rmMeshes.length; r++) {
        RM.turnMeshYellowLowAlpha(RM._rmMeshes[r]);
    }
}

/**
 * From outside view back, enter the torus
 * TODO: add a shake/rumble sound.
 * @param {*} $ringMesh 
 */
RM.enterTheTorusFinish = function($ringMesh, $delay=2, $time=2) {
    console.log("enterTheTorusFinish");
    var ease = "power1.in";
    var QO = RM.__getQO($ringMesh.position);
    var bCOMPLETE1 = false;
    var bCOMPLETE2 = false;
    var bFINALONETIME = false;

    var tween = gsap.to(RM.camera.position, {
        delay:$delay,
        z:$ringMesh.position.z,
        y:$ringMesh.position.y,
        x:$ringMesh.position.x,
        ease:ease,
        duration: $time,
        onUpdate: function() {
            RM.__updateSlerp(QO, tween.time());
        },
        onComplete: async()=> {  
            bCOMPLETE1=true;
            if (bCOMPLETE2==true && bFINALONETIME==false) {
                RM.show("main_logo");
                bFINALONETIME=true;
                RM.scene.remove(RM.logoWithoutBall);
                RM.prepForInnerTravel();
                RM.highlightNextIndex(RM._ring_camera_place_index);
                RM.stepThrough(.65, 1.6);
            } 
        }});

    gsap.to(RM.camera.rotation, {
        delay:$delay,
        x:Math.PI / 2,
        ease:ease,
        duration: $time,
        onUpdate: function() {
            //camera.updateProjectionMatrix(); 
        },
        onComplete: async()=> { 
            bCOMPLETE2=true;
            if (bCOMPLETE1==true && bFINALONETIME==false) {
                RM.show("main_logo");
                bFINALONETIME=true;
                RM.scene.remove(RM.logoWithoutBall);
                RM.prepForInnerTravel();
                RM.highlightNextIndex(RM._ring_camera_place_index);
                RM.stepThrough(.65, 1.6);
                //RM.flyThroughOnce();
            } 
        }});
}


RM.getPrevIndex=function($index) {
    if ($index==0) return RM._rmMeshes.length-1;
    return $index-1;
}
RM.getNextIndex=function($index) {
    if ($index==RM._rmMeshes.length-1) return 0;
    return $index+1;
}

/**
 * 
 * @param {*} $rObj 
 */
RM.travelToRingObj = function($rObj) {
    if (RM.bFLYTHRU==true) return;

    //we need to "go" the prev index of the ring we want to "view"
    var target_r_index = RM.getPrevIndex($rObj.r_index);
    var steps = 0;

    //count our steps fwd.. 
    //(we want to go fwd for the fx of the fast speed)
    var test_target_r_index = RM._ring_camera_place_index;
    while(test_target_r_index!=target_r_index) {
        test_target_r_index+=1;
        steps+=1;
        if (test_target_r_index >= (RM._rmMeshes.length)) {
            test_target_r_index=0;
        }
    }

    var next_index = RM.getNextIndex(RM._ring_camera_place_index);

    RM.flyThroughTo(next_index, 1, target_r_index, steps);
}


/**
 * Fly through the torus
 * @param {*} $next_index - next we are going to 
 * @param {*} $time - total time we want to take on all steps
 * @param {*} $target_r_index - the target index of the cam position we want to end up at
 * @param {*} $steps - the total number of rings we will pass through
 */
RM.flyThroughTo = function($next_index, $time=2, $target_r_index, $steps) {
    var timepart = $time/$steps;
    var meshes = [];
    
    while (meshes.length<$steps) {
        meshes.push(RM._rmMeshes[$next_index]);
        $next_index = RM.getNextIndex($next_index);
    }

    RM.bFLYTHRU=true;

    //loop tween to move cam to each step (no pause)
    //can't loop indexes as they can loop.
    for (var m=0; m < meshes.length; m++) {
        const mesh = meshes[m];
        var delay = timepart*m;
        var ease = "none";
        
        if (m==0) ease = "power1.out";        //first step ease out
        else if (m==meshes.length-1) {
            ease = "power.in"; //last step ease in
            RM._ring_camera_place_index =  $target_r_index; 
        }
        const tween = gsap.to(RM.camera.position, {
            delay:delay,
            x:mesh.position.x,
            y:mesh.position.y,
            z:mesh.position.z,
            ease:ease,
            duration: timepart
        });

        const tween2 = gsap.to(RM.camera.rotation, {
            delay:delay,
            y:mesh.rotation.y+RM.radStep,
            ease:ease,
            duration: timepart,
            onComplete: async()=> { 
                RM.bFLYTHRU=false;
            }
        });
    }

    RM._ring_camera_place_index = $target_r_index;  //safety setting
}


RM.lerpSet=function($start, $end, $time, $slices) {
    const incs = $time / $slices;
    const delta = ($start - $end) / incs;
    let currentValue = $start;
    const vals = [];
    for (let i = 0; i <= incs; i++) {
        vals.push(currentValue);
        currentValue += delta;
    }
    return vals;
}


/**
 * 
 */
RM.stepThrough = function($delay=1, $time=1) {
    //console.log("RM.stepThrough "+RM._ring_camera_place_index);

    if (RM._ring_camera_place_index >= (RM._rmMeshes.length-1)) {
        if (RM.bSTEPONETIMEPASS==true) {
            RM.loadDetails(RM._rmMeshes[0].rObj); //trigger the first item.
            RM.bINTRODONE=true;
            return;
        }
        else {
            RM.bSTEPONETIMEPASS=true;
            RM._ring_camera_place_index = 0;
        }
    }
    else if (RM._ring_camera_place_index<-1) {
        RM._ring_camera_place_index = 0;
    }
    else RM._ring_camera_place_index = RM._ring_camera_place_index+1;
    var ringMesh = RM._rmMeshes[RM._ring_camera_place_index];
    RM.nextRing(ringMesh, $delay, $time);

    //ring.position.x = Math.cos(RM.radStep*(r)) * RM.outter_r;
    //ring.position.y = Math.sin(RM.radStep*(r)) * RM.outter_r;
    //ring.rotation.x = Math.PI / 2;
    //ring.rotation.y = RM.radStep*(r);
}



/**
 * 
 * @param {*} $ringMesh 
 * @param {*} $delay 
 * @param {*} $time 
 */
RM.nextRing = function($ringMesh, $delay=1, $time=2) {
    var ease = "none";
    //var QO = RM.__getQO($ringMesh.position);
    $time = Math.max(.2, $time);

    const tween = gsap.to(RM.camera.position, {
        delay:$delay,
        z:$ringMesh.position.z,
        y:$ringMesh.position.y,
        x:$ringMesh.position.x,
        ease:ease,
        duration: $time,
        onComplete: async()=> {
            RM.highlightNextIndex(RM._ring_camera_place_index);
            RM.stepThrough($delay, $time); //-$time/(2*RM._rmMeshes.length)
        }});

    var finalRad = RM.camera.rotation.y+ANNE.RM.radStep;
   
    
    gsap.to(RM.camera.rotation, {
        delay:$delay,
        y:finalRad,
        ease:ease,
        duration: $time,
    });
}



/**
 * 
 */
RM.doRingsFlyThru = function() {
    var ring_mesh = RM._rmMeshes[RM._ring_camera_place_index];    
    gsap.to(RM.camera.position, {
        z:ring_mesh.position.z,
        y:ring_mesh.position.y,
        x:ring_mesh.position.x,
        ease:"none",
        duration: 1-(RM._ring_camera_place_index*.04),
        onUpdate: function() {
        },
        onComplete: async()=> {
            
            if (RM._ring_camera_place_index<RM._rmMeshes.length-1){
                RM._ring_camera_place_index+=1;
                RM.doRingsFlyThru();
            }
            else {
                RM.arrival();
            }
        }});
}


/**
 * we arrived.
 */
RM.arrival = function() {

}



/**
 * 
 * @param {*} $rObj 
 */
RM.goToRing = async($rObj)=>{

}


/**
 * gets start and end quaternion for use in slerps.
 * @param {*} $position 
 * @returns 
 */
RM.__getQO=function($position) {
    var QO = {sQ:null, eQ:null};
    QO.sQ = RM.camera.quaternion.clone(); 
    RM.camera.lookAt($position);
    QO.eQ = RM.camera.quaternion.clone(); 
    RM.camera.quaternion.copy(QO.sQ);
    return QO;
}
RM.__updateSlerp=function($qObj, $t) {
    var resultQuaternion = new THREE.Quaternion();
    resultQuaternion.slerp($qObj.sQ, $qObj.eQ, $t);
    RM.camera.quaternion.copy(resultQuaternion);
}


/**
 * 
 * @param {*} $rObj 
 * @returns 
 */
RM.loadDetails = async ($rObj) => {
    //console.log("LOAD DETAILS for "+$rObj.title);

    //turn off all buttons.
    var btns = document.querySelectorAll(".roadmap-btn");
    for (var b=0; b < btns.length; b++) {
        btns[b].classList.remove("active");
    }

    //hide any subtitle/textblurbs open
    var conts = document.querySelectorAll(".roadmap-item-container");
    for (var c=0; c < conts.length; c++) {
        RM.hide(conts[c]);
    }

    var rings = document.querySelectorAll(".ring-title");
    for (var r=0; r < rings.length; r++) {
        rings[r].classList.remove("active");
    }

    if (RM._lastGreenMesh!=null) {
        RM.turnMeshYellowLowAlpha(RM._lastGreenMesh);
    }

    //set lbl to active (it will be green)
    $rObj.ring.ringCSSObj.element.classList.add("active"); 
    RM.turnMeshGreen($rObj.ring);

    var btn_el = $rObj.el_btn;
    btn_el.classList.add("active");

    //show the subtitle/textblurb div
    if ($rObj.el_template!=null) {
        RM.show($rObj.el_template);
        return;
    }

    var xhr;
    if (window.XMLHttpRequest) xhr = new XMLHttpRequest();
    else xhr = new ActiveXObject("Microsoft.XMLHTTP");

    return new Promise((resolve, reject) => {
        xhr.onreadystatechange = function () {
            if (this.readyState == 4 && this.status == 200) {
                var resp = this.response;
                $rObj.el_template = document.createElement("div");
                $rObj.el_template.classList = "roadmap-item-container";
                $rObj.el_template.innerHTML = resp;
                btn_el.after($rObj.el_template);
                var items = $rObj.el_template.querySelectorAll('.dynamic');

                for (var i = 0; i < items.length; i++) {
                    var item = items[i];
                    if (item.id != null && item.id != "") {
                        item.id = item.id + "_" + $rObj.r_index;
                    }
                    var forattr = item.getAttribute("for");
                    if (forattr != null) item.setAttribute("for", forattr + "_" + $rObj.r_index);
                    var nameattr = item.getAttribute("name");
                    if (nameattr != null) item.setAttribute("name", nameattr + "_" + $rObj.r_index);
                }

                //set the category and estimate subtitle  (one-time)
                var el_cat = document.getElementById("cat_"+$rObj.r_index);
                el_cat.innerHTML = $rObj.cat+" - "+$rObj.estimate;
                
                var el_status = document.getElementById("status_"+$rObj.r_index);
                RM.classifyStatus(el_status, $rObj.status);
                return resolve($rObj.el_template);
            }
        };

        xhr.open("GET", "./roadmap_parts/"+$rObj.part + "?v=" + Math.random() * 10, true);
        xhr.responseType = "";
        xhr.send();
    });
}


/**
 * roadmap status colorization
 * @param {*} $el_status 
 * @param {*} $status 
 */
RM.classifyStatus = function($el_status, $status) {
    $el_status.innerHTML = "<span style='color:whitesmoke'>status</span>&nbsp;"+$status;
    var color_class = "color-white";
    if ($status=="done" || $status=="active" || $status=="draft" || $status=="historic") {
        color_class = "color-green";
    }
    else if ($status=="waiting activation" || $status=="planning" || $status=="ongoing" || $status=="soon") {
        color_class = "color-yellow";
    }
    else if ($status=="investigating" || $status=="brainstorming") {
        color_class="color-teal";
    }
    $el_status.classList.add(color_class)
}


/**
 * 
 * @param {*} $el 
 * @param {*} $displayProp 
 * @returns 
 */
RM.show = function ($el, $displayProp = "") {
    var el = RM.__getEl($el);
    if (el == null) return null;
    if (el.classList.contains("show") == false) el.classList.add("show");
    if (el.classList.contains("hide") == true) el.classList.remove("hide");
    if ($displayProp == "") {
        var restore = el.getAttribute("display_restore");
        if (restore) el.style.display = restore;
        else if (el.style.display == "none") el.style.display = "";
    }
    else el.style.display = $displayProp;
    return el;
}


/**
 * 
 * @param {*} $el 
 * @returns 
 */
RM.hide = function ($el) {
    var el = RM.__getEl($el);
    if (el == null) return null;
    if (el.classList.contains("hide") == false) el.classList.add("hide");
    if (el.classList.contains("show") == true) el.classList.remove("show");
    el.style.display = "none";
    return el;
}


/**
 * 
 * @param {*} $el 
 * @returns 
 */
RM.playSound = function ($el) {
    var el = this.__getEl($el);
    if (el == null) return null;
    el.play();
}


/**
 * 
 * @param {*} $el 
 * @returns 
 */
RM.__getEl = function ($el) {
    if (typeof $el === "string") {
        if ($el[0] == ".") return document.querySelector($el);
        else if ($el[0] == "#") return document.getElementById($el.substring(1));
        return document.getElementById($el);
    }
    return $el;
}


/**
 * 
 */
RM.onWindowResize=function() {
    RM.camera.aspect = window.innerWidth / window.innerHeight;
    RM.camera.updateProjectionMatrix();
    RM.renderer.setSize( window.innerWidth, window.innerHeight );
    RM.rendererCSS.setSize( window.innerWidth, window.innerHeight );
}


RM.startSample = function() {
    
}

RM.tally = function($arr, $name) {
    var keys = Object.keys($arr);
    var scoreWithMostHits = 0;
    var mostHits = 0;
    var highestScore = 0;
    var highestScoreHits = 0;
    for (var k=0; k < keys.length; k++) {
        var score = keys[k];
        var ticks = $arr[score];
        if (ticks.length > mostHits) {
            mostHits = ticks.length;
            scoreWithMostHits = score;
        }
        if (score>highestScore) {
            highestScore = score;
            highestScoreHits = ticks.length;
        }
    }
    console.log("tally fo "+$name);
    console.log("score with most hits: "+scoreWithMostHits + "hits: "+mostHits);
    console.log("high: "+highestScore + "hits of high: "+highestScoreHits);
}

//for use via the console and debugging.
RM.pa = function() {
    RM.tally(RM._b1s, "b1s");
    RM.tally(RM._b2s, "b2s");
    RM.tally(RM._m1s, "m1s");
    RM.tally(RM._m2s, "m2s");
}


/**
 * 
 */
RM.animate = function() {
    RM.tick+=1;
    const time = performance.now() * 0.001;

    requestAnimationFrame(RM.animate);

    if (RM.rotateBall==true) {
        RM.logoheartofanne.rotation.y+=.01;
    }

    if (RM.bINTRODONE==true) {
        
        RM.updateSoundVolume();

        const data = RM.analyser.getFrequencyData();
        const b1 = (data[0]+data[1]); 
        const b2 = (data[1]+data[2]);
        const m1 = (data[3]+data[4]); 

        let sum = 0;
        for (let i = 0; i < data.length; i++) {
            sum += data[i];
        }
        var beatsize = sum/RM.vol;
        RM.logoheartofanne.scale.set(beatsize*7, beatsize*7, beatsize*7);

        var spin = .001;
        //elevation should got to 90, from zero in 370secs
        //chill mix slows spin
        //and alters the sky..
        //no more color flashes...
        if (RM.bCHILLMIX==true || RM.bANNIE==true) {
            spin = .00033;
            RM.effectController.rayleigh = beatsize*3;
            RM.effectController.exposure = .4*(m1/510)+.5;
            RM.effectController.mieCoefficient = beatsize/10;
            RM.effectController.elevation = ANNE.RM.maintrack.context.currentTime / 370;
            RM.updateSun();
        }

        else {
            if (b2>=475 && b1==510) {
                RM.white_material_low_alpha.color.setHex(0x00ffff);     //turn whites teal   
                if (m1>=337) {
                    RM.yellow_material_low_alpha.color.setHex(0xfe17df);     //turn whites teal     
                }
                else RM.yellow_material_low_alpha.color.setHex(0xfefe17);     //turn it back to white
            }
            else {
                RM.yellow_material_low_alpha.color.setHex(0xfefe17); 
                RM.white_material_low_alpha.color.setHex(0xffffff);     //turn it back to white
            }
        }

        //rotation based on base 1/2 channel
        var fact = b1/510;
        var fact2 = b2/475;

        RM.giantRing.rotation.z-=(spin*fact);
        RM.skyboxMesh.rotation.z-=(spin/10*fact);

        if (RM.bCHILLMIX==true) {
            for (var m=0; m < RM._rmMeshes.length; m++) {
                RM._rmMeshes[m].rotation.z-=(.003*fact2);
                RM._rmMeshes[m].ringCSSObj.rotation.z+=(.003*fact2);
            }
        }
        else if (RM.bANNIE==true) {
            for (var m=0; m < RM._rmMeshes.length; m++) {
                RM._rmMeshes[m].rotation.z+=(.003*fact2);
                RM._rmMeshes[m].ringCSSObj.rotation.z-=(.003*fact2);
            }
        }
        else if (RM._lastGreenMesh!=null) {
            RM._lastGreenMesh.rotation.z-=(.003*fact2);
            RM._lastGreenMesh.ringCSSObj.rotation.z+=(.003*fact2);
        }

       // if (RM._b1s[b1]!=undefined) RM._b1s[b1].push(RM.tick);
       //else RM._b1s[b1]=[RM.tick];
       // if (RM._b2s[b2]!=undefined) RM._b2s[b2].push(RM.tick);
       // else RM._b2s[b2]=[RM.tick];
       // if (RM._m1s[m1]!=undefined) RM._m1s[m1].push(RM.tick);
       // else RM._m1s[m1]=[RM.tick];
       // if (RM._m2s[m2]!=undefined) RM._m2s[m2].push(RM.tick);
       // else RM._m2s[m2]=[RM.tick];
       // if (RM.tick % 2000 == 0) {
          //  RM.tally(RM._b1s, "b1s");
          //  RM.tally(RM._b2s, "b2s");
           // RM.tally(RM._m1s, "m1s");
           // RM.tally(RM._m2s, "m2s");
       // }
    }

    else {
        RM.giantRing.rotation.z-=.001;
        if (RM._lastGreenMesh!=null) {
            RM._lastGreenMesh.ringCSSObj.rotation.z+=(.003);
            RM._lastGreenMesh.rotation.z-=(.003);
        }
    }

    RM.renderer.render(RM.scene, RM.camera);
    RM.rendererCSS.render( RM.scene, RM.camera );
}


/**
 * 
 */
RM.updateSoundVolume=function() {
    const cameraDirection = RM.camera.getWorldDirection(new THREE.Vector3());
    const cameraToObject = RM.logoheartofanne.position.clone().sub(RM.camera.position);
    const angle = cameraDirection.angleTo(cameraToObject);
  
    //Adjust the volume based on the angle
    const maxAngle = Math.PI / 2;

    var volume = Math.cos(angle / maxAngle);
    volume = (volume - 0.65) / (1 - 0.65);
    volume = Math.max(volume, 0);
    volume = Math.min(volume, 1);

    // Set the volume of the sound RM.ff_source
    if (RM.bHOAPULSE==false && volume > .65) {
        if (RM.ff_source.context.state != "running") RM.ff_source.start(0);
    }
    
    RM.logoheartofannegainnode.gain.setValueAtTime(volume, audioContext.currentTime);
}