/***********************************************************************************************
                                                                                                                                            
                                     .....                                      
                            @@@@@@@@@@@@@@@@@@@@@@@@@@@@,                       
                       @@@@@@@@@@@@@@&%%&@@@@@@@@@@@@@@@@@@@@.                  
                    @@@@@@@@@,                        @@@@@@@@@@(               
                 @@@@@@@@@                                .@@@@@@@@             
               @@@@@@@@             %@@@@@@@@                @@@@@@@            
             @@@@@@@@            (@@@@@@@@@@@@@.               @@@@@@&          
           #@@@@@@@             @@@@@@@@@@@@@@@@.               @@@@@@/         
          @@@@@@@@              @@@@@@@@@@@@@@@@/               /@@@@@@         
         @@@@@@@@                @@@@@@@@@@@@@@%                 @@@@@@@        
        @@@@@@@@                    @@@@@@@@@                    @@@@@@@        
       @@@@@@@@                                                  @@@@@@@        
      @@@@@@@@@                                                  @@@@@@@*       
      @@@@@@@@#         @@ @    @&@   @.  @,@  @%  @@@@@@        @@@@@@@@       
      @@@@@@@@@        (@  /@   @  @  @.  @  @ @%  @@@(          @@@@@@@@       
      @@@@@@@@@@       @    @@  @   @,@.  @   @@%  @@@@@@        @@@@@@@@       
       @@@@@@@@@                                                  @@@@@@        
        @@@@@@@                                                                 
                   
***********************************************************************************************/
const TYPE_NAME = -1;

const TYPE_SYS = 0;             //Sytem info for neuron (sponsored_by, annother, owned_by).  Builds extra indexes.
const TYPE_BE = 1;              //semantic shothand for BE DIMENSION.  CLASS/INST RELNS create a DAG.
const TYPE_PARAM = 2;           //semantic shorthand for param (class attribute/property)
const TYPE_HAS = 3;             //semantic shorthand for param fulfilling has value (attribute/property)
const TYPE_AWARENESS = 4;       //for awareness space (emotions/feelz)

const TYPE_EXPERIENCE = 5;      //for experience (temporal) neuron space

//CLUMPS can function as a TYPE_BE, TYPE_HAS, OR TYPE_SYS.
//those Types are actually short-hand for clumps but they can get away with
//only using one TX to specify the triplet info.
//A full clump is for specifying more than the standard 3 pieces of semantic info
//in a single ANNE TX.
//** For ease of use we can put a TYPE_XXX in the clump
//to clue our interpreter to what dimension the clump lives in.
//The other CLUMPS of episodic, BRAIN and EXPERIENCE are already their own dimension
//so this is only relevant for semantic clumps that are handled in semantic memory space.
const TYPE_CLUMP = 6;

const TYPE_SYS_MSG = 7;        //only relevant to coms, not the neuron
const TYPE_BRAIN = 8;          //for neural-net space wirings a->wire->b
const TYPE_ROLE = 9;           //for specifiying  semantic roles inside of a clump

const TYPE_METAN = 10;         //for specifying meta information about roles used in a clump

const TYPE_EPISODIC = 11;      //for documenting a self-sensory/vitals data at run-time in a clump.
const TYPE_INTERNAL_PAY = 17;  //keyless neurons paying out.
const TYPE_BROADCAST_PAY = 18; //standard broadcast payment.

const TYPE_BRAIN0 = 20;        //20+ are for sequence in the brain dimension

//100+ are available for brain type/custom use.. assumed to be order if more than one in a neuron's outs.
const TYPE_BRAIN120 = 120;     //used in ASDF-> annode relons to store trigger pats. (can use how u like in other neurons)
const TYPE_BRAIN121 = 121;     //myelination measure (a->b->[to_type_number]#c)

const TO_TYPE_SEED = 0;        //relon to is an encrypted seed (all seed are encrypted)
const TO_TYPE_ANNE = 1;        //relon to is a .nid, the idenfier of another neuron.  This is a legit graph triplet.
const TO_TYPE_TXID = 2;        //relon to is a txid.
const TO_TYPE_URL = 3;         //relon to is a url.

//relon to is a number.  Note: our ANNODE stores numerics in a LONG field
//if a non-long/int is needed use the TO_TYPE_STRING

const TO_TYPE_NUMBER = 4;       
const TO_TYPE_DATE = 5;         //relon to is a date in the ANNE date format.    
const TO_TYPE_DATETIME = 6;     //relon to is a datetime in the ANNE datetime format.
const TO_TYPE_STRING = 7;       //relon to is a string

const TO_TYPE_TEXT = 7;         //alias in case u conflate

const TO_TYPE_BLOB = 8;         //relon to is a blob (bad donkey, not welcome in ANNE)
const TO_TYPE_TIME = 9;         //relon to is a time in the ANNE time format.
const TO_TYPE_BOOLEAN = 10;     //relon to is a boolean string which will be converted to true/false at runtime
const TO_TYPE_FILEHASH = 11;    //relon to is a filehash
const TO_TYPE_IPV4 = 12;    //relon to is a IPV4 address
const TO_TYPE_IPV6 = 13;    //relon to is a IPV6 address

//encrypted versions of the above + 30 in value from its respective type.
//expansion to 30 base types possible (don't foresee this as needed
//in a perfect-brain world, all are references to either neurons or other TXIDs

const TO_TYPE_ENCR_ANNE = 31;                
const TO_TYPE_ENCR_TXID = 32;
const TO_TYPE_ENCR_URL = 33;        
const TO_TYPE_ENCR_NUMBER = 34;
const TO_TYPE_ENCR_DATE = 35;
const TO_TYPE_ENCR_DATETIME = 36;
const TO_TYPE_ENCR_STRING = 37; 
const TO_TYPE_ENCR_BLOB = 38;
const TO_TYPE_ENCR_TIME = 39;
const TO_TYPE_ENCR_BOOLEAN = 40;
const TO_TYPE_ENCR_FILEHASH = 41;
const TO_TYPE_ENCR_IPV4 = 42;
const TO_TYPE_ENCR_IPV6 = 43;

const NTYPE_CLASS = 0;
const NTYPE_INST = 1;
const NTYPE_UNKNOWN = 2;
const NTYPE_CLUMP = 3;

const ORIGIN_CHAIN = 0;         //a real TX on-chain
const ORIGIN_NODE = 1;          //an emulated neuron created by the node
const ORIGIN_NODE_PRELOAD = 2;  //an emulated neuron preloaded by the node (with a real seed)
const ORIGIN_OWNERPROOF = 3;    
const ORIGIN_CONTRACTS = 4;     //emulated from contracts
const ORIGIN_KEYLESS = 5;
const ORIGIN_GLOBAL = 6;
const ORIGIN_CUSTODIAL  = 8;
const ORIGIN_EXTERNAL_ML = 12;  //if a ML is joined w/ the DB, that data is ML.

const RORIGIN_ANNODE = 0;         //origination made on the ANNODE in ANNE hypergraph only
const RORIGIN_BROADCAST = 1;    //origination broadcast as a TX
const RORIGIN_COLLAB_BROADCAST = 2;        //collab broadcast (waiting approval)
const RORIGIN_GHOST = 3;        //origination user local

const ANNE_ACCTID = "7777770163980908667";    
const ANNE_ADMIN_ACCTID = "7777771543345157118";
const ANNE_ADMIN_BACKUP_ACCTID = "7777044340734581180";

const TARGET_NID = 1;
const TARGET_TXID = 2;

const TXMODE_GHOST = 0;             //store to local DB only
const TXMODE_ANNE = 1;              //instant send

const CTYPE_COLLAB_ONLY = 0;
const CTYPE_COLLAB_HANAI = 1;
const CTYPE_PENDING_ANJ = 2;
const CTYPE_HANAI_ONLY = 3;
const CTYPE_MERGE_REQUEST = 4;

//Keyless neurons allow for easy creation of child relons w/o the management of another key
//They can be collaborated w/ by request or easy delegation rights set to other approved authors.
//Note: ANNE globals (date, numbers, locations, keywords) are keyless neurons and are created on-demand.
const NEWNEURONMODE_KEYLESS = 0;  

//Key'd neurons provide real a real/key that is only known by the identity that created it.
//Key'd neurons can be collaborated w/ by request but they cannot be delegated since only the
//Key holder can broadcast transactions from the key'd neuron.
//Note: ANNE preloaded neurons are key'd as they need pre-known neuron IDs to reference them in the base-protocol.
const NEWNEURONMODE_KEYED = 1;        

const NEURON_NOT_FROZE = 0;
const NEURON_FROZE = 1;

/**
 * Load up of neurons and relons
 */
class mem {
    constructor(){
        this.ANNE = null;
        this.MEMP = null;
        this.USER = null;
        this.ANNE_ACCTID = "7777770163980908667";    
        this.ANNE_ADMIN_ACCTID = "7777771543345157118";
        this.ANNE_ADMIN_BACKUP_ACCTID = "7777044340734581180";
        this.ANNE_DEV_FUND_ACCTID = "7770353147211057540";

        this.ANNE_SUPERS = [this.ANNE_ACCTID, this.ANNE_ADMIN_ACCTID, this.ANNE_ADMIN_BACKUP_ACCTID];

        this._bootupCompleteCallback = null;

        this._aNeurons_hashtbl = [];            //hash of nid w/ prop point to nid.
        this._aNeurons = [];                    //nid indexed nObjs
        this._aRelons_hashtbl = [];             //hash of txid w/ prop pointing to rid
        this._aRelons = [];                     //rid indexed rObj 
        
        this._valid_TI = [];                    //valid to types for our text index
        this.__bDOTEXTINDEX = false;            //flag to turn on general text indexes
        this._bDOPERSONINDEX = false;           //flat to turn on person indexes (not yet made, will be an optimization.)
        
        this._TEXT_INDEX = new Object();  
        this._TEXT_INDEX_PARTIALS = new Object();

        this.__bMINIEDGES = false;

        this.__INDENT = "&#160;└─";              //this will be moved to a website preference neuron :)
        this._emulatedTXIDLength = 20;       
        this._EMULATE_BE_TEXT = true;            //this emulates text neurons for the text strings of our hard-coded neurons.  (overkill for most usecases)
        this._EMULATE_TO_TYPES = true;           //this emulates the non bitcion to types fround in edges.  (a must for advanced utility features)
        this._EMULATE_BE_TEXT_NEURON = false;
        
        this._activeWorldViewTempNids = [];

        this.neuron_necrosis_list = [];
        this.tx_ap_inh_list = []; 
        this._localRunDelete = [];   
        this._thisRunNECROSIS = [];

        this._SYS_TYPE_CODES = [];
        this._SYS_TO_TYPE_CODES = [];

        //any neuron classified as an ANNE sys prop class.
        //these props auto populate javascript .prop values in neurons
        //that supply their value.
        this._SYS_PROPS_TNIDS = [];   

        
        
        this._RESERVED_PROPS = ["parent", 
                                "be", 
                                "bes", 
                                "_gt", 
                                "sponsored_by", 
                                "owned_by", "name", "text",
                                "nid",
                                "txid", "from_nid", "to_s", "reln", "type", "to_type",
                                "outs", "ins", "relon_sponsor"];

        //set to the active language
        //will looks for all text values from this.__TEXT relons as default (english)
        //else, will look for 
        this.__TEXT_ptr = null;     

        
        //confirm emulated boolean maps to existing boolean neuron values.
        this._emulatables = [];
                            
        //as we process local ghosts, look for the ghost_order
        //as we find values, bump this value up to one greater
        this._nextGhostOrder = 0;
        this._anne_ghosts = [];         //local ghost handling (not yet re-enabled)

        //used for loading emulated neurons run-time for apps.
        //big loads can be done then all proccessed instead of processing
        //each relon one at a time.
        this._sand_mark = 0;  
        
        this._active_lang_alpha2 = "";

        this._cascadePropsDownChildrenClasses_cnt = 0;

        this._monthMap=[];
        this._monthLookup=[];
    }


    setPersonIndex=function($bDOPERSONINDEX) {
        this._bDOPERSONINDEX=$bDOPERSONINDEX;
    }

    //this is getting inherited by neuron that is SHOULD NOT BE INHERITED BY???
    setNumber = function ($v) { 
        if ($v==undefined || $v==null) return;
        if (typeof $v != "string") {
            if ($v.text!=undefined) {
                if (ANNE.UTILS.isNum($v.text)==false) {
                    return;
                }  
                $v.setValue( parseFloat($v.text) );
                return;
            }
        }
        $v.setValue( parseFloat($v) );
    }
    

    /** 
     * Boot up the RAM ANNE memory.
     * 
     * Init the MEMP
     * Load up local indexeddb from the browser.
     * Grab preloaded neurons from the ANNODE, if DB mem is empty.
     * 
     * @author {Mr.Scatman} 
     */
     boot_up=async($ANNE, $bootupCompleteCallback)=>{
        //console.warn("(this.boot_up)");
        this.ANNE = $ANNE;

        //We need a few neurons sys level neurons and some protocol-using specific 
        //concepts when building ANNE local memory in the browser
        //we create them here in bare-bone capacity.  Their definition will build
        //as we process more of their relons.  Here we have the basics to overcome 
        //the chicken-and-egg problem that arrises when using an electronic brain to
        //build a brain.  The language used to build the brain itself is 
        //defined by the neurons and relons that are in it.
        var newSysTypeCode = function ($nObj, $sys_alias, $sys_enum_val, $sys_code="", $validator=null, $set_value=null) {
            $nObj.SYS_ALIAS=$sys_alias;
            $nObj.SYS_ENUM_VAL = $sys_enum_val
            $nObj.SYS_CODE=$sys_code;
            $nObj.format_validator=$validator;      //if exists, will run after set value -- to validate the format.
            $nObj.set_value=$set_value;             //if exists, will run to set the .value of this nObj
            ANNE.MEM._SYS_TYPE_CODES.push($nObj);
        }

        var newSysToTypeCode = function ($nObj, $sys_alias, $sys_enum_val, $sys_code="", $validator=null, $set_value=null) {
            $nObj.SYS_ALIAS=$sys_alias;
            $nObj.SYS_ENUM_VAL = $sys_enum_val
            $nObj.SYS_CODE=$sys_code;
            $nObj.format_validator=$validator;      //if exists, will run after set value -- to validate the format.
            $nObj.set_value=$set_value;             //if exists, will run to set the .value of this nObj
            ANNE.MEM._SYS_TO_TYPE_CODES.push($nObj);
        }

        this.___SELF___ = this.minBootNeuron({nid:"7777089034065106708", text:"___SELF___", ntype:NTYPE_CLASS});
        this.___SELF___.be_lvl = 0;

        this.__ANNE__ = this.minBootNeuron({nid:ANNE_ACCTID, text:"A.N.N.E.", ntype:NTYPE_INST});
        this.__ANNE__.addAkaManual("ANNE");

        //patterns for events and other things can be stored on the app for prefs etc.
        this.ANNE_EXT_APP = this.minBootNeuron({nid:"7772292993542690466", text:"ANNE EXT APP", ntype:NTYPE_INST});
        
        //Neurons will be an instance of a class/sublass or a subclass of a class.
        //All classifications exist in the BE Hierarchy parent/child DAG (TYPE_BE dimension).
        //if the neuron is a class and is intended to only be subclassed classes_only is set.
        //if the neuron is a class and is intended to only be instantiated insts_only is set.
        //if both subclasses and insts are ok, then neither classes_only, or insts_only is set.

        //not intended to be instantiated.
        this._GENERAL_PROP_CLASS = this.minBootNeuron({nid:"7770801601458804038", text:"general prop class", ntype:NTYPE_CLASS});
        this._GENERAL_PROP_CLASS.be_lvl = 1;
        
            this.unicode_start = this.minBootNeuron({nid:"7773282456393979069", text:"unicode start", ntype:NTYPE_CLASS});
            this.unicode_start.be_lvl = 2;
            this.unicode_end = this.minBootNeuron({nid:"7775740209310803806", text:"unicode end", ntype:NTYPE_CLASS});
            this.unicode_end.be_lvl = 2;
        
        //special for the semantic space and protocol enforcement
        this.__ANNE_SYS_CLASS = this.minBootNeuron({nid:"7775402688912060933", text:"ANNE sys prop class", ntype:NTYPE_CLASS});
        this.__ANNE_SYS_CLASS.be_lvl = 1;
            this.rez_pool_priority1 = this.minBootNeuron({nid:"7770223143918999134", text:"rez pool priority 1", ntype:NTYPE_CLASS});    
            this.rez_pool_priority1.be_lvl = 2;
            this.rez_pool_priority2 = this.minBootNeuron({nid:"7772071294239796669", text:"rez pool priority 2", ntype:NTYPE_CLASS});    
            this.rez_pool_priority2.be_lvl = 2;

            this.TYPE = this.minBootNeuron({nid:"7771390251954504665", text:"DATA_TYPE", ntype:NTYPE_CLASS});
            this.TYPE.be_lvl = 2;

                this.DATA_TYPE = this.minBootNeuron({nid:"7776990924864902362", text:"DATA_TYPE", ntype:NTYPE_CLASS});
                this.DATA_TYPE.be_lvl = 3;
            
                        
            this.CMD = this.minBootNeuron({nid:"7772212307781566495", text:"CND", ntype:NTYPE_CLASS});
            this.CMD.be_lvl = 2;
                this.SYS_MSG_CMD = this.minBootNeuron({nid:"7775628525395840248", text:"sys msg cmd", ntype:NTYPE_CLASS});
                this.SYS_MSG_CMD.be_lvl = 3;
                    this.ANNE_SYS_ANNOUNCEMENT = this.minBootNeuron({nid:"7777961960607822223", text:"ANNE system announcement", ntype:NTYPE_CLASS});
                    this.ANNE_SYS_ANNOUNCEMENT.be_lvl = 4;
                    
                    //extreme testing
                    this.ANNE_SUPER_POP = this.minBootNeuron({nid:"777654215432346068", text:"SuperPop", ntype:NTYPE_CLASS});
                    
                    this.ANNE_BRICK_POINT = this.minBootNeuron({nid:"777004069246568712", text:"ANNODE BrickPoint", ntype:NTYPE_CLASS});
                    this.ANNE_BRICK_POINT.be_lvl = 4;
                    this.ANNE_BRICK_PT_MSG = this.minBootNeuron({nid:"7771381953594362011", text:"ANNODE BrickPoint message", ntype:NTYPE_CLASS});
                    this.ANNE_BRICK_PT_MSG.be_lvl = 4;
                    this.ANNE_BRICK_PT_MIN_V = this.minBootNeuron({nid:"7776219282356411272", text:"ANNODE BrickPoint min version", ntype:NTYPE_CLASS});
                    this.ANNE_BRICK_PT_MIN_V.be_lvl = 4;
                    this.ANNE_BRICK_PT_ACTIVATION_BLOCK = this.minBootNeuron({nid:"7770141402154065935", text:"ANNODE BrickPoint activation block", ntype:NTYPE_CLASS});
                    this.ANNE_BRICK_PT_ACTIVATION_BLOCK.be_lvl = 4;


        this.SAFUID = this.minBootNeuron({nid:"7774440805627554213", text:"SAFUID", ntype:NTYPE_CLASS});
        this._default_class_nid = this.minBootNeuron({nid:"777178866083313518", text:"default class nid", ntype:NTYPE_CLASS});
        

        //special for the brain firing space
        this.__SYSBRAIN = this.minBootNeuron({nid:"777683243654836824", text:"SYSVAR", ntype:NTYPE_CLASS});
        this.__SYSBRAIN.be_lvl = 1;

            this.brain_assoc_strong = this.minBootNeuron({nid:"777837095094060702", text:"brain:assoc_strong", ntype:NTYPE_CLASS});    
            this.brain_assoc_strong.be_lvl = 2;
            this.brain_assoc_strong.value=2.22;

            this.brain_assoc = this.minBootNeuron({nid:"7772784241541954051", text:"brain:assoc", ntype:NTYPE_CLASS});    
            this.brain_assoc.be_lvl = 2;
            this.brain_assoc_strong.value=1.11;

            this.anbl_assoc = this.minBootNeuron({nid:"7774753656336827036", text:"brain:anbl_assoc", ntype:NTYPE_CLASS});
            
        //this occcurs in the TYPE_BRAIN dimension and throttles firing of the the TO concept it points to.
        this.__NOTACTIVE = this.minBootNeuron({nid:"7776496372488323950", text:"**notactive**", ntype:NTYPE_CLASS});
        this.__NOTACTIVE.be_lvl = 2;
        this.__NOTACTIVE.setValue(-2.22);

        //throttling neuron firing.
        this.__throttlemin = this.minBootNeuron({nid:"7777643683526919133", text:"**throttlemin**", ntype:NTYPE_CLASS});
        this.__throttlemin.be_lvl = 3;
        this.__throttlemin.setValue(-1.11);

        this.__throttlemax = this.minBootNeuron({nid:"7775174343458816319", text:"**throttlemax**", ntype:NTYPE_CLASS});
        this.__throttlemax.be_lvl = 3;
        this.__throttlemin.setValue(-2.22);

        this.__ACTIVE = this.minBootNeuron({nid:"7778810467984387174", text:"**active**", ntype:NTYPE_CLASS});
        this.__ACTIVE.be_lvl = 2;

        this.EC_CLASS = this.minBootNeuron({nid:"777585041016039777", text:"**EC**", ntype:NTYPE_CLASS});
        this.EC_CLASS.be_lvl = 1;

            this.class = this.minBootNeuron({nid:"7777942827590344670", text:"class", ntype:NTYPE_CLASS});
            this.class.be_lvl = 2;
        
            this.inst = this.minBootNeuron({nid:"7777109984382823688", text:"inst", ntype:NTYPE_CLASS});
            this.inst.prompt = "What are you?";
                    
            this.ec_action = this.minBootNeuron({nid:"7774365287013199911", text:"ec:action", ntype:NTYPE_INST});
            this.ec_action.be_lvl = 2;
            this.ec_action._uico = "🏃‍♂️";
            
            this.ec_thing = this.minBootNeuron({nid:"7778260411935216992", text:"ec:thing", ntype:NTYPE_CLASS});
            this.ec_thing.be_lvl = 2;
                this.entity = this.minBootNeuron({nid:"7775965067238290504", text:"entity", ntype:NTYPE_CLASS});
                this.entity.be_lvl = 3;
            
            this.ec_living = this.minBootNeuron({nid:"7772014754426478363", text:"ec:living", ntype:NTYPE_CLASS});
            this.ec_living.be_lvl = 3;
            
            //Person is to document a non-identity definition of a person.
            //Not to be confused with the identity class with are real users of the system.
            this.person = this.minBootNeuron({nid:"7773974326223336303", text:"person", ntype:NTYPE_CLASS});
            this.person.be_lvl = 4;
            this.person._uico = "<span style='color:yellow'>👤</span>";
            
            //we can do movies from TMDB with this one...
            this.movie = this.minBootNeuron({nid:"7771769778243524971", text:"movie", ntype:NTYPE_CLASS});  
            //all the TMDB relns we want should be here.  /director/actor/genre etc.

            this.named_after = this.minBootNeuron({nid:"7779919689813420887", text:"named after", ntype:NTYPE_CLASS});
            this.pop_demonym = this.minBootNeuron({nid:"777505779900849159", text:"pop demonym", ntype:NTYPE_CLASS});

            this.ec_nonliving = this.minBootNeuron({nid:"7773981355789251728", text:"ec:nonliving", ntype:NTYPE_CLASS});
            this.ec_nonliving.be_lvl = 3;

            this.ec_location = this.minBootNeuron({nid:"7773349123364275631", text:"ec:location", ntype:NTYPE_CLASS});    
            this.ec_location.be_lvl = 4;

            this.ec_state = this.minBootNeuron({nid:"7777439602338734976", text:"ec:state", ntype:NTYPE_CLASS});
            this.ec_state.be_lvl = 2;
            
            this.ec_measure = this.minBootNeuron({nid:"7770647771955623748", text:"Unit of Measure", ntype:NTYPE_CLASS});
            this.ec_measure.be_lvl = 3;
           
            this.unit_of_measure = this.minBootNeuron({nid:"777700476213440998", text:"Unit of Measure", ntype:NTYPE_CLASS});
            this.unit_of_measure.be_lvl = 4;
             
            this.unit_of_account = this.minBootNeuron({nid:"7774906754160016124", text:"Unit of Account", ntype:NTYPE_CLASS});
            this.unit_of_account.be_lvl = 5;
            
            this.unit_of_crypto= this.minBootNeuron({nid:"7771690081405920648", text:"Unit of Crypto", ntype:NTYPE_CLASS});
            this.unit_of_crypto.be_lvl = 6;

            this.annies = this.minBootNeuron({nid:"7771153305797255960", text:"annies", ntype:NTYPE_CLASS});
            this.annies.be_lvl = 7;

            //aka ROLE_BE.
            this.ec_is = this.minBootNeuron({nid:"7775323663099289436", text:"ec:is", ntype:NTYPE_INST});
            this.ec_is.be_lvl = 2;
            this.role_be = this.ec_is;    //same concept.

            this.ec_posn = this.minBootNeuron({nid:"777042615346522551", text:"ec:posn", ntype:NTYPE_INST});
            this.ec_posn.be_lvl = 2;
            
            this.ec_expection = this.minBootNeuron({nid:"7774919700244959416", text:"ec:expectation", ntype:NTYPE_INST});
            
            this.ec_opposite = this.minBootNeuron({nid:"7777676504812424745", text:"ec.opposite", ntype:NTYPE_CLASS});
        
            //7770250558819326868 mapper state

            this.ec_mode = this.minBootNeuron({nid:"7774567902469413739", text:"ec:mode", ntype:NTYPE_CLASS});    
            this.ec_mode.be_lvl = 2;


                this.location = this.minBootNeuron({nid:"7775670854606187707", text:"location", ntype:NTYPE_CLASS});
                this.location.be_lvl = 3;
                this.location.unit_tests =["Syracuse, New York, U.S."];
                this.location.unitTest=async()=>{
                    for (var u=0; u < ANNE.MEM.location.unit_tests.length; u++) {
                        await ANNE.MEM.location.testParse( ANNE.MEM.location.unit_tests[u]);
                    }
                }
                this.location.testParse=async($loc)=>{
                    this.defaultLocObj = new anne_location(this.misc.default_value);
                    await defaultLocObj.processFull();
                    defaultLoclObj.print();
                    defaultLocObj.printFullDebugString();
                }
                this.location._uico="📍";

                this.registrationPolity = this.minBootNeuron({nid:"7775972763622961295", text:"registration polity", ntype:NTYPE_CLASS}); 
                this.hqLocation = this.minBootNeuron({nid:"7775518395372784977", text:"Headquarters", ntype:NTYPE_CLASS}); 
                
            this.ec_theSpeaker = this.minBootNeuron({nid:"6130905652501722840", text:"ec:the speaker", ntype:NTYPE_CLASS});    
            this.ec_theSpeaker.be_lvl = 2;

            //is under ec_idiom
            this.idiom = this.minBootNeuron({nid:"7779579523222965446", text:"idiom", ntype:NTYPE_CLASS});
            this.idiom.be_lvl = 3;

        this.city = this.minBootNeuron({nid:"7779778654320523970", text:"city", ntype:NTYPE_CLASS});
            
        this.anbl = this.minBootNeuron({nid:"7773795520052184088", text:"anbl", ntype:NTYPE_CLASS});

        this.motto = this.minBootNeuron({nid:"7775617504524778859", text:"motto", ntype:NTYPE_CLASS});

        //for testing.  s.b. there already.
        this.ec_anvert_yin = this.minBootNeuron({nid:"7776879633620640959", text:"ec:anvert_yin", ntype:NTYPE_CLASS});    
        this.ec_anvert_yang = this.minBootNeuron({nid:"7773336624326965724", text:"ec:anvert_yang", ntype:NTYPE_CLASS});    
        
        this.eng_subj = this.minBootNeuron({nid:"8557921478923992894", text:"english subject", ntype:NTYPE_CLASS});    
        this.eng_obj = this.minBootNeuron({nid:"16826117856129032282", text:"english object", ntype:NTYPE_CLASS});    
        
        this.eng_grammar_rule = this.minBootNeuron({nid:"7770417256868673002", text:"ENG grammar rule", ntype:NTYPE_CLASS});    
        
        this.eng_action_to_thing = this.minBootNeuron({nid:"7772004807272349660", text:"action->thing", ntype:NTYPE_CLASS});    
        this.eng_action_to_state = this.minBootNeuron({nid:"7770228850164583833", text:"action->state", ntype:NTYPE_CLASS});    
        this.eng_thing_to_state = this.minBootNeuron({nid:"7773065301314646689", text:"thing->state", ntype:NTYPE_CLASS});    
        this.eng_thing_to_thing = this.minBootNeuron({nid:"7774092348316350665", text:"thing->thing", ntype:NTYPE_CLASS});    
        this.eng_state_to_thing = this.minBootNeuron({nid:"7777785038941313166", text:"state->thing", ntype:NTYPE_CLASS});    
        this.eng_state_to_action = this.minBootNeuron({nid:"7773073322491997535", text:"state->action", ntype:NTYPE_CLASS});    
        this.eng_state_to_manner = this.minBootNeuron({nid:"7777057456095646496", text:"state->manner", ntype:NTYPE_CLASS});    
        this.eng_action_to_action = this.minBootNeuron({nid:"7779372114410308475", text:"action->action", ntype:NTYPE_CLASS});    

        //for prounounish resolutions or any language's notion of placeholders that resolve to other concepts
        //that are in context.
        this.ec_resolve = this.minBootNeuron({nid:"777847931831336568", text:"ec:resolvable (thing)", ntype:NTYPE_CLASS});    
        
        this.ec_resolvableThing = this.minBootNeuron({nid:"12811539405663275121", text:"ec:resolvable (thing)", ntype:NTYPE_CLASS});    
        this.ec_resolvablePluralThing = this.minBootNeuron({nid:"1229903086758389843", text:"ec:resolvable plural (thing)", ntype:NTYPE_CLASS});    
        this.ec_resolvableMale = this.minBootNeuron({nid:"16239629656306601583", text:"ec:resolvable male (thing)", ntype:NTYPE_CLASS});    
        this.ec_resolvableFemale = this.minBootNeuron({nid:"13243147462547516359", text:"ec:resolvable female (thing)", ntype:NTYPE_CLASS});    
        this.ec_resolvableSelf = this.minBootNeuron({nid:"12040890514035853644", text:"ec:resolvable self (thing)", ntype:NTYPE_CLASS});    
        this.ec_resolves_to_concept = this.minBootNeuron({nid:"7776775212574226990", text:"ec:resolves to concept", ntype:NTYPE_CLASS});    
        this.ec_resolves_to_role = this.minBootNeuron({nid:"7773087588979181787", text:"ec:resolves to role", ntype:NTYPE_CLASS});    
        

        this.ec_negation = this.minBootNeuron({nid:"7772189067884368306", text:"ec:negation", ntype:NTYPE_CLASS});    
        this.ec_plural = this.minBootNeuron({nid:"7775067489908426734", text:"ec:plural", ntype:NTYPE_CLASS});    
        this.ec_singular = this.minBootNeuron({nid:"7778687569286032056", text:"ec:singular", ntype:NTYPE_CLASS});    
        this.ec_direction = this.minBootNeuron({nid:"7775225830148199820", text:"ec:direction", ntype:NTYPE_CLASS}); 
        this.ec_manner = this.minBootNeuron({nid:"7778089761207879407", text:"ec:manner", ntype:NTYPE_CLASS}); 

        this.ec_indefiniteness = this.minBootNeuron({nid:"7772688541621829136", text:"ec:indefiniteness", ntype:NTYPE_CLASS}); 

        this.vb_form = this.minBootNeuron({nid:"5489825729469988645", text:"vb forms", ntype:NTYPE_CLASS}); 
        
        this.vb_v1 = this.minBootNeuron({nid:"7771743005401130790", text:"V1 - eng", ntype:NTYPE_CLASS}); 
        this.vb_v1_infinitive = this.minBootNeuron({nid:"7778016318966708690", text:"V1 - eng infinitive", ntype:NTYPE_CLASS});
        this.vb_v2 = this.minBootNeuron({nid:"7771644245115783595", text:"V2 - eng", ntype:NTYPE_CLASS}); 
        this.vb_v3 = this.minBootNeuron({nid:"7775543337600043609", text:"V3 - eng", ntype:NTYPE_CLASS}); 
        this.vb_v4 = this.minBootNeuron({nid:"7777759725587481842", text:"V4 - eng", ntype:NTYPE_CLASS}); 
        this.vb_v5 = this.minBootNeuron({nid:"7776191686416614280", text:"V5 - eng", ntype:NTYPE_CLASS}); 
        
        this.eng_vb_forms = [this.vb_v1.temp_nid, this.vb_v1_infinitive.temp_nid, this.vb_v2.temp_nid, this.vb_v3.temp_nid, this.vb_v4.temp_nid, this.vb_v5.temp_nid];

        //LC is the language circuitry area of the brain.
        this.LC_prefix = this.minBootNeuron({nid:"7773986200038360248", text:"ec:prefix", ntype:NTYPE_CLASS});    
        this.LC_suffix = this.minBootNeuron({nid:"7778838134500110231", text:"ec:suffix", ntype:NTYPE_CLASS});    
        
        //primary one-schema relon data dimensions.  (relon type)
        //TYPE
        this.relon_dimension = this.minBootNeuron({nid:"7770927391413238961", text:"relon dimension", ntype:NTYPE_CLASS});
        this.ASCII = this.minBootNeuron({nid:"7771796572391639487", text:"ASCII", ntype:NTYPE_CLASS});
        this.ascii_code_num = this.minBootNeuron({nid:"77759961069237519", text:"ASCII code #", ntype:NTYPE_CLASS});
        this.ascii_eval = this.minBootNeuron({nid:"7775975832455259480", text:"ASCII eval", ntype:NTYPE_CLASS});

        this.__SYS = this.minBootNeuron({nid:"7775032170648219182", text:"TYPE_SYS", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__SYS, "TYPE_SYS", TYPE_SYS);
        this.__BE = this.minBootNeuron({nid:"7777873907414567647", text:"TYPE_BE", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__BE, "TYPE_BE", TYPE_BE);
        this.__PARAM = this.minBootNeuron({nid:"7777094485082477142", text:"TYPE_PARAM", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__PARAM, "TYPE_PARAM", TYPE_PARAM);
        this.__HAS = this.minBootNeuron({nid:"7776798782476847291", text:"TYPE_HAS", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__HAS, "TYPE_HAS", TYPE_HAS);
        this.__AWARENESS = this.minBootNeuron({nid:"7779521186374438598", text:"TYPE_AWARENESS", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__AWARENESS, "TYPE_AWARENESS", TYPE_AWARENESS);
        this.__EXPERIENCE = this.minBootNeuron({nid:"7776169509680303554", text:"TYPE_EXPERIENCE", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__EXPERIENCE, "TYPE_EXPERIENCE", TYPE_EXPERIENCE);
        this.__CLUMP = this.minBootNeuron({nid:"7770332898798547236", text:"TYPE_CLUMP", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__CLUMP, "TYPE_CLUMP", TYPE_CLUMP);
        this.__SYS_MSG = this.minBootNeuron({nid:"7778077217855587615", text:"TYPE_BRAIN", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__SYS_MSG, "TYPE_SYS_MSG", TYPE_SYS_MSG);
        this.__BRAIN = this.minBootNeuron({nid:"7778077217855587615", text:"TYPE_BRAIN", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__BRAIN, "TYPE_BRAIN", TYPE_BRAIN);
        this.__ROLE = this.minBootNeuron({nid:"7779056189210800473", text:"TYPE_ROLE", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__BRAIN, "TYPE_ROLE", TYPE_ROLE);
        this.__EPISODIC = this.minBootNeuron({nid:"7770774009406800345", text:"TYPE_EPISODIC", ntype:NTYPE_CLASS});
        newSysTypeCode(this.__BRAIN, "TYPE_EPISODIC", TYPE_EPISODIC);

        //THIS IS CURRENTLY NOT BEING USED ANYMORE.
        this.seed = this.minBootNeuron({nid:"7778643893074759184", text:"seed", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.seed, "TO_TYPE_SEED", TO_TYPE_SEED, "", null, null);
        this.seed.bTERMINATING_TYPE = true;  
        this.seed.be_lvl = 4;
        
        this.anne_id = this.minBootNeuron({nid:"7777571073929239436", text:"anne_id", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.anne_id, "TO_TYPE_ANNE", TO_TYPE_ANNE, "", null, null);
        this.anne_id.bTERMINATING_TYPE = false;  
        this.anne_id.be_lvl = 4;
        
        this.e_anne_id = this.minBootNeuron({nid:"7777582090111925141", text:"encrypted_anne_id", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_anne_id, "TO_TYPE_ENCR_ANNE", TO_TYPE_ENCR_ANNE, "", null, null);
        this.e_anne_id.bTERMINATING_TYPE = false;     //we want normal HASSFULL handling.
        this.e_anne_id.be_lvl = 4;
        
        this.blob = this.minBootNeuron({nid:"7772915831532867191", text:"blob", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.blob, "TO_TYPE_BLOB", TO_TYPE_BLOB, "", null, null);
        this.blob.bTERMINATING_TYPE = true; 
        this.blob.be_lvl = 4;

        this.e_blob = this.minBootNeuron({nid:"777644219286194944", text:"encrypted_blob", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_blob, "TO_TYPE_ENCR_BLOB", TO_TYPE_ENCR_BLOB, "", null, null);
        this.e_blob.bTERMINATING_TYPE = true;
        this.e_blob.be_lvl = 4;

        this.date = this.minBootNeuron({nid:"7776322997340577329", text:"date", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.date, "TO_TYPE_DATE", TO_TYPE_DATE, "", null, null);
        this.date.bTERMINATING_TYPE = true;
        this.date.be_lvl = 4;

        this.e_date = this.minBootNeuron({nid:"7770256565764089519", text:"encrypted_date", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_date, "TO_TYPE_ENCR_DATE", TO_TYPE_ENCR_DATE, "", null, null);
        this.e_date.bTERMINATING_TYPE = true;
        this.e_date.be_lvl = 4;

        this.date_gregorian = this.minBootNeuron({nid:"777137291158471827", text:"gregorian date", ntype:NTYPE_CLASS});
        this.date_gregorian.prompt = "Enter date YYYY-DD-MM";
        this.date_gregorian.frenlyPrint=function($parts, $era="AD"){
            if ($parts.length!=3) return "";
            var year = parseInt($parts[0]);
            var month = MEMG.getNeuronNameByClassAndValue(ANNE.MEM.month, parseInt($parts[1]), true);
            
            var day = ANNE.TALK.lazyDayIntEnding(parseInt($parts[2]));
            var era = $era;
            return "<b>"+month+"</b>, <b>"+day+"</b> <b>"+year+"</b> <b>"+era+"</b>";
        }


        //???????  nasty hack, gunna have to get rid of this.
        //this.date_gregorian.formatFromResolve=function($obj){
            //return $obj[this.year.temp_nid+"_text"]+"-"+$obj[this.day.temp_nid+"_text"]+"-"+$obj[this.month.temp_nid+"_text"];
        //}

        this.gregorian_date = this.date_gregorian; //tired of making errors that cost me debug time.
        this.gregorian_date._uico = "🗓️";
        
        
        this.Alma_mater = this.minBootNeuron({nid:"777854422638973861", text:"Alma mater", ntype:NTYPE_CLASS});
        this.religion = this.minBootNeuron({nid:"777925711019642749", text:"religion", ntype:NTYPE_CLASS});
        this.religion._uico = "🙏";

        this.kinfolk = this.minBootNeuron({nid:"7779726195767515267", text:"kinfolk", ntype:NTYPE_CLASS});
        
        
        this.arts = this.minBootNeuron({nid:"1825382717134509033", text:"arts", ntype:NTYPE_CLASS});
        this.mathematics = this.minBootNeuron({nid:"12047036016253144920", text:"mathematics", ntype:NTYPE_CLASS});
        this.military = this.minBootNeuron({nid:"7770621763180151408", text:"miliary", ntype:NTYPE_CLASS});
        this.science_and_medicine = this.minBootNeuron({nid:"17841506035681030284", text:"science and medicine", ntype:NTYPE_CLASS});
        this.sport = this.minBootNeuron({nid:"18255368834152888231", text:"sport", ntype:NTYPE_CLASS});
        this.politics_and_law = this.minBootNeuron({nid:"6474702777226473172", text:"politics and law", ntype:NTYPE_CLASS});

        //wikispooks lists this as religion "jewish" but that is not the right label for religion.
        //we remap it to judiasm.
        this.judaism = this.minBootNeuron({nid:"12609260194840621674", text:"Judaism", ntype:NTYPE_CLASS});
        this.zionism = this.minBootNeuron({nid:"16686116773752269042", text:"Zionism", ntype:NTYPE_CLASS})
        
        //a tribal identity..
        this.jewish = this.minBootNeuron({nid:"17344203630892052405", text:"Jewish", ntype:NTYPE_CLASS});
        
        this.occupation = this.minBootNeuron({nid:"7774593245582491319", text:"occupation", ntype:NTYPE_CLASS});
        this.occupation._uico = "🔨";
        
        this.position = this.minBootNeuron({nid:"7772884086015787050", text:"position", ntype:NTYPE_CLASS}); 
        this.position._uico = "🔨";
        
        this.career = this.minBootNeuron({nid:"10179302147304506923", text:"career", ntype:NTYPE_CLASS}); 

        this.service_term = this.minBootNeuron({nid:"7778128204103734929", text:"service term", ntype:NTYPE_CLASS});
        this.service_term._uico = "🔨";
        
        //used with company type.
        this.higher_education = this.minBootNeuron({nid:"77712777044716550", text:"Higher Education", ntype:NTYPE_CLASS});
        
        this.nationality = this.minBootNeuron({nid:"7772786288596504983", text:"nationality", ntype:NTYPE_CLASS});
        this.political_party = this.minBootNeuron({nid:"7771124117310918099", text:"political party", ntype:NTYPE_CLASS});
        this.tribal = this.minBootNeuron({nid:"7770927790880535113", text:"tribal identity", ntype:NTYPE_CLASS});
        this.citizenship = this.minBootNeuron({nid:"7773415791436217714", text:"citizenship", ntype:NTYPE_CLASS});
        
        //things much more loosely tagged -- ie, from wikipedia last name parse and 
        //tags that are NOT languages but still relevant.  (see "ashkenazi" for epstein)
        //we do not have an alt be parent/hierarchy setup for this yet as we do not want/need it.
        //lateral connections of TYPE_BE is -> ethnicity suffice.  Classification via observation of use.
        this.ethnicity = this.minBootNeuron({nid:"17140987543979601380", text:"ethnicity", ntype:NTYPE_CLASS});

        this.representative = this.minBootNeuron({nid:"6476312658984199463", text:"representative", ntype:NTYPE_CLASS});
        this.governor =this.minBootNeuron({nid:"3567666822242597329", text:"Governor", ntype:NTYPE_CLASS});
        this.lgovernor =this.minBootNeuron({nid:"2050992576214445805", text:"Lieutenant Governor", ntype:NTYPE_CLASS});
        this.president = this.minBootNeuron({nid:"9793621070511313956", text:"President", ntype:NTYPE_CLASS});
        this.mayor =this.minBootNeuron({nid:"13749506479367882199", text:"Mayor", ntype:NTYPE_CLASS});
        this.judge = this.minBootNeuron({nid:"11286563740955259900", text:"Judge", ntype:NTYPE_CLASS});
        //this.cjudge = this.minBootNeuron({nid:"", text:"Circuit Judge", ntype:NTYPE_CLASS});
        
        this.ceo = this.minBootNeuron({nid:"842151752508316689", text:"CEO", ntype:NTYPE_CLASS});
        this.chairman = this.minBootNeuron({nid:"14221162201076302843", text:"chairman", ntype:NTYPE_CLASS});
        
        this.WEF = this.minBootNeuron({nid:"7776951561979368390", text:"World Economic Forum", ntype:NTYPE_INST});
        
        //used to classify board of director and board of trustee neurons.
        //they are "company" neurons that tie back to the corporation or association they are connected to.
        //they themselves have a comany_type of governance board.
        this.governance_board = this.minBootNeuron({nid:"7779663457772195626", text:"governance board", ntype:NTYPE_CLASS});
        this.board_member  = this.minBootNeuron({nid:"7773721544735553188", text:"board member", ntype:NTYPE_CLASS});

        //will have to fine tune these spellings as we dupe them from wikispooks.
        this.us_republican = this.minBootNeuron({nid:"7773678707765331234", text:"US Republican", ntype:NTYPE_CLASS});
        this.us_democrat = this.minBootNeuron({nid:"7772420605042291734", text:"US Democat", ntype:NTYPE_CLASS});
        this.independent = this.minBootNeuron({nid:"7773736162587243688", text:"Independent", ntype:NTYPE_CLASS});
        this.communist = this.minBootNeuron({nid:"7779857133825607397", text:"Communist", ntype:NTYPE_CLASS});
        this.socialist = this.minBootNeuron({nid:"777564380965650179", text:"Socialist", ntype:NTYPE_CLASS});
        this.cpoc = this.minBootNeuron({nid:"7770258420272538620", text:"Communist Party of China", ntype:NTYPE_CLASS});

        //relations
        this.kinfolk_parent = this.minBootNeuron({nid:"7775760258501601282", text:"parent", ntype:NTYPE_CLASS});
        this.kinfolk_child = this.minBootNeuron({nid:"777299714238049539", text:"child", ntype:NTYPE_CLASS});
        this.spouse = this.minBootNeuron({nid:"7778826002852320790", text:"spouse", ntype:NTYPE_CLASS});

        //standout relationship that can be calculated but stands out if relevant to highlight.
        this.kinfolk_cousin = this.minBootNeuron({nid:"1548501709675585821", text:"kinfolk cousin", ntype:NTYPE_CLASS});
        
        this.kinfolk_grandparent = this.minBootNeuron({nid:"150810937878958292", text:"kinfolk grandparent", ntype:NTYPE_CLASS});
        this.grandfather = this.minBootNeuron({nid:"5668674581344300968", text:"grandfather", ntype:NTYPE_CLASS});
        this.grandmother = this.minBootNeuron({nid:"9317028650101357972", text:"grandmother", ntype:NTYPE_CLASS});
        
        //action used in marriage clumps.
        this.marry = this.minBootNeuron({nid:"6385755796053286049", text:"marry", ntype:NTYPE_CLASS});
        this.marriage = this.minBootNeuron({nid:"7772714390270034050", text:"marriage", ntype:NTYPE_CLASS});
        

        this.father = this.minBootNeuron({nid:"7770334697346006910", text:"father", ntype:NTYPE_CLASS});
        this.bio_father = this.minBootNeuron({nid:"7774430960064578792", text:"bilogical father", ntype:NTYPE_CLASS});
        this.adoptive_father = this.minBootNeuron({nid:"777421606944026826", text:"adoptive father", ntype:NTYPE_CLASS});
        this.father_in_law = this.minBootNeuron({nid:"7778846417673009620", text:"father in-lawr", ntype:NTYPE_CLASS});
        this.son = this.minBootNeuron({nid:"7774803688522754022", text:"son", ntype:NTYPE_CLASS});
        this.bio_son = this.minBootNeuron({nid:"7775737496175899538", text:"bio son", ntype:NTYPE_CLASS});
        this.adopted_son = this.minBootNeuron({nid:"7777944849660781543", text:"adopted son", ntype:NTYPE_CLASS});
        this.son_in_law = this.minBootNeuron({nid:"7772106931969635469", text:"son in-law", ntype:NTYPE_CLASS});
        this.daughter = this.minBootNeuron({nid:"7772451463343244417", text:"daughter", ntype:NTYPE_CLASS});
        this.bio_daughter = this.minBootNeuron({nid:"7777495899734158074", text:"bio daughter", ntype:NTYPE_CLASS});
        this.adopted_daughter = this.minBootNeuron({nid:"7778259908422705155", text:"adopted daughter", ntype:NTYPE_CLASS});
        this.daughter_in_law = this.minBootNeuron({nid:"7778965895626708730", text:"daughter in-law", ntype:NTYPE_CLASS});
        this.mother = this.minBootNeuron({nid:"7778367143803659148", text:"mother", ntype:NTYPE_CLASS});
        
        //generic sibbling
        this.kinfolk_sibbling = this.minBootNeuron({nid:"7775340410780712793", text:"sibbling", ntype:NTYPE_CLASS});

        this.brother = this.minBootNeuron({nid:"11056317496902226178", text:"brother", ntype:NTYPE_CLASS});
        this.sister = this.minBootNeuron({nid:"11390471507506307054", text:"sister", ntype:NTYPE_CLASS});

        this.kinfolk_tnids = [
            this.kinfolk_cousin.temp_nid, this.mother.temp_nid,
            this.kinfolk_sibbling.temp_nid, this.kinfolk_parent.temp_nid, this.kinfolk_child.temp_nid, this.spouse.temp_nid, this.father.temp_nid,
            this.bio_father.temp_nid, this.adoptive_father.temp_nid, this.father_in_law.temp_nid,  this.son.temp_nid,
            this.bio_son.temp_nid, this.adopted_son.temp_nid, this.son_in_law.temp_nid,  this.daughter.temp_nid, 
            this.bio_daughter.temp_nid, this.adopted_daughter.temp_nid, this.daughter_in_law.temp_nid
        ];

        this.comp_union = this.minBootNeuron({nid:"7771287958516484511", text:"comp union", ntype:NTYPE_CLASS});

        //0-day 10800BC, every day after is +1.
        //any real gregorian date has an ANNE EPOCH DAY.
        //NOTE: this does not take in to account calendar changes (since gregorian calendar didn't exist in 10800BC!)
        //This is a way to quickly (numerically) compare dates in the annode while searching relons.
        this.ANNE_EPOCH_DAY = this.minBootNeuron({nid:"7778857555786131806", text:"ANNE EPOCH DAY", ntype:NTYPE_CLASS});

        this.month = this.minBootNeuron({nid:"777842179295607978", text:"month", ntype:NTYPE_CLASS});
        this.year = this.minBootNeuron({nid:"7774796274147265857", text:"year", ntype:NTYPE_CLASS});
        this.day = this.minBootNeuron({nid:"7775054193871499071", text:"day", ntype:NTYPE_CLASS});
        this.daymonth = this.minBootNeuron({nid:"7772711710926955713", text:"daymonth", ntype:NTYPE_CLASS});
        this.monthyear = this.minBootNeuron({nid:"7778824928802286657", text:"monthyear", ntype:NTYPE_CLASS});
        
        this.date_any = [MEM.date_gregorian.temp_nid, MEM.daymonth.temp_nid, MEM.monthyear.temp_nid];
        this.date_any_nid = [MEM.date_gregorian.nid, MEM.daymonth.nid, MEM.monthyear.nid];
        this.isDateAny=function($nid) {
            if (this.date_any_nid.indexOf($nid)!=-1) return true;
            return false;
        }
        this.era = this.minBootNeuron({nid:"7775223774408988417", text:"era", ntype:NTYPE_CLASS});
        this.era_bc =  this.minBootNeuron({nid:"7773055704590862597", text:"BC", ntype:NTYPE_CLASS});
        this.era_ad = this.minBootNeuron({nid:"7772300057442531736", text:"AD", ntype:NTYPE_CLASS});

        this.doty = this.minBootNeuron({nid:"7776382262043367717", text:"day of the year", ntype:NTYPE_CLASS});
        this.dotw = this.minBootNeuron({nid:"7773395836100627599", text:"day of the week", ntype:NTYPE_CLASS});


        this.month_jan =  this.minBootNeuron({nid:"7779972248415370744", text:"Januar", ntype:NTYPE_CLASS});
        this.month_jan.value=1;
        this.month_feb =  this.minBootNeuron({nid:"7772796602534780742", text:"February", ntype:NTYPE_CLASS});
        this.month_feb.value=2;
        this.month_mar =  this.minBootNeuron({nid:"7771522067996789646", text:"March", ntype:NTYPE_CLASS});
        this.month_mar.value=3;
        this.month_apr =  this.minBootNeuron({nid:"7774465631312901598", text:"April", ntype:NTYPE_CLASS});
        this.month_apr.value=4;
        this.month_may =  this.minBootNeuron({nid:"7770857201478628723", text:"May", ntype:NTYPE_CLASS});
        this.month_may.value=5;
        this.month_jun =  this.minBootNeuron({nid:"7774339894223286536", text:"June", ntype:NTYPE_CLASS});
        this.month_jun.value=6;
        this.month_jul =  this.minBootNeuron({nid:"7774838671233944689", text:"July", ntype:NTYPE_CLASS});
        this.month_jul.value=7;
        this.month_aug =  this.minBootNeuron({nid:"7779263094123527469", text:"August", ntype:NTYPE_CLASS});
        this.month_aug.value=8;
        this.month_sep =  this.minBootNeuron({nid:"7777384954129664128", text:"September", ntype:NTYPE_CLASS});
        this.month_sep.value=9;
        this.month_oct =  this.minBootNeuron({nid:"7777395188527845949", text:"October", ntype:NTYPE_CLASS});
        this.month_oct.value=10;
        this.month_nov =  this.minBootNeuron({nid:"7775446664476737015", text:"November", ntype:NTYPE_CLASS});
        this.month_nov.value=11;
        this.month_dec =  this.minBootNeuron({nid:"7770036874097137801", text:"December", ntype:NTYPE_CLASS});
        this.month_dec.value=12;
        
        this.gps_lat = this.minBootNeuron({nid:"7773164283122952368", text:"lat", ntype:NTYPE_CLASS});
        this.gps_long = this.minBootNeuron({nid:"7776410985703118068", text:"long", ntype:NTYPE_CLASS});


        /**
         * DoDateFormat event handler.
         * Set any date text fields to use this for their keyup function. 
         * @author {Mr.Scatman}
         */
        this.doDateFormat=function($e) { 
            if (!$e.ctrlKey && !$e.metaKey && ($e.keyCode == 32 || $e.keyCode > 46)) {
                ANNE.MEM.date.formatter($e.target);
            }
        }

        /**
         * this will not work with BC.  Too bad, for now.
         * @param {*} $el 
         */
        this.date.formatter = function($el) {
            var format = "yyyy-mm-dd";
            var match = new RegExp(format
                .replace(/(\w+)\W(\w+)\W(\w+)/, "^\\s*($1)\\W*($2)?\\W*($3)?([0-9]*).*")
                .replace(/m|d|y/g, "\\d"));
            var replace = "$1-$2-$3$4"
                .replace(/\//g, format.match(/\W/));

            $el.value = $el.value.replace(/(^|\W)(?=\d\W)/g, "$10")   // padding
                    .replace(match, replace)             // fields
                    .replace(/(\W)+/g, "$1");            // remove repeats

            var parts = $el.value.split("-");
            if (parts[0].length>4) parts[0].substring(0, 3);
            if (parseInt( parts[1] ) > 12) parts[1] = 12;
            if (parseInt( parts[2] ) > 31) parts[2] = 31;
            $el.value = parts.join("-");
        }
        this.date.be_lvl = 4;
        
        //
        this.start_date = this.minBootNeuron({nid:"7774876474320373351", text:"start date", ntype:NTYPE_CLASS});
        this.end_date = this.minBootNeuron({nid:"7773588826071954087", text:"end date", ntype:NTYPE_CLASS});
        this.launch_date = this.minBootNeuron({nid:"1293921080709290782", text:"launch date", ntype:NTYPE_CLASS});
        
        this.dissolved_on = this.minBootNeuron({nid:"11098414578706519923", text:"dissolved on", ntype:NTYPE_CLASS});

        this.datetime = this.minBootNeuron({nid:"7770829136782284019", text:"datetime", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.datetime, "TO_TYPE_DATETIME", TO_TYPE_DATETIME, "", null, null);
        this.datetime.bTERMINATING_TYPE = true;
        this.datetime.be_lvl = 4;

        this.e_datetime = this.minBootNeuron({nid:"7778516895066405156", text:"encrypted_datetime", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_datetime, "TO_TYPE_ENCR_DATETIME", TO_TYPE_ENCR_DATETIME, "", null, null);
        this.e_datetime.bTERMINATING_TYPE = true;
        this.e_datetime.be_lvl = 4;

        this.number = this.minBootNeuron({nid:"777466536043042154", text:"number", ntype:NTYPE_CLASS});
        this.number.be_lvl = 4;
        newSysToTypeCode(this.number, "TO_TYPE_NUMBER", TO_TYPE_NUMBER,  "", null, null)
        
        
        this.number.set_value = this.setNumber;
        
        this.number.formatter = function ($v) {
            var vtype = typeof $v; 
            if (vtype ==='object') {
                this.number.set_value($v);
                return $v.value.toString().replace(/\B(?=(\d{3})+(?!\d))/g, ",");
            }
            else if (vtype==='number') {
                return $v.toString().replace(/\B(?=(\d{3})+(?!\.\d))/g, ",");
            }
            else return $v.toString().replace(/\B(?=(\d{3})+(?!\.\d))/g, ",");
        }

        this.e_number = this.minBootNeuron({nid:"7770791628311366616", text:"encrypted_number", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_number, "TO_TYPE_ENCR_NUMBER", TO_TYPE_ENCR_NUMBER, "", null, null);
        this.e_number.bTERMINATING_TYPE = true;
        this.e_number.be_lvl = 4;

        this.integer = this.minBootNeuron({nid:"7775505125310503749", text:"integer", ntype:NTYPE_CLASS});
        this.integer.set_value = function($n) { $n.setValue(parseInt($n.text)); }  //overrides the set_value function of number (parseFloat) for integers.
        this.integer.be_lvl = 4;

        this.n0 = this.minBootNeuron({nid:"7772759491040052489", text:"0", ntype:NTYPE_CLASS});
        this.n0.be_lvl = 5;
        this.n1 = this.minBootNeuron({nid:"7779936209861264085", text:"1", ntype:NTYPE_CLASS});
        this.n1.be_lvl = 5;
        this.n2 = this.minBootNeuron({nid:"7777466399639974815", text:"2", ntype:NTYPE_CLASS});
        this.n2.be_lvl = 5;
        this.n3 = this.minBootNeuron({nid:"7770513775335783280", text:"3", ntype:NTYPE_CLASS});
        this.n3.be_lvl = 5;
        this.n4 = this.minBootNeuron({nid:"7779613874588391142", text:"4", ntype:NTYPE_CLASS});
        this.n4.be_lvl = 5;
        this.n5 = this.minBootNeuron({nid:"7773838259673220295", text:"5", ntype:NTYPE_CLASS});
        this.n5.be_lvl = 5;
        this.n6 = this.minBootNeuron({nid:"7772956192305001444", text:"6", ntype:NTYPE_CLASS});
        this.n6.be_lvl = 5;
        this.n7 = this.minBootNeuron({nid:"7779032678869944184", text:"7", ntype:NTYPE_CLASS});
        this.n7.be_lvl = 5;
        this.n8 = this.minBootNeuron({nid:"7777087244990101066", text:"8", ntype:NTYPE_CLASS});
        this.n8.be_lvl = 5;
        this.n9 = this.minBootNeuron({nid:"7778263312619963662", text:"9", ntype:NTYPE_CLASS});
        this.n9.be_lvl = 5;
        this.n10 = this.minBootNeuron({nid:"777749396793604897", text:"10", ntype:NTYPE_CLASS});
        this.n10.be_lvl = 5;
        this.n11 = this.minBootNeuron({nid:"7774857001185551332", text:"11", ntype:NTYPE_CLASS});
        this.n11.be_lvl = 5;
        this.n12 = this.minBootNeuron({nid:"7777727428900691983", text:"12", ntype:NTYPE_CLASS});
        this.n12.be_lvl = 5;

        //referenced for direct "values" of neurons set with the nObj->eVal->xx pattern,
        //where xx is the value.  
        //example.   dozen ->eval ->12
        this.eval = this.minBootNeuron({nid:"7776011636867438465", text:"eval", ntype:NTYPE_CLASS});
        this.eval.set_value=function($v) {
            if (typeof $v !=="object") {
                console.warn("must use neuron to set_value");
                return;
            }
            if ($v["be_"+ANNE.MEM.eval.temp_nid]!=undefined) {   //ez road
                $v.setValue($v["be_"+ANNE.MEM.eval.temp_nid]);
                return $v.value;
            }

            //longer road..
            var evalRelon = MEMG.getRelonByTempR($v.outs, ANNE.MEM.eval.temp_nid);
            if (evalRelon!=null && evalRelon.type==TYPE_BE) {
                if (evalRelon.to_type==TO_TYPE_ANNE) {  //have to get val of this neuron!s
                    var valueNOBJ = ANNE.MEM.getNeuronByTempNid(polarityRelon.t);
                    if (valueNOBJ!=null) $v.polarity=ANNE.MEM.__MASTER_GET_VALUE(valueNOBJ);
                }
                else {
                    $v.setValue(evalRelon.to_s);
                    if (isNaN($v.value)==true) {
                        $v.value=null;
                    }
                    return $v.value;
                }
            }
            return null;
        }

        //Must override the default set_value for parent number class.
        //it converts text to number.
        //proportion has text as a fraction and the value can be eval'd from the text
        //but more accurately (and true) is the eval relon pointing to the decimal value.
        this.proportion = this.minBootNeuron({nid:"7779861280974626916", text:"proportion", ntype:NTYPE_CLASS});
        this.proportion.set_value = function($v) { 
            this.eval.set_value($v);
        }

        this.time = this.minBootNeuron({nid:"7771199207074827487", text:"time", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.time, "TO_TYPE_TIME", TO_TYPE_TIME, "", null, null);
        this.time.bTERMINATING_TYPE = true;
        this.time.be_lvl = 4;

        this.time.formatter = function($el) {
            if ($el==null) return;
            var val = $el.value;
            //20:17:40 GMT+00:00
        }
        this.doTimeFormat=function($e) { 
            if (!$e.ctrlKey && !$e.metaKey && ($e.keyCode == 32 || $e.keyCode > 46)) {
                ANNE.MEM.formatter($e.target);
            }
        }

        this.e_time = this.minBootNeuron({nid:"7776682659772901989", text:"encrypted_time", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_time, "TO_TYPE_ENCR_TIME", TO_TYPE_ENCR_TIME, "", null, null);
        this.e_time.bTERMINATING_TYPE = true;
        this.e_time.be_lvl = 4;

        this.txid = this.minBootNeuron({nid:"7774781823726482570", text:"txid", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.txid, "TO_TYPE_TXID", TO_TYPE_TXID, "",  null, null);
        this.txid.set_value = function ($v) { 
            $v.setValue(MEMG.lookupRelonByTXID($v.text));
        }
        this.txid.be_lvl = 4;

        this.e_txid = this.minBootNeuron({nid:"7774527078408052235", text:"encrypted_txid", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_txid, "TO_TYPE_ENCR_TXID", TO_TYPE_ENCR_TXID, "", null, null);
        this.e_txid.bTERMINATING_TYPE = true;
        this.e_txid.be_lvl = 4;

        this.url = this.minBootNeuron({nid:"7771426118341686899", text:"url", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.url, "TO_TYPE_URL", TO_TYPE_URL, "", null, null);
        this.url.bTERMINATING_TYPE = true;
        this.url.format_validator=ANNE.UTILS.__urlValidation;
        this.url.be_lvl = 4;

        this.e_url = this.minBootNeuron({nid:"7772967135250758561", text:"encrypted_url", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_url, "TO_TYPE_ENCR_URL", TO_TYPE_ENCR_URL, "", null, null);
        this.e_url.bTERMINATING_TYPE = true;
        this.e_url.be_lvl = 4;

        this.boolean = this.minBootNeuron({nid:"7777601580469353964", text:"boolean", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.boolean, "TO_TYPE_BOOLEAN", TO_TYPE_BOOLEAN, "",  this.__setValByBoolPolarity);
        this.boolean.bTERMINATING_TYPE = true;
        this.boolean.be_lvl = 4;

        this.boolean.set_value = function($n) { 
            var rObj = MEMG.getRelonByTempR($n.outs, ANNE.MEM.eval.temp_nid);
            if (rObj!=null) $n.setValue(parseInt(rObj.to_s)); 
        }
        
        this.e_boolean = this.minBootNeuron({nid:"7773554013708022388", text:"encrypted_boolean", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_boolean, "TO_TYPE_ENCR_BOOLEAN", TO_TYPE_ENCR_BOOLEAN);
        this.e_boolean.bTERMINATING_TYPE = true;
        this.e_boolean.be_lvl = 4;

        this.yesno= this.minBootNeuron({nid:"7775762158326148225", text:"yes/no", ntype:NTYPE_CLASS});
        this.yes= this.minBootNeuron({nid:"7773442870220826516", text:"yes", ntype:NTYPE_CLASS});
        this.yes.value=1;
        this.no = this.minBootNeuron({nid:"7775849543552546059", text:"no", ntype:NTYPE_CLASS});
        this.no.value=0;

        this.string = this.minBootNeuron({nid:"7777506478065484896", text:"string", ntype:NTYPE_CLASS});
        this.string.set_value = function($v) { 
            $v.setValue($v.text.trim()); 
        }
        newSysToTypeCode(this.string, "TO_TYPE_STRING", TO_TYPE_STRING, "", null);
        this.string.bTERMINATING_TYPE = true;
        this.string.be_lvl = 4;

        //will point to base of url so we can craft dynamic wikilinks w/o storing the entire string again and again.
        //this.wikispookpart = this.minBootNeuron({nid:"7777667326299864145", text:"wikispook part", ntype:NTYPE_CLASS});
        
        this.e_string = this.minBootNeuron({nid:"777615879896782812", text:"encrypted_string", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_string, "TO_TYPE_ENCR_STRING", TO_TYPE_ENCR_STRING);
        this.e_string.bTERMINATING_TYPE = true;
        this.e_string.be_lvl = 4;

        this.filehash = this.minBootNeuron({nid:"7771966589026475642", text:"filehash", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.filehash, "TO_TYPE_FILEHASH", TO_TYPE_FILEHASH, "", null);
        this.filehash.bTERMINATING_TYPE = true;
        this.filehash.be_lvl = 4;

        this.e_filehash = this.minBootNeuron({nid:"7773418046787332221", text:"encrypted_filehash", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_filehash, "TO_TYPE_ENCR_FILEHASH", TO_TYPE_ENCR_FILEHASH, "", null);
        this.e_filehash.bTERMINATING_TYPE = true;
        this.e_filehash.be_lvl = 4;

        this.ipv4 = this.minBootNeuron({nid:"7770686966970402490", text:"ipv4", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.ipv4, "TO_TYPE_IPV4", TO_TYPE_IPV4, "", null);
        this.ipv4.bTERMINATING_TYPE = true;
        this.ipv4.format_validator=ANNE.UTILS.__ipv4Validation;
        this.ipv4.be_lvl = 4;

        this.e_ipv4 = this.minBootNeuron({nid:"7776380890278424141", text:"encrypted_ipv4", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_filehash, "TO_TYPE_ENCR_IPV4", TO_TYPE_ENCR_IPV4, "", null);
        this.e_ipv4.bTERMINATING_TYPE = true;
        this.e_ipv4.be_lvl = 4;

        this.ipv6 = this.minBootNeuron({nid:"7774737336689250477", text:"ipv6", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.ipv4, "TO_TYPE_IPV6", TO_TYPE_IPV6, "", null);
        this.ipv6.bTERMINATING_TYPE = true;
        this.ipv6.format_validator=ANNE.UTILS.__ipv4Validation;
        this.ipv6.be_lvl = 4;

        this.e_ipv6 = this.minBootNeuron({nid:"7779915290430504", text:"encrypted_ipv4", ntype:NTYPE_CLASS});
        newSysToTypeCode(this.e_filehash, "TO_TYPE_ENCR_IPV6", TO_TYPE_ENCR_IPV6, "", null);
        this.e_ipv6.bTERMINATING_TYPE = true;
        this.e_ipv6.be_lvl = 4;


        this.sequence = this.minBootNeuron({nid:"7773197197458854483", text:"sequence", ntype:NTYPE_CLASS});
        this.sequence.set_value = function($v) { 
            $v.outs.sort(ANNE.MEM.__sortSeqRelons); 
        }

        this.factor = this.minBootNeuron({nid:"7778025035144069284", text:"factor", ntype:NTYPE_CLASS});
        this.factor.set_value = function($uomNOBJ) {    //the factor class has a set_value function which will correctly set the value of its instances.
            if ($uomNOBJ==null) return;
            
            //if we have done this, don't do it again.
            if ($uomNOBJ.uom!=undefined) return $uomNOBJ.uom;

            var cRelon = MEMG.getRelonByTempR($uomNOBJ.outs, ANNE.MEM.conversion_unit.temp_nid);
            
            //this is just a factor (like a prefix to alter a value without a conversion
            //unit or conversion relon.. it will have a factor relon.
            if (cRelon==null) {
                var uom = {base_unit:$uomNOBJ};

                var unit_lbl_rel = MEMG.getRelonByTempRBeLine($uomNOBJ.outs, [TYPE_HAS], ANNE.MEM.unit_symbol.temp_nid, [ANNE.MEM.html_symbol.temp_nid]);
                if (unit_lbl_rel!=null) {
                    uom.unit_lbl = unit_lbl_rel.to_s;
                }
                else uom.unit_lbl="";
                
                var factor_rel = MEMG.getRelonByTempR($uomNOBJ.outs, ANNE.MEM.factor.temp_nid);
                if (factor_rel==null) return null;
                var parts = factor_rel.to_s.split('^');

                if (parts.length!=2) {
                    uom.factored = uom.factor_rel.to_s;
                    $uomNOBJ.factored = uom.factor_rel.to_s;
                    $uomNOBJ.uom = uom;
                    return uom;
                }
                var base = parseFloat(parts[0]);
                var exponent = parseFloat(parts[1]);
                $uomNOBJ.factored = Math.pow(base, exponent);
                uom.factored = $uomNOBJ.factored;
                $uomNOBJ.uom = uom;
                return uom;
            }
            
            var cUnitNOBJ = MEMG.getNeuronByTempNid(cRelon.t);
            if (cUnitNOBJ==null) return;
            
            var uom = {unit:cUnitNOBJ, base_unit:$uomNOBJ};
    
            var conversionRel = MEMG.getRelonByTempR($uomNOBJ.outs, ANNE.MEM.conversion.temp_nid);
            if (conversionRel==null) return;

            uom.conversion = conversionRel.to_s;

            //this should get "any" child of symbol..  so symbol or also something like si_unit_symbol.. etc
            var unit_lbl_rel = MEMG.getRelonByTempRBeLine(cUnitNOBJ.outs, [TYPE_HAS], ANNE.MEM.unit_symbol.temp_nid, [MEM.html_symbol.temp_nid]);
            if (unit_lbl_rel!=null) {
                uom.unit_lbl = unit_lbl_rel.to_s;
            }
            else uom.unit_lbl="";  

            var parts = uom.conversion.split('^');
            if (parts.length!=2) {
                uom.factored = uom.conversion;
                $uomNOBJ.factored = $uomNOBJ.conversion;
                $uomNOBJ.uom = uom;
                //$uomNOBJ.value = uom.factored;
                return uom;
            }
            var base = parseFloat(parts[0]);
            var exponent = parseFloat(parts[1]);
            $uomNOBJ.factored = Math.pow(base, exponent);
            uom.factored = $uomNOBJ.factored;
            $uomNOBJ.uom = uom;

            var symbolRel = MEMG.getRelonByTempR(cUnitNOBJ.outs, ANNE.MEM.html_symbol.temp_nid);
            if (symbolRel!=null) {
                uom.symbolHTML = symbolRel.to_s;
            }

            return uom;
        }

        this.param_list = this.minBootNeuron({nid:"7773017254938949678", text:"param list", ntype:NTYPE_CLASS});
        this.param_list.set_value = function ($v) {
            console.warn("THIS HAS NOT BEEN TESTED");
            $v.value = this.buildArrOfValueKeyPairs(ANNE.MEM.__PARAM, [TYPE_HAS]);//ACQL.execute_circuit($v, ACQL._RCPgenTextListOfAllRelnsOf_params);
        }

        this.name = this.minBootNeuron({nid:"7770034521431820596", text:"name", ntype:NTYPE_CLASS});
        this.name.prompt = "What is your name?";
        
        //extra info on names (not required for persons, but nice to have for L1 grpah linkage)
        this.firstName = this.minBootNeuron({nid:"7773598701821046892", text:"first name", ntype:NTYPE_CLASS});
        this.lastName = this.minBootNeuron({nid:"7770897545226351159", text:"last name", ntype:NTYPE_CLASS});
        this.middleName = this.minBootNeuron({nid:"7770526629291509703", text:"middle name", ntype:NTYPE_CLASS});
        this.name_variant = this.minBootNeuron({nid:"16280936195756863604", text:"name variant", ntype:NTYPE_CLASS});
        
        this.polarity = this.minBootNeuron({nid:"7778811780002344235", text:"polarity", ntype:NTYPE_CLASS});
        this.polarity.set_value=function($v) {
            //fast version.
            if ($v["be_"+ANNE.MEM.polarity.temp_nid]!=undefined) {
                var pol = MEMG.getNeuronByTempNid($v["be_"+ANNE.MEM.polarity.temp_nid]);
                if (pol) $v.polarity=pol.polarity;
                return $v.polarity;
            }
            //longer road if no index was set.
            var polarityRelon = MEMG.getRelonByTempR($v.outs, ANNE.MEM.polarity.temp_nid);
            if (polarityRelon!=null && polarityRelon.type==TYPE_BE) {
                var polarity = ANNE.MEM.getNeuronByTempNid(polarityRelon.t);
                if (polarity!=null) {
                    $v.polarity=polarity.polarity;
                    return $v.polarity;
                }
            }
            return null;
        }

        this.positive = this.minBootNeuron({nid:"7771641909001981874", text:"positive", ntype:NTYPE_CLASS});
        this.positive.polarity=1;
        this.positive.pol="pos";
        this.negative = this.minBootNeuron({nid:"777797306450052899", text:"negative", ntype:NTYPE_CLASS});
        this.negative.polarity=-1;
        this.negative.pol="neg";
        this.neutral = this.minBootNeuron({nid:"7772223634951560391", text:"neutral", ntype:NTYPE_CLASS});
        this.neutral.polarity=0;
        this.neutral.pol="neu";
        
        this.sponsored_by = this.minBootNeuron({nid:"7777123609427148390", text:"sponsor", ntype:NTYPE_CLASS});
        this.owned_by = this.minBootNeuron({nid:"7770602908121069199", text:"owner", ntype:NTYPE_CLASS});
        
        //this.annother = this.minBootNeuron({nid:"7779834425608376119", text:"annother", ntype:NTYPE_CLASS});
        this.necrosis = this.minBootNeuron({nid:"7773636908287156942", text:"necrosis", ntype:NTYPE_CLASS});
        this.arbor_prune = this.minBootNeuron({nid:"777918832015451554", text:"arbor prune", ntype:NTYPE_CLASS});
        this.inhibit_relon = this.minBootNeuron({nid:"777581383667732382", text:"inhibit relon", ntype:NTYPE_CLASS});
        this.code = this.minBootNeuron({nid:"7770043314931693691", text:"code", ntype:NTYPE_CLASS});
        this.primary_key = this.minBootNeuron({nid:"7770289090923457729", text:"primary key", ntype:NTYPE_CLASS});
        this.primary_key2 = this.minBootNeuron({nid:"7772038105301798922", text:"primary key2", ntype:NTYPE_CLASS});
        this.aka = this.minBootNeuron({nid:"7777180587425253122", text:"aka", ntype:NTYPE_CLASS});
        this.fka = this.minBootNeuron({nid:"7775797982596188653", text:"formerly known as", ntype:NTYPE_CLASS});
         
        this.PARAM_NAME = this.minBootNeuron({nid:"7772502840575856680", text:"PARAM NAME", ntype:NTYPE_CLASS});
        this.REL_PARAM_NAME = this.minBootNeuron({nid:"7774837507134294874", text:"RELN PARAM NAME", ntype:NTYPE_CLASS});
        
        this.ANNE_DDS_SRC = this.minBootNeuron({nid:"7778133976189354085", text:"ANNE DDS SRC", ntype:NTYPE_CLASS});
        this.require = this.minBootNeuron({nid:"7774954569060780571", text:"require", ntype:NTYPE_CLASS});
        this.validation = this.minBootNeuron({nid:"7778126176197655430", text:"validation", ntype:NTYPE_CLASS});
        this._multiples_ok = this.minBootNeuron({nid:"7777251338142322352", text:"multiples ok", ntype:NTYPE_CLASS});
        //praam name for this is "_multiples_ok" and the gui automatically sets this prop for us on the param relon
        //which the TYPE_SYS relon with reln of _multiples_ok points to.  :)

        this.BE_CHILDREN =  this.minBootNeuron({nid:"7775936058919519512", text:"BE_CHILDREN", ntype:NTYPE_CLASS});
        this.BE_PARENT =  this.minBootNeuron({nid:"7775942183860735787", text:"BE_PARENT", ntype:NTYPE_CLASS});
        

        //role mappers to extract facts from clumps via conveyors.
        this.DEFINES =  this.minBootNeuron({nid:"7778557818829959373", text:"defines", ntype:NTYPE_CLASS});
        this.DEFINED_BY =  this.minBootNeuron({nid:"7772056616837510132", text:"defined by", ntype:NTYPE_CLASS});
        
        this.language = this.minBootNeuron({nid:"7777662227658988771", text:"language", ntype:NTYPE_CLASS});
        this.language.be_lvl = 4;
     
        //build a tester based on unicode_start and unicode_end values.
        this.language.ascertainLang=function() {
            var tests = ["Киану Ривз", "키아누 리브스", "キアヌ・リーブス", "เคอานู รีฟส์",
            "基努·里维斯", "קיאנו ריבס", "Keanu Charles Reeves", "Κιάνου Ριβς", 
            "Κιάνου Τσαρλς Ριβς", "Кіану Рівз", "كيانو ريفز", "基努·李維"];
            for (var t=0; t < tests.length; t++) {
                var langNOBJ = ANNE.MEM.LANG.ascertain_lang(tests[t]);
                if (langNOBJ!=null) console.log(tests[t]+" is "+langNOBJ.text);
                else console.log(tests[t]+" is unknown (fix ucode start/finish)");
            }
        }


        this.lang_english = this.minBootNeuron({nid:"777722641707973679", text:"english", ntype:NTYPE_CLASS});
        this.language.be_lvl = 5;

        this.country = this.minBootNeuron({nid:"7773279276859581115", text:"country", ntype:NTYPE_CLASS});
        this.territory = this.minBootNeuron({nid:"7779776635715128867", text:"territory", ntype:NTYPE_CLASS}); 
        this.province = this.minBootNeuron({nid:"7773009519142577067", text:"province", ntype:NTYPE_CLASS}); 
        this.empire = this.minBootNeuron({nid:"7774987933806776971", text:"empire", ntype:NTYPE_CLASS});
        this.state = this.minBootNeuron({nid:"7772399593363773393", text:"state", ntype:NTYPE_CLASS});
        this.cownty = this.minBootNeuron({nid:"7771790616029822741", text:"county", ntype:NTYPE_CLASS});
        this.district = this.minBootNeuron({nid:"7770117249951034246", text:"district", ntype:NTYPE_CLASS});
        
        this.polity =  this.minBootNeuron({nid:"7773730208317207007", text:"polity", ntype:NTYPE_CLASS});
        this.polity.prompt = "What level of goverence is this location known to be?";

        this.gender =  this.minBootNeuron({nid:"7776266836705549230", text:"gender", ntype:NTYPE_CLASS});
        this.gender.be_lvl = 2;
        this.gender.ai_ico_class="ai-gender";
        this.gender._uico="⚥";
        
        this.male =  this.minBootNeuron({nid:"7774389889917145269", text:"male", ntype:NTYPE_CLASS});
        this.male.be_lvl = 3;
        this.female =  this.minBootNeuron({nid:"7775312609915005294", text:"female", ntype:NTYPE_CLASS});
        this.female.be_lvl = 3;

        // These are used in WHITELIST GARDENS (owner-collab)
        // and BLACKLIST GARDENS (communual)
        // They define collabability  (ability to receive collaboration requests)
        // to save time from requests of a neuron owner
        // that are not desired/welcome.
        this.let_wv = this.minBootNeuron({nid:"7779279922094725823", text:"let wv", ntype:NTYPE_CLASS});
        this.deny_wv = this.minBootNeuron({nid:"7770245499442134915", text:"deny wv", ntype:NTYPE_CLASS});
        this.viplet = this.minBootNeuron({nid:"7776626551577333735", text:"vip let", ntype:NTYPE_CLASS});

        // This adds in neurons to a worldview that are otherwise not apart of it.
        // It does not change ownership of the neuron and is permissionless.
        // anyone can adopt any neuron into their worldview
        this.trusted_adopt = this.minBootNeuron({nid:"7771877111711255601", text:"adopt", ntype:NTYPE_CLASS});
        
        //offering ur neuron into another worldview for them to own
        //this builds a consistent more clean (less noisy) world-view based resource.
        //The giver becomes the sponsor, the receiver becomes the owner.
        //ONLY KEYLESS NEURONS CAN BE HANAI'd.
        this.hanai_out = this.minBootNeuron({nid:"7778381533903269856", text:"Hānai out", ntype:NTYPE_CLASS});
        this.hanai_in = this.minBootNeuron({nid:"7773785541375535981", text:"Hānai in", ntype:NTYPE_CLASS});
        this.hanai_in_all = this.minBootNeuron({nid:"7773285486970861931", text:"Hānai in all", ntype:NTYPE_CLASS});
        
        //response to decline a hanai request.
        //to accept the hanai_in is used.
        this.hanai_in_reject = this.minBootNeuron({nid:"7772421775758796352", text:"hanai in - reject", ntype:NTYPE_CLASS});
        
        this.hanai_and_collab= this.minBootNeuron({nid:"7777343524618811798", text:"hanai & Collab", ntype:NTYPE_CLASS});
        this.hanai_and_collab_reject= this.minBootNeuron({nid:"7776217829049139253", text:"hanai & Collab - reject", ntype:NTYPE_CLASS});
        this.hanai_and_collab_accept= this.minBootNeuron({nid:"7776242842114517972", text:"hanai & Collab - accept", ntype:NTYPE_CLASS});
        
        //"a fund" (boosties bucket is a fund
        this.fund=this.minBootNeuron({nid:"7776570505699865118", text:"fund", ntype:NTYPE_CLASS});
        
        this.boosties = this.minBootNeuron({nid:"7772889444404452615", text:"boosties", ntype:NTYPE_CLASS});
        this.boosties_bucket = this.minBootNeuron({nid:"7776774728242910650", text:"boosties bucket", ntype:NTYPE_CLASS});
        this.winner_payout =this.minBootNeuron({nid:"7778589396232770515", text:"winner payout", ntype:NTYPE_CLASS});
        
        this.annex = this.minBootNeuron({nid:"7775093227033351572", text:"annex", ntype:NTYPE_CLASS});
        this.__setImgThumbDefaultClassImg(this.annex, ANNE.___root+"/imgs/annex_white_ico.png");

        this.annexed_by = this.minBootNeuron({nid:"7771428403886879153", text:"annexed by", ntype:NTYPE_CLASS});
        
        // This removes a trusted adopt.  It is essentially an arbor prune.
        this.abandon = this.minBootNeuron({nid:"7777722740177893415", text:"abandon", ntype:NTYPE_CLASS});

        //this is essentially an inst of an inst.  If a neuron is made outside your worldview and 
        //you want to edit it but the owner isn't collaborating -- you can mirror their neuron
        //and get all data values (reflected from the mirror) into a new neuron.  They are not copied.
        //the mirror is a pointer.  You can then add your own values to your own neuron at will.
        this.mirror = this.minBootNeuron({nid:"7771857764028242565", text:"mirror", ntype:NTYPE_CLASS});

        //this is the garden type for the class neuron.  It cascades down to all sub-classes
        this.garden_type = this.minBootNeuron({nid:"7773397126684636358", text:"garden type", ntype:NTYPE_CLASS});
        
        //this is the garden type for instances.  It is defined on classes and cascades down to instances.
        //this will let the protocol L1 know if an instance can be collaborated on.
        this.inst_garden_type = this.minBootNeuron({nid:"7771542651914330289", text:"inst garden type", ntype:NTYPE_CLASS});
        
        //anyone not blacklist can submit collab requests and ADOPT_OUT requests/responses
        //anyone not blacklisted can create children 
        //GUI will show children of class owned by the same owner or the logged in user.
        this.gt_communal = this.minBootNeuron({nid:"7775536037932928919", text:"communual", ntype:NTYPE_CLASS});
        
        //anyone on the whitelist can submit collab requests and ADOPT_OUT requests/responses
        //anyone on the whitelist can create children 
        //GUI will show children of class owned by the same owner or the logged in user.
        this.gt_owner_collab = this.minBootNeuron({nid:"7778367819213636391", text:"owner collab", ntype:NTYPE_CLASS});
        
        //no collab requests, no delegation
        //GUI will show children of class owned by the same owner.
        this.gt_owner_only = this.minBootNeuron({nid:"7776218453394292289", text:"owner only", ntype:NTYPE_CLASS});
        

        this.andy_collab_accepted = this.minBootNeuron({nid:"7776209936855135438", text:"Collab - accepted", ntype:NTYPE_CLASS});
        this.andy_collab_rejected = this.minBootNeuron({nid:"7775459356522660302", text:"Collab - rejected", ntype:NTYPE_CLASS});
        
        this.anj_accepted= this.minBootNeuron({nid:"7774382206911325219", text:"anjucator - accept", ntype:NTYPE_CLASS});
        this.anj_rejected= this.minBootNeuron({nid:"777184939498367203", text:"anjucator - reject", ntype:NTYPE_CLASS});
        
        this.hanai_and_collab_accept_all =this.minBootNeuron({nid:"7773734743483119116", text:"hanai & Collab - accept ALL", ntype:NTYPE_CLASS});
        
        this.collabRespRelns = [
             this.andy_collab_accepted.temp_nid,
             this.andy_collab_rejected.temp_nid,
             this.hanai_and_collab_reject.temp_nid,
             this.hanai_and_collab_accept.temp_nid,
             this.hanai_and_collab_accept_all.temp_nid
        ];

        //delegationrights
        //can alter class or any sub-class with same owner as issuer of rights.
        this.dr_andy_super_class = this.minBootNeuron({nid:"7776049367153763537", text:"ANDY super class", ntype:NTYPE_CLASS});
        this.dr_andy_super_class.ai_ico_class = "ai-andy";

        //can alter the neuron that issued the rights
        this.dr_andy_defn = this.minBootNeuron({nid:"7776053324438397729", text:"ANDY chg definition", ntype:NTYPE_CLASS});
        this.dr_andy_defn.ai_ico_class = "ai-andy";
        
        //can alter any insts in the down be-line of the issuer of rights
        this.dr_andy_super_insts = this.minBootNeuron({nid:"7776120177414977781", text:"ANDY super insts", ntype:NTYPE_CLASS});
        this.dr_andy_super_insts.ai_ico_class = "ai-andy";
        
        //can alter any insts or sub-classes in the down be-line of the issuer of rights
        this.dr_andy_super_all= this.minBootNeuron({nid:"7776131279600795401", text:"ANDY super all", ntype:NTYPE_CLASS});       
        this.dr_andy_super_all.ai_ico_class = "ai-andy";
        
        this.dr_andy_hanai_in = this.minBootNeuron({nid:"7776131985740735420", text:"Hanai admin", ntype:NTYPE_CLASS});
        this.dr_andy_hanai_in.ai_ico_class = "ai-hanai";
        
        //would be nice to have a superall admin.
        //can hanai and collab ANYTHING. (just like ANNE)
        this.dr_super_super = this.minBootNeuron({nid:"7779079235159468841", text:"ANNE super super admin", ntype:NTYPE_CLASS});
        this.dr_super_super.ai_ico_class = "ai-andy";
        

        //restrictions on delegation ability.
        //only specific are granted rights to delegate to
        this.dr_permitted_reln = this.minBootNeuron({nid:"7771471428900947007", text:"delegation permitted reln", ntype:NTYPE_CLASS});
        this.dr_permitted_reln.ai_ico_class = "ai-andy";
        
        this.garden_rule = this.minBootNeuron({nid:"7779338776490778825", text:"garden rule", ntype:NTYPE_CLASS}); 
        
        //restrictions on rights limiting them to the scope of the permitted nid only.
        this.gt_permit_reln = this.minBootNeuron({nid:"7777831065415748606", text:"garden permitted reln", ntype:NTYPE_CLASS});
          
        this.delegation_rights = this.minBootNeuron({nid:"7776042304763512185", text:"delegation rights", ntype:NTYPE_CLASS});
        this.delegation_reln_tnids = [
            this.dr_andy_super_class.temp_nid, 
            this.dr_andy_defn.temp_nid, 
            this.dr_andy_super_all.temp_nid,
            this.dr_andy_super_insts.temp_nid, 
            this.dr_andy_hanai_in.temp_nid, 
            this.dr_permitted_reln.temp_nid,
            this.dr_super_super.temp_nid
        ];


        //used to document un-conceptualized sentences for later,
        //or to audit against our conceptualization and then verbalization.
        this.original_sentence = this.minBootNeuron({nid:"7777509361790960520", text:"original sentence", ntype:NTYPE_CLASS});
        
        this.PENDING_HANAI = this.minBootNeuron({nid:"7776275301750156832", text:"pending hanai", ntype:NTYPE_CLASS});
        
        
        this.ANNE_THEME = this.minBootNeuron({nid:"7778978696584757203", text:"ANNE theme", ntype:NTYPE_CLASS});
            this.THEME_default = this.minBootNeuron({nid:"7779532849877969632", text:"ANNE theme (default)", ntype:NTYPE_CLASS});
        this.preference = this.minBootNeuron({nid:"7777013225235554698", text:"preference", ntype:NTYPE_CLASS});
        
        
        this.meme = this.minBootNeuron({nid:"7779959823859193016", text:"meme", ntype:NTYPE_CLASS});
        
        this.neuronImg = this.minBootNeuron({nid:"7772973406618599102", text:"neuron img", ntype:NTYPE_CLASS});
        this.neuronImg.ai_ico_class="ai-neuron-img";

        //external dataURL
        this.dataUrl = this.minBootNeuron({nid:"7775367751718991425", text:"data url", ntype:NTYPE_CLASS});

        //path to internal/ships-on-annode file location.
        this.dataRelativePath = this.minBootNeuron({nid:"7772478183226777305", text:"data relative path", ntype:NTYPE_CLASS});
        
        this.primaryNeuronImg = this.minBootNeuron({nid:"7776249509320308559", text:"primary neuron img", ntype:NTYPE_CLASS});
        
        this.imgThumbDefaultClass = this.minBootNeuron({nid:"7776781300433802741", text:"img Thumb Default Class", ntype:NTYPE_CLASS});
        
        this.fileX = this.minBootNeuron({nid:"7775682115334476760", text:"file", ntype:NTYPE_CLASS});
        
        this.photo_taken_on = this.minBootNeuron({nid:"7776305132990229691", text:"photo taken on", ntype:NTYPE_CLASS});
       
        this.fileFingerPrint = this.minBootNeuron({nid:"7774472646258496586", text:"file finger print", ntype:NTYPE_CLASS});
        this.fileSize = this.minBootNeuron({nid:"7770613549369528475", text:"file size", ntype:NTYPE_CLASS});

        this.neuronVid = this.minBootNeuron({nid:"7772892807996216912", text:"neuron video", ntype:NTYPE_CLASS});
        this.neuronVid.ai_ico_class="ai-neuron-video";

        this.media_author = this.minBootNeuron({nid:"7779863026234386372", text:"media author", ntype:NTYPE_CLASS});
        
        this.neuronAudio = this.minBootNeuron({nid:"7777637061760581244", text:"neuron audio", ntype:NTYPE_CLASS});
        this.neuronAudio.ai_ico_class="ai-neuron-audio";
        //this.neuronAudioDataUrl = this.minBootNeuron({nid:"7774188933350587286", text:"neuron audio data", ntype:NTYPE_CLASS});

        this.url_part = this.minBootNeuron({nid:"777764039944765162", text:"url part", ntype:NTYPE_CLASS});
        this.base_url = this.minBootNeuron({nid:"7779987300332115752", text:"base url", ntype:NTYPE_CLASS});
        this.ai_ico_class = this.minBootNeuron({nid:"7774405233694204252", text:"ai ico class", ntype:NTYPE_CLASS});
        this.logo_path_with_base = this.minBootNeuron({nid:"7772311622096722944", text:"logo path with base", ntype:NTYPE_CLASS});
        
        this.wcImgUriPart_Flag= this.minBootNeuron({nid:"7772262837359918982", text:"wc flag part", ntype:NTYPE_CLASS});
        this.wcImgUriPart_Map= this.minBootNeuron({nid:"7771813649898124264", text:"wc map part", ntype:NTYPE_CLASS});
        this.wcImgUriPart_Seal= this.minBootNeuron({nid:"7776167254082509668", text:"wc seal part", ntype:NTYPE_CLASS});
        this.wcImgUriPart_Skyline= this.minBootNeuron({nid:"7774254145083522907", text:"wc skyline part", ntype:NTYPE_CLASS});
        this.wcImgUriPart_thumbnail= this.minBootNeuron({nid:"777446366696254759", text:"wc thumbnail part", ntype:NTYPE_CLASS});

             
        //pts to a neuron img.
        this.logo = this.minBootNeuron({nid:"7774775920254990285", text:"logo", ntype:NTYPE_CLASS});
        
        this.website = this.minBootNeuron({nid:"7772702516735048412", text:"website", ntype:NTYPE_CLASS});
        
        this.mimetype = this.minBootNeuron({nid:"7777176012795245449", text:"mime type", ntype:NTYPE_CLASS});

        //sub of entity
        this.identity = this.minBootNeuron({nid:"7778460139233034749", text:"identity", ntype:NTYPE_CLASS});
        this.identity.wvBIPASS = true;
        this.identity.be_lvl = 4;

            //sub of identity
            this.individual_identity = this.minBootNeuron({nid:"7776971213981656361", text:"individual", ntype:NTYPE_CLASS}); 
            this.individual_identity.ai_ico_class = "ai-individual";
            this.individual_identity.wvBIPASS = true;
            this.individual_identity.be_lvl = 5;
            //this.individual_identity._uico = "👤";
            
            this.application_identity = this.minBootNeuron({nid:"7778868011102771570", text:"application", ntype:NTYPE_CLASS});
            this.application_identity.ai_ico_class = "ai-application";
            this.application_identity.wvBIPASS = true;
            this.application_identity.be_lvl = 5;
            this.application_identity._uico = "📱";

            this.corporation_identity = this.minBootNeuron({nid:"7775126160437279358", text:"corporation (identity)", ntype:NTYPE_CLASS}); 
            this.corporation_identity.ai_ico_class = "ai-corporation";
            this.corporation_identity.wvBIPASS = true;
            this.corporation_identity.be_lvl = 5;
            this.corporation_identity._uico="🏢";

            //any neuron can be an oracle, in theory with the ASDF
            this.oracle_identity = this.minBootNeuron({nid:"777941170343283715", text:"oracle", ntype:NTYPE_CLASS});
            this.oracle_identity.ai_ico_class = "ai-oracle";
            this.oracle_identity.wvBIPASS = true;
            this.oracle_identity.be_lvl = 5;
            this.oracle_identity._uico = "🧙‍♂️";
            
            this.miner_identity = this.minBootNeuron({nid:"7774067842216408975", text:"miner", ntype:NTYPE_CLASS});
            this.miner_identity.ai_ico_class = "ai-miner";
            this.miner_identity.wvBIPASS = true;
            this.miner_identity.be_lvl = 5;
            this.miner_identity._uico = "⚒️";
            this.__setImgThumbDefaultClassImg(this.miner_identity, ANNE.___root+"/imgs/miner_ico.png");

            
            this.annode_identity = this.minBootNeuron({nid:"7777777985638153890", text:"annode", ntype:NTYPE_CLASS});
            this.annode_identity.ai_ico_class = "ai-annode";
            this.annode_identity.wvBIPASS = true;
            this.annode_identity.be_lvl = 5;
            this.annode_identity._uico="💻";
            
            this.emerging_intelligence = this.minBootNeuron({nid:"7777798527035018972", text:"emerging intelligence", ntype:NTYPE_CLASS});
            this.emerging_intelligence.ai_ico_class = "ai-ei";
            this.emerging_intelligence.wvBIPASS = true;
            this.emerging_intelligence.be_lvl = 5;
            this.emerging_intelligence._uico = "⭕";
            

        this.established = this.minBootNeuron({nid:"7774165034042346848", text:"established", ntype:NTYPE_CLASS}); 

        //a class used to instaniate events with.
        //this will inherit a ROLE_ACTION of 
        this.group_event =  this.minBootNeuron({nid:"7777943654909778097", text:"group event", ntype:NTYPE_CLASS});
        
        this.group =  this.minBootNeuron({nid:"7774037854818954199", text:"group", ntype:NTYPE_CLASS});
        this.group.be_lvl = 4;

        this.lolita_express_flights = this.minBootNeuron({nid:"7778886709340031497", text:"lolital express flights", ntype:NTYPE_CLASS});
        
        this.key_person = this.minBootNeuron({nid:"7771668729141593676", text:"key person", ntype:NTYPE_CLASS});

        //from dbpedia .honors
        this.military_honours = this.minBootNeuron({nid:"7775417006091441836", text:"military honours", ntype:NTYPE_CLASS});
        this.military_honours._uico="🎖️";
        this.military_rank = this.minBootNeuron({nid:"7770904604329112073", text:"military rank", ntype:NTYPE_CLASS});
        this.military_rank._uico="🪖";
        this.military_office = this.minBootNeuron({nid:"7771182245027462135", text:"military office", ntype:NTYPE_CLASS});
        this.military_office._uico="🪖";

        this.military_academy = this.minBootNeuron({nid:"7774422634022818293", text:"Military Academy", ntype:NTYPE_CLASS});
        this.military_academy._uico="🛡️";
        

        
        //parent of corp_entity
        this.company = this.minBootNeuron({nid:"7772013259370769039", text:"company", ntype:NTYPE_CLASS});
        this.company._uico="🏢";
        this.__setImgThumbDefaultClassImg(this.company, ANNE.___root+"/imgs/default_location.png");

        this.company_type = this.minBootNeuron({nid:"7771419270983974071", text:"company type", ntype:NTYPE_CLASS});
        
        //parent of company type VALUES... do not instantiate this!
        //this is a the parent class of corp entity types.
        //NOTE: we don't have to do this like this.  We could just create a list of non-living things
        //and then tag them as lateral TYPE_BE corp types  (like we do with polity->country for locations)
        //I want corp entity types to be a bit more tightly controlled within their own heirarchy..
        //so they are more than just a lateral tie-in, they are a tie-in to a structure.
        this.corp_entity_thetype = this.minBootNeuron({nid:"7776097111646937092", text:"corp entity type", ntype:NTYPE_CLASS});
        this.corp_entity_thetype._uico="🏢";
        this.university = this.minBootNeuron({nid:"7771577574214483635", text:"University", ntype:NTYPE_CLASS});
        this.school = this.minBootNeuron({nid:"7770306012760616108", text:"school", ntype:NTYPE_CLASS});
        
        this.athletics_nickname = this.minBootNeuron({nid:"777756177663289842", text:"athletics nickname", ntype:NTYPE_CLASS});

        this.corp_legal_id = this.minBootNeuron({nid:"7779223539556635762", text:"corp legal id", ntype:NTYPE_CLASS});
        this.setCik = this.minBootNeuron({nid:"7778590084799932279", text:"corp legal id", ntype:NTYPE_CLASS});
        this.product = this.minBootNeuron({nid:"7779910224164104021", text:"product", ntype:NTYPE_CLASS});
        this.industry = this.minBootNeuron({nid:"7777507679900547743", text:"industry", ntype:NTYPE_CLASS});
        this.industry._uico="🏭";

        this.service = this.minBootNeuron({nid:"7778977228411484020", text:"service", ntype:NTYPE_CLASS});

        this.valid_identity_classes=[this.application_identity.nid, this.corporation_identity.nid, this.individual_identity.nid, 
                                    this.miner_identity.nid, this.annode_identity.nid, this.emerging_intelligence.nid];
 
        this.opinionz = this.minBootNeuron({nid:"7770982642614293125", text:"opinionz", ntype:NTYPE_CLASS});
        this.feeling = this.minBootNeuron({nid:"7777627441463363769", text:"feeling", ntype:NTYPE_CLASS});
        this.feelz = this.minBootNeuron({nid:"7770855606276797271", text:"feelz", ntype:NTYPE_CLASS});
        
        this.neg_op_bucket = this.minBootNeuron({nid:"7772878965553112082", text:"feelz", ntype:NTYPE_CLASS});
        this.pos_op_bucket = this.minBootNeuron({nid:"7775045073476665086", text:"feelz", ntype:NTYPE_CLASS});

        //this.thought = this.minBootNeuron({nid:"7777402732566502296", text:"thought", ntype:NTYPE_CLASS});
        this.rememberz = this.minBootNeuron({nid:"7779275038752038222", text:"rememberz", ntype:NTYPE_CLASS});

        //not used.  We now just arbor prune the rememberz.  This will unremember it.
        this.forgetz = this.minBootNeuron({nid:"7770910938663002967", text:"forgetz", ntype:NTYPE_CLASS});

        this.email_address = this.minBootNeuron({nid:"7770280080042939437", text:"email address", ntype:NTYPE_CLASS});
        this.email_address.format_validator=ANNE.UTILS.__emailValidation;
        this.email_address.prompt = "What is your email address? (this will be publicly seen)";

        this.KB_PGP_identity =  this.minBootNeuron({nid:"7779531655061632346", text:"KB PGP identity", ntype:NTYPE_CLASS});
        this.KB_PGP_identity.prompt = "What is your kebybase user name?"

        //this.key = this.minBootNeuron({nid:"7772066628629112453", text:"key", ntype:NTYPE_CLASS});

        this.handler = this.minBootNeuron({nid:"7776885682127776948", text:"handler", ntype:NTYPE_CLASS});
        this.clump_semantic = this.minBootNeuron({nid:"7773872021180785983", text:"semantic clump", ntype:NTYPE_CLASS});
        
        //important MAKE cmds
        //ANNODES know what these are.
        //ANY instance can be made without a custom make cmd but 
        //ANNE creates neurons in ANNE worldview in some special cases and these
        //are important to pay special attention to for ANNE.
        this.MAKE_INTEGER = this.minBootNeuron({nid:"7773497986078093746", text:"MAKE INTEGER", ntype:NTYPE_CLASS});
        this.MAKE_DATE = this.minBootNeuron({nid:"7771903652223599868", text:"MAKE DATE", ntype:NTYPE_CLASS});
        this.MAKE_ANNDLE = this.minBootNeuron({nid:"777623943332475085", text:"MAKE ANNDLE", ntype:NTYPE_CLASS});
        this.MAKE_ACTION = this.minBootNeuron({nid:"7779000837012326152", text:"MAKE ACTION (verb)", ntype:NTYPE_CLASS});
        this.MAKE_LOCATION = this.minBootNeuron({nid:"7777564219572819129", text:"MAKE LOCATION", ntype:NTYPE_CLASS});
        this.MAKE_ANORG = this.minBootNeuron({nid:"7771642700241611064", text:"MAKE ANORG", ntype:NTYPE_CLASS});
        this.MAKE_COMPANY = this.minBootNeuron({nid:"777950857929447723", text:"MAKE COMPANY", ntype:NTYPE_CLASS});
        this.MAKE_IMG = this.minBootNeuron({nid:"7779601435233589280", text:"MAKE IMG", ntype:NTYPE_CLASS});
        this.MAKE_PERSON = this.minBootNeuron({nid:"7771801135056150128", text:"MAKE PERSON", ntype:NTYPE_CLASS});
        this.MAKE_BOUNTY = this.minBootNeuron({nid:"7775295870250788931", text:"MAKE BOUNTY", ntype:NTYPE_CLASS});
        
        //note: events will instantiate their action verb.
        this.MAKE_EVENT = this.minBootNeuron({nid:"7776832446666400296", text:"MAKE EVENT", ntype:NTYPE_CLASS});

        
        this.ACTIVATE = this.minBootNeuron({nid:"7770767425934383201", text:"ACTIVATE", ntype:NTYPE_CLASS});
        this.DEACTIVATE = this.minBootNeuron({nid:"7771229050067657498", text:"ACTIVATE", ntype:NTYPE_CLASS});
        

        this.legal_entity = this.minBootNeuron({nid:"7777212189219449042", text:"legal entity", ntype:NTYPE_CLASS});
        this.site = this.minBootNeuron({nid:"7775256219385612317", text:"site", ntype:NTYPE_CLASS});
        this.site.ai_ico_class="ai-site" 
        this.geo_site = this.minBootNeuron({nid:"7772492123548075783", text:"geo site", ntype:NTYPE_CLASS});
        this.geo_site.ai_ico_class = "ai-geo-site";
        
        //all claims (any user-supplied relon) that wishes to invoke a service for validation
        //will use a validation request relon pointing to the txid they want to validate.
        //the service (annode or other 3rd party) will do its magics (black box or using on-chain proof)..
        //and then burn a valid/invalid relon from their service to the txid of the claim.
        //This can be done using the RRV (request relon validation) annvoice..
        //The service burns the valid/invalid relon and also references the annovice txid
        //which paid for the service in the built in reference field. (TODO piggy back signum field)
        this.validation_request = this.minBootNeuron({nid:"7776161597907601765", text:"validation request", ntype:NTYPE_CLASS});

        //set a default location class img url.
        this.__setImgThumbDefaultClassImg(this.location, ANNE.___root+"/imgs/default_location.png");

        this.constituent_of = this.minBootNeuron({nid:"7771195041236647730", text:"constituent of", ntype:NTYPE_CLASS});
        this.constituent = this.minBootNeuron({nid:"7775238842023699728", text:"constituent", ntype:NTYPE_CLASS});

        this.participant =  this.minBootNeuron({nid:"7772308481505259319", text:"participant", ntype:NTYPE_CLASS});
        this.participant_of =  this.minBootNeuron({nid:"7776908840642789745", text:"participant of", ntype:NTYPE_CLASS});
            
        this.founder =  this.minBootNeuron({nid:"7773741333644308987", text:"founder", ntype:NTYPE_CLASS});
        this.founder_of =  this.minBootNeuron({nid:"7778112941848478136", text:"founder of", ntype:NTYPE_CLASS});
        
        //test for ANNE, we don't need diff relns for direction.. we use TYPE_BE and TYPE_HAS for that.
        this.member =  this.minBootNeuron({nid:"7774234737693337904", text:"member", ntype:NTYPE_CLASS});
        this.member_of =  this.minBootNeuron({nid:"7771323162068539738", text:"member of", ntype:NTYPE_CLASS});
        
        this.board_member =  this.minBootNeuron({nid:"7773721544735553188", text:"board member", ntype:NTYPE_CLASS});
        this.board_member_of =  this.minBootNeuron({nid:"7775205884503390304", text:"board member of", ntype:NTYPE_CLASS});
        
        this.corpSponsor =  this.minBootNeuron({nid:"7777670247182851992", text:"corp sponsor", ntype:NTYPE_CLASS});
        this.corpSponsor_of =  this.minBootNeuron({nid:"7772087542288975131", text:"corp sponsor of", ntype:NTYPE_CLASS});
        
        //never instantiated, only a gen prop to point to other companies...
        this.subsidary = this.minBootNeuron({nid:"777364764299065896", text:"subsidary", ntype:NTYPE_CLASS}); 
        this.subsidary_of = this.minBootNeuron({nid:"77754005942822853", text:"subsidary of", ntype:NTYPE_CLASS}); 

        this.division = this.minBootNeuron({nid:"17285996896823940540", text:"division", ntype:NTYPE_CLASS}); 
        this.division_of = this.minBootNeuron({nid:"11438744566110935559", text:"division of", ntype:NTYPE_CLASS}); 

        this.appointed = this.minBootNeuron({nid:"2677740920383363638", text:"appointed", ntype:NTYPE_CLASS}); 
        
        //THIS IS A GOOD CONCEPTUALIZATION EXAMPLE OF HOW TO BUILD THIS CONCEPT FOR LATER NLP USE.
        this.appointed_by = this.minBootNeuron({nid:"18061812464206165953", text:"appointed by", ntype:NTYPE_CLASS}); 

        this.preceded_by = this.minBootNeuron({nid:"18231916307500268961", text:"preceded by", ntype:NTYPE_CLASS});
        this.succeeded_by = this.minBootNeuron({nid:"12775004814305050777", text:"succeeded by", ntype:NTYPE_CLASS});

        //things TYPE_BE relationships of this.
        //it a short-hand clump where the to is really the ROLE_THEME_THING
        //x be founder of y
        //x be participant of y
        this.BE_ofs = [this.participant_of.nid, this.founder_of.nid, this.member_of.nid, this.board_member_of.nid,
            this.corpSponsor_of.nid, this.subsidary_of.nid, this.division_of.nid];

        //things TYPE_HAS relationships of this.
        //y has founder x
        //y has participant x
        this.HAS_no_ofs = [this.participant.nid, this.founder.nid, this.member.nid, this.board_member.nid,
            this.corpSponsor.nid, this.subsidary.nid, this.division.nid];

        //a type of corp
        this.conglomerate = this.minBootNeuron({nid:"8072940268512505445", text:"conglomerate", ntype:NTYPE_CLASS});

        this.ideology = this.minBootNeuron({nid:"17020879412571846679", text:"ideology", ntype:NTYPE_CLASS});
        
        //alumni should makea HAS relon pointing to alumnis.
        //harvard->has->alumni->Bob Smith
        this.alumni = this.minBootNeuron({nid:"7775772515552382303", text:"alumni", ntype:NTYPE_CLASS});

        //alumni_of should make a BE RELON from the alumni to what they are an alumni of
        this.alumni_of = this.minBootNeuron({nid:"7770037193997679017", text:"alumni of", ntype:NTYPE_CLASS});

        this.wikispooks_desc = this.minBootNeuron({nid:"7775709052747670582", text:"wikispooks desc", ntype:NTYPE_CLASS}); 

        //thise are essentially actionvb+when role
        this.foundedOn = this.minBootNeuron({nid:"7772745521469079754", text:"founded on", ntype:NTYPE_CLASS}); 
        
        //for clumps with date OR location roles included.
        this.founded = this.minBootNeuron({nid:"7618523810988349729", text:"founded", ntype:NTYPE_CLASS}); 
        
        this.power_family = this.minBootNeuron({nid:"777842302370981935", text:"power family", ntype:NTYPE_CLASS}); 
        
        this.MAKE_SYMBOL = this.minBootNeuron({nid:"7771149320550437941", text:"MAKE KEYWORD", ntype:NTYPE_CLASS});


        //person id
        this.TMDB_id = this.minBootNeuron({nid:"777811874948779992", text:"TMDB pid", ntype:NTYPE_CLASS});
        this.__setClassBaseURL(this.TMDB_person_id, "https://www.themoviedb.org/");
        this.ai_ico_class="ai-TMDB";

        //
        this.bioguide_id = this.minBootNeuron({nid:"7775032836974177659", text:"buioguide", ntype:NTYPE_CLASS});
        this.thomas_id = this.minBootNeuron({nid:"7778923864354766137", text:"thomas_id", ntype:NTYPE_CLASS});
        this.youtube_id = this.minBootNeuron({nid:"7771610674492367662", text:"youtube_id", ntype:NTYPE_CLASS});
        this.govtrack = this.minBootNeuron({nid:"7770925240349684802", text:"govtrack", ntype:NTYPE_CLASS});
        this.icpsr_prez = this.minBootNeuron({nid:"7773850913328907555", text:"icpsr_prez", ntype:NTYPE_CLASS});
        this.opensecrets = this.minBootNeuron({nid:"7770478812806785802", text:"opensecrets", ntype:NTYPE_CLASS});
        this.fec = this.minBootNeuron({nid:"7779064998469501513", text:"fec", ntype:NTYPE_CLASS});
        this.votesmart = this.minBootNeuron({nid:"7777143372995579784", text:"votesmart", ntype:NTYPE_CLASS});
        this.lis = this.minBootNeuron({nid:"7775005862771329923", text:"lis", ntype:NTYPE_CLASS});
        this.ballotpedia = this.minBootNeuron({nid:"7770151981975829494", text:"ballotpedia", ntype:NTYPE_CLASS});
        this.maplight = this.minBootNeuron({nid:"7770298099837584694", text:"maplight", ntype:NTYPE_CLASS});
        this.cspan = this.minBootNeuron({nid:"7776274437771740534", text:"cspan", ntype:NTYPE_CLASS});
        this.icpsr = this.minBootNeuron({nid:"7776338279415159744", text:"icpsr", ntype:NTYPE_CLASS});
        this.google_entity_id = this.minBootNeuron({nid:"7772247736870247808", text:"google_entity_id", ntype:NTYPE_CLASS});
        this.house_history = this.minBootNeuron({nid:"7772280668716687175", text:"house_history", ntype:NTYPE_CLASS});
        this.wikidata = this.minBootNeuron({nid:"7775289459331965701", text:"wikidata", ntype:NTYPE_CLASS});
        
        //is this a dupe?
        //this.wikipediaDEAD = this.minBootNeuron({nid:"7772905047713397630", text:"wikipedia", ntype:NTYPE_CLASS});
    
        this.wikipart = this.minBootNeuron({nid:"7772333349880064178", text:"wikiurl", ntype:NTYPE_CLASS});
        
        //meaning, from wikipedia.
        this.wiki_meaning = this.minBootNeuron({nid:"17504329198913088809", text:"meaning", ntype:NTYPE_CLASS});
        
        this.dbpedia_id = this.minBootNeuron({nid:"777992912015974318", text:"dbpedia id", ntype:NTYPE_CLASS});
        this.dbpedia_respart = this.minBootNeuron({nid:"7779668581171663468", text:"dbpedia resource part", ntype:NTYPE_CLASS});
        
        this.wikispooks_respart = this.minBootNeuron({nid:"7778158657377162702", text:"wikispooks resource part", ntype:NTYPE_CLASS});
        
        //person id!
        this.IMDB_id = this.minBootNeuron({nid:"7777167264347475422", text:"IMDB pid", ntype:NTYPE_CLASS});
        //https://www.imdb.com/name/
        this.IMDB_id.ai_ico_class="ai-IMDB";
        
        this.instagram_id = this.minBootNeuron({nid:"7771145768785067039", text:"Instagram id", ntype:NTYPE_CLASS});
        this.__setClassBaseURL(this.TMDB_person_id, "https://www.instagram.com/");
        this.instagram_id.ai_ico_class="ai-instagram";

        this.twitter_id = this.minBootNeuron({nid:"7771320139782962995", text:"Twitter id", ntype:NTYPE_CLASS});
        this.__setClassBaseURL(this.twitter_id, "https://twitter.com/");
        this.twitter_id.ai_ico_class="ai-twitter";

        this.facebook_id = this.minBootNeuron({nid:"7779997270324532488", text:"Facebook id", ntype:NTYPE_CLASS});
        this.facebook_id.ai_ico_class="ai-facebook";

        this.birthname = this.minBootNeuron({nid:"7770585205993013954", text:"birth name", ntype:NTYPE_CLASS});
        
        this.birthday = this.minBootNeuron({nid:"7774793041983944904", text:"birthday", ntype:NTYPE_CLASS});
        this.birthday._uico="🎂";
        
        this.birthplace = this.minBootNeuron({nid:"7777813598381855599", text:"birth place", ntype:NTYPE_CLASS});
        
        //used in full clumps -- WHERE is the birthplace, WHEN is the birthday
        this.born = this.minBootNeuron({nid:"7772914512930290384", text:"born", ntype:NTYPE_CLASS});
       
        //for short hand-relons
        this.deathday = this.minBootNeuron({nid:"7775451663389638990", text:"deathday", ntype:NTYPE_CLASS});
        this.deathday.ai_ico_class="ai-death";
        this.deathday._uico="💀";

        this.deathplace = this.minBootNeuron({nid:"7776799915004256366", text:"death place", ntype:NTYPE_CLASS});
        this.deathplace.ai_ico_class="ai-death";
        this.deathplace._uico="🪦";

        this.apprehended = this.minBootNeuron({nid:"7773509819489469268", text:"apprehended", ntype:NTYPE_CLASS});
        this.apprehended._uico = "🚨";

        //used in full clumps -- WHERE is the birthplace, WHEN is the birthday
        this.die  = this.minBootNeuron({nid:"7772869077031626256", text:"die", ntype:NTYPE_CLASS}); 

        this.streampay = this.minBootNeuron({nid:"7779403914189757295", text:"♼ streampay ♼", ntype:NTYPE_CLASS}); 
        this.block_frequency = this.minBootNeuron({nid:"7772686674875797392", text:"block frequency", ntype:NTYPE_CLASS});
        this.min_start_block = this.minBootNeuron({nid:"7776489666160494893", text:"min start block", ntype:NTYPE_CLASS});
        
        this.streampay_src = this.minBootNeuron({nid:"77738031779292675", text:"streampay src", ntype:NTYPE_CLASS});
        this.streampay_src_theFund = this.minBootNeuron({nid:"7773274639808292524", text:"streampay src - the fund", ntype:NTYPE_CLASS});
        this.streampay_src_owner = this.minBootNeuron({nid:"7774646684344886968", text:"streampay src - owner of fund", ntype:NTYPE_CLASS});
        
        this.keep_min_balance = this.minBootNeuron({nid:"777687915780771745", text:"keep min balance", ntype:NTYPE_CLASS});
        this.max_total_payments = this.minBootNeuron({nid:"7778146027628050531", text:"max total payments", ntype:NTYPE_CLASS});
        this.min_streampay_payment = this.minBootNeuron({nid:"7777690415839350242", text:"min streampay payment", ntype:NTYPE_CLASS});
        
        this.ANNE_DDS =  this.minBootNeuron({nid:"7779217783975400919", text:"ANNE DDS", ntype:NTYPE_CLASS});

        //new method to make new neurons keys.  This RELN is used to instruct the ANNODE.
        //Keyless neurons are able to have delegation set.
        //They also can be collab'd with like key'd neurons if their garden type permits
        //collaboration.
        this.MAKE_KEYLESS = this.minBootNeuron({nid:"7776157600356963165", text:"MAKE KEYLESS", ntype:NTYPE_CLASS});
            
        //this.ANNODE_BAN_NID = this.minBootNeuron({nid:"", text:"ANNODE ban nid", ntype:NTYPE_CLASS});
        
        //irrevokable one-time freeze of neuron that will prevent any future OUTS from
        //being broadcast..  and it will also prevent any future ARBOR PRUNES.
        //To extend a froze neuron any further requires a mirror neuron....
        this.CMD_FREEZE = this.minBootNeuron({nid:"7776758007315842114", text:"freeze neuron", ntype:NTYPE_CLASS});
        this.CMD_FREEZE.ai_ico_class="ai-freeze";

        this.CMD_KEYLESS_PAYOUT = this.minBootNeuron({nid:"7771064906243877097", text:"keyless payout", ntype:NTYPE_CLASS});
        
        //will trigger a merge neuron request and must be sent as a collab (not auto-delegatable)
        this.CMD_MERGE_NEURON = this.minBootNeuron({nid:"7775453604074825079", text:"MERGE NEURON", ntype:NTYPE_CLASS});
        this.MERGED_WITH = this.minBootNeuron({nid:"7775681672672970556", text:"MERGED WITH", ntype:NTYPE_CLASS});
        
        //for testing.. in console.. NOT NEEDED IN MEM on release.
        this.usa = this.minBootNeuron({nid:"7774152290984691776", text:"United States of America", ntype:NTYPE_INST});
        this.american = this.minBootNeuron({nid:"10076243270600789032", text:"American", ntype:NTYPE_INST});
        
        this.role = this.minBootNeuron({nid:"7778531759128370850", text:"ec:role", ntype:NTYPE_CLASS});

        //first pass resolution into thing and thing mod
        this.role_thing = this.minBootNeuron({nid:"7770394591629606840", text:"role_thing", ntype:NTYPE_CLASS});
        this.role_thing.ai_ico_class="ai-thing";

        this.role_thing_mod = this.minBootNeuron({nid:"7773772721114312848", text:"role_thing_mod", ntype:NTYPE_CLASS});
        

        this.role_actor = this.minBootNeuron({nid:"7777573229173770374", text:"role_actor", ntype:NTYPE_CLASS}); 
        this.role_actor.ai_ico_class="ai-thing";

        this.role_actor_mod = this.minBootNeuron({nid:"7777519193840188065", text:"role_actor_mod", ntype:NTYPE_CLASS}); 
        this.role_affected_thing = this.minBootNeuron({nid:"777744227568420012", text:"role_affected_thing", ntype:NTYPE_CLASS});
        this.role_affected_thing.ai_ico_class="ai-aff-thing";
        this.role_affected_thing_mod = this.minBootNeuron({nid:"7774964023545941379", text:"role_affected_thing_mod", ntype:NTYPE_CLASS});
        
        //note: in ANNE all verbs are encoded into the R_ACTION role.  They can be thinking, feeling, doing, or being verbs.
        //it doesn't matter.  Be verbs (in english) are children of ec_is (beingness) and they have some differences.
        
        //aux vbs that alter progressive/tense. 
        this.role_aux_action = this.minBootNeuron({nid:"777983066090703356", text:"role_aux_action", ntype:NTYPE_CLASS});
        this.role_mode = this.minBootNeuron({nid:"7774040954807558904", text:"role_mode", ntype:NTYPE_CLASS});
        this.role_tense = this.minBootNeuron({nid:"7770229184388498871", text:"role_tense", ntype:NTYPE_CLASS});
        this.role_action = this.minBootNeuron({nid:"7777055149623754971", text:"role_action", ntype:NTYPE_CLASS});
        this.role_action.ai_ico_class="ai-action";

        this.role_action_thing = this.minBootNeuron({nid:"7776460781235543752", text:"role_action_thing", ntype:NTYPE_CLASS});
        this.role_be_thing = this.minBootNeuron({nid:"7773949836736341630", text:"role_be_thing", ntype:NTYPE_CLASS});
        this.role_be_thing_mod = this.minBootNeuron({nid:"777751343740502205", text:"role_be_thing_mod", ntype:NTYPE_CLASS});
        
        //a role that points to another clump with no particular semantic relationships specified
        this.role_clump = this.minBootNeuron({nid:"7770332898798547236", text:"role_clump", ntype:NTYPE_CLASS});

        this.role_connector = this.minBootNeuron({nid:"777763967723953677", text:"role_connector", ntype:NTYPE_CLASS});
        this.role_posn_thing = this.minBootNeuron({nid:"7773433027241820655", text:"role_posn_thing", ntype:NTYPE_CLASS});
        this.role_posn_thing_mod = this.minBootNeuron({nid:"7776712633779543606", text:"role_posn_thing_mod", ntype:NTYPE_CLASS});
        this.role_manner = this.minBootNeuron({nid:"7777019863026566731", text:"role_manner", ntype:NTYPE_CLASS});
        this.role_theme_thing = this.minBootNeuron({nid:"7777401229711945348", text:"role_theme_thing", ntype:NTYPE_CLASS});
        this.role_theme_thing_mod = this.minBootNeuron({nid:"7776071931241757674", text:"role_theme_thing_mod", ntype:NTYPE_CLASS});
       
        this.role_where = this.minBootNeuron({nid:"7779297186094235437", text:"role_where", ntype:NTYPE_CLASS});
        this.role_where.ai_ico_class="ai-where";
        this.role_where_mod = this.minBootNeuron({nid:"7776640823686512237", text:"role_where_mod", ntype:NTYPE_CLASS});
        this.role_source = this.minBootNeuron({nid:"7779889734760631010", text:"role_source", ntype:NTYPE_CLASS});
        this.role_path = this.minBootNeuron({nid:"7778169513595010118", text:"role_path", ntype:NTYPE_CLASS});
        this.role_goal = this.minBootNeuron({nid:"777522762176315595", text:"role_goal", ntype:NTYPE_CLASS});
       
        this.role_when = this.minBootNeuron({nid:"777581408555193886", text:"role_when", ntype:NTYPE_CLASS});
        this.role_when.ai_ico_class="ai-when";
        this.role_when_mod = this.minBootNeuron({nid:"7774341790656657922", text:"role_when_mod", ntype:NTYPE_CLASS});
        this.role_start = this.minBootNeuron({nid:"7774749544848267976", text:"role_start", ntype:NTYPE_CLASS});
        this.role_end = this.minBootNeuron({nid:"7776097378912616014", text:"role_end", ntype:NTYPE_CLASS});

        this.role_instrument = this.minBootNeuron({nid:"7779418330103123081", text:"role_instrument", ntype:NTYPE_CLASS});
        this.role_instrument_mod = this.minBootNeuron({nid:"7772775616468645341", text:"role_instrument", ntype:NTYPE_CLASS});
        this.role_resulting_new_thing = this.minBootNeuron({nid:"7770394682139219297", text:"role_resulting_new_thing", ntype:NTYPE_CLASS});
        this.role_reason = this.minBootNeuron({nid:"7777005928865157015", text:"role_reason", ntype:NTYPE_CLASS});
        this.role_purpose = this.minBootNeuron({nid:"7772840210245397923", text:"role_purpose", ntype:NTYPE_CLASS});
        this.role_question = this.minBootNeuron({nid:"7778784064660352579", text:"role_question", ntype:NTYPE_CLASS});
                
        //link to other clumps
        this.role_cause = this.minBootNeuron({nid:"777104212581317832", text:"role_cause", ntype:NTYPE_CLASS});
        this.role_cause.ai_ico_class="ai-why";
        this.role_effect = this.minBootNeuron({nid:"7772214179202770375", text:"role_effect", ntype:NTYPE_CLASS});
        this.role_effect.ai_ico_class="ai-effect";
        this.role_next = this.minBootNeuron({nid:"7775884336586790023", text:"role_next", ntype:NTYPE_CLASS});
        this.role_prev = this.minBootNeuron({nid:"7773564479359111744", text:"role_prev", ntype:NTYPE_CLASS});
        this.role_cdx = this.minBootNeuron({nid:"7772055522386399843", text:"role_cdx", ntype:NTYPE_CLASS});
        this.role_cdx_result = this.minBootNeuron({nid:"7776763659152580485", text:"role_cdx_result", ntype:NTYPE_CLASS});
        this.role_why = this.minBootNeuron({nid:"7779965580540643037", text:"role_why", ntype:NTYPE_CLASS});
        this.role_why.ai_ico_class="ai-why";
        
        this.role_feelz = this.minBootNeuron({nid:"777531590761518279", text:"role_feelz", ntype:NTYPE_CLASS});
        this.role_thought = this.minBootNeuron({nid:"7776728107601212876", text:"role_thought", ntype:NTYPE_CLASS});
        this.role_fact = this.minBootNeuron({nid:"7777226313918390844", text:"role_fact", ntype:NTYPE_CLASS});
        this.role_state = this.minBootNeuron({nid:"7777228404354952479", text:"role_state", ntype:NTYPE_CLASS});
        this.role_implies = this.minBootNeuron({nid:"777391174437512597", text:"role_implies", ntype:NTYPE_CLASS});
        this.role_infer = this.minBootNeuron({nid:"7779731604484117151", text:"role_infer", ntype:NTYPE_CLASS});

        //documenting information source
        this.role_src_ref = this.minBootNeuron({nid:"7779273238068191847", text:"role_src_ref", ntype:NTYPE_CLASS});
        this.role_src_ref.ai_ico_class="ai-ref-src";
        
        this.role_meta_meaning = this.minBootNeuron({nid:"7778288624518302407", text:"role_meta_meaning", ntype:NTYPE_CLASS});
        
        this.role_standards=[
            this.role_action.temp_nid, 
            this.role_manner.temp_nid, 
            this.role_actor.temp_nid, 
            this.role_actor_mod.temp_nid, 
            this.role_tense.temp_nid,
            this.role_when.temp_nid, 
            this.role_when_mod.temp_nid, 
            this.role_where.temp_nid, 
            this.role_where_mod.temp_nid
        ];

        //ec patterns use ANNE ops. (see ANNEX)
        this.ec_pattern = this.minBootNeuron({nid:"7770697322248948996", text:"ec.pattern", ntype:NTYPE_CLASS});    
        this.name_pattern = this.minBootNeuron({nid:"7770271905530827053", text:"name_pattern", ntype:NTYPE_CLASS});    
        

        this.__max =  this.minBootNeuron({nid:"7772748822322176774", text:"__max", ntype:NTYPE_CLASS});    
        this.__min =  this.minBootNeuron({nid:"7772077984786233524", text:"__min", ntype:NTYPE_CLASS});    
        this.__min_digits =  this.minBootNeuron({nid:"7771179556152284330", text:"__min_digits", ntype:NTYPE_CLASS});    
        this.__optional =  this.minBootNeuron({nid:"7771382453261270234", text:"__optional", ntype:NTYPE_CLASS});    
        this.__required =  this.minBootNeuron({nid:"7772207950419895534", text:"__required", ntype:NTYPE_CLASS});    
        this.__value =  this.minBootNeuron({nid:"7776251561494105914", text:"__value", ntype:NTYPE_CLASS});    
        this.__next_active =  this.minBootNeuron({nid:"7775608436552493714", text:"__next_active", ntype:NTYPE_CLASS});    
        this.__resolves_to =  this.minBootNeuron({nid:"7776251561494105914", text:"__resolves_to", ntype:NTYPE_CLASS});    
        this.__make_new =  this.minBootNeuron({nid:"7770730581584298096", text:"__make_new", ntype:NTYPE_CLASS});    
        

        this.ec_part =  this.minBootNeuron({nid:"7770730581584298096", text:"ec:part", ntype:NTYPE_CLASS});    
        this.ec_whole =  this.minBootNeuron({nid:"7773902746956577014", text:"ec:whole", ntype:NTYPE_CLASS});    
        
        //these are words/phrases we toss away after we get the signal.  We do not store them.  They help us
        //know what is coming.
        
        this.ec_expect_actor = this.minBootNeuron({nid:"7770746507409045605", text:"x_actor", ntype:NTYPE_CLASS});    
        this.ec_expect_action = this.minBootNeuron({nid:"7778899167440690152", text:"x_action", ntype:NTYPE_CLASS});    
        this.ec_expect_manner = this.minBootNeuron({nid:"7772660770283904989", text:"x_manner", ntype:NTYPE_CLASS});    
        this.ec_expect_affected_thing = this.minBootNeuron({nid:"8150309968616768260", text:"x_affected_thing", ntype:NTYPE_CLASS});    
        this.ec_expect_clump = this.minBootNeuron({nid:"9146769968576781671", text:"x_clump", ntype:NTYPE_CLASS});    
        this.ec_expect_concept = this.minBootNeuron({nid:"13998275032128243027", text:"x_concept", ntype:NTYPE_CLASS});    
        this.ec_expect_condition = this.minBootNeuron({nid:"12319126297907506798", text:"x_condition", ntype:NTYPE_CLASS});    
        this.ec_expect_goal = this.minBootNeuron({nid:"4322029487966010252", text:"x_goal", ntype:NTYPE_CLASS});    
        this.ec_expect_posn = this.minBootNeuron({nid:"13698091595456770706", text:"x_posn", ntype:NTYPE_CLASS});    
        this.ec_expect_question = this.minBootNeuron({nid:"777713567747830967", text:"x_question", ntype:NTYPE_CLASS});    
        this.ec_expect_role = this.minBootNeuron({nid:"3437301647300744555", text:"x_role_*", ntype:NTYPE_CLASS});    
        this.ec_expect_state = this.minBootNeuron({nid:"2704422592051914885", text:"x_state", ntype:NTYPE_CLASS});    
        this.ec_expect_theme_thing = this.minBootNeuron({nid:"3069042795682288556", text:"x_theme_thing", ntype:NTYPE_CLASS});    
        this.ec_expect_thing = this.minBootNeuron({nid:"614649676553071566", text:"x_thing", ntype:NTYPE_CLASS});    
        this.ec_expect_when = this.minBootNeuron({nid:"1703459443289362399", text:"x_when", ntype:NTYPE_CLASS});    
        this.ec_expect_where = this.minBootNeuron({nid:"1875762255298275190", text:"x_where", ntype:NTYPE_CLASS});    
        this.ec_expect_actionthing = this.minBootNeuron({nid:"7774775137559625245", text:"x_actionthing", ntype:NTYPE_CLASS});    
        this.ec_expect_more_of_prev= this.minBootNeuron({nid:"7776848702697001980", text:"x_more_of_prev", ntype:NTYPE_CLASS});    
        this.ec_expect_thing_mod= this.minBootNeuron({nid:"8754326448223055429", text:"x_thing_mod", ntype:NTYPE_CLASS});    


        this.role_mapper = this.minBootNeuron({nid:"7770273269579595019", text:"role mapper", ntype:NTYPE_CLASS});
        this.role_mapper_actor = this.minBootNeuron({nid:"7772508118194483571", text:"role mapper - actor (agent/subj)", ntype:NTYPE_CLASS});
        this.role_mapper_be_tense = this.minBootNeuron({nid:"7778791128313669071", text:"mapper - be tense", ntype:NTYPE_CLASS});
        this.role_mapper_when = this.minBootNeuron({nid:"7779712508649805541", text:"role mapper - when", ntype:NTYPE_CLASS});
        this.role_mapper_where = this.minBootNeuron({nid:"7770345493567182036", text:"role mapper - where", ntype:NTYPE_CLASS});
        this.role_mapper_action = this.minBootNeuron({nid:"7775935060062601273", text:"role mapper - action", ntype:NTYPE_CLASS});
        this.role_mapper_theme = this.minBootNeuron({nid:"7779583700584149129", text:"role mapper - theme", ntype:NTYPE_CLASS});
        this.role_mapper_affected_thing = this.minBootNeuron({nid:"7770417822110584754", text:"role mapper - affected thing", ntype:NTYPE_CLASS});
        this.role_mapper_tense = this.minBootNeuron({nid:"7774822487994741720", text:"role mapper - tense", ntype:NTYPE_CLASS});
        
        this.role_mappers = [
            this.role_mapper_actor.temp_nid,
            this.role_mapper_be_tense.temp_nid,
            this.role_mapper_when.temp_nid,
            this.role_mapper_where.temp_nid,
            this.role_mapper_action.temp_nid,
            this.role_mapper_theme.temp_nid,
            this.role_mapper_affected_thing.temp_nid,
            this.role_mapper_tense.temp_nid
        ];
        this.role_map = [];
        this.role_map[this.role_mapper_actor.temp_nid]= this.role_actor.temp_nid;
        this.role_map[this.role_mapper_when.temp_nid]= this.role_when.temp_nid;
        this.role_map[this.role_mapper_where.temp_nid]= this.role_where.temp_nid;
        this.role_map[this.role_mapper_action.temp_nid]= this.role_action.temp_nid;
        this.role_map[this.role_mapper_theme.temp_nid]= this.role_theme_thing.temp_nid;
        this.role_map[this.role_mapper_affected_thing.temp_nid]= this.role_affected_thing.temp_nid;
        this.role_map[this.role_mapper_tense.temp_nid]= this.role_tense.temp_nid;


        //pts to the from slot of a relon used in role mappers.
        this.relon_from = this.minBootNeuron({nid:"777780868832582872", text:"relon from", ntype:NTYPE_CLASS});
        this.relon_to = this.minBootNeuron({nid:"7772898090768696320", text:"relon to", ntype:NTYPE_CLASS});
        this.relon_reln = this.minBootNeuron({nid:"7773696674962705615", text:"relon reln", ntype:NTYPE_CLASS});
        
        
        this.SMART_SEARCH_ONLY = this.minBootNeuron({nid:"7773734358383597712", text:"smart search only", ntype:NTYPE_CLASS});

        //private data neuron (definitions will be on-chain when preload occurs, as with all other data
        //here in the bare-bones neural pre-load)
        this.ENCRYPED_DATA_TYPE = this.minBootNeuron({nid:"7777437385154819845", text:"encrypted data type", ntype:NTYPE_CLASS});
        this.PDN = this.minBootNeuron({nid:"7776729867696386401", text:"private data neuron", ntype:NTYPE_CLASS});
        this.PDN.wvBIPASS = true;
        this.PDN_PUBLIC_ACCESS = this.minBootNeuron({nid:"7777374627033925008", text:"PDN public access", ntype:NTYPE_CLASS});
        this.PDN_PUBLIC_ACCESS.wvBIPASS = true;
        this.PDN_VIP = this.minBootNeuron({nid:"777865917767872287", text:"PDN VIP", ntype:NTYPE_CLASS});
        this.PDN_VIP.wvBIPASS = true;
        this.PDN_BEARER_TOKEN = this.minBootNeuron({nid:"7772868456433715520", text:"PDN Bearer Token", ntype:NTYPE_CLASS});
        this.PDN_BEARER_TOKEN.wvBIPASS = true;
        //bipass needed ins PDNS are owner only-- so general public won't find
        //them in downline searches (gardenEnforce fails).
        //we want them to stay owner-only for NO COLLAB thus we bipass WV to allow
        //searches to find them w/o user putting them in a white list.
        this.bearer_token = this.minBootNeuron({nid:"7772559250440764390", text:"bearer token", ntype:NTYPE_CLASS});
        this.PDN_OP_QUESTION = this.minBootNeuron({nid:"777461361898715860", text:"PDN OP Q", ntype:NTYPE_CLASS});
        this.PDN_OP_REVEAL_ALL = this.minBootNeuron({nid:"7778580360323145476", text:"PDN OP REVEAL ALL", ntype:NTYPE_CLASS});
        this.PDN_EQUALITY_QUESTION = this.minBootNeuron({nid:"7771390471443803065", text:"PDN EQUALITY Q", ntype:NTYPE_CLASS});
        this.PDN_SECRET = this.minBootNeuron({nid:"7771501455189830305", text:"PDN secret", ntype:NTYPE_CLASS});
        this.ANNODE_URI = this.minBootNeuron({nid:"7775200563510845790", text:"ANNODE URI", ntype:NTYPE_CLASS});

        this.PDN_ANSWER = this.minBootNeuron({nid:"7778569016061145146", text:"PDN answer", ntype:NTYPE_CLASS});


        //limited set of XOPs used in PDNs that are helpful to have a var in MEM memory.
        this.XOP = this.minBootNeuron({nid:"777559007866194998", text:"ANNE X OPERATORS", ntype:NTYPE_CLASS});
        
        this.XOP_equal = this.minBootNeuron({nid:"7775618696518156032", text:"==", ntype:NTYPE_CLASS});
        this.XOP_equal.asdf_date_disp = "on";
        this.XOP_notequal = this.minBootNeuron({nid:"7775628753205217910", text:"!=", ntype:NTYPE_CLASS});
        this.XOP_notequal.asdf_date_disp = "not on";
        this.XOP_lessthan= this.minBootNeuron({nid:"7775667956029190870", text:"<", ntype:NTYPE_CLASS});
        this.XOP_lessthan.asdf_date_disp = "before";
        this.XOP_lessthanequal= this.minBootNeuron({nid:"7775675477949841201", text:"<=", ntype:NTYPE_CLASS});
        this.XOP_lessthanequal.asdf_date_disp = "on or before";
        this.XOP_greaterthan= this.minBootNeuron({nid:"7775682071577396678", text:">", ntype:NTYPE_CLASS});
        this.XOP_greaterthan.asdf_date_disp = "after";
        this.XOP_greaterthanequal = this.minBootNeuron({nid:"7775684994215435705", text:">=", ntype:NTYPE_CLASS});
        this.XOP_greaterthanequal.asdf_date_disp = "on or after";
        
        this.TERMINATING_TYPE = this.minBootNeuron({nid:"7777852852171534168", text:"TERMINATING TYPE", ntype:NTYPE_CLASS});

        this.prompt = this.minBootNeuron({nid:"7777440647376760473", text:"prompt", ntype:NTYPE_CLASS});
        this.conversion_unit = this.minBootNeuron({nid:"7770720868717377200", text:"conversion unit", ntype:NTYPE_CLASS});
        this.conversion = this.minBootNeuron({nid:"7771003415909957137", text:"conversion", ntype:NTYPE_CLASS});
        //this.symbol = this.minBootNeuron({nid:"7771062370252084332", text:"symbol", ntype:NTYPE_CLASS});
        this.unit_symbol = this.minBootNeuron({nid:"7772145224575039225", text:"symbol", ntype:NTYPE_CLASS});
        this.html_symbol = this.minBootNeuron({nid:"7775192422854000995", text:"html symbol", ntype:NTYPE_CLASS});
        
        //used to tune children of a class (if class permissable, if inst permissable).
        //also used in pattern NOBJ->classesonly/instsonly->TO_TYPE_TXID->TXID OF PARAM
        //to inform ANNE gui that the param intends to connect only to insts or classes..
        this.classes_only = this.minBootNeuron({nid:"7775050740207039709", text:"classes only", ntype:NTYPE_CLASS});
        this.insts_only = this.minBootNeuron({nid:"7778060997442299009", text:"insts only", ntype:NTYPE_CLASS});
        this.no_new_children = this.minBootNeuron({nid:"7771418654491356905", text:"no children", ntype:NTYPE_CLASS});
        this.uo = this.minBootNeuron({nid:"7770928393157957794", text:"unique on", ntype:NTYPE_CLASS});

        this.anndle = this.minBootNeuron({nid:"7772143308845525283", text:"anndle", ntype:NTYPE_CLASS});
        this.CMD_MAKE_SUBANNDLE = this.minBootNeuron({nid:"7777257285707664978", text:"MAKE subanndle", ntype:NTYPE_CLASS});
        this.anndle_lvl = this.minBootNeuron({nid:"7775100368007980447", text:"anndle-lvl", ntype:NTYPE_CLASS});
        this.subanndle = this.minBootNeuron({nid:"7775386871234172677", text:"subanndle", ntype:NTYPE_CLASS});
        //pts up to parent anndle/subanndle (lvl-1)
        this.subanndle_of = this.minBootNeuron({nid:"7775674902351449345", text:"subanndle of", ntype:NTYPE_CLASS});
        
        
        //anndle resolution now is equal to the sponsor of the ANNDLE.
        //anndles CANNOT be annexed.
        this.anndle_sellback = this.minBootNeuron({nid:"77793877955602199", text:"sellback anndle", ntype:NTYPE_CLASS});
        this.anndle_renew = this.minBootNeuron({nid:"7777763623704051918", text:"renew anndle", ntype:NTYPE_CLASS});

        this.anndle_offer = this.minBootNeuron({nid:"7775310066328986130", text:"anndle offer", ntype:NTYPE_CLASS});
        this.anndle_offer_refund = this.minBootNeuron({nid:"7777288463269446287", text:"anndle offer", ntype:NTYPE_CLASS});
        this.anndle_accept_offer = this.minBootNeuron({nid:"7771162452567479785", text:"accept anndle offer", ntype:NTYPE_CLASS});
        this.anndle_reject_offer = this.minBootNeuron({nid:"7778362583058494579", text:"reject anndle offer", ntype:NTYPE_CLASS});
        this.anndle_revoke_offer = this.minBootNeuron({nid:"7777984968890672286", text:"revoke anndle offer", ntype:NTYPE_CLASS});
        this.anndle_offer_nullified = this.minBootNeuron({nid:"7775865098209225416", text:"anndle offer nullified", ntype:NTYPE_CLASS});
        
        this.resolves_to_anneid= this.minBootNeuron({nid:"7770557671924077616", text:"resolves to anneid", ntype:NTYPE_CLASS});
        this.resovles_to_uri = this.minBootNeuron({nid:"7778625875487717827", text:"resolves to uri", ntype:NTYPE_CLASS});
        this.resolves_to_ipv4 = this.minBootNeuron({nid:"7770701407567396183", text:"resolves to ipv4", ntype:NTYPE_CLASS});
        this.resolves_to_ipv6 = this.minBootNeuron({nid:"7772083318109703659", text:"resolves to ipv6", ntype:NTYPE_CLASS});

        this.tense = this.minBootNeuron({nid:"7774888866193797726", text:"tense", ntype:NTYPE_CLASS});
        this.tense_present = this.minBootNeuron({nid:"7775362585331918618", text:"present tense", ntype:NTYPE_CLASS});
        this.tense_past = this.minBootNeuron({nid:"7772326017534981884", text:"past tense", ntype:NTYPE_CLASS});
        this.tense_future = this.minBootNeuron({nid:"7770373507125829431", text:"future tense", ntype:NTYPE_CLASS});
        
        //(action verbs)
        this.action_give = this.minBootNeuron({nid:"77705947298070034", text:"give", ntype:NTYPE_CLASS});
        this.action_receive = this.minBootNeuron({nid:"7773006403613657951", text:"receive", ntype:NTYPE_CLASS});
        this.action_delegate = this.minBootNeuron({nid:"7773093116715894615", text:"delegate", ntype:NTYPE_CLASS});
        
        //now is a base identity type..
        this.anorg = this.minBootNeuron({nid:"7777778897441556324", text:"Anorg", ntype:NTYPE_CLASS});
        
        this.description = this.minBootNeuron({nid:"7779025607680638924", text:"description", ntype:NTYPE_CLASS});
        this.class_desc = this.minBootNeuron({nid:"7777732071930766985", text:"class description", ntype:NTYPE_CLASS});
        
        this.anorg_type = this.minBootNeuron({nid:"777725395794844067", text:"anorg type", ntype:NTYPE_CLASS});
        
        //for payouts of keyless neuFrons.
        this.governance = this.minBootNeuron({nid:"7771671873508876461", text:"governance", ntype:NTYPE_CLASS});
        this.governance.ai_ico_class="ai-anjucator";

        this.gov_dictatorship = this.minBootNeuron({nid:"7779412447342618259", text:"dictatorship", ntype:NTYPE_CLASS});
        this.gov_anj_majority = this.minBootNeuron({nid:"777068234629750280", text:"anjucator majority", ntype:NTYPE_CLASS});
        this.gov_anj_super_majority = this.minBootNeuron({nid:"777771743001523175", text:"anjucator super majority", ntype:NTYPE_CLASS});
        this.gov_anj_jury = this.minBootNeuron({nid:"777132582625210184", text:"anjucator jury", ntype:NTYPE_CLASS});
          
        
        this.DIG = this.minBootNeuron({nid:"7778765030709506308", text:"DIG", ntype:NTYPE_CLASS});
        this.annies_per_aon = this.minBootNeuron({nid:"7774922870627231232", text:"annies per aon", ntype:NTYPE_CLASS});

        this.MAKE_CUSTODIAL = this.minBootNeuron({nid:"7778593465311902303", text:"make custodial", ntype:NTYPE_CLASS});
        
        /***** ASDF ************************************************/
        //ASDF is like bounty neurons.  They will be defined, built in the user's worldview and then FROZEN
        //Only a FROZEN ASDF can accept cmds.
        //Funds will be locked in the ASDF neuron.
        this.ASDF = this.minBootNeuron({nid:"7774123821045389572", text:"ASDF", ntype:NTYPE_CLASS});
        this.ASDF.wvBIPASS=true; //we accept other worldviews as default view for ASDFs (ANNE doesn't adopt them)
        this.ASDF_ORACLE = this.minBootNeuron({nid:"7772363726012774091", text:"ASDF-ORACLE", ntype:NTYPE_CLASS});                       //RELN specifying the oracle of an ASDF neuron
        //RELN SIGNAL tht the future event is underway and futures cannot be accepted.
        this.ASDF_ORACLE_NO_MORE_TAKERS = this.minBootNeuron({nid:"777332540881126681", text:"ASDF-NO-MORE-TAKERS", ntype:NTYPE_CLASS});           
        
        this.ASDF_ORACLE_RELN = this.minBootNeuron({nid:"7775619843467308405", text:"ASDF-RELN", ntype:NTYPE_CLASS});                    //the reln of the oracle that will trigger the AUTO SETTLEMENT
        this.ASDF_FUTURE_TO = this.minBootNeuron({nid:"7775674213293630742", text:"ASDF-FUTURE-TO", ntype:NTYPE_CLASS});
        this.ASDF_REWARD = this.minBootNeuron({nid:"7772537879312541472", text:"ASDF-REWARD", ntype:NTYPE_CLASS});           //always relative to 1.
        this.ASDF_STATUS =  this.minBootNeuron({nid:"7776761752802692203", text:"ASDF-STATUS", ntype:NTYPE_CLASS});
        this.ASDF_EXPIRE_AT_BLOCK =  this.minBootNeuron({nid:"7774654388305965960", text:"ASDF-EXPIRE-AT-BLOCK", ntype:NTYPE_CLASS});
        this.ASDF_EXPIRE_AT_BLOCK.ai_ico_class="ai-clock";

        this.ASDF_RISK_FUNDED =  this.minBootNeuron({nid:"7772927452444197703", text:"ASDF-RISK-FUNDED", ntype:NTYPE_CLASS});
        this.ASDF_RISK_REFUNDED =  this.minBootNeuron({nid:"7779055060461771493", text:"ASDF-RISK-REFUNDED", ntype:NTYPE_CLASS});

        this.ASDF_STATUS_EXPIRED =  this.minBootNeuron({nid:"7775414101854878186", text:"ASDF status expired", ntype:NTYPE_CLASS});
        this.ASDF_STATUS_SETTLED =  this.minBootNeuron({nid:"7777642930344078665", text:"ASDF status settled", ntype:NTYPE_CLASS});
        this.ASDF_STATUS_OPEN =  this.minBootNeuron({nid:"7771432588868502834", text:"ASDF status open", ntype:NTYPE_CLASS});
        this.ASDF_STATUS_FULL =  this.minBootNeuron({nid:"777404000421391248", text:"ASDF status full", ntype:NTYPE_CLASS});
        this.ASDF_STATUS_SETTLEMENT_ERR =  this.minBootNeuron({nid:"7773494473412466357", text:"ASDF settlement err", ntype:NTYPE_CLASS});
        this.ASDF_STATUS_NOMORE_RISK = this.minBootNeuron({nid:"7777584087450006238", text:"ASDF status owner cutoff", ntype:NTYPE_CLASS});           
       
        //set to max available risk reamining w/ respect to reward ratio
        //FUTURE TO is the condition which pays out the event reward.
        //If FUTURE TO is not observed on the ORACLE RELN (and something else is), 
        //the funds are distributed to the ASDF owner.
        this.ASDF_RISK_REMAINING = this.minBootNeuron({nid:"7777229642227033086", text:"ASDF-RISK-REMAINING", ntype:NTYPE_CLASS});  
        this.ASDF_FUTURE_TO_CDX = this.minBootNeuron({nid:"7778014966634263522", text:"ASDF-FUTURE-TO-CDX", ntype:NTYPE_CLASS});  
        this.ASDF_FUTURE_TO_CDX_NONANNE_VAL = this.minBootNeuron({nid:"777791626878838486", text:"ASDF-FUTURE-TO-NONANNE-CDX-VAL", ntype:NTYPE_CLASS});  
        this.ASDF_FUTURE_TO_ANNE_CDX_ANNE_DATE = this.minBootNeuron({nid:"7775993685838455770", text:"ASDF-FUTURE-TO-ANNE-CDX-ANNE-DATE", ntype:NTYPE_CLASS});  
        this.ASDF_FUTURE_TO_NID_ATTR = this.minBootNeuron({nid:"7773689919601064934", text:"ASDF-FUTURE-TO-NID-ATTR", ntype:NTYPE_CLASS});  

        //user taking on the risk specified in the ASDF
        this.ASDF_ACCEPT_RISK = this.minBootNeuron({nid:"7777809231670492713", text:"ASDF-ACCEPT-RISK", ntype:NTYPE_CLASS});  
        this.ASDF_ADD_RISK = this.minBootNeuron({nid:"777390341076576182", text:"ASDF-ADD-RISK", ntype:NTYPE_CLASS});  

        this.ASDF_RESULT = this.minBootNeuron({nid:"7772822327738373678", text:"ASDF-RESULT", ntype:NTYPE_CLASS});  
        this.ASDF_RESULT_FUTURE_OBSERVED = this.minBootNeuron({nid:"7770617469176432515", text:"ASDF-RESULT-FUTURE-OBSERVED", ntype:NTYPE_CLASS});  
        this.ASDF_RESULT_FUTURE_NOT_OBSERVED = this.minBootNeuron({nid:"7770390220377781103", text:"ASDF-RESULT-FUTURE-NOT-OBSERVED", ntype:NTYPE_CLASS});  

        //annode managed neuron to hold block triggers.
        this.ANNE_BLOCK_TRIGGER = this.minBootNeuron({nid:"7770514447679553139", text:"Block Trigger", ntype:NTYPE_CLASS});  

        //these relons are only put in local RAM, never stored in the IDB tbls --- as they are temp (but on the ANNODE)
        //and reflect an intra-block state change pending.  We don't want them stored as they will muck about when they
        //linger and are no longer accurate statuses.  The aplex obj will pull a .has_pending boolean for quicker access.
        this.INTRA_BLOCK_PENDING = this.minBootNeuron({nid:"7772178508525445062", text:"INTRA BLOCK PENDING", ntype:NTYPE_CLASS});  

        this.REDLIST = this.minBootNeuron({nid:"7773260417288127971", text:"REDLIST", ntype:NTYPE_CLASS});  
        
        /**************************************************/

        this.BLOCK_WINNER  = this.minBootNeuron({nid:"7776164025119639648", text:"Block Winner", ntype:NTYPE_CLASS});  
        this.BLOCK_SHARE_WINNER  = this.minBootNeuron({nid:"777072988166742136", text:"Block Share Winner", ntype:NTYPE_CLASS});  
        
        this.achat = this.minBootNeuron({nid:"7772867043812936263", text:"A𝄐chat", ntype:NTYPE_CLASS});
        
        this.bounty = this.minBootNeuron({nid:"7772768880266899814", text:"bounty", ntype:NTYPE_CLASS});
        this.bounty.wvBIPASS=true;
        this.bounty_deadline = this.minBootNeuron({nid:"777443587645819431", text:"bounty deadline", ntype:NTYPE_CLASS});
        
        //bounty types:
        //anyone chase it, complete it, claim it
        //this.bounty_public_race = this.minBootNeuron({nid:"777283655934389925", text:"claim when complete", ntype:NTYPE_CLASS});
        
        //potential bounty shows interest
        //contact offline (later on p2p chat)
        //vendor selected (too big of a project to tackle w/o transparency of selected vendor)
        //this.bounty_vendor_select = this.minBootNeuron({nid:"7779289286905306093", text:"vendor select", ntype:NTYPE_CLASS});

        //the ANNEID chosen to take on the bounty.
        //they must complete the bounty to get the payout.
        //this.bounty_vendor = this.minBootNeuron({nid:"7770289084266469525", text:"vendor", ntype:NTYPE_CLASS});
        
        //for both types, this is the official payment.
        this.bounty_payout = this.minBootNeuron({nid:"7779986301683332147", text:"bounty payout", ntype:NTYPE_CLASS});

        //make a public declaration you have completed a bounty.
        //if the bounty has specified a chatnode, the request will auto switch to the 
        //chat annode, handshake, and create a chat session that can be accessed by the anjucators.
        //there things can be discussed p2p.. (user->annode annode->user)
        //The shared keys are used (not signing/agreement) as to not leak data.
        //A new PGP key could be created for this purpose but as of now is not.
        //The chat is NOT put on-chain.  It is local ANNODE encrypted chat so it is wise
        //for the bounty creator to select an ANNODE they trust or they own.
        this.bounty_claim = this.minBootNeuron({nid:"7777462743773526396", text:"bounty claim", ntype:NTYPE_CLASS});

        //a clump class which already has 2 roles defined..
        //instantiate this and provide the missing 3 roles (of the params) to create a 
        //delegation request.
        this.delegateDistVoteClump =this.minBootNeuron({nid:"7776157359286235395", text:"delegate distribution voting rights", ntype:NTYPE_CLASS});
        
        this.anjucate_vote = this.minBootNeuron({nid:"7771703615799316076", text:"anjucate vote", ntype:NTYPE_CLASS});
        this.anjucate_vote.ai_ico_class="ai-anjucator";
        this.anj_vote_yay = this.minBootNeuron({nid:"7775592309515630471", text:"vote - yay", ntype:NTYPE_CLASS});
        this.anj_vote_nay = this.minBootNeuron({nid:"7770525053411833701", text:"vote - nay", ntype:NTYPE_CLASS});
        this.anj_vote_abstain = this.minBootNeuron({nid:"777916843002801549", text:"vote - abstain", ntype:NTYPE_CLASS});
        this.anj_vote_hasnot = this.minBootNeuron({nid:"7778382797157395681", text:"vote - has not voted", ntype:NTYPE_CLASS});
        
        this.anj_invitation = this.minBootNeuron({nid:"7774437122736557228", text:"anjucator invitation", ntype:NTYPE_CLASS});
        
        //delegate your vote
        this.anj_delegate = this.minBootNeuron({nid:"777888550313830205", text:"anjucate - delegate vote", ntype:NTYPE_CLASS});
        
        this.vote_relns = [this.anj_vote_yay.temp_nid, this.anj_vote_nay.temp_nid, 
                        this.anj_vote_abstain.temp_nid, this.anj_vote_hasnot.temp_nid];
        this.valid_vote = [this.anj_vote_yay.temp_nid, this.anj_vote_nay.temp_nid, 
                            this.anj_vote_abstain.temp_nid];
    
        //helpful calc stored on the vote neuron.
        this.anj_votes_to_pass = this.minBootNeuron({nid:"7775094174782013630", text:"anj votes to pass", ntype:NTYPE_CLASS});
        this.vote_result = this.minBootNeuron({nid:"7770885647556226532", text:"vote result", ntype:NTYPE_CLASS});
        this.vote_passed = this.minBootNeuron({nid:"7771391628585996940", text:"vote passed", ntype:NTYPE_CLASS});
        this.vote_didnt_pass = this.minBootNeuron({nid:"7773739150549833304", text:"vote did not pass", ntype:NTYPE_CLASS});
        
        //distro requests  (call for a vote)
        this.distro_req = this.minBootNeuron({nid:"7778840860841443809", text:"distro request", ntype:NTYPE_CLASS});
        //this.per_d_vote_by_hodler = this.minBootNeuron({nid:"7773920230765932714", text:"% distro vote by hodler", ntype:NTYPE_CLASS});
        //this.per_d_vote_by_list = this.minBootNeuron({nid:"7776254538552572051", text:"% distro vote by list", ntype:NTYPE_CLASS});
        //this.specific_d_vote_by_hodler = this.minBootNeuron({nid:"7772460455121008551", text:"specific distro vote by hodler", ntype:NTYPE_CLASS});
        this.specific_d_vote_by_list = this.minBootNeuron({nid:"7770256274602944478", text:"specific distro vote by list", ntype:NTYPE_CLASS});
        
        this.mr_scatman = this.minBootNeuron({nid:"777776301542347271", text:"Mr.Scatman", ntype:NTYPE_CLASS});
        this.johnny5 = this.minBootNeuron({nid:"7777798716640079675", text:"johnny#5", ntype:NTYPE_CLASS});
        this.nikt = this.minBootNeuron({nid:"7777111323427337713", text:"nikt", ntype:NTYPE_CLASS});
        
        this.distro_src = this.minBootNeuron({nid:"7776409001442599365", text:"distro source", ntype:NTYPE_CLASS});
        this.distro_to_whom = this.minBootNeuron({nid:"7776415552030872607", text:"benefactor", ntype:NTYPE_CLASS});
        this.distro_quantity = this.minBootNeuron({nid:"7773018074680059484", text:"distro quantity", ntype:NTYPE_CLASS});
        this.distro_deadline = this.minBootNeuron({nid:"7774366570419816048", text:"distro deadline", ntype:NTYPE_CLASS});
        this.distro_block_deadline = this.minBootNeuron({nid:"7773505632843152624", text:"distro block deadline", ntype:NTYPE_CLASS});
        this.distro_quantity_type = this.minBootNeuron({nid:"77719030987172936", text:"distro quantity type", ntype:NTYPE_CLASS});
        this.distro_qt_percent = this.minBootNeuron({nid:"7774613566577121605", text:"distro qt percent", ntype:NTYPE_CLASS});
        this.distro_qt_specific = this.minBootNeuron({nid:"7772653274078556397", text:"distro qt specific", ntype:NTYPE_CLASS});
        this.distro_to_type = this.minBootNeuron({nid:"7770704584830358901", text:"distro to type", ntype:NTYPE_CLASS});
        this.distro_tt_hodler = this.minBootNeuron({nid:"7779256405810042395", text:"distro to type hodler", ntype:NTYPE_CLASS});
        this.distro_tt_list = this.minBootNeuron({nid:"7776184053255403454", text:"distro to type list", ntype:NTYPE_CLASS});

        //used to call a vote from a self-neuron, with reln ADV, pointing to a valid distro req nid.
        this.distro_vote = this.minBootNeuron({nid:"7772171285820605524", text:"Annorg distribution vote", ntype:NTYPE_CLASS});
        this.anjucator = this.minBootNeuron({nid:"7775009149036337280", text:"anjucator", ntype:NTYPE_CLASS});
        this.anjucator.ai_ico_class="ai-anjucator";

        this.AOE = this.minBootNeuron({nid:"7774534916375278299", text:"asset of exchange", ntype:NTYPE_CLASS});
        
        //no collab on these when collab is available.
        this.admin_tnids = [];
        this.admin_tnids.push(this.garden_type.temp_nid);
        this.admin_tnids.push(this.inst_garden_type.temp_nid);
        this.admin_tnids.push(this.let_wv.temp_nid);
        this.admin_tnids.push(this.deny_wv.temp_nid);
        this.admin_tnids.push(this.viplet.temp_nid);
        this.admin_tnids.push(this.dr_andy_defn.temp_nid);
        this.admin_tnids.push(this.dr_andy_hanai_in.temp_nid);
        this.admin_tnids.push(this.dr_andy_super_class.temp_nid);
        this.admin_tnids.push(this.dr_andy_super_insts.temp_nid);
        this.admin_tnids.push(this.dr_andy_super_all.temp_nid);
        this.admin_tnids.push(this.dr_permitted_reln.temp_nid);
        this.admin_tnids.push(this.gt_permit_reln.temp_nid);
        this.admin_tnids.push(this.garden_rule.temp_nid);
        

        /******************************************/
        this.HARDCODED_BOOT_SIZE = this._aNeurons.length;
        /******************************************/

        this.USER = ANNE.USER;
        
        this.annovice_tnids = [];

        //make poa (request to annode)
        //->poa-> (burned by ANNODE only) 
        this.nodecmds_tnids = [];

        this.pdnactions_tnids = [
            this.PDN_OP_REVEAL_ALL.temp_nid,
            this.XOP_equal.temp_nid,
            this.XOP_notequal.temp_nid,
            this.XOP_lessthan.temp_nid,
            this.XOP_lessthanequal.temp_nid,
            this.XOP_greaterthan.temp_nid,
            this.XOP_greaterthanequal.temp_nid
        ];
        this.pdnOpTnids =[this.XOP_equal.temp_nid, 
            this.XOP_notequal.temp_nid,
            this.XOP_lessthan.temp_nid,
            this.XOP_lessthanequal.temp_nid,
            this.XOP_greaterthan.temp_nid,
            this.XOP_greaterthanequal.temp_nid];

        this._bootupCompleteCallback = $bootupCompleteCallback;

        //Some ez references so I don't have to type things out so long
        //wen referencing them.
        if (typeof MEMP != "undefined") {
            this.ANNE.MEMP = MEMP;
            this.MEMP = MEMP
            this.MEMP.init(MEM);
            this.ANNE.MEMP = MEMP;
        }
        if (typeof MEMG != "undefined") {
            this.MEMG = MEMG;
            this.MEMG = MEMG;
            this.MEMG.init(MEM);
            this.ANNE.MEMG = MEMG;
        }
        if (typeof MEMZ != "undefined") {
            this.ANNE.MEMZ = MEMZ;
            this.MEMZ = MEMZ;
            this.MEMZ.init(MEM);
            this.ANNE.MEMZ = MEMZ;
        }

        //set alt ptrs in memp/memg/memz
        if (this.MEMP) {
            if (this.MEMG) this.MEMP.MEMG=this.MEMG;
            if (this.MEMZ) this.MEMP.MEMZ=this.MEMZ;
        }
        if (this.MEMG) {
            if (this.MEMP) this.MEMG.MEMP=this.MEMP;
            if (this.MEMZ) this.MEMG.MEMZ=this.MEMZ;
        }
        if (this.MEMZ) {
            if (this.MEMG) this.MEMZ.MEMG=this.MEMG;
            if (this.MEMP) this.MEMZ.MEMP=this.MEMP;
        }

        if (this.ANNE.GUI_STUFF!=null) {
            this.ANNE.GUI_STUFF.init(this.ANNE, true);
        }

        this._initDBs(this.continueBootUp); //no need to await it.
    }


    getNeuronByNid = function($nid) {
        if (ANNE.MEMG!=null) {
            console.warn("user MEMG, not MEM for getNeuronByNid");
            return ANNE.MEMG.getNeuronByNid($nid);
        }
    }

    printPreloadBuilder = function() {

    }


    /**
     * The 3 top admins in ANNE which assign admin rights to the users 
     * so that they can manage ANNE admin duties in various areas of the system.
     * @param {*} $nid 
     * @returns 
     */
    isANNEADMIN=function($nid) {
        if (this.ANNE_SUPERS.indexOf($nid)!=-1) return true;
        return false;
    }

    
    /**
     * Init local DB
     * This will create the indexDB for persistent local memory storage.  
     * If it already exists it will load it in to local RAM. 
     * 
     * @param {function} - callback when done
     * @author {Mr.Scatman}     
     *            
     */              
     _initDBs = async($doneCallback=null) => {

        //if not INDEXDB available, just call the doneCallback straightaway
        //nothing should be awaiting this function since it uses the callbacks.
        if (!('indexedDB' in window)) {
            console.log('This browser doesn\'t support IndexedDB');
            ANNE.DB.bSUPPORTED=false;
            if ($doneCallback!=null) $doneCallback();   //give back control of thread to callback.
            return;
        }

        //if we do have INDEXDB supported, we will call the _finishPreInit function
        //and pass throught he donecallback there.
        else {	
            ANNE.DB.bSUPPORTED=true;

            var result = await ANNE.DB.___init();

            if (result==null) return;

            if (result.status!=undefined && result.status=="no-node") {}
            if (ANNE.SETTINGS && ANNE.SETTINGS.debug()==true) console.log(result);

            var rez = result.results;
            this._finishPreInit(rez.relons, rez.history, rez.settings, $doneCallback); 
        }
        return;
    }

    
    /**
     * aSettings are saved settings local that can be pre-loaded back into
     * the app. you likely have a way to handle this alreayd in the 
     * phone/backoffice.     
     * 
     * @param {array} - an array of relons extracted from indexeddb memory.
     * @param {array} - an array of annode api use history (for optimizations).
     * @param {array} - an array of local (non-saved-onchain OR saved on-chain) settings.
     * @param {function} - callback when done
     * @author {Mr.Scatman}      
     *                           
     */
     _finishPreInit=async($aRelons, $aFullPullHistory, $aSettings, $doneCallback=null)=> {
        
        //pre-load our query history (for optimizing future queries)
        if ($aFullPullHistory!==null) {
            if (this.ANNE.HISTORY!=null) this.ANNE.HISTORY.processPullHistory($aFullPullHistory);
        }

        //pre-load all our graph edges (these are the ANNE TXs we broadcast which made it to the ANNE middle memory layer in Mongo)
        if ($aRelons!=null && $aRelons.length) {
            await MEMP.santizeSavedRelonsForReprocessing($aRelons);
            var rRETURNALL=false;
            var bDOLOCALSAVE=false;

            //do not try to store these in local memory, as we just pulled them from where we store them!
            if ($aRelons.length) {

                //silly test -->
                await MEMP.processNeuronsFromRelons($aRelons, rRETURNALL, bDOLOCALSAVE, "relons", "_finishPreInit");
            }
        }

        //process the ANNE default settings here.
        if ($aSettings.length) {
            if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.aSettingsToProcess = $aSettings;
        }

        if ($doneCallback!=null) $doneCallback();
        return;
    }


    /** 
     * Load up required min memory from the node if we haven't yet.
     *  
     * @author {Mr.Scatman} 
     */
    continueBootUp=async()=> {
                
        //check memory length.  If none, do the preload.
        //we will have MORE neurons than JS preload if we've pulled preload from ANNODE.
        if (this._aNeurons.length<=this.HARDCODED_BOOT_SIZE) {
            //load local memory first.
            //if none, load min ANNODE memory.
            DB.DO_THE_ONETIME_PRELOAD=true;
            await ANNE.API.getAllPreloads();
            DB.DO_THE_ONETIME_PRELOAD=false;
        }

        else {
            if (ANNE.AWARE) await ANNE.AWARE.init(); //this will require some nuanced handling.
        }

        await this.specialsBuild();

        //we are done with bootup.
        //we can now safely check localStorage for login info.
        if (this._bootupCompleteCallback!=null) {
            this._bootupCompleteCallback();
        }
        return;
    }


    /** 
     * Wire up important functions and relons.
     * 
     * In past version of the memory manager we preloaded hundreds and even
     * thousands of neurons and relons in files.  This was helpful for testing.
     * In those files we were able to wire up things to setup some functionality
     * of the memory handler because we had direct access to variables to do so.
     * 
     * What we have not done in the preloaded relons we need to do here now..
     * at bootup.
     * 
     * Note: we can still load supporting .js files that have extended helper
     * functions will can be used in conjunction with these dynamic module mem loads.
     * This will add to the module if it is found in this.
     * 
     * NOTE: THIS WILL ONLY WORK IF ALL THESE NEURONS ARE PRESENT AT LOADUP!!!!
     * 
     * @author {Mr.Scatman} 
     */
    specialsBuild = async()=> {

        //ASCII
        //ascii_code_num
        //ascii_eval
        var _n0 = MEMG.getNeuronByNid("7772759491040052489");
        var _n1 = MEMG.getNeuronByNid("7779936209861264085");
        var _n2 = MEMG.getNeuronByNid("7777466399639974815");
        var _n3 = MEMG.getNeuronByNid("7770513775335783280");
        var _n4 = MEMG.getNeuronByNid("7779613874588391142");
        var _n5 = MEMG.getNeuronByNid("7773838259673220295");
        var _n6 = MEMG.getNeuronByNid("7772956192305001444");
        var _n7 = MEMG.getNeuronByNid("7779032678869944184");
        var _n8 = MEMG.getNeuronByNid("7777087244990101066");
        var _n9 = MEMG.getNeuronByNid("7778263312619963662");


        var numHack = [_n0, _n1, _n2, _n3, _n4, _n5, _n6, _n7, _n8, _n9];

        await this.buildAndInjectMod(this.ASCII, "ASCII");

        //nums are be classed as integers and i didn't do a 
        //BE PARENT / BE ASCII for this heirarchy.
        for (var nhack=0; nhack < numHack.length; nhack++) {
            var numNOBJ = numHack[nhack];
            if (numNOBJ==null) continue;
            this.ASCII["p_"+(nhack+48)] = numNOBJ;
            numNOBJ.charCode=(nhack+48)
            numNOBJ.char=numNOBJ.text;
            numNOBJ.pk2_name="char";
            numNOBJ.pk2_val=numNOBJ.text;
            numNOBJ.pk_name="charCode";
            numNOBJ.pk_val=(nhack+48);
        }

        await this.doLangStuff();

        //let's do the creative countries..
        await this.buildAndInjectMod(this.country, "COUNTRY");

        //await this.buildAndInjectMod(this.state, "STATE");
        this.STATE = {"status":"pre-init", "name":"STATE"};
        this.STATE._aChildren=this.getChildrenDynamic(this.state);
        
        for (var a=0; a < this.STATE._aChildren.length;a++) {
            var stateNOBJ = this.STATE._aChildren[a];
            if (stateNOBJ.aka!=null && stateNOBJ.aka.length>0) {
                for (var ak=0; ak < stateNOBJ.aka.length; ak++) {
                    if (stateNOBJ.aka[ak].length==2) this.STATE[ stateNOBJ.aka[ak] ] = stateNOBJ;
                }
            }
        }

        await this.buildAndInjectMod(this.month, "MONTH");

        var months = [this.month_jan, this.month_feb, this.month_mar, this.month_apr, this.month_may, this.month_jun,
                    this.month_jul, this.month_aug, this.month_sep, this.month_oct, this.month_nov, this.month_dec];
        for (var m=0; m < months.length; m++) {
            var monthNOBJ = months[m];
            var str = monthNOBJ.text.toLowerCase();
            this._monthMap[ str ] = monthNOBJ;
            this._monthLookup.push(str);
            var str = str.substring(0,3);
            this._monthMap[ str ] = monthNOBJ;
            this._monthLookup.push(str);
            this._monthMap[ (m+1) ] = monthNOBJ;
        }
        return;
    }

    doLangStuff = async()=>{
        //let's do a language module.
        await this.buildAndInjectMod(this.language, "LANG");
        MEM.LANG.urange = [];

        //set the urange based on known unicode start/ends
        for (var c=0; c<MEM.LANG._aChildren.length; c++) {
            var childNOBJ= MEMG.getNeuronByTempNid(MEM.LANG._aChildren[c]);
            var rel_u_start = MEMG.getRelonByTempR(childNOBJ.outs, MEM.unicode_start.temp_nid);
            if (rel_u_start!=null) {
                var rel_u_end = MEMG.getRelonByTempR(childNOBJ.outs, MEM.unicode_end.temp_nid);
                if (rel_u_end!=null) {
                    MEM.LANG.urange.push({s:rel_u_start.to_s, e:rel_u_end.to_s, lang:childNOBJ.temp_nid, lang_debug:childNOBJ.text});
                }
            }
        }

        //sort the urange by start val
        MEM.LANG.urange.sort( function($a, $b) {
            if ($a.s > $b.s) return 1;
            else if ($a.s < $b.s) return -1;
            return 0;
        });

        //ascertain lang script by ucode
        MEM.LANG.ascertain_lang=function($str, $maxbreak=5){
            var tally = [];
            var max = 0;
            var winning = null;
            for (var s=0; s < $str.length; s++) {
                var test = $str[s].charCodeAt(0);
                for (var u=0; u < ANNE.MEM.LANG.urange.length; u++) {
                    var ur = MEM.LANG.urange[u];
                    if (test >= ur.s && test <= ur.e) {
                        if (tally[ur.lang]==undefined) tally[ur.lang]=1;
                        else tally[ur.lang]+=1;
                        if (tally[ur.lang]>max) {
                            max = tally[ur.lang];
                            winning = MEMG.getNeuronByTempNid(ur.lang);
                        }
                        if (max>=$maxbreak) return winning;
                        break;
                    }
                }
            }
            return winning;
        }
    }


    /**
    * USED FOR MAKING NEW CLASSES AND NEURONS ----> 
    * THIS IS FOR KEY'D NEURONS (OR GHOSTS), NOT KEYLESS.
    * WE MAKE THEM REAL HERE, WITH A SEED, 
    *
    * DO NOT CREATE TEMP NEURONS YOU DO NOT INTEND TO KEEP AROUND.  
    * IF U STORE RELONS ON-CHAIN WITH THESE TEMP NEURONS THEY GO AWAY 
    * AND THEN YOU HAVE ORPHANED RELONS. 
    * @param {string}
    * @param {int}
    * @param {string}
    * @param {anne_neuron}
    * @param {boolean}
    * @param {boolean}
    */
    makeTempNObjOnly = function ($owner, $type, $newText, $parentObj=null, $bPUSH=true){
        if (ANNE.USER==null) {
            console.error("must have a USER object");
            return null;
        }

        var temp_nid = UTILS.randTXID(this._emulatedTXIDLength-2);
        var temp_nObj = new anne_neuron($newText, $type, temp_nid);
        temp_nObj.seed = "";
        temp_nObj.pubKey = "";  
        temp_nObj.ins = [];
        temp_nObj.outs = [];	
        temp_nObj.sponsored_by = "";

        //hash the neuron.
        //mrscatman 04/24/2022 -- monitor this.. could lead to nasty mistakes.
        MEMP.__hashNeuron(temp_nObj);

        if ($owner) temp_nObj.setOwner($owner);

        temp_nObj.temp=true;
        temp_nObj.mine = true;

        if (ANNE.USER!=null && ANNE.USER._userNObj.nid) {
            if(ANNE.USER._userNObj.owns==undefined) ANNE.USER._userNObj.owns=[];
            if (ANNE.USER._userNObj.owns.indexOf(temp_nObj.temp_nid)==-1) ANNE.USER._userNObj.owns.push(temp_nObj.temp_nid);
        }

        if ($type==NTYPE_INST) temp_nObj.bTERMINATING_TYPE=true;
        else if ($parentObj!=null && $parentObj!="") temp_nObj.bTERMINATING_TYPE=$parentObj.bTERMINATING_TYPE;

        if ($parentObj!=null && $parentObj!="" && ($parentObj instanceof anne_neuron) == true) {
            temp_nObj.addUniqueBe_fastIndex($parentObj);
            if ($parentObj.children.indexOf(temp_nObj.temp_nid)==-1) {
                $parentObj.children.push(temp_nObj.temp_nid);
            }

            temp_nObj.__functInherit($parentObj, "mem.makeTempNObjOnly");

            //emulate the BE relon...
            var reln_nid = "";
            if ($type==NTYPE_INST) reln_nid = this.inst.nid;
            else reln_nid = this.class.nid;

            //bPUSH indicates pushing this into the master arrays
            //makeTempRelon=function($fNOBJ, $type, $reln, $to_type, $to_s, $bPUSH=true, $bUNIQUE=true) 
                
            var parentRelon = this.makeTempRelon(temp_nObj, TYPE_BE, reln_nid, TO_TYPE_ANNE, $parentObj.nid, $bPUSH);
            parentRelon.parent = $parentObj.nid;
            parentRelon.owned_by = temp_nObj.owned_by;

            //make a temp be text relon for name!
            if ($newText!="") {
                //console.log("Make a new name relon "+$newText);
                var beTextRelon = this.makeTempRelon(temp_nObj, TYPE_BE, this.name.nid, TO_TYPE_STRING, $newText, $bPUSH);
                if ($parentObj==null || $parentObj.nid==undefined) {
                    console.log("WHAAAAAAAAAAAT?");
                }
                beTextRelon.parent = $parentObj.nid;
                beTextRelon.owned_by = temp_nObj.owned_by;    
                //if (this._EMULATE_BE_TEXT_NEURON==true) {
                    //this.makeAndPushIfUniqueEMULATED($newText, TO_TYPE_STRING, beTextRelon);
                //}
            }
        }

        temp_nObj.ghost=false;

        return temp_nObj;
    }


    /**
     * 
     * @param {*} $owner 
     * @param {*} $type 
     * @param {*} $newText 
     * @param {*} $parentObj 
     * @param {*} $bPUSH 
     * @returns 
     */
    makeGhostNObj = async ($owner, $type, $newText, $parentObj=null, $bPUSH=true)=> {
        if (ANNE.USER==null) {
            console.error("must have a USER object");
            return null;
        }

        var temp_nid = UTILS.randTXID(this._emulatedTXIDLength-2);
        var temp_nObj = new anne_neuron($newText, $type, temp_nid);
        temp_nObj.seed = "";
        temp_nObj.pubKey = "";  
        temp_nObj.ins = [];
        temp_nObj.outs = [];	
        temp_nObj.sponsored_by = "";

        //hash the neuron.
        //mrscatman 04/24/2022 -- monitor this.. could lead to nasty mistakes.
        MEMP.__hashNeuron(temp_nObj);

        if ($owner) temp_nObj.setOwner($owner);

        temp_nObj.temp=true;
        temp_nObj.mine = true;

        if (ANNE.USER!=null && ANNE.USER._userNObj.nid) {
            if(ANNE.USER._userNObj.owns==undefined) ANNE.USER._userNObj.owns=[];
            if (ANNE.USER._userNObj.owns.indexOf(temp_nObj.temp_nid)==-1) ANNE.USER._userNObj.owns.push(temp_nObj.temp_nid);
        }

        if ($type==NTYPE_INST) temp_nObj.bTERMINATING_TYPE=true;
        else if ($parentObj!=null && $parentObj!="") temp_nObj.bTERMINATING_TYPE=$parentObj.bTERMINATING_TYPE;

        if ($parentObj!=null && $parentObj!="" && ($parentObj instanceof anne_neuron) == true) {
            temp_nObj.addUniqueBe_fastIndex($parentObj);
            if ($parentObj.children.indexOf(temp_nObj.temp_nid)==-1) {
                $parentObj.children.push(temp_nObj.temp_nid);
            }

            temp_nObj.__functInherit($parentObj, "mem.makeGhostNObj");

            //emulate the BE relon...
            var reln_nid = "";
            if ($type==NTYPE_INST) reln_nid = this.inst.nid;
            else reln_nid = this.class.nid; 
                
            var parentRelon = this.makeTempRelon(temp_nObj, TYPE_BE, reln_nid, TO_TYPE_ANNE, $parentObj.nid, $bPUSH);
            parentRelon.parent = $parentObj.nid;
            parentRelon.owned_by = temp_nObj.owned_by;

            //make a temp be text relon for name!
            if ($newText!="") {
                //console.log("Make a new name relon "+$newText);
                var beTextRelon = this.makeTempRelon(temp_nObj, TYPE_BE, this.name.nid, TO_TYPE_STRING, $newText, $bPUSH);
                if ($parentObj==null || $parentObj.nid==undefined) {
                    console.log("WHAAAAAAAAAAAT?");
                }
                beTextRelon.parent = $parentObj.nid;
                beTextRelon.owned_by = temp_nObj.owned_by;    
            }
        }

        for (var g = 0; g < temp_nObj.outs.length; g++) {
            temp_nObj.outs[g].from_text = $newText;
        }
        await GHOSTDATA.localGhostTheRelons(temp_nObj.outs);
        temp_nObj.ghost = true;
        temp_nObj.ghost_owner = USER._userNObj.nid;

        return temp_nObj;
    }


    /** 
     * Create a new mod in MEM, or snag an existing one.
     * 
     * Existing ones could be loaded from supporting files to have
     * supporting code that does complex things.
     * 
     * @param {anne_neuron} - The classNeuron
     * @param {string} - desired prop name in the MEM space for the MOD.  (ie, this.modeName)
     * @author {Mr.Scatman} 
     * 
     */
    buildAndInjectMod = async($classNObj, $modName="")=> {
        if (($classNObj instanceof anne_neuron) == false) return;
        var MOD = null;
        if ($modName=="") {
            MOD={"status":"pre-init"};
            var modName = $classNObj.text.trim().toUpperCase();
            modName.replace(/[^a-z]/gi, '');
            MOD.name = modName;
            MEM[modName]=MOD;
            MOD._aChildren=[];
        }
        else {
            MOD = {"status":"pre-init", "name":$modName};
            MEM[$modName]=MOD;
            MOD._aChildren=[];
        }
        
        //this feature is ONLY intended for classes that have specified
        //a priary key.  We will use this primary key as the var name
        //in the mod.

        //we should expect to see "iso_639_1" for the language class.
        var pk = $classNObj.pk;
        if (pk==undefined) return;
        var pk_param_name = "";
        var pk2_param_name = "";
        var pkNObj = MEMG.getNeuronByTempNid(pk);
        if (pkNObj!=null) {
            var param_name_relon = MEMG.getParamNameRelon(pkNObj.outs);
            if (param_name_relon!=null) {
                pk_param_name = param_name_relon.to_s;
                //pk_be_lookup = "be_"+pkNObj.temp_nid;
            }
            else return;
        }
        else return;

        var pk2 = $classNObj.pk2;
        if (pk2!=undefined) {
            if(Array.isArray(pk2)==true) {
                for (var a=0; a < pk2.length; a++) {
                    var pk2NObj = MEMG.getNeuronByTempNid(pk2[a]);
                    if (pk2NObj!=null) {
                        var param_name2_relon = MEMG.getParamNameRelon(pk2NObj.outs);
                        if (param_name2_relon!=null) {
                            pk2_param_name = param_name2_relon.to_s;
                            break;
                        }
                    }   
                }
            }
                
            else {
                var pk2NObj = MEMG.getNeuronByTempNid(pk2);
                if (pk2NObj!=null) {
                    var param_name2_relon = MEMG.getParamNameRelon(pk2NObj.outs);
                    if (param_name2_relon!=null) {
                        pk2_param_name = param_name2_relon.to_s;
                    }
                }   
            }
        }
 
        //alt-be quick props..
        var childrenNOBJS = this.getChildrenDynamic($classNObj);
       
        for (var c=0; c < childrenNOBJS.length; c++) {
            var childNOBJ = childrenNOBJS[c];
            MOD._aChildren.push(childNOBJ.temp_nid);
            var pk_quick= childNOBJ[pk_param_name];
            if (pk_quick==undefined) continue;
            MOD['p_'+pk_quick] = childNOBJ;
            childNOBJ["pk_val"] = pk_quick;
            childNOBJ["pk_name"] = pk_param_name;
            if (pk2_param_name!="") {
                var pk2_quick= childNOBJ[pk2_param_name];
                childNOBJ["pk2_val"] = pk2_quick;
                childNOBJ["pk2_name"] = pk2_param_name;
                MOD['p_'+pk2_quick] = childNOBJ;
                var val = parseInt(childNOBJ.value);
            }
        }

        MOD.status="init";
        if (ANNE.SETTINGS.bCONSOLE_DEBUG==true) console.log("** LOADED dynamic MOD **"+$modName, MOD);


    }


    /** 
     * @author {Mr.Scatman} 
     */
     init=async()=> {
    }


    /** 
     * Add in the user neuron into mem.
     * 
     * @param {string} - The nid longs string of the neuron
     * @param {boolean} - flag indicating this is the logged-in user's neuron. (they own it)
     * @author {Mr.Scatman} 
     * 
     */
     addInUserNeuron=function($nid, $mine=false) {
        var nObj = MEMG.getNeuronByNid($nid);
        if (nObj!=null) {
            nObj.mine=$mine;
            nObj.ntype = NTYPE_INST;

            //sneaky temp add -- audit what this does when we have an
            //actual REAL bes
            //we know this to be true, but don't know which subclass yet.
            //it is important to denote that this IS an idenity tho. 
            if (nObj.bes==undefined) nObj.bes=[];
            //if (nObj.bes.length==0) nObj.bes.push(this.identity.temp_nid);
            //nObj.be = "identity"; 
        
            if (nObj.adopted_with_trust==undefined) nObj.adopted_with_trust = [];
            return nObj;
        }

        nObj = new anne_neuron("", NTYPE_INST, $nid);
        nObj.adopted_with_trust = [];      //a list of adopted worldviews.
        nObj.mine = $mine;

        //default for any user of ANNE if not more specified
        nObj.setBeTempNid(this.identity.temp_nid);   
        nObj.be = "identity"; 

        return nObj;
    }


    /** 
     * Add to active worldview.
     * 
     * @param {string} - the nid  long string of the worldview we are adding.
     * @author {Mr.Scatman} 
     */
    addToAppWideActiveWorldViews = function ($worldviewTNid) {
        var wv_tnid = parseInt($worldviewTNid);
        if (this._activeWorldViewTempNids.indexOf(wv_tnid)==-1) this._activeWorldViewTempNids.push(wv_tnid);
    }


    /** 
     * Get children depending on how the class was defined.
     * 
     * Some classes do not have hierarchical children.  Other systems
     * incorrectly pigeon whole their data into a locked-in view.
     * ANNE does not.  Something like a country is not actually a child
     * of Country.  It is a location and a Country is a polity.  The child
     * of location points with a reln of Polity to the neuron Country using
     * a TYPE_BE relon.  This is the practical children of country.
     * 
     * Get the right children and return an array.
     * 
     * @param {anne_neuron} $classNObj - The classNeuron
     * @param {Boolean} $bDOTNIDS - return tnids not nObjs
     * @param {Array} $tnids_existing - an array of tnids we already have as children
     * 
     * @author {Mr.Scatman} 
     */
    getChildrenDynamic = function($classNObj, $bDOTNIDS=false, $tnids_existing=[]) {
        var children = [];
        if ($classNObj['be_'+this.BE_CHILDREN.temp_nid]!=undefined) {
            var be_reln = $classNObj['be_'+this.BE_CHILDREN.temp_nid];
            if ($classNObj['be_'+this.BE_PARENT.temp_nid]!=undefined) {
                var parent_restrict = $classNObj['be_'+this.BE_PARENT.temp_nid];
                var relnNOBJ = ANNE.MEMG.getNeuronByTempNid(be_reln);
                //console.warn("CHILDREN ARE (FROM CHILD) TYPE_BE->reln of '"+relnNOBJ.text+"'");
                if (parent_restrict) {
                    var parentNOBJ = ANNE.MEMG.getNeuronByTempNid(parent_restrict);
                    //console.warn("CHILDREN MUST HAVE REAL PARENT OF '"+parentNOBJ.text+"'");
                    if (parentNOBJ.children==null) return children;
                    for (var c=0; c < parentNOBJ.children.length; c++) {
                        var maybeChildNOBJ = MEMG.getNeuronByTempNid(parentNOBJ.children[c]);
                        if (maybeChildNOBJ==undefined || maybeChildNOBJ==null) continue;
                        if (maybeChildNOBJ.enforceGardenUp()==false) continue;
                        var beAltRelon = ANNE.MEMG.getRelonByTempR(maybeChildNOBJ.outs, relnNOBJ.temp_nid);
                        if (beAltRelon!=null && beAltRelon.to_type==TO_TYPE_ANNE && beAltRelon.t==$classNObj.temp_nid) {
                            if ($tnids_existing.length && $tnids_existing.indexOf(maybeChildNOBJ.temp_nid)!=-1) continue;
                            if ($bDOTNIDS==false) {
                                children.push(maybeChildNOBJ);
                            }
                            else {
                                children.push(maybeChildNOBJ.temp_nid);
                            }
                        }
                    }
                }
            }
            else {
                for (var r=0; r < this._aRelons.length; r++) {
                    var rObj = this._aRelons[r];
                    if (rObj.type!=TYPE_BE) continue;
                    if (rObj.r!=this.polity.temp_nid) continue;
                    if (rObj.t!=$classNObj.temp_nid) continue;
                    if ($tnids_existing.length && $tnids_existing.indexOf(rObj.f)!=-1) continue;                    
                    var nObj2 = ANNE.MEMG.getNeuronByTempNid(rObj.f);
                    if (nObj2.enforceGardenUp()==false) continue;
                    if ($bDOTNIDS==false) children.push(nObj2);
                    else children.push(rObj.f);
                }
            }
        }
        
        //standard hierarchical children.
        else {
            if ($classNObj.children==null) return children;
            for (var c=0; c < $classNObj.children.length; c++) {
                if ($tnids_existing.length && $tnids_existing.indexOf($classNObj.children[c])!=-1) continue;
                var childNOBJ = ANNE.MEMG.getNeuronByTempNid($classNObj.children[c]);
                if (childNOBJ.enforceGardenUp()==false) continue;
                if ($bDOTNIDS==false) children.push(childNOBJ);
                else children.push(rObj.f);
            }
        }
        return children;
    }



    /**
     * 
     * MAY BE DEPRECATED 
     * 
     * 
     * Build a dropdown of valid children.
     * 
     * This needs to build a list of instances of the reln, subclasses, or  *
     * any instances of subclasses.  Basically EVERYTHING below this in the  *
     * BE TREE.   The from is not relevant unless restricting to existing    *
     * params on the class definition.                                       *
     * 
     * @param {anne_neuron} - The from neuron
     * @param {string} - The nid long string of the reln.
     * @param {boolean} - Flag indicating downline instances will also be retrieved.
     * @author {Mr.Scatman}  
     */
    buildValidHasChildrenTempNids=function($fNObj, $reln_nid, $bINSTS=false) {
        var aMatchingParams = MEMG.getParentLineParams($fNObj, $reln_nid);
        var aList = [];
        if (aMatchingParams.length) aList = this.extractTempNidsFromRelonsTo(aMatchingParams);
        var children_of_parent_params = [];
        if (aList.length) children_of_parent_params = this.extractAllChildrenOfTempNidList(aList);
        if (children_of_parent_params.length) aList = aList.concat(children_of_parent_params);
        var relnNObj = MEMG.getNeuronByNid($reln_nid);
        var childrenDownline = [];
        relnNObj.getAllDownlineChildren(childrenDownline, 0, $bINSTS); //this should get all subs, all children of subs.. (all insts of)
        var validClassTempNids = aList.concat(childrenDownline);
        validClassTempNids = validClassTempNids.filter(leg => leg !=$fNObj.temp_nid);
        return validClassTempNids;
    }


    /**
    * Extract all children nids.
    * 
    * This will honor creative hierarchies that use BE_PARENT and BE_CHILDREN
    * to define their primary hierarchy as something non-parent/child.
    *  
    * @param {anne_neuron} - The parent neuron
    * @author {Mr.Scatman} 
    */
    extracAllChildrenNidsDynamic = function($nObj) {
        var childrenNOBJS = this.getChildrenDynamic($nObj);
        var children_nids = [];
        for (var c=0; c < childrenNOBJS.length; c++) {
            children_nids.push(childrenNOBJS[c].nid);
        }
        return children_nids;
    }


    /**
     * 
     * @param {array} - An array of relon objs.
     * @author {Mr.Scatman} 
     */
    extractTempNidsFromRelonsFrom=function($aRelons) {
        if ($aRelons==undefined || $aRelons==null || Array.isArray($aRelons)==false) return [];
        var aTempNids = [];
        for (var r=0; r < $aRelons.length; r++) {
            var rObj = $aRelons[r];
            if (rObj.f==undefined || rObj.f==null || rObj.f==-1) {
                var fromNOBJ = MEMG.getNeuronByNid(rObj.from_nid);
                if (fromNOBJ!=null) {
                    if (aTempNids.indexOf(fromNOBJ.temp_nid)===-1) {
                        console.log("extracting "+fromNOBJ.text);
                        aTempNids.push(fromNOBJ.temp_nid);
                    }
                }
                continue;
            }
            if (aTempNids.indexOf(rObj.f)===-1) {
                var fromNOBJ = MEMG.getNeuronByTempNid(rObj.f);
                aTempNids.push(rObj.f);
                console.log("extracting "+fromNOBJ.text);
            }
        }
        return aTempNids;
    }
    extractTempNidsFromRelonsTo=function($aRelons) {
        var aTempNids = [];
        for (var r=0; r < $aRelons.length; r++) {
            var rObj = $aRelons[r];
            if (rObj.t==undefined || rObj.t==null || rObj.t==-1) continue;
            if (aTempNids.indexOf(rObj.t)===-1) aTempNids.push(rObj.t);
        }
        return aTempNids;
    }


    /**
     * Get all the children neurons of a list of temp nids.
     * 
     * @param {array} - An array of temp nid ints
     * @author {Mr.Scatman} 
     */
    extractAllChildrenOfTempNidList=function($arr) {
        var aChildren = [];
        for (var i=0; i < $arr.length; i++) {
            var nObj = MEMG.getNeuronByTempNid($arr[i]);
            if (nObj==null) continue;
            var children = nObj.children;
            if (children.length==0) continue;
            for (var c=0; c < children.length; c++) {
                var child = children[c];
                if (aChildren.indexOf(child)!==-1) continue;
                aChildren.push(child);  //unique child for our summary list.
            }
        }
        return aChildren;
    }



                            
    /**
     * This establishes a base URL for a class.
     *
     * @param {anne_neuron} - The class neuron
     * @param {string} - The base url to use as "part" of the url.
     * @author {Mr.Scatman} - 
     */
    __setClassBaseURL=function($classNOBJ, $base_url_string) {
        return this.makeTempRelon($classNOBJ, TYPE_HAS, this.base_url.nid, TO_TYPE_STRING, $base_url_string, true);
    }


    /**
     * 
     * Registers a img for a class to expect to be used as the reln for a given 
     * class.. If a base_url is supplied it will also be created as a relon. 
     * see moviesDB_api::                                                        
     * this.__setImgThumbRelonForClass(this.media_credit, this.profile_path_with_base, "https://image.tmdb.org/t/p/w185");
     *
     * @param {anne_neuron} - The class neuron
     * @param {anne_neuron} - The img thumb reln's full neuron.
     * @param {string} - An optional base url to use as "part" of the url.
     * @author {Mr.Scatman} 
     * 
     */
    __setImgThumbRelonForClass=function($classNOBJ, $imgThumbRelnNOBJ, $base_url="") {
        this.makeTempRelon($classNOBJ, TYPE_HAS, this.imgThumbDefaultClass.nid, TO_TYPE_ANNE, $imgThumbRelnNOBJ.nid, true);
        if ($base_url!="") {
            this.makeTempRelon($classNOBJ, TYPE_HAS, this.base_url.nid, TO_TYPE_STRING, $base_url, true);
        }
    }


    /**
     * local desperado img url for classes and instances that we have no img of.
     * 
     * @param {anne_neuron} - The classneuron neuron.
     * @param {string} - The url.
     * @author {Mr.Scatman} 
     * 
     */
    __setImgThumbDefaultClassImg = function($classNOBJ, $url) {
        $classNOBJ.desperado_img_url=$url;
    }


    /**
     * Set a primary key for a class.
     * 
     * These are properties used by the class in additional to the nid.
     * They are external in nature but are primary properties and can
     * be used to resolve and identify the neuron.
     * 
     * ALL ANNE NEURONS have a true primary key of their .nid.  That is 
     * a given.  This is in addition to that.
     * 
     * @param {anne_neuron} - The target neuron.
     * @param {anne_neuron} - The primary key neuron.
     * @author {Mr.Scatman} 
     * 
     */
    __setPrimaryKey=function($fromNOBJ, $toNOBJ) {
        this.makeTempRelon($fromNOBJ, TYPE_HAS, this.primary_key.nid, TO_TYPE_ANNE, $toNOBJ.nid, true);
    }


    /**
     * Set a secondary key for a class.
     * 
     * This is not a foreign key, it is an alt primary key.
     * Classes such as countries will supply unique ISO codes of 2 digits
     * as well as 3 digits.  Secondary key is used in these such cases.
     * 
     * Note: nid is the ACTUAL ANNE PRIMARY KEY for everything.  Primary and secondary
     * dennote outside ANNE data source data keys.
     * 
     * @param {anne_neuron} - The target neuron.
     * @param {anne_neuron} - The secondary key neuron.
     * @author {Mr.Scatman} 
     * 
     */
    __setSecondaryKey2=function($fromNOBJ, $toNOBJ) {
        this.makeTempRelon($fromNOBJ, TYPE_HAS, this.primary_key2.nid, TO_TYPE_ANNE, $toNOBJ.nid, true);
    }


    /**
     * Make a new neuron and push it, if unique.
     * 
     * @param {Object} - The relon obj.
     * @param {string} - A string of the prop name to extract the nid from.
     * @param {string} - A string name of the prop to create to map to the temp_nid of the new neuron.
     * @param {string} - We will delete this after we use it to make chrome debugging more visually easy.
     * @author {Mr.Scatman} 
     * 
     */
    makeAndPushIfUniqueNid = function ($rObj, $prop='from_nid', $setprop='f', $temptext="") {
        var lookingForNID = $rObj[$prop];
        if ($rObj[$prop+"TNOBJ"]!=undefined) $rObj[$prop+"TNOBJ"] = $rObj[$prop+"TNOBJ"];
        else $rObj[$prop+"TNOBJ"] = MEMG.getNeuronByNid(lookingForNID);

        //(rObj, 'from_nid', 'f', 'from_txt_temp');    

        if ($rObj[$prop+"TNOBJ"]!=null) {
            $rObj[$prop+"TNOBJ"].temp=false;
            if ($rObj[$prop+"TNOBJ"].nid[0]=="z") $rObj[$prop+"TNOBJ"].ghost=true;
            if ($setprop) $rObj[$setprop]=$rObj[$prop+"TNOBJ"].temp_nid;
            if ($temptext!="") $rObj[$temptext]=$rObj[$prop+"TNOBJ"].text;
            return $rObj[$prop+"TNOBJ"];
        }

        //we don't have it -- make it.
        $rObj[$prop+"TNOBJ"] = this.makeNeuronFromANYRelon($rObj, $prop);
        $rObj[$prop+"TNOBJ"].temp=false;
        if ($setprop) $rObj[$setprop]=$rObj[$prop+"TNOBJ"].temp_nid;
        if ($temptext!="") $rObj[$temptext]=$rObj[$prop+"TNOBJ"].text;
        return $rObj[$prop+"TNOBJ"];
    }


    /**
     * Make and push an emulated neuron from a non ANNE ID to.                 
     * 
     * final rObj is for context when the TO_TYPE is TO_TYPE_ENCRYPTED.        
     * we need to look at the RELN to contextualize what type of encrypted   
     * value it is.    
     * TO DO --> use rObj to contextuals the encrypt 
     * (reln will tell us what it is)  
     *
     * @author {Mr.Scatman} 
     */
    makeAndPushIfUniqueEMULATED=function($text, $to_type, $rObj=null, $bFREEPASS=false) {
        if (this._EMULATE_TO_TYPES == false && $bFREEPASS==false) return null;
        if (ANNE.USER==null) return null;

        //not an emulatable to type.
        if ($bFREEPASS==false && (this._emulatables.length==0 || this._emulatables.indexOf($to_type)==-1)) return null;

        //if we are emulating text neurons... only do so in non sys relons//$rObj.type==TYPE_BE || 
        if ($bFREEPASS==false && $to_type==TO_TYPE_STRING && $rObj!=null && ($rObj.type==TYPE_SYS)) return null;


        //we do not emulate neurons that are bigger than 64 chars in length
        //this allows for us to max out with emulation of a TXID as its own neuron. (when it is the to)
        if ($to_type==TO_TYPE_STRING && $text.len > 64) return null;

        var text = ($text+"").trim();
        var type = NTYPE_INST;
        var TTSYSOBJ = MEMG.getSysToTypeObjByTypeAndBe($to_type, this.DATA_TYPE);
        var bes=[];
        var be="";
        var rNObj = null;
        var rNObj_TTOBJ =null;

        if ($rObj!=null) {
            rNObj = MEMG.getNeuronByTempNid($rObj.r);
            if (rNObj==undefined || rNObj==null) {
                console.log($rObj);
                console.log("debug this!!!!");
                rNObj = MEMG.getNeuronByNid($rObj.reln);
                if (rNObj!=null) $rObj.r = rNObj.temp_nid;
            }
            if (rNObj!=null) rNObj_TTOBJ = MEMG.getTerminatingType(rNObj);
        }

        var beNOBJ = null;

        //use the reln if we have it, else use the to_type default.
        if (rNObj_TTOBJ==null) rNObj_TTOBJ = TTSYSOBJ;

        //dates are dates.. that's the end of it.
        if (rNObj_TTOBJ.temp_nid==this.date.temp_nid) {
            bes=[rNObj_TTOBJ.temp_nid];
            beNOBJ = rNObj_TTOBJ;
            be=rNObj_TTOBJ.text;
            var nObj = MEMG.getEmulatedNeuronByClassAndValue(beNOBJ, $text);
            if (nObj!=null)  {
                //we need to make a new emulated relon pointing to this new emulated neuron
                //we can use this more properly in the all the neuron/relon functions.
                this.emulateEdgeToEmulated($rObj, nObj);
                return nObj;
            }
        }

        else if (rNObj_TTOBJ!=null && TTSYSOBJ!=null && rNObj_TTOBJ.SYS_ENUM_VAL==TTSYSOBJ.SYS_ENUM_VAL && rNObj.nid!=this.name.nid) {
            bes=[rNObj.temp_nid];
            beNOBJ = rNObj;
            be=rNObj.text;
            var nObj = MEMG.getEmulatedNeuronByClassAndValue(beNOBJ, $text);
            if (nObj!=null)  {
                //we need to make a new emulated relon pointing to this new emulated neuron
                //we can use this more properly in the all the neuron/relon functions.
                this.emulateEdgeToEmulated($rObj, nObj);
                return nObj;
            }
        }
        
        //super critical here to classify the text neuron as a class of its relon.
        //we accept any existing downliners of the same.
        else if (rNObj_TTOBJ==null && TTSYSOBJ.SYS_ENUM_VAL==TO_TYPE_STRING) {
            bes=[rNObj.temp_nid];
            beNOBJ = rNObj;
            be=rNObj.text;
            nObj = MEMG.getEmulatedNeuronByClassAndValue(rNObj, $text);
            if (nObj!=null)  {
                //we need to make a new emulated relon pointing to this new emulated neuron
                //we can use this more properly in the all the neuron/relon functions.
                this.emulateEdgeToEmulated($rObj, nObj);
                return nObj;
            }
        }

        //a number value
        else if (TTSYSOBJ.SYS_ENUM_VAL==TO_TYPE_NUMBER && MEM_NUMBER!=undefined) {
            bes=[this.number.temp_nid];
            beNOBJ = this.number;
            be=this.number.text;
            nObj = MEM_NUMBER.getNumberNeuron($text);
            if (nObj!=null)  {
                //we need to make a new emulated relon pointing to this new emulated neuron
                //we can use this more properly in the all the neuron/relon functions.
                this.emulateEdgeToEmulated($rObj, nObj);
                return nObj;
            }
        }

        else {
            bes=[TTSYSOBJ.temp_nid];  //bes twe can do.. this to_type was used on a rogue reln which was not defined as sub of a to type
            beNOBJ = TTSYSOBJ;
            be=TTSYSOBJ.text;
            var nObj = this.getEmulatedNeuronByTypeAndText($to_type, $text);
            if (nObj!=null)  {
                //we need to make a new emulated relon pointing to this new emulated neuron
                //we can use this more properly in the all the neuron/relon functions.
                this.emulateEdgeToEmulated($rObj, nObj);
                return nObj;
            }
        }

        //we do not make emulated neurons for numbers unless
        //they are special numbers.
        //special numbers are in the downline of number_notation or the conversion class.
        //WE MAY NEED TO RETHINK THIS.... for the LIVE-FIRING-NN version.
        if ($to_type==TO_TYPE_NUMBER) {
            var downline_nids = [this.number_notation.temp_nid];
            this.number_notation.getAllDownlineChildren(downline_nids, 0, false);
            if (downline_nids.indexOf(rNObj.temp_nid)==-1) {
                downline_nids = [this.conversion.temp_nid];
                this.conversion.getAllDownlineChildren(downline_nids, 0, false);
                if (downline_nids.indexOf(rNObj.temp_nid)==-1) return null;
            }
        }
            
        var nid = ANNE.USER.getRandomEmulatedANNEAccountId();

        var emulatedNOBJ = new anne_neuron(text, type, nid);
        emulatedNOBJ.be = be;
        emulatedNOBJ.setBeTempNid(bes[0]);
        emulatedNOBJ.emulated = true;
        emulatedNOBJ.bTERMINATING_TYPE=true;

        //we need to make a new emulated relon pointing to this new emulated neuron
        //we can use this more properly in the all the neuron/relon functions.
        this.emulateEdgeToEmulated($rObj, emulatedNOBJ);

        //emulate the be relon of the emulated to,
        //it points to what we chose as the beNOBJ.
        var emulateBeROBj = this.makeEmulatedRelon(emulatedNOBJ, TYPE_BE, this.class.nid, TO_TYPE_ANNE, beNOBJ.nid, true, false);
        
        emulatedNOBJ.__setValueOfNeuron(beNOBJ, TTSYSOBJ);

        return emulatedNOBJ;
    }



    /**
     * Set the text name with the first valid relon.
     * 
     * This is done when we arbor or inhibit a name relon.  The name is lost.
     * We then try to set the name (rollback) to something else we have in memory.
     * 
     * @param {anne_neuron} - the Neuron object
     * @author {Mr.Scatman} 
     * 
     */
    setTextWithFirstValidAssigner=function($nObj) {
        if ($nObj==undefined || $nObj==null || $nObj.outs.length==0) return "";
        var len = 0;
        var str = "";

        //try for any other BE NAME relons.
        for (var r=0; r < $nObj.outs.length; r++) {
            var rObj = $nObj.outs[r];
            if (rObj.active==false) continue;   //why is this still even here?
            if (rObj.type==TYPE_BE && rObj.reln==this.name.nid && rObj.to_type==TO_TYPE_STRING) {
                if (rObj.to_s.length > len) {
                    len = rObj.to_s.length;
                    str = rObj.to_s;
                }
            }
        }

        if ($nObj.short!=undefined && $nObj.short!="") return $nObj.short;
        if ($nObj.aka!=undefined && $nObj.aka!="") return $nObj.aka;

        //take any be.
        if (str=="") {
            for (var r=0; r < $outs.length; r++) {
                var rObj = $outs[r];
                if (rObj.active==false) continue;   //why is this still even here?
                if (rObj.type==TYPE_BE && rObj.to_type==TO_TYPE_STRING) {
                    if (rObj.to_s.length > len) {
                        len = rObj.to_s.length;
                        str = rObj.to_s;
                    }
                }
            }
        }
        return str;
    }

    /** 
     * sort tnids 
     */
    sortTnidsAlpha=function($tnids) {
        var nObjs = MEMG.getNObjsFromTempNidList($tnids);
        nObjs.sort(ANNE.UTILS.sortByTextNoCase);
        return MEMG.getTnidsFromNObjs(nObjs);
    }


    /** 
     * Create an emulated edge.
     * 
     * Sometimes edges point to non-encoded TERMINATING TYPES and we want to emulate
     * a neuron for that type.  In doing so we also need to then emulate a relon that
     * points to that emulated neuron since the original relon will be in the wrong 
     * TO_TYPE space.
     * 
     * @param {Object} - the original relon we want to create an emulated relon for
     * @param {anne_neuron} - the emulated Neuron we will point to with our emulated relon. 
     * @author {Mr.Scatman} 
     */
    emulateEdgeToEmulated = function ($rObj, $emulatedNOBJ) {
        var fNOBJ = MEMG.getNeuronByTempNid($rObj.f);
        var emulated_relon = this.makeEmulatedRelon(fNOBJ, $rObj.type, $rObj.reln, TO_TYPE_ANNE, $emulatedNOBJ.nid, true, false);
        emulated_relon.emulated_from_txid = $rObj.txid;
        $rObj.active=false;
    }


    /**
     * Make an emulated relon
     * 
     * Emulated relons are NOT ghosts.  They are made up based upon 
     * converting non-neuron values to neurons -- and the emulated relons 
     * support them as-if-they-were-real. 
     * example: turn a date string into a date neuron -- that doesn't 
     * actually exist on-chain, but now exists in mem, as an emulated date 
     * neuron. 
     * 
     * Temp relons are NOT ghosts. but they can become ghosts if they are 
     * sent to the DB.
     * 
     * @author {Mr.Scatman} 
     */
    makeEmulatedRelon=function($fNOBJ, $type, $reln, $to_type, $to_s, $bPUSH=true, $bTEMP=false) {
        var rObj = this.makeTempRelon($fNOBJ, $type, $reln, $to_type, $to_s, $bPUSH);
        rObj.emulated=true;
        rObj.temp=$bTEMP;
        rObj.ghost=false;
        if ($reln==this.require.nid && $type==TYPE_PARAM) rObj.required=true;
        return rObj;
    }


    /**
     * @author {Mr.Scatman} 
    */
    makeTempRelon=function($fNOBJ, $type, $reln, $to_type, $to_s, $bPUSH=true) {
        if ($fNOBJ==null || $fNOBJ.nid==undefined) return null;
        if (typeof $reln !=='string') {
            console.warn("bad data in make temp relon $reln");
            if (typeof $reln ==="object") {
                $reln = $reln.nid;
                if (this.__isValidNid($reln)==true) {
                    console.warn("we fixed it, but u should pass the .nid, not the object!");
                }
                else return null;
            }
            else return null;
        }
        if (typeof $to_s !=='string' && typeof $to_s !=='number') {
            console.log("bad data in make temp relon $to_s");
            if ($to_type==TO_TYPE_ANNE && typeof $to_s ==="object") {
                $to_s = $to_s.nid;
                if (this.__isValidNid($to_s)==true) {
                    console.warn("we fixed it, but u should pass the .nid, not the object!");
                }
                else return null;
            }
            else return null;
        }
        if (parseInt($type)>255 || parseInt($to_type)>255) {
            console.log("bad data in make temp relon $type/$to_type");
            return null;
        }


        var r = MEMG.getNeuronByNid($reln);
        if (r) r=r.temp_nid;
        else r = -1;
        var t = -1;
        var tObj = null;
        if ($to_type==TO_TYPE_ANNE) {
            tObj = MEMG.getNeuronByNid($to_s);
            if (tObj) {
                t = tObj.temp_nid;
                if ($type==TYPE_BE) $fNOBJ["be_"+r]=t;
            }
        }

        $to_type = parseInt($to_type);
        if ($to_type==TO_TYPE_NUMBER) $to_s = parseFloat($to_s);

        //temp relons are 21 length with a leading z.
        var rObj = {
                    "txid":ANNE.UTILS.randTXID(this._emulatedTXIDLength),
                    "type":parseInt($type),
                    "reln":$reln,
                    "to_type":$to_type,
                    "to_s":$to_s,
                    "from_nid":$fNOBJ.nid,
                    "f":$fNOBJ.temp_nid,
                    "r":r,
                    "t":t,
                    "temp":true,
                    "emulated":false,
                    "preloaded":false,
                    "time":-1,
                    "firstseen":0,
                    "rorigin":RORIGIN_GHOST,
                    "ghost":false
                    }
        if ($bPUSH==false) return rObj;

        //puts relons in a master hash array and a master array with rid.
        MEMP.__hashRelon(rObj);
        $fNOBJ.outs.push(rObj);

        if (rObj.from_nid=="3140226982098870734") {
            console.log("test this.");
        }
        if (rObj.txid=="10654123745552326374") {
            console.log("test this");
        }

        if (tObj!=null) {
            if (tObj.ins) tObj.ins.push(rObj);
        }

        //we need to emulate the to as well, if it is a bTERMINATING_TYPE
        else if (rObj.to_type!=TO_TYPE_ANNE &&  rObj.to_type!=TO_TYPE_SEED) {
            var tObj = this.makeAndPushIfUniqueEMULATED(rObj.to_s, rObj.to_type, rObj);
            if (tObj) {
                tObj.ins.push(rObj);  //track that we have an in to this emulated neuron ?
                //we do not emulate the owns and owned relons for emulated to type neurons.
                //we do, however, set the owned_by of the first occurance we observe of any 
                //emulated tos.  We set it equal to the owner of the from that pointed to it.  (if that exists)
                if (tObj.owned_by=="" && $fNOBJ.owned_by!="") tObj.owned_by = $fNOBJ.owned_by;
                rObj.t = tObj.temp_nid;     
            }
        }
        if (tObj!=null && this.require!=undefined && $reln==this.require.nid && tObj.temp_nid==this.validation.temp_nid) {
            $fNOBJ.requires_validation=true; 
        }
        return rObj;
    }


    /**
     * Assess if a relon has met the class PARAM definition.
     * 
     * returns true if the paramRelon we are reviewing is in the array_has
     * we match the be downline....  (not the upline) 
     * example:                                                              
     * identity class has a param of HOUSE
     * instance of identity "Mr.Scatman" has a HAS of house->"house for ants"
     * That "hosue for ants" is an instance of "tiny house", which is a 
     * sub-class of house.   Be looking at the PARAM reln we find out what 
     * was desired to supply.   We can look at its DOWNLINE to gnenerate
     * other viable classes of information.   Tiny House is a type of house,
     * thus the parameter has been supplied in the MR SCATMAN HAS reln. 
     * 
     * @author {Mr.Scatman} 
     */
    hasMetClassDefn=function($relons, $paramRelon, $paramToNObj=null) {
        var paramRelnNObj = MEMG.getNeuronByTempNid($paramRelon.r);
        if (paramRelnNObj==null) return false; //must have this.
        
        //console.log(paramRelnNObj.text);
        
        var paramToNObj = $paramToNObj;
        if ($paramRelon.t && paramToNObj==null) paramToNObj = MEMG.getNeuronByTempNid($paramRelon.t);
        
        var TTOBJ = MEMG.getTerminatingType(paramRelnNObj);

        var TO_TTOBJ = MEMG.getTerminatingType(paramToNObj);

        //loop what we have to see if we have this param!  :)
        //we also satisfy PARAMs with BEs -- so allow these to pass as well.
        for (var a=0; a <$relons.length; a++) {
            var relon = $relons[a];

            //if the relon is a TYPE_ROLE.. 
            //and the reln is a role... it is good whatever the to is.
            if (relon.type==TYPE_ROLE && relon.r==$paramRelon.r) {
                var relNOBJ = MEMG.getNeuronByTempNid(relon.r);
                if (relNOBJ.bes[0]==this.role.temp_nid) {
                    return true;
                }
            }


            //we may have some relon types in this array we don't want to consider.
            //Only a has/sys/be/role can satisfy the PARAM and we already checked the TYPE_ROLE.
            if (relon.type!=TYPE_HAS && relon.type!=TYPE_BE && relon.type!=TYPE_SYS) continue;    //these are supposed to be has onl.. sometimes we are lazy

        
            if (relon.reln!=paramRelnNObj.nid) {
                if (this.matchesBeDownLine(relon.r, paramRelnNObj)!=true) continue;
            }

            //quick exits
            if (TTOBJ!=null) {
                if (relon.to_type==TTOBJ.SYS_ENUM_VAL) return true;
            }

            if (TO_TTOBJ!=null && paramToNObj!=null) {
                if (relon.to_type==TO_TTOBJ.SYS_ENUM_VAL) return true;
            }
            

            //this paramreln is already in the has (or be) ....
            //see if the to matches the param's to class (or a parent);
            //if (paramRelnNObj.bTERMINATING_TYPE == true) EXPECT an emulated neuron in the has...
            if (relon.t!=undefined && relon.t!=-1) {
                var toNObj = MEMG.getNeuronByTempNid(relon.t); //there will be no hasTONObj (unless emulated)
                if (toNObj==null) return false;

                if (toNObj.bes[0] == $paramToNObj.temp_nid) return true;
                else {
                    if ($paramToNObj.temp_nid==this.class.temp_nid) return true;
                    
                    var BENOBJ = MEMG.getNeuronByTempNid(toNObj.bes[0]);
                    //is the has BE a descendent of the param's to nobj?
                    //this.matchesBeDownLine(BENOBJ.temp_nid, $paramToNObj)
                    //does the bas BE upline run into the $paramToNObj?
                    //alt version (testing)
                    if (BENOBJ && this.matchesBeDownLine(BENOBJ.temp_nid, $paramToNObj)==true) return true;  //check the paramTOObj's upline...
                    else {
                        //console.log("param reln matches, to class doesn't...");
                        if (relon.type==TYPE_BE) return true;  //ok for be, not for param/has
                    }
                }
            }

            //if no to class is known, assume this is ok.
            else return true;

        
        }
        return false;
    }


    /**
     * Count how many times a neuron has been used as the reln in relon memory space.
     * 
     * @params {int} - The temp_nid of the relon's reln we want to look for.
     * @author {Mr.Scatman} 
     * 
     */
    countRelnUsage=function($r) {
        var cnt = 0;
        for (var r; r<this._aRelons.length;r++) {
            if (this._aRelons[r].r=$r) cnt+=1;
        }
        return cnt;
    }

    //UNTESTED
    //misc.child_cdx = [{from_nid:for_nOBJ.nid, reln:this.neuronImg.nid, to_s:null, type:null, to_type:TO_TYPE_ANNE, target_part:"to_s"}];
    enforceCDXS=function($cdxs, $nObj) {
        for (var c=0; c <$cdxx.length; c++) {
            var cdx_rel = $cdxs[c];
            if (cdx_rel.target_part=="to_s") {
                var from_nid = cdx_rel.from_nid;
                var reln = cdx_rel.reln;
                var type = cdx_rel.type;
                var to_type = cdx_rel.to_type;
                var exists = this.enforceCDXIn($nObj, type, from_nid, reln, to_type); 
                if (exists==false) return false;
            }
            else if (cdx_rel.target_part=="from_nid") {
                var reln = cdx_rel.reln;
                var type = cdx_rel.type;
                var to_type = cdx_rel.to_type;
                var to_s = cdx_rel.to_s;
                var exists = this.enforceCDXOut($nObj, type, reln, to_type, to_s); 
                if (exists==false) return false;
            }
            else return false;
        }
        return true;
    }

    //UNTESTED
    enforceCDXIn = function ($nObj, $type, $from_nid, $reln, $to_type) {
        for (var r=0; r < $nObj.ins.length; r++) {
            var rel = $nObj.ins[r];
            if ($from_nid!=null && rel.from_nid!=$from_nid) continue;
            if ($reln!=null && rel.reln!=$reln) continue;
            if ($type!=null && rel.type!=$type) continue;
            if ($to_type!=null && rel.to_type!=$to_type) continue;
            return true;
        }
        return false;
    }

    //UNTESTED
    enforceCDXOut = function ($nObj, $type, $reln, $to_type, $to_s) {
        for (var r=0; r < $nObj.ins.length; r++) {
            var rel = $nObj.ins[r];
            if ($reln!=null && rel.reln!=$reln) continue;
            if ($type!=null && rel.type!=$type) continue;
            if ($to_type!=null && rel.to_type!=$to_type) continue;
            if ($to_s!=null && rel.to_s!=$to_s) continue;
            return true;
        }
        return false;
    }


    /**
     * TODO: AUDIT THIS AGAIN
     * 
     * a bit of overkill, we could just walk the list as we build it and     *
     * look for a match..                                                    *
     * (this will handle multiple be lines like GROUP and FEELZ)             *
     * 
     * @author {Mr.Scatman} 
     */
    matchesBeDownLine=function($target_temp_nid, $nObj) {
        if ($nObj==null) return false;
        if ($nObj.temp_nid==this.class.temp_nid) return true;

        if ($nObj.children && $nObj.children.length) {
            var children = $nObj.children;
            for (var c=0; c < children.length; c++) {
                var child = children[c];
                if (child==$target_temp_nid) return true;
                var childNObj = MEMG.getNeuronByTempNid(child);
                if (childNObj.children!=undefined && childNObj.children.length) {
                    var bFOUND = this.matchesBeDownLine($target_temp_nid, childNObj);
                    if (bFOUND==true) return true;
                } 
            }

        }
        return false;
    }

    
    /**
     * Test for inclusion in lineage.
     * 
     * Is this temp_nid or any parent of this be equal to the target_nid ? 
     * In other words, is the $targetNid in the parental TYPE_BE dag lineage of the $temp_nid? 
     * 
     * @param {int} - Temp nid for the neuron we want to test
     * @param {int} - The temp nid of the ancestor we are looking to see if is in the temp_nid line.
     * @Author {Mr.Scatman}
     * 
     */
    tempBeIsA = function($temp_nid, $target_tnid, $ntype=null) {
        var nObj = null;
        nObj = MEMG.getNeuronByTempNid($temp_nid); 
        if ($ntype==null) {
            if ($temp_nid==$target_tnid) return true;       //quick exit
        }
        else {
            if (nObj==null) return false;
            if ($ntype==NTYPE_INST && nObj.ntype!=NTYPE_INST) return false; 
            else if ($ntype==NTYPE_CLASS && nObj.ntype!=NTYPE_CLASS) return false;
            else if ($temp_nid==$target_tnid) return true;  //quick exit
        }

        if (nObj==null) return false;
        if (nObj.bes[0]==$target_tnid) return true;
        var depth=0;
        var parentNObj = MEMG.getNeuronByTempNid(nObj.bes[0]);
        var loops = 0;
        while (depth<10 && parentNObj!=null) {
            loops+=1;
            if (loops>20) break;//safey
            if (parentNObj.be_lvl==0 || parentNObj.be_lvl==-1) break;
            depth++;
            if (parentNObj.temp_nid==$target_tnid) return true;
            parentNObj = MEMG.getNeuronByTempNid(parentNObj.bes[0]);
            if (parentNObj==null) return false;
        }
        return false;
    }


    /** 
     * @author {Mr.Scatman} 
     */
    makeNeuronFromANYRelon = function ($rObj, $prop='from_nid') {
        if ($rObj[$prop] == undefined) {
            console.warn('NO '+$prop+' prop on the relon!!  Process all relons in (processNeuronsFromRelons)');
        }
        return new anne_neuron("", "", $rObj[$prop]);
    }


    /** 
     * Load up a min neuron for use in the brain chicken and egg loadup challenge.
     * note: This only runs once during the js class initialization phase.
     * 
     * @param {Object} - The quick json obj used to preload a neuron to have as a permanent reference
     * used in the chicken & egg loadup challenge.
     * @author {Mr.Scatman} 
     * 
     */
    minBootNeuron=function($nObj_min, $parentNOBJ=null) {
        var nOBJ = new anne_neuron($nObj_min.text, $nObj_min.ntype, $nObj_min.nid);
        if ($parentNOBJ==null) return nOBJ;
        nOBJ.setBeTempNid($parentNOBJ.temp_nid);
        return nOBJ;
    }


    /** 
     * Standard relon sort by value.
     * 
     * @param {Object} - relon a.
     * @param {Object} - relon b.
     * @author {Mr.Scatman} 
     * 
     */
    __sortSeqRelons = function($a, $b) {
        var a = MEMG.getNeuronByTempNid($a.r);
        var b = MEMG.getNeuronByTempNid($b.r);
        if (a.value==undefined) {
            if (b.value!=undefined) return 1;
            else return 0;
        }
        else if (b.value==undefined) return -1;
        if (a.value > b.value) return 1;
        else if (b.value > a.value) return -1;
        return 0;
    }

    /**
     * THIS WILL BE DEPRECATED.
     * Done by the context pool handling of a phrase.
     */
    addModsToGhostNeuron=function($role, $chaseClassNid) {
        if ($role.pre_mods.length==0) return;
        if ($role.new==false) return;
        var nObj =$role.nObj;
        if (nObj.ghost==false) return;
        
        //constituent relons are key here.
        if ($chaseClassNid==this.location.nid) {
            for (var r=0; r < $role.pre_mods.length; r++) {
                var mod = $role.pre_mods[r];
                var nObjmod = mod.nObj;
                if (nObjmod==null) continue;

                //check for relon
                var rel = MEMG.getRelonByTempR(nObj.outs, this.constituent_of.temp_nid);
                if (rel==null) {
                    GHOSTDATA.new_ghost_be(nObj, this.constituent_of.nid, TO_TYPE_ANNE, nObjmod.nid);
                }
                else {
                    //we have a problem!  This location is a constituent of a different location!
                    if (rel.to_s!=nObj.nid) {
                        alert("trouble!!!!");
                        return;
                    }
                    continue;
                }
                nObj = nObjmod;
            }
        }
    }


    /**
     * A class has specified it is a terminating type.
     * We do this when a class wants to piggy-back terminating type logics
     * but it is classified as its own unique thing in the TYPE_BE tree.
     * 
     * Example:  annie(s).
     * They are a unit of measure (which is a measure) but that is a class,
     * not a terminating type.  The GUI will try to look for children of annies
     * when params are used to point to annies, but what we really want is a
     * numerical entry.  
     * 
     * Any class that uses the TERMINATYPE_TYPE relon can specify its actual
     * terminating type.  We can then map that here to the terminating type object.
     */
     mapSysCode=function($nObj, $to_type_temp_nid) {
        for (var s=0; s < this._SYS_TO_TYPE_CODES.length; s++) {
            var sysNOBJ = this._SYS_TO_TYPE_CODES[s];
            if (sysNOBJ.temp_nid==$to_type_temp_nid) {
                $nObj.SYS_ALIAS= sysNOBJ.SYS_ALIAS;
                $nObj.SYS_ENUM_VAL = sysNOBJ.SYS_ENUM_VAL;
                $nObj.SYS_CODE= sysNOBJ.SYS_CODE;
                //console.warn("successfully mapped new TT ", $nObj);
                return;
            }
        }
        console.warn("unsuccessful mapping of new TT ", $nObj);
    }


    /** 
     * THIS WAS DONE MANY VERSIONS AGO WHEN IT WASN'T PREOPTIMIZING..
     * BUT NOW IT IS A JUMBLEDMESS.  IGNORE IT.
     * @author {Mr.Scatman} 
     */
    markChasedIns = function($nObj, $types) {
        return [];
        if ($nObj==null || $nObj.emulated==true) return [];
        if ($nObj.preloaded==true && MEM_UDF._lookupPreloadedIns==false) return [];
        var post = "ins";
        return this.markChasedArr($nObj, $types, post);
    }


    /** 
     * Optimize ANNODE queries by marking what we lookuped up.
     * 
     * THIS WAS DONE MANY VERSIONS AGO WHEN IT WASN'T PREOPTIMIZING..
     * BUT NOW IT IS A JUMBLEDMESS.  IGNORE IT.
     * 
     * @param {anne_neuron} - The neuron we just queried the ANNODE for.
     * @param {array} - An array of ints for the types we just queried.
     * * @author {Mr.Scatman} 
     */ 
    markChasedOuts = function($nObj, $types) {
        return [];
        if ($nObj==null || $nObj.emulated==true) return [];
        if ($nObj.preloaded==true && MEM_UDF._lookupPreloadedOuts==false) return [];
        var post = "outs";
        return this.markChasedArr($nObj, $types, post);
    }


    /** 
     * Optimize ANNODE queries by marking what we lookuped up.
     * 
     * This marks parts of more complex queries that we sometimes cannot
     * optimize.  The pieces can be optimized and prevent us from doing repeat
     * work on future complicated queries that include these smaller parts.
     * 
     * @param {anne_neuron} - The neuron we just queried the ANNODE for.
     * @param {array} - An array of ints for the types we just queried.
     * @param {string} - Either "outs" or "ins" for indicating which direction we chased.
     * @author {Mr.Scatman} 
     * 
     */
    markChasedArr=function($nObj, $types, $post) {
        var arr = $nObj["chased_"+$post];
        if (arr==undefined) {
            var arr = [];
            $nObj["chased_"+$post]=arr;
        }

        //it is new.. update our arr in mem and return the original types array
        if (arr.length==0) {
            for (var s=0; s < 7; s++) {
                if ($types.indexOf(s)!=-1) arr[s]=1;
                else arr[s]=0;  //make a new slot for this type to track it.
            }
            return $types;
        }   

        //we alreayd have done something.. let's see what.
        //return a new array of the net diff of what we have not done.
        var types = [];
        for (var s=0; s < types.length; s++) {
            var sch = types[s];
            var tracker_slot = arr[sch];
            if (tracker_slot==1) continue;
            types.push(s);
        }
        return types;
    }


    /** 
     * Set a neuron's value by its polarity relon.
     * 
     * TODO: revisit this.
     * 
     * @param {anne_neuron} - the neuron you want to set value for.
     * @author {Mr.Scatman} 
     */
    __setValByBoolPolarity = function ($nObj) {
        if ($nObj == null) return;
        var polarity_relon = MEMG.getRelonFromWithRToType($nObj.outs, TYPE_HAS, this.polarity.nid, TO_TYPE_ANNE);
        if (polarity_relon!=null) {
            if (polarity_relon.to_s==this.positive.nid) $nObj.setValue(true);
            else $nObj.setValue(false);
        }
    }

   /** 
     * Potentially call an external function.
     * Maybe it was loaded.  Maybe not.  If loaded, call it.
     * 
     * ALL OF THESE HAVE NO PARAMS.
     * @param {Object} - The mod (ie, MEM_UDF, NF_UDF, LANGUAGE, COUNTRY) 
     * @param {string} - A string of the function name in the mod.
     * 
     * @author {Mr.Scatman} 
     */
    _MOD_FUNC_CALLER = function($MOD, $FUNCSTR="") {
        if ($MOD!=null && $FUNCSTR!="" && $MOD[$FUNCSTR]!=undefined && typeof MOD[$FUNCSTR]==="function") {
            $MOD[$FUNC]();
        }
    }

    /**
    * Get the relns to chase for a set of nObjs.
    * 
    * The nObjs should be classes here.
    * 
    * @param {array} - an array of anne_neuron class neurons (ntype = NTYPE_CLASS).
    * @author {Mr.Scatman} 
    *
    */
    __GET_THUMB_RELNS_TO_CHASE=function($nObjs) {
        var relns = [];
        var handled_class_nids = [];
        for (var n=0; n < $nObjs.length; n++) {
            var nObj = $nObjs[n];
            if (nObj.ntype!=NTYPE_CLASS) continue;
            if (handled_class_nids.indexOf(nObj.nid)!=-1) continue;
            handled_class_nids.push(nObj.temp_nid);
            var thumb_relon = MEMG.getParamRelon(nObj.outs, this.primaryNeuronImg.temp_nid);
            if (thumb_relon!=null) {
                if (relns.indexOf(thumb_relon.to_s)!=-1) relns.push(thumb_relon.to_s);
            }
        }
        return relns;
    }


    
    /**
     * Get the URL
     * 
     * Intended to be generalized for any to_type TO_TYPE_URL to return the url.
     * Sometimes the URL is a URL part and requires a base.  Othertimes it is ready
     * to go as is.
     * 
     * When we look for a prop, store the prop with a p_TEMPNID in the neuron.
     * Use it next time without looking it up. 
     *    
     * @param {Object} - the relon object.                            
     * @author {Mr.Scatman} 
     * 
     */
    __GET_URL=function($relon) {
        if ($relon.to_type!=TO_TYPE_URL) {
            if (typeof $relon ==='string') return $relon;
            else return "";
        }
        var fNOBJ = MEMG.getNeuronByTempNid($relon.f);
        if (fNOBJ['p_'+$relon.r]!=undefined) return fNOBJ['p_'+$relon.r];
        var relnNOBJ = MEMG.getNeuronByTempNid($relon.r);
        var base_url_relon = this.__isBaseUrl(relnNOBJ);
        if (base_url_relon!=null) {
            var fNOBJ = MEMG.getNeuronByTempNid($relon.f);
            fNOBJ['p_'+$relon.r] = base_url_relon.to_s+$relon.to_s;
            return fNOBJ['p_'+$relon.r];
        }
        var fNOBJ = MEMG.getNeuronByTempNid($relon.f);
        fNOBJ['p_'+$relon.r] = $relon.to_s;
        return fNOBJ['p_'+$relon.r];
    }


    /**
     * This is for neuron imgs/vid/audio and files that have urls for their data.
     */
    __GET_URL_FROM_TNID=function($temp_nid){
        var assetNOBJ =  MEMG.getNeuronByTempNid($temp_nid);
        if (assetNOBJ==null) return "";
        var fdu = this.dataUrl.temp_nid;
        for (var r=0; r < assetNOBJ.outs.length; r++) {
            if(assetNOBJ.outs[r].to_type!=TO_TYPE_URL) continue;
            if(assetNOBJ.outs[r].r!=fdu) continue;
            return assetNOBJ.outs[r].to_s;
        }
        return "";
    }


    /**
     * ALL images must be vetted for fingerprint to be displayed in the extension.
     * 
     * @param {anne_neuron} - The parent class neuron.
     * @param {anne_neuron} - The target neuron we are trying to get the thumb URL for.
     * @author {Mr.Scatman} 
     * 
     */
     __GET_DEFAULT_IMG_NEURON=function($classNOBJ, $nObj) {
        
        //get it and get out!
        if ($nObj['p_'+this.primaryNeuronImg.temp_nid]!=undefined) {
            return MEMG.getNeuronByNid($nObj['p_'+this.primaryNeuronImg.temp_nid]);
        }

        //set the quick helper ref for next time, get out!
        var defaultSelfImg = MEMG.getRelonByTempR($nObj.outs, this.primaryNeuronImg.temp_nid, [TYPE_PARAM]);
        if (defaultSelfImg!=null) {
            $nObj['p_'+this.primaryNeuronImg.temp_nid] = defaultSelfImg.to_s;
            return MEMG.getNeuronByNid($nObj['p_'+this.primaryNeuronImg.temp_nid]);
        }

        //if no class or if this is the class of class... get out.
        if ($classNOBJ==undefined || $classNOBJ==null || $classNOBJ.temp_nid==this.class.temp_nid) return null;

        //check for class-based default image
        var defaultClassImgForAll = MEMG.getRelonByTempR($classNOBJ.outs, this.primaryNeuronImg.temp_nid, [TYPE_PARAM]);
        if (defaultClassImgForAll!=null) {
            $nObj['p_'+this.primaryNeuronImg.temp_nid] = defaultClassImgForAll.to_s;
            return MEMG.getNeuronByNid($nObj['p_'+this.primaryNeuronImg.temp_nid]);
        }
        return null;
    }


    /**
     * 
     * @param {*} $temp_nid 
     * @returns 
     */
    __isAnyReln = function($temp_nid) {
        for (var r=0; r < this._aRelons.length; r++) {
            if(this._aRelons[r].r==$temp_nid) return true;
        }
        return false;
    }


    /**    
     * 
     * Is this a base url relon?
     * 
     * @param {anne_neuron} - the neuron of a relon's reln.
     * @author {Mr.Scatman} 
     * 
     */
    __isBaseUrl=function($relnNOBJ) {
        var base_url_relon = MEMG.getRelonByTempR($relnNOBJ.outs, this.base_url.temp_nid);
        if (base_url_relon!=null) return base_url_relon;
        return null;
    }


    /**   
     * Is this a valid nid? 
     * 
     * This function is not such a big deal now in ANNE CHAIN.
     * In prior iteratons the ID wasn't a long string and it had more robust 
     * checking for conformity.  With a string for a long we can get shorter values
     * and thus cannot check for string length other than knowing it is 20 or less.
     * 
     * @param {string} - The long nid string.
     * @author {Mr.Scatman} 
     */
    __isValidNid = function($nid) {
        if (($nid+"").trim().length>20) return false;
        var test = parseInt($nid);
        if (test!=NaN && test && Number.isInteger(test)==true) return true;
        return false;
    } 


    /**    
     * Is this txid valid?
     * 
     * Since the txid and nid data types in ANNODE are the same (BIGINT(20))
     * this is the same as isValidNid.
     * 
     * @param {string} - The txid long string of a relon.
     * @author {Mr.Scatman} 
     * 
     */
    __isValidTXID = function($txid) {
        var regx = /^[0-9]+$/;
        if (regx.test($txid)==false) return false;

        var len = ($txid+"").trim().length;
        if (len>20 || len <10) return false;  // < 10 possible but rather rare.
        var test = parseInt($txid);
        if (test!=NaN && test && Number.isInteger(test)==true && (test+"").length()==$txid.length) return true;
        return false;
    }

    
	__isMalformedRelon=function($relon) {
		if ($relon==undefined || $relon.from_nid==undefined || $relon.type==undefined || $relon.reln==undefined || $relon.to_type==undefined || $relon.to_s==undefined ) {
			console.warn("FIX THIS!!! this should NEVER happen.");
			return true;
		}
		if ($relon.from_nid==null || $relon.type==null || $relon.reln==null || $relon.to_type==null || $relon.to_s==null ) {
			console.warn("FIX THIS!!! this should NEVER happen.");
			return true;
		}
		return false;
	}


    /**    
     * Validate the format of a neuron's value property.
     * 
     * @param {anne_neuron} - The ANNE neuron you want to validate format for.
     * @author {Mr.Scatman} 
     */
    __MASTER_FORMAT_VALIDATOR=function($nObj) {
        $nObj.format_validator($nObj);
    }


    /**  
     * Use the set_value function and get the value.
     *   
     * We call get value when dealing directly with nObjs for various reasons
     * 
     * @param {anne_neuron} - The ANNE neuron you want to get the run-time value for. 
     * @author {Mr.Scatman} 
     */
    __MASTER_GET_VALUE = function ($nObj) {
        if (typeof $nObj!=="object") {
            $nObj = MEMG.getNeuronByNid($nObj);
            if ($nObj==null) return "";
        }
        if ($nObj.bes[0]==this.integer.temp_nid) {
            $nObj.setValue(parseInt($nObj.text));
            return $nObj.value;
        }

        if ($nObj.set_value!=undefined && typeof $nObj.set_value==="function" && $nObj.ntype==NTYPE_INST) {
            this.__MASTER_SET_VALUE($nObj);
        }

        //try for an eval, value.  nOBJ->eval->xyz
        else {
            this.eval.set_value($nObj);
        }

        if ($nObj.value!=undefined) return $nObj.value;
        return $nObj.text;
    }


    /**  
     * Set the value of a neuron.
     * 
     * Use the master set value.  If there is a validator call it.
     * 
     * @param {anne_neuron} - The ANNE neuron you want to set the run-time value for. 
     * @author {Mr.Scatman} 
     */
    __MASTER_SET_VALUE = function ($nObj) {
        //if ($nObj.ntype!=NTYPE_INST) return;
        if ($nObj.set_value!=undefined) $nObj.set_value($nObj);
        if ($nObj.validator!=undefined) {
            this.__MASTER_FORMAT_VALIDATOR($nObj);
        }
    }


    /**  
     * Get a nid array from the worldview temp nid array.
     * 
     * We store the temp_nids array because it is a smaller data footprint.
     * @author {Mr.Scatman} 
     */
     getWorldViewNidList = function () {
        var wvs=[];
        for (var w=0; w < this._activeWorldViewTempNids.length; w++) {
            var wvOBJ = MEMG.getNeuronByTempNid(this._activeWorldViewTempNids[w]);
            wvs.push(wvOBJ.nid);
        }
        return wvs;
    }


    /**
     * just a convienance.. makes it easier than inline
     */
    isECAction=function($nObj, $bINSTONLY=true) {
        if ($nObj==null) return false;
        if ($bINSTONLY==true && $nObj.ntype!=NTYPE_INST) return false;
        return MEMG.besOf($nObj, this.ec_action.temp_nid);
    }
    isVbInst=function($nObj){
        if ($nObj==null) return false;
        if ($nObj.ntype!=NTYPE_INST) return false;
        if ($nObj.is(this.ec_action)==true) return true;
        if ($nObj.is(this.ec_is)==true) return true;
        if ($nObj.is(this.ec_posn)==true) return true;
    }
    isVb=function($nObj){
    }
    isVbClass=function($nObj){
    }

    /**
     */
     isAnorg=function($nObj) {
        if ($nObj==null) return false;
        if ($nObj.ntype!=NTYPE_INST) return false;
        return MEMG.besOf($nObj, this.anorg.temp_nid);
    }

    isRole=function($nObj) {
        if ($nObj==null) return false;
        return MEMG.besOf($nObj, this.role.temp_nid);
    }
    isRoleR=function($r) {
        var nOBJ = MEMG.getNeuronByTempNid($r);
        if (nOBJ==null) return false;
        return MEMG.besOf(nOBJ, this.role.temp_nid);
    }


    /**
     * Some things are just silly to do an ANUM display on....
     * @param {*} $nObj 
     * @returns 
     */
    annumable = function($nObj) {
        if ($nObj==undefined || $nObj==null) return false;
        if (!$nObj.text) return false;
        if (($nObj.is(this.EC_CLASS)==true && $nObj.is(this.ec_thing)==false && $nObj.is(this.ec_action)==false)  
        || $nObj.is(this.__ANNE_SYS_CLASS)==true 
        || $nObj.is(this._GENERAL_PROP_CLASS)==true
        || ($nObj.text.indexOf("USA Rep")!=-1)
        || ($nObj.text.indexOf("USA Sen")!=-1)
        || ($nObj.is(this.service_term)==true)
        || ($nObj.is(this.bounty)==true)
        || ($nObj.is(this.PDN)==true)
        || ($nObj.is(this.ASDF)==true)
        || ($nObj.is(this.distro_req)==true)
        || ($nObj.is(this.service_term))) return false;

        return true;
    }


    /**
     * Check if a nObj is a child of any of the $bes
     */
    isInBeLines=function($nObj, $bes) {
        for (var b=0; b < $bes.length; b++) {
            var be_tnid = $bes[b];
            if (MEMG.besOf($nObj, be_tnid)==true) return true;
        }
        return false;
    }

    /**
     */
    isIdentity=function($nObj) {
        if ($nObj==null) return false;
        return MEMG.besOf($nObj, this.identity.temp_nid);
    }

    /**
     */
    isIdentityINST=function($nObj) {
        if ($nObj==null) return false;
        if ($nObj.ntype!=NTYPE_INST) return false;
        return MEMG.besOf($nObj, this.identity.temp_nid);
    }

    /**
     */
    unitOfMeasure = function($psblUOM) {
        //conversion unit method.. 
        if (MEMG.besOf($psblUOM, this.unit_of_measure.temp_nid)==false) return null;

        //will return a uom object if we can make one.
        return this.factor.set_value($psblUOM);
    }

    convertUOM = function($uom, $uom_start_val, $altToUnits=null) {
        //$uom.unit = nObj of unit
        //$uom.value = factor already done.
        if ($uom==null) return "";
        if ($uom.unit==null) return "";
        if ($uom.factored==null) return "";
        var unit_lbl = "";
        if ($uom.unit_lbl!="") unit_lbl = $uom.unit_lbl;
        return parseFloat($uom_start_val)*parseFloat($uom.factored)+unit_lbl;
    }



    /**
     * 
     * @param {*} $textOld 
     * @param {*} $textNew 
     * @param {*} $nid 
     * @returns 
     */
    editTextIndex = function($textOld, $textNew, $nid) {
        if ($textOld==null) return;
        $textOld = $textOld.trim()    
        if ($textOld=="") return;
        $textOld = $textOld.replace(/[^\w\s]/gi, '').replace(/\s+/g, ' ');
        if ($textOld==" ") return;

        if ($textNew==null) return;
        $textNew = $textNew.trim()    
        if ($textNew=="") return;
        $textNew = $textNew.replace(/[^\w\s]/gi, '').replace(/\s+/g, ' ');
        if ($textNew==" ") return;

        if (this._TEXT_INDEX[ $textOld ] == undefined) {
            console.log("huh???");
        }
        else {
            var entryArr = this._TEXT_INDEX[ $textOld ];
            if (entryArr.length==1) delete this._TEXT_INDEX[ $textOld ];
            else {
                for (var s=0; s < entryArr.length; s++) {
                    if (entryArr[s]==$nid) {
                        entryArr.splice(s, 1);
                        break;
                    }
                }
            }
        }
        this.addToTextIndex($textNew, $nid);
    }


    /**
     * 
     * @param {*} $nids 
     */
    removeTextIndexByNids=async($nids)=> {
        var keys = Object.keys( this._TEXT_INDEX);
        for(var k=0; k < keys.length; k++) {
            var key = keys[k];
            var arr = this._TEXT_INDEX[key];
            for(var a=0; a < arr.length; a++) {
               if ($nids.indexOf(arr[a])!=-1) {
                   if (arr.length>1) {
                      a--;
                      arr.splice(a,1);
                      continue;
                   }
                   else delete this._TEXT_INDEX[key];
               }
            }
        }
    }


    /**
     * 
     * @param {*} $nid 
     */
    removeTextIndexByNid=function($nid) {
        var keys = Object.keys( this._TEXT_INDEX);
        for(var k=0; k < keys.length; k++) {
            var key = keys[k];
            var arr = this._TEXT_INDEX[key];
            for(var a=0; a < arr.length; a++) {
               if (arr[a]==$nid) {
                   if (arr.length>1) {
                      a--;
                      arr.splice(a,1);
                      continue;
                   }
                   else delete this._TEXT_INDEX[key];
               }
            }
        }
    }


        
    /**
     * typically used when deleting ghosts. 
     * @param {*} $text 
     * @param {*} $nid 
     * @returns 
     */
    removeTextIndex = function($text, $nid) {
        if ($text==null) return;
        $text = $text.trim()    
        if ($text=="") return;
        $text = $text.replace(/[^\w\s]/gi, '').replace(/\s+/g, '');
        if ($text==" ") return;

        if (this._TEXT_INDEX[ $text ] == undefined) {
            if (ANNE.SETTINGS.debug()==true) console.log("no text index for: "+$text);    
            return;
        }
        else {
            var entryArr = this._TEXT_INDEX[ $text ];
            if (entryArr.length==1) {
                delete this._TEXT_INDEX[ $text ];
            }
            else {
                for (var s=0; s < entryArr.length; s++) {
                    if (entryArr[s]==$nid) {
                        entryArr.splice(s, 1);
                        break;
                    }
                }
            }
        }
    }


    /**
     * 
     * @param {*} $text 
     * @param {*} $nid 
     * @param {*} $bDOPARTIALS 
     * @returns 
     */
    addToTextIndex = function($text, $nid, $bDOPARTIALS=true) {
        if ($text==null) return;
        $text = $text.trim()    
        if ($text=="") return;
        $text = $text.replace(/[^\w\s]/gi, '').replace(/\s+/g, '');
        if ($text==" ") return;
        $text = $text.toLowerCase();
            
        if ($text=="" || $text==" ") return;

        if (ANNE.MEM._TEXT_INDEX[ $text ] == undefined) {
            ANNE.MEM._TEXT_INDEX[ $text ] = [ $nid ];
        }
        else if (ANNE.MEM._TEXT_INDEX[ $text].indexOf($nid) == -1) {
            ANNE.MEM._TEXT_INDEX[ $text ].push($nid);
        }

        //this will blowout ties in our index for any "wd" 
        //experimental for lukat search finding partials..
        //at present it must find full entry in the index (of name/alias or anything referenced)
        if ($bDOPARTIALS==true) {
            var parts = $text.split( /[_\s]+/);
            if (parts.length>1) {
                var partial = ""
                for (var p=0; p < parts.length; p++) {
                    partial+=parts[p];
                    if (partial=="" || partial==" ") {
                        continue;
                    }
                    //if (ANNE.MEM._TEXT_INDEX_PARTIALS[ partial ] == undefined) {
                        //ANNE.MEM._TEXT_INDEX_PARTIALS[ partial ] = [ $nid ];
                    //}
                   // else if (ANNE.MEM._TEXT_INDEX_PARTIALS[ partial].indexOf($nid) == -1) {
                       // ANNE.MEM._TEXT_INDEX_PARTIALS[ partial ].push($nid);
                   // } 

                    if (ANNE.MEM._TEXT_INDEX[ partial ] == undefined) {
                        ANNE.MEM._TEXT_INDEX[ partial ] = [ $nid ];
                    }
                    else if (ANNE.MEM._TEXT_INDEX[ partial].indexOf($nid) == -1) {
                        ANNE.MEM._TEXT_INDEX[ partial ].push($nid);
                    }
                    //partial+=" ";
                }
            }
        }
    }



    /**
     * 
     * @param {*} $text 
     * @returns 
     */
    getTnidsByText = function($text) {
        if ($text==null) return [];
        $text = $text+"";
        if ($text=="") return [];
        $text=$text.trim();
        $text = $text.replace(/[^\w\s]/gi, '').replace(/\s+/g, '');
        if ($text=="") return [];
        $text = $text.toLowerCase();
        var arr = ANNE.MEM._TEXT_INDEX [ $text ];
        if (arr==undefined || arr.length==0) return [];
        return ANNE.MEMG.getTnidsFromNids(arr);
    }


    searchTextIndex = async($text, $bPULLIDB=true, $bEXACT=false)=> {
        if ($text==null) return [];
        $text = $text+"";
        if ($text=="") return [];
        $text=$text.trim();
        $text = $text.replace(/[^\w\s]/gi, '').replace(/\s+/g, '');
        if ($text=="") return [];
        var exactText = $text;
        $text = $text.toLowerCase();
        var arr = this._TEXT_INDEX [ $text ];
        if (arr==undefined || arr.length==0) return [];

        var nObjs = [];
        var nids_to_pull = [];
        for (var a=0; a < arr.length; a++) {
            let nObj = MEMG.getOrMakeNid(arr[a]);

            if ($bEXACT==true && nObj.text!=exactText) {
                console.log(nObj.text+" is not an exact match, skipping")
                continue;
            }

            if (nObj==null || nObj.text=="" || nObj.bes.length==0) nids_to_pull.push(arr[a]);
            else nObjs.push(nObj);
        }

        if (nids_to_pull.length>0 && $bPULLIDB==true) {
            await ANNE.IDB.lookupNidsGetAllOuts(null, nids_to_pull);
            for(var n=0; n < nids_to_pull.length; n++) {
                let nid = nids_to_pull[n];
                let nObj = MEMG.getNeuronByNid(nid);
                if (nObj!=null) nObjs.push(nObj);
            }
        }
        return nObjs;
    }


    //for loads at run-time of neurons not on-chain.
    //mostly used for testing at present
    _MARKSAND=function() {
        this._sand_mark = this._aRelons.length;
    }
    _CLEANSAND=async()=> {
        await MEMP.reprocessRelonsFrom(this._sand_mark);
        this._sand_mark = 0;
    }

}

const MEM = new mem ();